Package: LACE
Version: 2.16.0
Date: 2026-03-09
Title: Longitudinal Analysis of Cancer Evolution (LACE)
Authors@R: c(person("Daniele", "Ramazzotti", role=c("aut"),email="daniele.ramazzotti@unimib.it",
                comment = c(ORCID = "0000-0002-6087-2666")),
    person("Fabrizio","Angaroni",role=c("aut"),email="fabrizio.angaroni@unimib.it"),
    person("Davide","Maspero",role=c("cre","aut"),email="d.maspero@campus.unimib.it"),
    person("Alex","Graudenzi",role=c("aut"),email="alex.graudenzi@unimib.it"),
    person("Luca", "De Sano", role=c("aut"), email="luca.desano@gmail.com",
                comment = c(ORCID = "0000-0002-9618-3774")),
    person("Gianluca","Ascolani",role=c("aut"),email="gianluca.ascolani@unimib.it"))
Depends: R (>= 4.2.0)
Imports: curl, igraph, foreach, doParallel, sortable, dplyr, forcats,
        data.tree, graphics, grDevices, parallel, RColorBrewer, Rfast,
        stats, SummarizedExperiment, utils, purrr, stringi, stringr,
        Matrix, tidyr, jsonlite, readr, configr, DT, tools, fs,
        data.table, htmltools, htmlwidgets, bsplus, shinyvalidate,
        shiny, shinythemes, shinyFiles, shinyjs, shinyBS,
        shinydashboard, biomaRt, callr, logr, ggplot2, svglite
Suggests: BiocGenerics, BiocStyle, testthat, knitr, rmarkdown
Name: LACE: an R package for the inference of longitudinal cancer
        evolution models
Description: 
    LACE is an algorithmic framework that processes single-cell somatic mutation profiles from cancer samples collected at different time points and in distinct experimental settings, to produce longitudinal models of cancer evolution. The approach solves a Boolean Matrix Factorization problem with phylogenetic constraints, by maximizing a weighed likelihood function computed on multiple time points.
Encoding: UTF-8
License: file LICENSE
URL: https://github.com/BIMIB-DISCo/LACE
BugReports: https://github.com/BIMIB-DISCo/LACE
biocViews: BiomedicalInformatics, SingleCell, SomaticMutation
RoxygenNote: 7.3.3
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/LACE
git_branch: RELEASE_3_23
git_last_commit: 5636e19
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 22:36:30 UTC; biocbuild
Author: Daniele Ramazzotti [aut] (ORCID:
    <https://orcid.org/0000-0002-6087-2666>),
  Fabrizio Angaroni [aut],
  Davide Maspero [cre, aut],
  Alex Graudenzi [aut],
  Luca De Sano [aut] (ORCID: <https://orcid.org/0000-0002-9618-3774>),
  Gianluca Ascolani [aut]
Maintainer: Davide Maspero <d.maspero@campus.unimib.it>
Built: R 4.6.0; ; 2026-04-29 04:48:09 UTC; unix

# Generated by roxygen2: do not edit by hand

export(LACE)
export(LACEview)
export(compute.mutation.distance)
export(compute.variants.error.rates)
export(lacedata)
export(longitudinal.tree.plot)
import(DT)
import(RColorBrewer)
import(SummarizedExperiment, except = c("trim","shift", "show"))
import(biomaRt, except = c("show","select"))

import(bsplus)
import(callr)
import(configr)
import(curl, except = c("parse_date"))
import(data.table, except = c("first","last", "between"))
import(data.tree)
import(doParallel)
import(dplyr)
import(forcats)
import(foreach)
import(fs)
import(ggplot2)
import(grDevices)
import(graphics, except = c("box"))
import(htmltools)
import(jsonlite)
import(logr)
import(parallel)
import(readr)
import(shiny, except = c("runExample", "dataTableOutput", "renderDataTable","validate"))
import(shinyBS)
import(shinyFiles)
import(shinydashboard)
import(shinyjs)
import(shinythemes)
import(shinyvalidate)
import(sortable)
import(stringi)
import(stringr)
import(svglite)
import(tidyr)
import(tools)
import(utils)
importFrom(Matrix,rankMatrix)
importFrom(Rfast,rowMaxs)
importFrom(grDevices,colorRampPalette)
importFrom(jsonlite,toJSON)
importFrom(purrr,detect_index)
importFrom(purrr,is_null)
importFrom(stats,dist)
importFrom(stats,hclust)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(tidyr,gather)

Lua Access Control Engine - Lace
================================

Lace is a simple access control engine modelled on Squid's acl syntax.
It provides a parser of rulesets and an engine to execute the parsed
rulesets.  It relies on the calling application to provide access
control types and then Lace runs the boolean logic and returns an
allow/deny result along with the location of the decision and any
description provided by it.  Lace also handles errors in the control
callbacks to always return gracefully in the form:

    local result, reason = engine:run(context)
    
    if result == nil then
       report_error(reason)
    elseif result == false then
       handle_deny(reason)
    else
       handle_allow(reason)
    end

Lace is designed to allow a ruleset loaded into an engine to be run
multiple times with different contexts, each time unaffected by the
last.  Of course, this relies on various idempotency requirements
being placed on the control type callbacks, but that is covered in the
usage documentation.

For some examples of using Lace, please see the examples/ tree.

Thanks
======

Thanks go to Codethink Limited for sponsoring development by means of tea,
biscuits and long lunch hours.

lace
====

[![version](https://img.shields.io/pypi/v/lace?style=flat-square)][pypi]
[![python versions](https://img.shields.io/pypi/pyversions/lace?style=flat-square)][pypi]
[![version](https://img.shields.io/pypi/l/lace?style=flat-square)][pypi]
[![build status](https://img.shields.io/circleci/project/github/lace/lace/master?style=flat-square)][circle]

Polygonal mesh library developed at Body Labs.

The library is under active maintenance, and the goals are compatible with that:

- Keep the library working in current versions of Python and other tools.
- Make bug fixes.
- Provide API stability and backward compatibility with the upstream version.
- Respond to community contributions.

The eventual goal is to perform a rewrite of the loader and core mesh
functionality with quad support as part of a ["lace-core" project][lacecore]
and consider how to handle the extensive mesh manipulation functions which
remain.

[circle]: https://circleci.com/gh/lace/lace
[pypi]: https://pypi.org/project/lace/
[lacecore]: https://github.com/lace/lacecore-sketches


Installation
------------

### Install dependencies

Mac OS:
```sh
brew update && brew install boost
pip install numpy==1.13.1
pip install lace
```

Linux:
```sh
apt-get install -y --no-install-recommends libsuitesparse-dev libboost-dev
pip install numpy==1.13.1
pip install lace
```

Docker:
```
docker build .
```

### Install the library

```sh
pip install lace
```


Development
-----------

```sh
pip install -r requirements_dev.txt
pip install -e .
rake test
rake lint
```


Contribute
----------

- Issue Tracker: https://github.com/lace/lace/issues
- Source Code: https://github.com/lace/lace

Pull requests welcome!


Support
-------

If you are having issues, please let us know.


Acknowledgements
----------------

This library was refactored from legacy code at Body Labs by [Alex Weiss][],
with portions by [Eric Rachlin][], [Paul Melnikow][], [Victor Alvarez][],
and others. It was extracted from the Body Labs codebase and open-sourced by
[Guillaume Marceau][]. In 2018 it was [forked by Paul Melnikow][fork] and
published as [metabolace][fork pypi]. Thanks to a repository and package
transfer from Body Labs, the fork has been merged back into the original.

[alex weiss]: https://github.com/algrs
[eric rachlin]: https://github.com/eerac
[paul melnikow]: https://github.com/paulmelnikow
[victor alvarez]: https://github.com/yangmillstheory
[guillaume marceau]: https://github.com/gmarceau
[fork]: https://github.com/metabolize/lace
[fork pypi]: https://pypi.org/project/metabolace/


Similar projects
----------------

There is an unrelated permissively licensed mesh manipulation library called
[Trimesh][] which provides some similar functionality.

[trimesh]: https://github.com/mikedh/trimesh


License
-------

The project is licensed under the two-clause BSD license.

This project uses the RPly library to read and write PLY files, by Diego Nehab,
IMPA, distributed under the MIT License.
 * http://www.impa.br/~diego/software/rply

# lace-stubble

`lace-stubble` is a small, (mostly) memory-safe stub bootloader for running the Linux kernel on UEFI, implemented using the libraries provided by Project Lace.

It is a Rust implementation of the earlier [stubble](https://github.com/ubuntu/stubble) project, which in turn is a minimal derivative of systemd-stub with fewer features.

## Overview

`lace-stubble` is a "stub" EFI binary that can be specialized into a bootable image by embedding resources (kernel, initrd, etc.) into PE sections.

## Features

- Loading of Linux kernel, initramfs, command line, and device trees from a single executable.
- Automatic Device Tree Selection:
  - Determines the platform's compatible string by inspecting the firmware-provided DTB or by matching Computer Hardware IDs (CHIDs) against an embedded `.hwids` database.
  - Selects and installs the correct Device Tree Blob (DTB) from embedded `.dtbauto` sections based on the platform compatible string.
- Command Line Handling: Supports embedded command lines via the `.cmdline` section, with fallback to UEFI Load Options.
- Memory Safety: Written in Rust, leveraging the `uefi` crate and `lace-util` for safe parsing and execution.

## PE Sections

`lace-stubble` looks for the following sections in its own PE image:

| Section Name | Description | Requirement |
|--------------|-------------|-------------|
| `.linux`     | The Linux kernel image. | **Mandatory** |
| `.initrd`    | The initial RAM disk (initramfs). | Optional** |
| `.cmdline`   | Kernel command line arguments. | Optional* |
| `.hwids`     | JSON/Binary mapping of CHIDs to compatible strings. | Optional |
| `.dtbauto`   | Device Tree Blob (DTB). Multiple sections allowed. | Optional |

\* If `.cmdline` is missing, `lace-stubble` will attempt to use the UEFI Load Options (arguments passed to the EFI binary).

\** If `.initrd` is missing, an external initrd will be used if available (via UEFI Load File2 Protocol).

## Build

`lace-stubble` is part of the Lace workspace. To build it:

```none
cargo build -p lace-stubble --target MY_TARGET [--release]
```

Note that `lace-stubble` targets `x86_64-unknown-uefi` or `aarch64-unknown-uefi`. Ensure you have the appropriate target installed:

```none
rustup target add x86_64-unknown-uefi
# or
rustup target add aarch64-unknown-uefi
```

## Usage

To create a bootable `lace-stubble` image, you use the `pewrap` tool (part of Project Lace) to inject the necessary sections into the base `stubble.efi` binary.

### Example with `pewrap`

```none
# Wrap resources into the stub
cargo run -p pewrap -- \
    --stub target/x86_64-unknown-uefi/release/lace-stubble.efi \
    --output bootable.efi \
    --linux /boot/vmlinuz-linux \
    --initrd /boot/initramfs-linux.img \
    --cmdline "console=ttyS0 console=tty0 root=UUID=... rw" \
    --hwids data/hwids/json \
    --dtbauto path/to/board1.dtb \
    --dtbauto path/to/board2.dtb
```

This command creates `bootable.efi`, which contains the kernel, initrd, command line, and DTBs. When executed on a UEFI system, `lace-stubble` will automatically select the correct DTB for the hardware and boot the kernel.

# xtask

Build automation tasks for the lace project.

## Usage

Run xtask tasks with:
```none
cargo run -p xtask -- <command>
```

## Available commands

### `mangen`

Generate manual pages for all tools that have CLI interfaces.

```none
cargo run -p xtask -- mangen
```

This generates man pages in the `man/` directory at the repository root:
- `man/pewrap.1` - Man page for pewrap
- `man/collect-hwids.1` - Man page for collect-hwids

By default, man pages are written to `./man`. To specify a different output directory, use:

```none
cargo run -p xtask -- mangen --output-dir /path/to/output
```

To view the generated man pages, use the `man` command:
```none
man ./man/pewrap.1
man ./man/collect-hwids.1
```

## Adding new tools

To add man page generation for a new tool:

1. Ensure the tool uses `clap` with the `derive` feature.
2. Extract the CLI argument struct to a separate `cli.rs` module.
3. Create a `lib.rs` that re-exports the cli module: `pub mod cli;`.
4. Add the tool as a dependency in `tools/xtask/Cargo.toml`.
5. Add man page generation call in `tools/xtask/src/main.rs`.

See pewrap or collect-hwids for examples.

# `fdt`

A pure-Rust `#![no_std]` crate for parsing Flattened Devicetrees, with the goal of having a
very ergonomic and idiomatic API.

[![crates.io](https://img.shields.io/crates/v/fdt.svg)](https://crates.io/crates/fdt) [![Documentation](https://docs.rs/fdt/badge.svg)](https://docs.rs/fdt) ![Build](https://github.com/repnop/fdt/actions/workflows/test.yml/badge.svg?branch=master&event=push)

## License

This crate is licensed under the Mozilla Public License 2.0 (see the LICENSE file).

## Example

```rust
static MY_FDT: &[u8] = include_bytes!("../dtb/test.dtb");

fn main() {
    let fdt = fdt::Fdt::new(MY_FDT).unwrap();

    println!("This is a devicetree representation of a {}", fdt.root().model());
    println!("...which is compatible with at least: {}", fdt.root().compatible().first());
    println!("...and has {} CPU(s)", fdt.cpus().count());
    println!(
        "...and has at least one memory location at: {:#X}\n",
        fdt.memory().regions().next().unwrap().starting_address as usize
    );

    let chosen = fdt.chosen();
    if let Some(bootargs) = chosen.bootargs() {
        println!("The bootargs are: {:?}", bootargs);
    }

    if let Some(stdout) = chosen.stdout() {
        println!("It would write stdout to: {}", stdout.name);
    }

    let soc = fdt.find_node("/soc");
    println!("Does it have a `/soc` node? {}", if soc.is_some() { "yes" } else { "no" });
    if let Some(soc) = soc {
        println!("...and it has the following children:");
        for child in soc.children() {
            println!("    {}", child.name);
        }
    }
}
```

### annotation.R
###
### Code for ANNOVAR annotations.
###
### See file LICENSE in the main folder for licensing and copyright
### information.
###

### Notes:
### MA: questo file assume UN*X.  Modificare.

av_compute <- function(anovar_convert,
                       anovar_annot,
                       anovar_db,
                       vcf_in_dir,
                       vcf_annot_dir,
                       pattern = '.*.vcf$',
                       v_ref = 'hg38',
                       InFilesToDo = NULL,
                       OutFilesToRm = NULL) {
    ## library("foreach")
    ## library("doParallel")
    
    numCores <- detectCores()
    registerDoParallel(ceiling(numCores / 2))
    
    dir.create(vcf_annot_dir, showWarnings = FALSE)
    stdout_log <- file.path(vcf_annot_dir, 'stdout.log') 
    stderr_log <- file.path(vcf_annot_dir, 'stderr.log') 
    #log_bash <- paste0(' >> ', stdout_log, ' 2>> ', stderr_log) # UN*X! E Windows?
    tmp_dir <- tempdir()
    stdout_log_tmp <- file.path(tmp_dir, "av_stdout.log")           # UN*X! E Windows?
    stderr_log_tmp <- file.path(tmp_dir, "av_stderr.log")           # UN*X! E Windows?
    
    cat("", file = stdout_log, append=FALSE) # comment to not erase log
    cat("", file = stderr_log, append=FALSE) # comment to not erase log
    
    
    vcf_filtered = list.files(path = vcf_in_dir, pattern = pattern)
    log2_print(InFilesToDo)
    if (!is.null(InFilesToDo)) {
        log2_print('In')
        log2_print(vcf_filtered)
        vcf_filtered <- vcf_filtered[vcf_filtered %in% basename(InFilesToDo)]
        log2_print(vcf_filtered)
    }
    if (!is.null(OutFilesToRm)) {
        log2_print('Out')
        ## OutFilesToRm <- file.path(vcf_annot_dir, OutFilesToRm)
        log2_print(OutFilesToRm)
        ## file.remove(OutFilesToRm)
        log2_print(list.files(path = vcf_annot_dir))
    }
    
    ## res <- foreach(f = vcf_filtered, .combine=c) %dopar% {
    for (f in vcf_filtered) {
        
        f_in = file.path(vcf_in_dir,f)
        f_out = sub('\\.vcf$', '', f)
        f_out = file.path(vcf_annot_dir,paste0(f_out, '.anninput'))
        
        ## cmd_exp=paste0(anovar_convert,' -format vcf4 ',f_in,' --outfile ',f_out, ' --includeinfo', log_bash)
        ## exit_status<- <- system(cmd_exp, intern = FALSE)
        cmd_exp = paste0('-format vcf4 ',
                         f_in,
                         ' --outfile ',
                         f_out,
                         ' --includeinfo')
        log2_print(cmd_exp)
        exit_status <-
            system2(anovar_convert,
                    args = cmd_exp,
                    stdout = stdout_log_tmp,
                    stderr = stderr_log_tmp,
                    wait = TRUE)
        now <- date()
        cat(paste0("\nExecuted: ",
                   now,
                   '\t',
                   anovar_convert,
                   "\n",
                   "Exit status: ",
                   exit_status,
                   "\n"),
            file = stdout_log,
            append = TRUE)
        file.append(stdout_log, stdout_log_tmp)
        cat(paste0("\nExecuted: ",
                   now,
                   '\t',
                   anovar_convert,
                   "\n",
                   "Exit status: ",
                   exit_status,
                   "\n"),
            file = stderr_log,
            append=TRUE)
        file.append(stderr_log, stderr_log_tmp)
    }
    
    ## res <- foreach(f = vcf_filtered, .combine=c) %dopar% {
    for (f in vcf_filtered) {
        f_out = sub('\\.vcf$', '', f)
        f_out = file.path(vcf_annot_dir, paste0(f_out, '.anninput'))
        f_in = f_out
        
        ## cmd_exp=paste0(anovar_annot, ' -out ', f_out, ' -exonicsplicing -build hg38 ', f_in, ' ',anovar_db, log_bash)
        ## exit_status<- <- system(cmd_exp, intern = FALSE)
        cmd_exp = paste0(' -out ', f_out,
                         ' -exonicsplicing -build ',
                         v_ref, ' ', f_in, ' ', anovar_db)
        log2_print(cmd_exp)
        exit_status <- system2(anovar_annot,
                               args = cmd_exp,
                               stdout = stdout_log_tmp,
                               stderr = stderr_log_tmp,
                               wait = TRUE)
        now <- date()
        cat(paste0("\nExecuted: ",
                   now,
                   '\t',
                   anovar_annot,
                   "\n",
                   "Exit status: ",
                   exit_status,
                   "\n"),
            file = stdout_log,
            append = TRUE)
        file.append(stdout_log, stdout_log_tmp)
        
        cat(paste0("\nExecuted: ", now, '\t', anovar_annot, "\n",
                   "Exit status: ",exit_status, "\n"),
            file = stderr_log,
            append = TRUE)
        file.append(stderr_log, stderr_log_tmp)
    }
}


### end of file -- annotation.R

#################################################
##########  NA2  ###########
#################################################
NA_compute2_load <- function(data_dir, depth_dir, out_dir) {

  browser()


  dir.create(out_dir, showWarnings = FALSE)
  if (!dir.exists(out_dir))
    showNotification(paste("Filtered data output folder,", out_dir, ", cannot be created"), duration = 10, type = "warning")

  # load data
  cells_aggregate_info <- NULL
  snpMut_filt_freq <- NULL
  depth <-NULL
  if (file.exists(file.path( data_dir, "cells_aggregate_info.rds")))
    cells_aggregate_info <- readRDS(file=paste0(file.path( data_dir, "cells_aggregate_info.rds")))
  if (file.exists(file.path( out_dir, "snpMut_filt_freq_reduced.rds")))
    snpMut_filt_freq <- readRDS(file=paste0(file.path( out_dir, "snpMut_filt_freq_reduced.rds")))
  if (file.exists(file.path( depth_dir, "final_data_depth.txt")))
    depth <- as.matrix(read.table( file.path(depth_dir, "final_data_depth.txt")))
  return( list(cells_aggregate_info=cells_aggregate_info, snpMut_filt_freq=snpMut_filt_freq, depth=depth))
}

NA_compute2 <- function(depth_minimum, minumum_median_total, minumum_median_mutation, data_dir, depth_dir, out_dir, time_points, verified_genes, non_NA_genes, files) {

  browser()

  cells_aggregate_info <- files$cells_aggregate_info
  snpMut_filt_freq <- files$snpMut_filt_freq
  depth <- files$depth

  # set working directory
  #baseDir = "/data_directory/BimiB/share/LACE/MELANOMA/"
  #setwd(baseDir)

  # load required libraries
  # library("TRONCO")


  # dir.create(out_dir, showWarnings = F)
  # if (!dir.exists(out_dir))
  #   showNotification(paste("Filtered data output folder,", out_dir, ", cannot be created"), duration = 10, type = "warning")


  # list of manually verified mutations
  #verified_genes = c("ARPC2","CCT8","COL1A2","CYCS","HNRNPC","PCBP1","PRAME","RPL5")

  #depth_minimum = 3 # minimum depth to set values to NA
  #minumum_median_total = 10 # minimum median depth for total reads
  #minumum_median_mutation = 4 # minimum median depth for reads supporting mutations

  # load data
  # cells_aggregate_info <- readRDS(file=paste0(file.path( data_dir, "cells_aggregate_info.rds")))
  # snpMut_filt_freq <- readRDS(file=paste0(file.path( data_dir, "snpMut_filt_freq.rds")))
  # depth = as.matrix(read.table( file.path(depth_dir, "final_data_depth.txt")))

  # select only verified mutations
  if (!is.null(non_NA_genes))
    snpMut_filt_freq = snpMut_filt_freq[
      which(snpMut_filt_freq$Gene %in% non_NA_genes),
      c("scID","Time","Gene","Chr","PosStart","PosEnd","REF",
        "ALT","MutType","depth","Allele_Ratio")
      ]
  

  snpMut_filt_freq = snpMut_filt_freq[order(snpMut_filt_freq[,3],snpMut_filt_freq[,4],snpMut_filt_freq[,5],snpMut_filt_freq[,6],snpMut_filt_freq[,7],snpMut_filt_freq[,8],snpMut_filt_freq[,9],snpMut_filt_freq[,1],snpMut_filt_freq[,2],snpMut_filt_freq[,10],snpMut_filt_freq[,11]),]

  # compute frequency of each mutation
  #distinct_mutations = unique(snpMut_filt_freq[,c("Gene","Chr","PosStart","PosEnd","REF","ALT","ALT","ALT","ALT","ALT","ALT","ALT")])
  distinct_mutations = unique(snpMut_filt_freq[,c("Gene","Chr","PosStart","PosEnd","REF","ALT")])
  
  if (nrow(distinct_mutations)==0){
    num_columns <- sapply(distinct_mutations, is.numeric)
    distinct_mutations[num_columns] <- lapply(distinct_mutations[num_columns], round, 3)
    if (file.exists(file.path( out_dir, "D.RData")))
      file.remove(file.path(out_dir,"D.RData"))
    return(list("distinct_mutations"=distinct_mutations, "g"=NULL))
  }
    
  
  #time_points=c("before treatment", "4d on treatment", "28d on treatment", "57d on treatment")
  #for (t in seq(1,length(time_points)) )
  #  distinct_mutations[[paste0('FreqT',t)]] = NA
  #distinct_mutations[['MedianDepth']] = NA
  #distinct_mutations[['MedianDepthMut']] = NA
  #colnames(distinct_mutations)[7] = "FreqT1"
  #colnames(distinct_mutations)[8] = "FreqT2"
  #colnames(distinct_mutations)[9] = "FreqT3"
  #colnames(distinct_mutations)[10] = "FreqT4"
  #colnames(distinct_mutations)[11] = "MedianDepth"
  #colnames(distinct_mutations)[12] = "MedianDepthMut"
  #cells_timepoints = unique(snpMut_filt_freq[,c("scID","Time")])
  #distinct_mutations$FreqT1 = NA
  #distinct_mutations$FreqT2 = NA
  #distinct_mutations$FreqT3 = NA
  #distinct_mutations$FreqT4 = NA
  #distinct_mutations$MedianDepth = NA
  #distinct_mutations$MedianDepthMut = NA

  #times = list()
  #for (t in time_points)
  #  times[[t]] = as.numeric(table(cells_timepoints$Time)[t])

  # browser()
  # 
  # for(i in 1:nrow(distinct_mutations)) {
  #   curr = snpMut_filt_freq[
  #     which(
  #       snpMut_filt_freq$Gene %in% distinct_mutations[i,"Gene"] &
  #         snpMut_filt_freq$Chr %in% distinct_mutations[i,"Chr"] & 
  #         snpMut_filt_freq$PosStart %in% distinct_mutations[i,"PosStart"] & 
  #         snpMut_filt_freq$PosEnd %in% distinct_mutations[i,"PosEnd"] & 
  #         snpMut_filt_freq$REF %in% distinct_mutations[i,"REF"] & 
  #         snpMut_filt_freq$ALT %in% distinct_mutations[i,"ALT"]),]
  #   
  #   for (t in seq(1,length(time_points)) )
  #     distinct_mutations[i,paste0('FreqT',t)] = as.numeric(table(curr$Time)[time_points[[t]]]) / times[[t]]
  # 
  #   #distinct_mutations[i,"FreqT1"] = as.numeric(table(curr$Time)["before treatment"]) / t1
  #   #distinct_mutations[i,"FreqT2"] = as.numeric(table(curr$Time)["4d on treatment"]) / t2
  #   #distinct_mutations[i,"FreqT3"] = as.numeric(table(curr$Time)["28d on treatment"]) / t3
  #   #distinct_mutations[i,"FreqT4"] = as.numeric(table(curr$Time)["57d on treatment"]) / t4
  #   #print('a')
  #   #print(curr$depth)
  #   #print(median(curr$depth))
  #   #print(median(curr$Allele_Ratio))
  #   distinct_mutations[i,"MedianDepth"] = as.numeric(median(curr$depth))
  #   distinct_mutations[i,"MedianDepthMut"] = as.numeric(median(curr$Allele_Ratio))
  # }
  
  distinct_mutations <- snpMut_filt_freq %>%
    group_by(Time) %>% 
    mutate(tot_per_time=n_distinct(scID)) %>% #cells per time point
    group_by(Gene,Chr,PosStart,PosEnd,REF,ALT,tot_per_time) %>% 
    add_count(Time) %>% #cells per mutations
    mutate(fr=n/tot_per_time) %>% 
    group_by(Gene,Chr,PosStart,PosEnd,REF,ALT) %>%
    mutate(MedianDepth=median(depth), MedianDepthMut=median(Allele_Ratio)) %>%
    ungroup() %>%
    select(-c(scID,MutType, Allele_Ratio, depth, n, tot_per_time)) %>% group_by_all() %>%
    mutate(Time= as.integer(factor(Time, levels = time_points, ordered = TRUE))) %>%
    distinct() %>%
    #summarise() %>%  #arrange(Time) %>%
    pivot_wider(names_from = Time, values_from = fr, names_prefix = "FreqT") 
  
  
  
#   distinct_mutations %>% add_count(Time, name= tot_per_time) %>%
#     arrange(name) %>%
#     group_by(Gene,Chr,PosStart,PosEnd,REF,ALT,Time) %>% 
#     pivot_wider(names_from = wool, values_from = Time, 
#                 
#                 mutate( )%>%
#                   
#                   group_by(Gene,Chr,PosStart,PosEnd,REF,ALT) %>%
#                   mutate(MedianDepth"] = median(depth)) %>%
# mutate(MedianDepth"] = median(Allele_Ratio))
#                 
  
  #distinct_mutations = distinct_mutations[-c(3,5,6,8),]
  # if (
  #   length(
  #     sort(
  #       unique(
  #         c(
  #           which(distinct_mutations$MedianDepth<minumum_median_total),
  #           which(distinct_mutations$MedianDepthMut<minumum_median_mutation)
  #           )
  #         )
  #       )
  #     )>0)
  #   distinct_mutations = distinct_mutations[
  #     -sort(
  #       unique(c(
  #         which(distinct_mutations$MedianDepth<minumum_median_total),
  #         which(distinct_mutations$MedianDepthMut<minumum_median_mutation)
  #         ))
  #       ),]
  
  distinct_mutations <- distinct_mutations %>% 
    filter(MedianDepth>=minumum_median_total & MedianDepthMut>=minumum_median_mutation) %>%
    ungroup()
  
  if (nrow(distinct_mutations)==0){
    num_columns <- sapply(distinct_mutations, is.numeric)
    distinct_mutations[num_columns] <- lapply(distinct_mutations[num_columns], round, 3)
    if (file.exists(file.path( out_dir, "D.RData")))
      file.remove(file.path(out_dir,"D.RData"))
    return(list("distinct_mutations"=distinct_mutations, "g"=NULL))
  }
  
  #rownames(distinct_mutations) = 1:nrow(distinct_mutations)
  
  valid_distinct_mutations = distinct_mutations
  if (!is.null(verified_genes))
    valid_distinct_mutations = subset(distinct_mutations,distinct_mutations$Gene %in% verified_genes) %>% ## mod
    filter(REF!="-") ## mod
  
  if (nrow(valid_distinct_mutations)==0){
    num_columns <- sapply(distinct_mutations, is.numeric)
    distinct_mutations[num_columns] <- lapply(distinct_mutations[num_columns], round, 3)
    if (file.exists(file.path( out_dir, "D.RData")))
      file.remove(file.path(out_dir,"D.RData"))
    return(list("distinct_mutations"=distinct_mutations, "g"=NULL))
  }
  
  
  valid_distinct_mutations_values = NULL
  for(i in 1:nrow(valid_distinct_mutations)) {
    valid_distinct_mutations_values = c(valid_distinct_mutations_values,paste0(valid_distinct_mutations[i,c("Chr","PosStart")],collapse="_"))
  }
  save(valid_distinct_mutations,file=file.path(out_dir,"valid_distinct_mutations.RData"))

  # make final mutations data structures
  #load("valid_clones_mapping.RData")
  #mutations = array(0,c(length(unique(valid_clones_mapping$Run)),6))
  #rownames(mutations) = sort(unique(valid_clones_mapping$Run))
  #colnames(mutations) = paste0(valid_distinct_mutations$Gene,"_",valid_distinct_mutations_values,"_",valid_distinct_mutations$REF,"_",valid_distinct_mutations$ALT)
  mutations = array(0,c(length(unique(colnames(depth))),nrow(valid_distinct_mutations)))
  rownames(mutations) = sort(unique(colnames(depth)))
  colnames(mutations) = paste0(valid_distinct_mutations$Gene,"_",valid_distinct_mutations_values,"_",valid_distinct_mutations$REF,"_",valid_distinct_mutations$ALT)


  
  #mutations = array(0,c(nrow(valid_distinct_mutations), length(unique(o_snpMut_filt_freq$scID))))
  #colnames(mutations) = sort(unique(o_snpMut_filt_freq$scID))
  #rownames(mutations) = valid_distinct_mutations_values
  
  #mutations = mutations[valid_distinct_mutations_values,]
  #browser()
  # set NA values
  depth = t(depth)
  #dim(depth)
  #depth = depth[colnames(mutations), rownames(mutations), drop=FALSE]
  depth = depth[sort(unique(snpMut_filt_freq$scID)), valid_distinct_mutations_values, drop=FALSE]
  
  
  mutations <- inner_join(cells_aggregate_info, valid_distinct_mutations) %>% 
    filter( scID %in% rownames({{depth}}))%>% 
    mutate(idx = paste(Gene,Chr,PosStart,REF,ALT,sep='_')) %>% 
    mutate(val=1) %>%
    distinct() %>%  
    #select(scID,idx,val) %>%
    pivot_wider(id_cols=scID, names_from = idx, values_from=val, values_fill=0) %>%
    full_join(data.frame(scID=rownames({{depth}})))
  
  mutations[is.na(mutations)] <- 0
  mutations=data.frame(mutations,row.names = 1)
  
  
  
  
  col_n <- 
    apply(str_split(colnames(mutations), pattern = "_", simplify = TRUE)[,2:3, drop=FALSE], 1, paste, collapse="_")
  
  depth = depth[rownames(mutations), col_n, drop=FALSE]
  colnames(depth) = colnames(mutations)
  mutations[which(depth<=depth_minimum,arr.ind=TRUE)] = NA # missing values rate equals to 359/2850, that is approx 12.6%
  
  
  cells_aggregate_info = cells_aggregate_info[which(cells_aggregate_info$scID%in%rownames(mutations)), , drop=FALSE]
  mycellsdata = list()
  for ( t in time_points )
    mycellsdata[[t]] = as.matrix.data.frame(mutations[sort(unique(cells_aggregate_info$scID[which(cells_aggregate_info$Time==t)])), , drop=FALSE], rownames.force = TRUE)
  D = mycellsdata
  #browser()
  save(D,file=file.path(out_dir,"D.RData"))

  
  ####
  #browser()
  g <- plot_D(D)
  
  
  
  ####
  
  
  # make Oncoprint
  n_rows=0
  for (t in seq(1, length(time_points)))
    n_rows=n_rows+nrow(D[[t]])
  clusters = array(NA,c(n_rows,1))

  init_i=1
  end_i=0
  r_names= NULL
  data = NULL
  for (t in seq(1, length(time_points))) {
    if(nrow(D[[t]])>0) {
      end_i <- end_i+nrow(D[[t]])
      r_names=c(r_names, rownames(D[[t]]))
      clusters[init_i:end_i,1] = time_points[[t]]
      init_i <- end_i+1
      data <- rbind(data, D[[t]])
    }
  }
  rownames(clusters) <- r_names
  #clusters[1:nrow(D[[1]]),1] = "T1_before_treatment"
  #clusters[(nrow(D[[1]])+1):(nrow(D[[1]])+nrow(D[[2]])),1] = "T2_4_days_treatment"
  #clusters[(nrow(D[[1]])+nrow(D[[2]])+1):(nrow(D[[1]])+nrow(D[[2]])+nrow(D[[3]])),1] = "T3_28_days_treatment"
  #clusters[(nrow(D[[1]])+nrow(D[[2]])+nrow(D[[3]])+1):(nrow(D[[1]])+nrow(D[[2]])+nrow(D[[3]])+nrow(D[[4]])),1] = "T4_57_days_treatment"
  #rownames(clusters) = c(rownames(D[[1]]),rownames(D[[2]]),rownames(D[[3]]),rownames(D[[4]]))
  #data = rbind(D[[1]],D[[2]],D[[3]],D[[4]])
  #data[which(is.na(data))] = 0
  #data = import.genotypes(data)
  # data = annotate.stages(data,clusters)
  ## if (ncol(data$genotypes)>1 || length(unique(data$genotypes[,1]))>1) #errore generico
  ##  oncoprint(data,excl.sort=FALSE,group.by.stage=TRUE)

  #distinct_mutations[['ResultantMeanDepth']]<-apply(depth,2,mean)
  #distinct_mutations[['ResultantVarDepth']]<-apply(depth,2,var)
  #browser()
  num_columns <- sapply(distinct_mutations, is.numeric)
  distinct_mutations[num_columns] <- lapply(distinct_mutations[num_columns], round, 3)
  return(list("distinct_mutations"=distinct_mutations, "g"=g))
}


plot_D <-function(D) {
  D <- D %>% 
    setNames({names(.) %>% 
        str_replace_all(pattern = "_", replacement = " ") %>% 
        data.frame("times"=.) %>% 
        mutate(
          times=ifelse(str_starts(times, pattern = "T[0-9] ", negate = TRUE), 
                       paste0('T', row_number(), " ",times), times)
        )}$times)
  
  #names(D) <- t_order$times
  
  
  D1 <- lapply(D, FUN=function(x){as_tibble(x,rownames = "cell")}) %>%
    bind_rows(.id = "time")  %>%
    mutate(time=forcats::fct_inorder(factor(time, ordered=TRUE))) %>%
    relocate(cell, .after = last_col()) %>% ##
    arrange(across(everything())) %>%
    mutate(ordered=factor(cell, levels = unique(cell, fromLast=TRUE), ordered = TRUE))
  
  
  #D1 <- D1 %>%
  #  mutate(ordered=cell) %>%
  #  arrange(across(everything())) %>%
  #  mutate(ordered=factor(ordered, levels = unique(ordered, fromLast=TRUE), ordered = TRUE))
  
  
  s <- D1 %>% group_by(time) %>% 
    summarise(across(where(is.numeric), ~sum(.,na.rm = TRUE))) %>% 
    pivot_longer(-1) %>% 
    pivot_wider(names_from = 1, values_from = value) #%>% 
    #arrange(across(everything())) not needed
  
  Dw <- D1 %>%
    pivot_longer( col = -c(time, cell,ordered), names_to = "mutation", values_to = "value") %>%
    mutate(value = factor(as.integer(value), levels=c(0,1,"NA"), exclude = NULL, ordered = T)) %>%
    mutate(mutation = factor(mutation, levels=s$name, ordered=TRUE))
  
  
  
  
  per_mut_site <- as.data.frame(Dw) %>% group_by(mutation)
  n_na <- (per_mut_site %>% mutate(cnt=sum(is.na(value))) )$cnt
  n_one <- (per_mut_site %>% mutate(cnt=sum(value==1, na.rm = TRUE)) )$cnt
  n_zero <- (per_mut_site %>% mutate(cnt=sum(value==0, na.rm = TRUE)) )$cnt
  str__na <- paste0("# \"NA\" = [", paste(range(n_na), collapse = ", "),"]")
  str__zero <- paste0("# \"0\" = [", paste(range(n_zero), collapse = ", "),"]")
  str__one <- paste0("# \"1\" = [", paste(range(n_one), collapse = ", "),"]")
  
  
  g <- ggplot(Dw, aes(mutation, y=ordered, fill= value)) +
    geom_tile()+
    
    scale_fill_manual(values = c("black", "white", "orange"), na.value="gray50",
                      aesthetics = c("colour", "fill")#, 
                      #labels = c("passing filter(s)", "filtered out", "NA")
    )+ #orange=2
    
    facet_grid(
      rows = vars(time),
      scales = "free",
      space = "free",
      labeller = label_wrap_gen(width=15)
    ) +
    scale_x_discrete(expand = c(0,0))+
    theme(axis.text.x = element_text(angle=45, hjust=1), ##it is shifted why?
          axis.text.y = element_blank(),
          axis.ticks.y= element_blank()
    )+
    theme(
      strip.background = element_rect(
        color="black", fill="#FC8207", size=1.5, linetype="solid"
      ),
      panel.border = element_rect(color = "#FC8207", fill = NA, size = 2)
    ) +
    theme(legend.key=element_rect(colour="black")) +
    theme(panel.spacing = unit(0, "pt"))+
    labs(y = "cells")+
    labs(x = "selected mutations")+
    #labs(title = paste("Pre-inference input matrix {",str__na, ";", str__one, ";", str__zero,"}"))+
    labs(title = paste("Filtered binary data matrix"))+
    theme(text = element_text(size = 14))  
  
  return(g)
}

## version 1118

### LACEview2.R
###
### Add lace_interface function

#' @title LACE Interface
#' @description This function generates a longitudinal clonal tree and a
#' graphic interface to explore the data using as input the clonal tree
#' formatted in the same way as the one produced by LACE during the imputation
#' steps
#'
#' @param B_mat (Required). B is the clonal tree matrix where columns are the
#' clonal mutations and and rows are the clones. The clonal tree matrix should
#' contain a column and a row named "Root" representing the root of the tree and
#' the wild type, respectively.
#' B is a binary matrix where 1 are the mutations associated to the clones.
#' The wild type column has all ones
#'
#' @param clones_prevalence (Required) The clonal prevalence matrix
#'
#' @param C_mat (Required) The corrected clonal attachment
#'
#' @param error_rates (Required) The false positive alpha and false negative beta
#' error rates used to infer the clonal tree
#'
#' @param info (Optional). HTML formatted text with information regarding
#' the experiments
#'
#' @param width (optional) Size of the window interafce
#'
#' @param height (optional) Size of the window interafce
#'
#' @param elementId (optional) Element id
#'
#' @return  An implementation of the htmlwidgets
#'
#' @importFrom  Matrix rankMatrix
#' @importFrom jsonlite toJSON
#' @importFrom grDevices colorRampPalette
#' @importFrom purrr is_null
#' @importFrom tidyr gather
#' @importFrom purrr detect_index
#' @import stringr
#' @import stringi
#' @import RColorBrewer
#'
#'
#'
# #' @export


lace_interface <- function (B_mat,
                            clones_prevalence,
                            C_mat,
                            error_rates,
                            width = NULL,
                            height = NULL,
                            elementId = NULL,
                            info = '') {

    ## library(purrr, quietly = T)
    ## library(stringr, quietly = T)
    ## library(stringi, quietly = T)
    ## library(Matrix, quietly = T)

    if (!exists('width')) {width=NULL}
    if (!exists('height')) {height=NULL}
    if (!exists('elementId')) {elementId=NULL}
    if (!exists('info')) {info=''}
    
    if (is_null(height)) {
        height=900
        height_fishplot=300
    } else if (height<800) {
        height_fishplot=300
    } else {
      height_fishplot=300
    }
  
    if (is_null(width)) {
        width=900
        width_fishplot=width/2
    } else if (width<900) {
        width_fishplot=width/2
    } else {
      width_fishplot=width/2
    }
    if (is_null(info)) {
        info = 'No dataset or experimental info available'
    }


    B = B_mat
    B


    make_binary<-function(B) {
        if (!is.numeric(B)){
          log2_print('B is not a valid matrix')
            return(NULL)
        }

        is_binary=sum(B==1)+sum(B==0)==ncol(B)*nrow(B)
        if (!is_binary) {
          log2_print('B is not binary')
            B[B==1]=1
            B[B==0]=0
        }
        else {
          log2_print('B is already binary')
        }
        return(B)
    }


    find_roots <- function(B) {
                                        #find roots
        set_of_roots = rownames(B)
        set_of_roots
        for (i in set_of_roots) {
            tmp_set_of_roots=setdiff(set_of_roots,i)
            for (j in tmp_set_of_roots) {
                proj=sum(B[i,]*B[j,])
                if (sum(B[i,])==proj) {
                    set_of_roots=setdiff(set_of_roots,j)
                    log2_print(paste(i,'is parent of', j))
                                        #print(set_of_roots)
                }
                if (sum(B[j,])==proj) {
                    set_of_roots=setdiff(set_of_roots,i)
                    log2_print(paste(j,'is parent of', i))
                                        #print(set_of_roots)
                }
            }
        }
        set_of_roots
        return(set_of_roots)
    }

    forest_check <- function(B) {
                                        #check if it is a forest
        check_forest=matrix(data = 0, nrow=nrow(B),ncol=1)
        check_forest
        check_forest_control=matrix(data = 1, nrow=nrow(B),ncol=1)
        check_forest_control[1,1]=0

        for(i in 1:nrow(B)) {
            check_forest_tmp_=0
            for(j in i:nrow(B)) {
                check_forest_tmp=(sum(abs(B[j,]-B[i,]))==1)
                check_forest[j,1]=check_forest[j,1]+check_forest_tmp
                                        #check_forest_tmp_ = check_forest_tmp_+ check_forest_tmp
                log2_print(paste(rownames(B)[i], rownames(B)[j], check_forest[j,1]))
            }

                                        #check_forest=rbind(check_forest, check_forest_tmp_)

        }
        log2_print(check_forest)
        log2_print(check_forest_control)
        log2_print(all(check_forest==check_forest_control))
        is_forest=all(check_forest==check_forest_control)
        if (is_forest) {
          log2_print("B represents a forest")
        } else {
          log2_print("B is not a forest or a tree")
        }
        return(is_forest)
    }

    add_unique_root <- function(B, root_name='Root') {
                                        #add unique root if necessary

        B_rank=rankMatrix(B)[1]

        set_of_roots = find_roots(B)

                                        #check if it has one root
        num_roots=length(set_of_roots)
        if (num_roots==1) {
          log2_print("nodes in B are sprouting from a single node")
        }
        else {
          log2_print("there are multiple roots")
        }

        if (B_rank==nrow(B) || B_rank==ncol(B)) {
          log2_print('B is a full-rank matrix.')
            if (forest_check(B)) {
              log2_print('B has a forest like structure')

                if (nrow(B)>ncol(B)) {
                  log2_print('B is not a square matrix. B has not enough mutations.')
                                        #print('B is not a full rank matrix. B has not enough mutations.')
                    if (nrow(B)==ncol(B)+1) {
                                        #check if there is a root
                        if (num_roots!=1  && root_col_name!=root_name) {
                            root_col_name=root_name
                            B = cbind(xxx=1,B)
                            colnames(B)[colnames(B) == 'xxx'] <- root_name
                        }
                    }
                } else if (nrow(B)<ncol(B)) {
                  log2_print('B is not a square matrix. B has not enough clones.')
                    if (nrow(B)+1==ncol(B)) {
                        lower_tri_column_index=sort(colSums(B),decreasing = TRUE)
                                        #number of clones with such mutation
                        clones_per_mutation=table(lower_tri_column_index)
                        max_clones_per_mutation_idx=(
                            max(as.integer(names(clones_per_mutation))))
                        max_clones_per_mutation_idx

                        if (max_clones_per_mutation_idx<B_rank && root_row_name!=root_name) {
                            root_row_name='Root'
                            B = rbind(xxx=0,B)
                            rownames(B)[rownames(B) == 'xxx'] <- root_name
                            B[1,1]=1
                        }
                    }
                }
            }
        }

        B_rank=rankMatrix(B)[1]
        B_rank
        if (!(B_rank==ncol(B) && B_rank==nrow(B))) {
          log2_print('Adding a root does not resolve the problem, B is till not a full rank matrix')
        }

        return(B)
    }

    make_lowtri<-function(B) {
                                        #make B lower triangular
        B_tmp=matrix(nrow=0, ncol=ncol(B))
        colnames(B_tmp)=colnames(B)
        B_tmp
        B2=B #B_row_reduced
        B3=B2 #B_row_reduced_col_ordered
        B4=B3 #B_row_col_reduced_col_ordered
        set_of_roots = rownames(B2)

        for(i in set_of_roots) {
            i=nrow(B)-length(set_of_roots)+1
            i
            first_r=names(which.min(rowSums(B4)))
            first_r

            B_tmp=rbind(B_tmp, B3[first_r, ,drop=FALSE])
            B_tmp
            if (i==nrow(B)){
                break
            }

            if (i>1) {
                one_col=B_tmp[first_r,-1:-i+1]==1
                zero_col=B_tmp[first_r,-1:-i+1]==0
            } else {
                one_col=B_tmp[first_r,]==1
                zero_col=B_tmp[first_r,]==0
            }
            one_col=names(one_col[one_col])
            one_col
            zero_col=names(zero_col[zero_col])
            zero_col
            B2=B2[rownames(B2) !=first_r, , drop=FALSE]
            B2
            B3=cbind(B2[,colnames(B_tmp[,1:i-1,drop=FALSE]),drop=FALSE],
                     B2[,one_col,drop=FALSE],B2[,zero_col,drop=FALSE])
            B3
            B4=cbind(B2[,one_col,drop=FALSE],B2[,zero_col,drop=FALSE])
            B4
                                        #B2=B2[,-1]
            B_tmp=cbind(B_tmp[,1:i-1,drop=FALSE], B_tmp[,one_col,drop=FALSE],
                        B_tmp[,zero_col,drop=FALSE])
            B_tmp
            set_of_roots=setdiff(set_of_roots,first_r)
        }

        B=B_tmp
        return(B)
    }

#####
    B=make_binary(B)
                                        #B_rank=rankMatrix(B)[1]
                                        #B_rank

                                        #is_forest = forest_check(B)
                                        #is_forest

    lower_tri_column_index=sort(colSums(B),decreasing = TRUE)
    root_col_name=names(which.max(lower_tri_column_index))
    root_row_name=names(which.min(sort(rowSums(B),decreasing = TRUE)))

    lower_tri_column_index
    root_col_name
    root_row_name

    B<-add_unique_root(B)
    B<-make_lowtri(B)

    lower_tri_column_index=sort(colSums(B),decreasing = TRUE)
    root_col_name=names(which.max(lower_tri_column_index))
    root_row_name=names(which.min(sort(rowSums(B),decreasing = TRUE)))

    lower_tri_column_index
    root_col_name
    root_row_name



                                        #check all
    is_binary = sum(B==1)+sum(B==0)==ncol(B)*nrow(B)
    is_square_mat = ncol(B)==nrow(B)
    is_full_rank = rankMatrix(B)[1]==nrow(B)
    is_forest = forest_check(B)
    has_single_root = length(find_roots(B))==1
    if (is_binary) {log2_print('B is a binary matrix')}
    if (is_square_mat) {log2_print('B is a square matrix')}
    if (is_full_rank) {log2_print('B is a full rank matrix')}
    if (is_forest) {log2_print('B is a forest ')}
    #browser()
    if (has_single_root) {log2_print('B has single root')}
    if (is_binary && is_square_mat && is_full_rank && is_forest && has_single_root) {
      log2_print('all checks done!')
      log2_print('continue...')
    }


                                        #this if else is not necessary in this form. Instead, it should set root names
    if ("Root" %in% rownames(clones_prevalence)){
        prevalence=clones_prevalence
    } else {
        prevalence=rbind(Root=0,clones_prevalence)
    }
    prevalence=prevalence[rownames(B),]

    clone_labels=as.list(str_split(colnames(B), pattern='_', n=2, simplify = TRUE)[,1])
    ##clone_labels=str_split(colnames(B), pattern='_', n=2, simplify = T)[,1]
    names(clone_labels)<-rownames(B)
    clone_labels

                                        #library("RColorBrewer")
    #colours = as.list(brewer.pal(n = nrow(prevalence), name = "Paired"))
    colours =  as.list(colorRampPalette(brewer.pal(12, "Paired"))(nrow(prevalence)))
    names(colours)<-rownames(B)
    colours






                                        #create adjacency matrix
    adj_matrix = array(0L, c((dim(B)[1]), (dim(B)[2])))
    rownames(adj_matrix) <- colnames(B)[(1:ncol(B))]
    colnames(adj_matrix) <- colnames(B)[(1:ncol(B))]
    for (rP in seq(1, nrow(B), by=1) ) {
        for (rC in seq((rP + 1), nrow(B), length = max(0,nrow(B)-(rP + 1)+1))) {
            if (all(B[rP, 1:rP] == B[rC, 1:rP]) && sum(B[rP, ]) == (sum(B[rC, ]) - 1)) {
                adj_matrix[(rP ), (rC )] <- 1
            }
        }
    }

    log2_print('clonal tree:')
    log2_print(adj_matrix)

    adj_to_B <- function( adj_matrix, clone_labels ) {
        B=matrix(0, nrow=nrow(adj_matrix), ncol=ncol(adj_matrix))
        colnames(B)=colnames(adj_matrix)
        rownames(B)=rownames(adj_matrix)
        B[1,1]=1
        for (s in rownames(adj_matrix)) {
            for (t in colnames(adj_matrix)) {
                if (adj_matrix[s,t]==1) {
                    B[t,]=B[s,]
                    B[t,t]=1
                }
            }
        }
        rownames(B)=names(clone_labels)
        return(B)
    }
    log2_print(adj_to_B(adj_matrix, clone_labels))


                                        #adjMatrix_base <- adj_matrix

####

                                        #it resolves all the incongruences resulting from not considering the time at all
    chrono_occurences <- function(adj_matrix, prevalence) {
                                        #create first time mutations appearance based on clonal prevalence only
        times= apply((prevalence)>0, 1, function(x) detect_index(x, ~ .x == TRUE) )
                                        #times=times+1 #test
        times[1]=1 #setting root

        times_pre<-times

        adj_matrix_t_pre<-adj_matrix
        colnames(adj_matrix_t_pre) <-paste0(rownames(prevalence), '_t', times)
        rownames(adj_matrix_t_pre) <-paste0(rownames(prevalence), '_t', times)


                                        #create first time mutations appearance based on chronological order
        sources=apply( t(adj_matrix)>0, 1, function(x) detect_index(x, ~ .x == TRUE) )
        sources[1]=1
        for(target in 1: length(rownames(prevalence))) {
            source=sources[target]
            if (times[target] < times[source]) {
              log2_print(times[target],times[source])
                times[source]=times[target]
            }
        }

        adj_matrix_t<-adj_matrix
        colnames(adj_matrix_t) <-paste0(rownames(prevalence), '_t', times)
        rownames(adj_matrix_t) <-paste0(rownames(prevalence), '_t', times)



        if (all(times==times_pre)) {
          log2_print('No incongruences found due to the chronological order')

          log2_print('first time mutation occurrences based the chronological order of samples:')
          log2_print(times)
          log2_print('adjacent matrix with first time occurrences:')
          log2_print(adj_matrix_t)

        } else {
          log2_print('The longitudinal clonal tree has been corrected')

          log2_print('pre-computation of first time mutation occurrences based only on prevalence values different from zero:')
          log2_print(times_pre)
          log2_print('adjacent matrix with pre-computed first time occurrences:')
          log2_print(adj_matrix_t_pre)

          log2_print('first time mutation occurrences based the chronological order of samples:')
          log2_print(times)
          log2_print('adjacent matrix with first time occurrences:')
          log2_print(adj_matrix_t)
        }

        return(times)
    }

    times <- chrono_occurences(adj_matrix,prevalence)

    adj_matrix_t<-adj_matrix
    colnames(adj_matrix_t) <-paste0(rownames(prevalence), '_t', times)
    rownames(adj_matrix_t) <-paste0(rownames(prevalence), '_t', times)




    ## adding the other elements of the direct product
    times_n= 1:(length(colnames(prevalence))-1)
    times_n
    for (t in times_n) {
        for (col in 1:ncol(adj_matrix)) {
            col_ = paste0(rownames(prevalence)[col],'_t',t)

            if (!(col_ %in% colnames(adj_matrix_t))) {
                adj_matrix_t=cbind(adj_matrix_t, xxx=0)
                colnames(adj_matrix_t)[colnames(adj_matrix_t) == 'xxx'] <- col_
            }
        }
        for (row in 1: nrow(adj_matrix)) {
            row_ = paste0(rownames(prevalence)[row],'_t',t)

            if (!(row_ %in% rownames(adj_matrix_t))) {
                adj_matrix_t=rbind(adj_matrix_t, xxx=0)
                rownames(adj_matrix_t)[rownames(adj_matrix_t) == 'xxx'] <- row_
            }
        }
    }

    ## Root as wild type should be considered present in all the next time steps:
    ## hence an accessory self link is used
    adj_matrix_t['Root_t1','Root_t1']=1

    ## setting links of clones in following times after first occurrence
  for (t_1 in times_n[1:(length(times_n)-1)]) {
    log2_print(c('time=',t_1))
    for (source in 1:nrow(adj_matrix)) {
      for (target in 1:ncol(adj_matrix)) {
        for (t_2 in times_n[t_1:(length(times_n)-1)]) { # r fa cacare
          #source mut_time
          source_t1 = paste0(rownames(prevalence)[source],'_t',t_1)
          #target mut_time
          target_t2 = paste0(rownames(prevalence)[target],'_t',t_2)
          #print(c(source_t1, target_t2, adj_matrix_t[source_t1,target_t2]))
          #source mut_time
          source_t2 = paste0(rownames(prevalence)[target],'_t',t_2)
          #target mut_time+1
          target_t2_next=paste0(rownames(prevalence)[target],'_t',t_2+1)

          if (adj_matrix_t[source_t1,target_t2]>=1 && adj_matrix_t[source_t1,target_t2]<=3) {

            if (target==source && prevalence[target,t_2]==0){
              value=3 #for dashed in case undetected clones

              log2_print(paste0(source_t1, "->", target_t2,"=", adj_matrix_t[source_t1,target_t2], ", prev_",target_t2, "=", prevalence[target,t_2], ": ", source_t1, "->", target_t2, "=", value))
              adj_matrix_t[source_t1, target_t2]<-value # we set the link before t2
              log2_print(paste0(source_t1, "->", target_t2,"=", adj_matrix_t[source_t1,target_t2], ", prev_",target_t2, "=", prevalence[target,t_2], ": ", source_t2, "->", target_t2_next, "=", value))
              adj_matrix_t[source_t2, target_t2_next]<-value # we set the link after t2

            }
            else {
              value=2 #for time persistence

              log2_print(paste0("1st ",source_t1, "->", target_t2,"=", adj_matrix_t[source_t1,target_t2], ", prev_",target_t2, "=", prevalence[target,t_2], ": ", source_t2, "->", target_t2_next, "=", value))
              adj_matrix_t[source_t2,target_t2_next]<-value
            }

            if (target==source && t_2+1==length(times_n) && prevalence[target,t_2+1]==0){
              value=3 #for dashed in case undetected clones
              log2_print(paste0(source_t1, "->", target_t2,"=", adj_matrix_t[source_t1,target_t2], ", prev_",target_t2, "=", prevalence[target,t_2], ": ", source_t2, "->", target_t2_next, "=", value))
              adj_matrix_t[source_t2, target_t2_next]<-value # we set the link after t2
            }
          }
          #if (target==source && t_1+1==t_2 && prevalence[target,t_2]==0) {
            #target_t1 = paste0(rownames(prevalence)[target],'_t',t_1) #t1=t2-1
            # = paste0(rownames(prevalence)[source],'_t',t_2)
            #target_t2 = paste0(rownames(prevalence)[target],'_t',t_2)
            #target mut_time+1
            #target_t2_next=paste0(rownames(prevalence)[target],'_t',t_2+1)

            #value=3 #for dashed in case undetected clones

            # print(c(source_t1, target_t2, source_t2, target_t2_next, value))
            # print(c(source_t1, target_t2, source_t1, target_t2, value))
            # adj_matrix_t[source_t2, target_t2_next]<-value # we set the link after t2
            # adj_matrix_t[source_t1, target_t2]<-value # we set the link before t2
          #}
        }
      }
    }
  }

                                        #dead breaches
                                        #in time reversal
                                        #prevalence=prevalence
    for (t_1 in times_n[(length(times_n)):1]) { # instead of 2 as smaller point
      log2_print(c('time=',t_1))
        for (source in 1:nrow(adj_matrix)) {
            target=source # correlation only at distance 0

            source_t1_prev =paste0(rownames(prevalence)[source],'_t',t_1-1)
            target_t1 = paste0(rownames(prevalence)[target],'_t',t_1)

            source_t1 =paste0(rownames(prevalence)[source],'_t',t_1)
            target_t1_next = paste0(rownames(prevalence)[target],'_t',t_1+1)

            if (t_1==times_n[(length(times_n))]  && prevalence[target,t_1]==0 ){ #&& prevalence[source,t_1-1]==0) {
                value=0 #for dead breaches
                log2_print(c(source_t1, target_t1, source_t1_prev, target_t1, value))
                adj_matrix_t[source_t1_prev,target_t1]<-value # we set the link after t2

                                        #prevalence[prevalence[,t_1]==0,t_1]=-1
                prevalence[target,t_1]=-1

            } else if (prevalence[source,t_1]==0  && adj_matrix_t[source_t1,target_t1_next]==0) {
                                        #target mut_time+1
                value=0 #for dead breaches
                log2_print(c('removed', target_t1))
                log2_print(c(source_t1, target_t1, source_t1, target_t1_next, value))
                adj_matrix_t[source_t1,target_t1_next]<-value # we set the link after t2

                if (t_1!=1){
                  log2_print(c(source_t1, target_t1, source_t1_prev, target_t1, value))
                    adj_matrix_t[source_t1_prev,target_t1]<-value # we set the link after t2
                }
                                        #prevalence[prevalence[,t_1]==0,t_1]=-1
                if (!(stri_detect_fixed(target_t1,"Root") && t_1==1)){
                    prevalence[target,t_1]=-1
                }
            }
        }
    }
    ## removal of the accessory self link
    adj_matrix_t['Root_t1','Root_t1']=0


    colnames(prevalence)

    elements = list()
    elements$nodes = list()
    elements$edges = list()

    for (row in 1:nrow(adj_matrix_t)) {
        for (col in 1:ncol(adj_matrix_t)) {
            if (adj_matrix_t[row, col] == 1) {
                data = list()
                data$data = list()

                clone_=stri_reverse(str_split(
                    stri_reverse(colnames(adj_matrix_t)[col]),pattern='_', n=2)[[1]][2])
                clone_idx=match(clone_,rownames(B))
                log2_print(c('idx',clone_))

                data$data$source = sprintf("%s", rownames(adj_matrix_t)[row])
                data$data$target = sprintf("%s", colnames(adj_matrix_t)[col])
                data$data$name = sprintf("%s", clone_labels[clone_][[1]])

                                        #data$data$color = colours[clone_idx]
                data$data$color = colours[clone_][[1]]

                data$data$id = sprintf("%s_%s", data$data$source, data$data$target)
                data$data$linestyle = "solid"
                elements$edges = c(elements$edges, list(data))
            }
            if (adj_matrix_t[row, col] == -200) {  #if (adj_matrix_t[row, col] == 2) {
                data = list()
                data$data = list()

                clone_=stri_reverse(str_split(stri_reverse(
                    colnames(adj_matrix_t)[col]),pattern='_', n=2)[[1]][2])
                                        #color_idx=match(clone_,rownames(B))

                data$data$source = sprintf("%s", rownames(adj_matrix_t)[row])
                data$data$target = sprintf("%s", colnames(adj_matrix_t)[col])
                data$data$name = sprintf("")

                data$data$color = colours[clone_][[1]]

                data$data$id = sprintf("%s_%s", data$data$source, data$data$target)
                data$data$linestyle = "solid" # only if next clonal_prev==0 else solid
                elements$edges = c(elements$edges, list(data))
            }
            if (adj_matrix_t[row, col] >= 2) {  #if (adj_matrix_t[row, col] == 3) {
                data = list()
                data$data = list()

                clone_=stri_reverse(str_split(stri_reverse(
                    colnames(adj_matrix_t)[col]),pattern='_', n=2)[[1]][2])
                                        #color_idx=match(clone_,rownames(B))

                data$data$source = sprintf("%s", rownames(adj_matrix_t)[row])
                data$data$target = sprintf("%s", colnames(adj_matrix_t)[col])
                data$data$name = sprintf("")

                data$data$color = colours[clone_][[1]]

                data$data$id = sprintf("%s_%s", data$data$source, data$data$target)
                data$data$linestyle = "dashed" # only if next clonal_prev==0 else solid
                elements$edges = c(elements$edges, list(data))
            }
        }
    }

                                        #nodes
    for (col in 1:(ncol(prevalence) - 1)) {
        data = list()
        data$data = list()
        data$data$id = sprintf("T%s", col)
        data$data$name = sprintf("T%s", col) #experiment2...
        data$data$color = "#FFFFFF"
        data$data$size = sprintf("%spx", 50)
        data$data$prev = sprintf("%s [%s]", data$data$name, round(prevalence[1, col], 2))
        parent = data$data$name
        elements$nodes = c(elements$nodes, list(data))
        for (row in 1:nrow(prevalence)) {
            if (prevalence[row,col]!=-1) {
                data = list()
                data$data = list()
                data$data$name = rownames(prevalence)[row]
                data$data$color = colours[row][[1]]
                data$data$id = sprintf("%s_t%s", rownames(prevalence)[row], col)

                data$data$size = sprintf("%spx", 5 + as.integer(prevalence[row, col] * 150))
                data$data$prev = sprintf("%s", round(prevalence[row, col], 2))
                data$data$parent = parent
                elements$nodes = c(elements$nodes, list(data))
            }
        }
    }


                                        #-------------------
                                        #library(tidyr)

    if ("Root" %in% rownames(clones_prevalence)){
        prevalence=clones_prevalence
    } else {
        prevalence=rbind(Root=0,clones_prevalence)
    }
    prevalence=prevalence[rownames(B),]

    fishplot_clonal_prev=as.data.frame(prevalence)
    fishplot_clonal_prev = fishplot_clonal_prev[,!(names(fishplot_clonal_prev) %in% c("Total"))]
    fishplot_clonal_prev=subset(fishplot_clonal_prev)
    colnames(fishplot_clonal_prev)=c(paste0('T',times_n))
    fishplot_clonal_prev['clone_id']=rownames(fishplot_clonal_prev)
    fishplot_clonal_prev


    fishplot_clonal_prev=gather(fishplot_clonal_prev,
                                'timepoint','clonal_prev', -'clone_id' )
    fishplot_clonal_prev=fishplot_clonal_prev[c('timepoint',
                                                'clone_id','clonal_prev')]
    fishplot_clonal_prev['clonal_prev']=fishplot_clonal_prev['clonal_prev']+0.0000001

    fishplot_clonal_prev


    fishplot_tree_edges=as.data.frame(adj_matrix)
                                        #fishplot_tree_edges['source']=rownames(fishplot_tree_edges)
    fishplot_tree_edges['source']=rownames(prevalence)
    colnames(fishplot_tree_edges)=c(fishplot_tree_edges[['source']],'source')
    fishplot_tree_edges=gather(fishplot_tree_edges, 'target','value', -source)
    fishplot_tree_edges=fishplot_tree_edges[fishplot_tree_edges['value']==1,][c('source', 'target')]

    fishplot_tree_edges

    fishplot_colour=t(as.data.frame(colours))
    fishplot_colour=data.frame(clone_id=rownames(fishplot_colour), colour=fishplot_colour[,1])


    log2_print(getwd())
                                        #source(timescape.R)
                                        #source(paste0(basedir,"/peter/LACEView/R/timescape.R"))
    x=timescape(clonal_prev = fishplot_clonal_prev,
                tree_edges = fishplot_tree_edges,  height=height_fishplot,
                width=width_fishplot, alpha=0, genotype_position="stack",
                clone_colours = fishplot_colour)

    tabula=data.frame(T=c(paste0('T',times_n)), alpha=error_rates$alpha,
                      beta=error_rates$beta,
                      m_tilde=length(rownames(prevalence))-1,
                      n=unlist(lapply(C_mat,length)))
    rownames(tabula)=c(paste0('T',times_n))
    tabula
    headings=c("<p style=' font-size: 14px'>T</p>",
               "<p style=' font-size: 14px'>alpha</p>
               <p style='font-weight:normal; margin-top: -20px; margin-bottom: -20px; font-size: 12px'>est.</p>",
               "<p style=' font-size: 14px'>beta</p>
               <p style='font-weight:normal; margin-top: -20px; margin-bottom: -20px; font-size: 12px'>est.</p>",
               paste("<p style=' font-size: 14px'>m_tilde</p>",
               #<p style='font-weight:normal; margin-top: -20px; margin-bottom: -20px; font-size: 12px'>n candidate drivers</p>",
               "<p style='font-weight:normal; margin-top: -20px; margin-bottom: -20px; font-size: 12px'>n drivers</p>"),
               "<p style=' font-size: 14px'>n</p>
               <p style='font-weight:normal; margin-top: -20px; margin-bottom: -20px; font-size: 12px'>single cells</p>"
               )
    tab={}
    tab$headings=headings
    tab$data=t(t(tabula))

                                        #library(jsonlite)
    toJSON(tab)
                                        #-------------------



    clonalprev_df <- prevalence
    clonalprev_df = clonalprev_df[,!(colnames(clonalprev_df) %in% c("Total"))]
    clonalprev_df <- subset(clonalprev_df)
    clonalprev_df_names <- rownames(clonalprev_df)
    stream_df <- data.frame(Time = c(1:ncol(clonalprev_df)))
    for (k in clonalprev_df_names) {
        temp <- data.frame(k = clonalprev_df[k, ], stringsAsFactors = FALSE)
        stream_df <- cbind(stream_df, temp)
    }
    names(stream_df) <- c("Time", clonalprev_df_names)
    stream_df_colors = list()
    for (K in 1:length(colours)) {
        stream_df_colors[[clonalprev_df_names[K]]] = colours[K]
    }



    log2_print("info")
    log2_print(info)
    jsdata = list(elements = elements, data = toJSON(stream_df),
                  columns = toJSON(names(stream_df)),
                  colors = toJSON(stream_df_colors),
                  clone_labels = toJSON(unlist(clone_labels)[-1]),
                  x=x,
                  info=toJSON(list(info=info)),
                  tab=toJSON(tab))

    results<-list()
    results[["prevalence"]] <- prevalence
    results[["colours"]] <- colours
    results[["html"]] <- htmlwidgets::createWidget(name = "LACE", jsdata, width = width,
                              height = height, package = "LACE",
                              elementId = elementId,
                              sizingPolicy = htmlwidgets::sizingPolicy(
                                                              viewer.suppress = TRUE,
                                                              knitr.figure = FALSE,
                                                              browser.fill = TRUE,
                                                              browser.padding = 0,
                                                              knitr.defaultWidth = 300,
                                                              knitr.defaultHeight = 600)
                              )
    results
}


####################


####################
#add modified LACEview_html function
####################
LACE_html<-function (id, style, class, ...) {
    htmltools::tags$div(id = id, class = class, style = style,
                        htmltools::div(id = "wrapper",
                                       style = "height:100%; width:100%;",
                                       htmltools::div(id = "navbar"),
                                       htmltools::div(id = "container",
                                                      style = "flex-flow: row wrap; display: flex; height:800px; width:100%;",
                                                      htmltools::div(
                                                                     style = "flex:0 0 48%; align-items: center; height:64%; margin: 5px; padding:5px; border: 1px solid black;",
                                                                     id = "cy_",
                                                                     htmltools::div(
                                                                                    style = "flex:100%; align-items: center; height:5%;",
                                                                                    id = "cy_title",
                                                                                    htmltools::p(
                                                                                                   style = "font-size:20px; text-align: center; font-weight: bold;", "Longitudinal clonal tree"
                                                                                               ),
                                                                                    ),
                                                                     htmltools::div(
                                                                                    style = "flex:100%; align-items: center; height:95%;",
                                                                                    id = "cy"
                                                                                ),
                                                                     ),
                                                      htmltools::div(
                                                                     style = "flex: 0 0 48%; align-items: center; height:64%; margin: 5px; padding:5px; border: 1px solid black;",
                                                                     id = "streamgraph_",
                                                                     htmltools::div(
                                                                                    style = "flex:100%; align-items: center; height:5%;",
                                                                                    id = "streamgraph_title",
                                                                                    htmltools::p(
                                                                                                   style = "font-size:20px; text-align: center; font-weight: bold;", "Fishplot"
                                                                                               ),
                                                                                    ),
                                                                     htmltools::div(
                                                                                    style = "flex:100%; align-items: center; height:95%; font-size:20px; font-weight: normal;",
                                                                                    id = "streamgraph"
                                                                                ),
                                                                     ),
                                                      htmltools::div(
                                                        style = "flex:100%; height:8%; display:flex; align-items: center; justify-content:flex-start; margin-left: 40px;",
                                                        id = "mutations"
                                                      ),
                                                      htmltools::div(
                                                                     style = "flex:100%; height:35%; display:flex; align-items: center; justify-content:center;",
                                                                     id = "lacetable",
                                                                     htmltools::div(
                                                                                    style = "flex:50%; height:100%; display:flex; align-items: flex-end; justify-content:flex-start; font-size:14px;",
                                                                                    id = "legend"
                                                                                ),
                                                                     htmltools::div(
                                                                                    style = "flex:50%; height:100%; display:flex; align-items: center; justify-content:center; font-size:14px;",
                                                                                    id = "table",
                                                                                    class='table'
                                                                                ),
                                                                     htmltools::div(
                                                                                    style = "flex:100%; height:100%; display:none; align-items: flex-end; justify-content:center;",
                                                                                    id = "other"
                                                                                )
                                                                 )
                                                      )
                                   )
                        )
}
####################



## Shiny bindings for LACE
##
## Output and render functions for using LACEview within Shiny
## applications and interactive Rmd documents.
# #' @export
LACEOutput <- function(outputId, width = '100%', height = '400px') {
    htmlwidgets::shinyWidgetOutput(outputId,
                                   'LACE',
                                   width,
                                   height,
                                   package = 'LACE')
}


# #' @rdname LACE-shiny
# #' @export
renderLACE <- function(expr, env = parent.frame(), quoted = FALSE) {
    if (!quoted) { expr <- substitute(expr) } # force quoted
    htmlwidgets::shinyRenderWidget(expr, LACEOutput, env, quoted = TRUE)
}


### end of file -- LACEview2.R

#################################################
##########  NA  ###########
#################################################
NA_compute <- function(depth_minimum, missing_values_max, data_dir, depth_dir, out_dir, time_points) {
  # set working directory
  browser()
  #library(stringr)

  #depth_minimum = 3 # minimum depth to set values to NA
  #missing_values_max = 0.40 # maximum number of considered missing data per gene



  dir.create(out_dir, showWarnings = FALSE)
  if (!dir.exists(out_dir))
    showNotification(paste("Filtered data output folder,", out_dir, ", cannot be created"), duration = 10, type = "warning")

  # load data
  snpMut_filt_freq <- readRDS(file=paste0(file.path(data_dir, "snpMut_filt_freq.rds")))
  mutations = as.matrix(read.table(file.path(depth_dir, "final_data_mut.txt")))
  depth = as.matrix(read.table(file.path(depth_dir, "final_data_depth.txt")))

  # set NA values
  check_ids <- snpMut_filt_freq$scID %in% colnames(mutations)
  error_ids <- sort(unique(snpMut_filt_freq$scID[!check_ids]))
  snpMut_filt_freq  <- snpMut_filt_freq[check_ids,]


  depth = depth[,colnames(mutations)]
  mutations[which(mutations>1,arr.ind=TRUE)] = 1
  mutations[which(depth<=depth_minimum,arr.ind=TRUE)] = NA

  mycellsdata = list()
  #time_points=c("before treatment", "4d on treatment", "28d on treatment", "57d on treatment")
  check_time_points <- time_points %in% unique(snpMut_filt_freq$Time)

  time_points <- time_points[check_time_points]


  for (i in time_points) {
    mycellsdata[[i]] = t(mutations[,sort(unique(snpMut_filt_freq$scID[which(snpMut_filt_freq$Time==i)]))])
  }

  # evaluate number of missing data per time point
  time_points_NA = list()
  #for (t in time_points)
  #  time_points_NA[[t]] = NULL
  #
  ##browser()
  #for (t in time_points)
  #  for(i in 1:ncol(mycellsdata[[1]]))
  #    time_points_NA[[t]] = c(time_points_NA[[t]],length(which(is.na(mycellsdata[[t]][,i])))/nrow(mycellsdata[[t]]))

  time_points_NA = list()
  for (t in time_points) 
    time_points_NA[[t]] <- colSums(is.na(mycellsdata[[t]]))/nrow(mycellsdata[[t]])
  
  ##valid_genes = sort(unique(colnames(mycellsdata[[1]])[which(all(time_points_NA<=missing_values_max))]))
  #valid_genes = colSums(do.call(rbind, lapply(time_points_NA, function(x){x<=missing_values_max})))
  
  #valid_genes <- colnames(mycellsdata[[1]])[apply(do.call(rbind, lapply(time_points_NA, function(x){x<=missing_values_max})), FUN = all, MARGIN = 2)]
  # browser()
  valid_genes <- colnames(mycellsdata[[1]])[apply(bind_rows(lapply(time_points_NA, function(x){x<=missing_values_max})), FUN = all, MARGIN = 2)]
  #valid_genes <- colnames(mycellsdata[[1]])[(data.frame(bind_rows(time_points_NA)<=missing_values_max) %>% rowwise() %>% mutate(y=all(c_across(everything())) ))$y]
  valid = NULL

  
  #for(i in valid_genes) {
  #  valid = c(valid,unique(snpMut_filt_freq$Gene[grep(i,snpMut_filt_freq$UIDsnp)]))
  #}
  
  #valid = paste0(valid,"_",valid_genes)
  valid = unique(grep(paste0(valid_genes, collapse = "|"), unique(snpMut_filt_freq$UIDsnp), value = T ))

  # get names of valid genes
  valid_genes_names = NULL
  #for(i in valid) {
  #  valid_genes_names = c(valid_genes_names,strsplit(i,split="_")[[1]][[1]])
  #}
  #browser()
  valid_genes_names <- str_split_fixed(valid,"_", n=2 )[,1]

  log2_print(length(valid_genes_names), msg = "NA_compute: # valid genes")
  log2_print(length(unique(valid_genes_names)))
  log2_print(sort(valid_genes_names[which(duplicated(valid_genes_names))]))

  # make final list of candidate selected variants
  #snpMut_filt_freq_reduced = unique(snpMut_filt_freq[which(snpMut_filt_freq$Gene%in%valid_genes_names),c("Gene","Chr","PosStart","PosEnd","REF","ALT")])
  snpMut_filt_freq_reduced = unique(snpMut_filt_freq[which(snpMut_filt_freq$Gene%in%valid_genes_names), ])
  snpMut_filt_freq_reduced = snpMut_filt_freq_reduced[order(snpMut_filt_freq_reduced[,1],snpMut_filt_freq_reduced[,2],snpMut_filt_freq_reduced[,3],snpMut_filt_freq_reduced[,4],snpMut_filt_freq_reduced[,5],snpMut_filt_freq_reduced[,6]),]

  saveRDS(snpMut_filt_freq_reduced, file=file.path(out_dir,"snpMut_filt_freq_reduced.rds"))
  #write.table(snpMut_filt_freq_reduced,file=file.path(out_dir,"snpMut_filt_freq_reduced.txt"),append=FALSE,quote=FALSE,sep="\t",row.names=FALSE,col.names=TRUE)
  
  #browser()
  
  return(valid_genes_names)

}

# Main function for phylogeny reconstruction with longitudinal data
learn.longitudinal.phylogeny <- function( D, 
                                          lik_w = rep(1/length(D),length(D)), 
                                          alpha = 10^-3, 
                                          beta = 10^-2, 
                                          initialization = NULL, 
                                          keep_equivalent = FALSE, 
                                          num_rs = 50, 
                                          num_iter = 10000, 
                                          n_try_bs = 500, 
                                          learning_rate = 1, 
                                          marginalize = FALSE, 
                                          error_move = FALSE, 
                                          num_processes = 1, 
                                          seed = NULL, 
                                          verbose = TRUE, 
                                          log_file = "") {
    
    # Set the seed
    set.seed(seed)
    
    # Do we have datasets with different error rates?
    if(length(alpha)>1) {
        different.error.rates <- TRUE
    } else {
        different.error.rates <- FALSE
    }

    # First of all, we remove any NA value from data
    for(i in 1:length(D)) {
        if(any(is.na(D[[i]]))) {
            D[[i]][which(is.na(D[[i]]),arr.ind=TRUE)] <- -3
        }
        storage.mode(D[[i]]) <- "integer"
    }
    
    equivalent_solutions <- list()
    
    if(num_processes > 1) {
        
        parallel_cl <- makeCluster(num_processes,outfile=log_file)
        
        clusterExport(parallel_cl,varlist=c("D","lik_w","alpha","beta","initialization","different.error.rates","keep_equivalent","num_rs","num_iter","n_try_bs","learning_rate","marginalize","error_move","verbose","log_file"),envir=environment())
        clusterExport(parallel_cl,c("MCMC","initialize.B","move.B","relabeling", "prune.and.reattach", "compute_C"),envir=environment())
        clusterSetRNGStream(parallel_cl,iseed=round(runif(1)*100000))
        
        mcmc_res <- parLapply(parallel_cl,1:num_rs,function(x) { 
            
            MCMC(D = D, 
                 lik_w = lik_w,
                 alpha = alpha,
                 beta = beta,
                 different.error.rates = different.error.rates,
                 initialization = initialization,
                 keep_equivalent = keep_equivalent, 
                 num_rs = num_rs, 
                 num_iter = num_iter,
                 n_try_bs = n_try_bs,
                 learning_rate = learning_rate,
                 marginalize = marginalize,
                 error_move = error_move,
                 verbose = verbose,
                 rs_i = x,
                 log_file = log_file
            )
            
        })
        
        stopCluster(parallel_cl)
        
    } else {
        
        # Repeat for num_rs number of restarts
        mcmc_res <- list()
        
        for(i in 1:num_rs) {
            
            mcmc_res[[i]] <- MCMC(D = D, 
                                  lik_w = lik_w,
                                  alpha = alpha,
                                  beta = beta,
                                  different.error.rates = different.error.rates,
                                  initialization = initialization,
                                  keep_equivalent = keep_equivalent, 
                                  num_rs = num_rs,
                                  num_iter = num_iter,
                                  n_try_bs = n_try_bs,
                                  learning_rate = learning_rate,
                                  marginalize = marginalize,
                                  error_move = error_move,
                                  verbose = verbose,
                                  rs_i = i,
                                  log_file = log_file)
            
        }
    }

    if(num_rs > 1) {
        best_joint_lik <- c()
        
        for(i in  1:num_rs) {
            best_joint_lik <- c(best_joint_lik, mcmc_res[[i]]$joint_lik)
        }
        idx_bjl <- which(best_joint_lik == max(best_joint_lik))
        
        best_mcmc <- mcmc_res[[idx_bjl[1]]]
        for(i in 2:length(idx_bjl)) {
            best_mcmc$equivalent_solutions <- c(best_mcmc$equivalent_solutions,mcmc_res[[idx_bjl[i]]]$equivalent_solutions)
        }
        
    }
    else {
        best_mcmc <-  mcmc_res[[1]]
    }

    return(best_mcmc)

}

# perform MCMC
MCMC <- function(D, 
                 lik_w = rep(1/length(D),length(D)), 
                 alpha = 10^-3, 
                 beta = 10^-2, 
                 different.error.rates = FALSE,
                 initialization = NULL, 
                 keep_equivalent = FALSE, 
                 num_iter = 10000,
                 num_rs = 50, 
                 n_try_bs = 500, 
                 learning_rate = 1, 
                 marginalize = FALSE, 
                 error_move = FALSE, 
                 rs_i = NULL,
                 verbose = TRUE, 
                 log_file = "" ) {

    # initialize B
    if(is.null(initialization)) {
        B <- initialize.B(D)
        
    }
    else {
        B <- initialization
        storage.mode(B) <- "integer"
        if(rs_i > 1) {
            # Restars after the first one start from a different point if 
            # a initial tree was passed (it could be generated by TRAIT or by user)
            for(i in seq(1,ncol(B))) {
                B <- move.B(B=B,alpha=alpha,beta=beta,error_move=error_move)$B
            }
        }
    }

    res <- compute_C(B,D,lik_w,alpha,beta,different.error.rates,marginalize)
    C <- res$C
    lik <- res$lik
    joint_lik <- res$joint_lik

    # Initialize result variables (best solution for the current restart)
    B_best <- B
    C_best <- C
    alpha_best <- alpha
    beta_best <- beta
    lik_best <- lik
    joint_lik_best <- joint_lik
    count_lik_best_cons <- 0
    equivalent_solutions <- list()
    if(keep_equivalent) {
        equivalent_solutions[[1]] <- list(B=B_best,C=C_best,alpha=alpha_best,beta=beta_best,relative_likelihoods=lik_best,joint_likelihood=joint_lik_best)
    }
    
    # Repeat until num_iter number of iterations is reached
    j = 1
    while(j <= num_iter) {
        
        if(verbose && (j %% 500)==0) {
            
            cat("\r",
                "Current best lik. = ",format(joint_lik_best, digit = 2, nsmall = 2), 
                " | Restart # ",rs_i,"/",num_rs," | Iter # ",j, " | Likelihood not improved for ", count_lik_best_cons,"/",n_try_bs," iterations",
                "     ", 
                sep='', 
                file = log_file, 
                append = TRUE)
            
        }
        
        # Try move on B
        move_tmp <- move.B(B=B,alpha=alpha,beta=beta,error_move=error_move)
        B_tmp <- move_tmp$B
        alpha_tmp <- move_tmp$alpha
        beta_tmp <- move_tmp$beta
        
        # Compute C at maximun likelihood given B_tmp and returns its likelihood
        res <- compute_C(B_tmp,D,lik_w,alpha_tmp,beta_tmp,different.error.rates,marginalize)
        C_tmp <- res$C
        lik_tmp <- res$lik
        joint_lik_tmp <- res$joint_lik

        # If likelihood at current step is better than best likelihood, replace best model with current
        if(joint_lik_tmp > joint_lik_best) {
            
            B_best <- B_tmp
            C_best <- C_tmp
            alpha_best <- alpha_tmp
            beta_best <- beta_tmp
            lik_best <- lik_tmp
            joint_lik_best <- joint_lik_tmp
            count_lik_best_cons <- 0
            
            B <- B_tmp
            C <- C_tmp
            alpha <- alpha_tmp
            beta <- beta_tmp
            lik <- lik_tmp
            joint_lik <- joint_lik_tmp
            equivalent_solutions <- list()
            if(keep_equivalent) {
                equivalent_solutions[[1]] <- list(B=B_best,C=C_best,alpha=alpha_best,beta=beta_best,relative_likelihoods=lik_best,joint_likelihood=joint_lik_best)
            }
            
        }
        else {
            
            if(joint_lik_tmp == joint_lik_best) {
                if(keep_equivalent) {
                    equivalent_solutions[[(length(equivalent_solutions)+1)]] <- list(B=B_tmp,C=C_tmp,alpha=alpha_tmp,beta=beta_tmp,relative_likelihoods=lik_tmp,joint_likelihood=joint_lik_tmp)
                }
            }
            
            count_lik_best_cons <- count_lik_best_cons + 1
            if(count_lik_best_cons > n_try_bs) {
                
                # Skipping to the next restart
                if(verbose) {
                    cat("\r",
                        "Current best lik. = ",format(joint_lik_best, digit = 2, nsmall = 2), 
                        " | Restart # ",rs_i,"/",num_rs," | Iter # ",j, " | Likelihood not improved for ", (count_lik_best_cons-1),"/",n_try_bs," iterations",
                        "     ", 
                        "\n",
                        sep='')
                }
                break
                
            }
            
            # Take the current state with a probability proportional to the ratio of the two likelihoods
            rho <- min(exp(((joint_lik_tmp-joint_lik)/learning_rate)),1)
            if(runif(1)<=rho) {
                
                B <- B_tmp
                C <- C_tmp
                alpha <- alpha_tmp
                beta <- beta_tmp
                lik <- lik_tmp
                joint_lik <- joint_lik_tmp
                
            }
            
        }
        j = j + 1
    }
    
    return(list(B=B_best,C=C_best,alpha=alpha_best,beta=beta_best,lik=lik_best,joint_lik=joint_lik_best,equivalent_solutions=equivalent_solutions))
    
}

# Initialize B
initialize.B <- function( D, seed = NULL ) {
    
    # Define number of mutations
    m <- ncol(D[[1]])
    
    # Initialize B
    B <- array(0L,c((m+1),(m+1)))
    rownames(B) <- c('r',1:m)
    colnames(B) <- c('r',sample(1:m))
    diag(B) <- 1L
    B[,1] <- 1L
    
    # Add arcs with probability 50%
    p <- 0.50
    for(i in 2:(nrow(B)-1)) {
        
        if(runif(1)<p) {
            B[(i+1),] <- B[i,] + B[(i+1),]
        }
        
    }
    
    B[which(B>1)] <- 1L
    
    return(B)
    
}

# Performing either relabeling or edge changing moves on B
move.B <- function( B, alpha, beta, error_move = FALSE, seed = NULL ) {
    
    # Random probability of choosing a move
    p <- runif(1)
    
    # With no estimation of error rates within the MCMC
    if(error_move==FALSE) {
        # Perform pairwise relabeling with 55% probability
        if(p<0.55) {
            
            B <- relabeling(B=B)
            
        }
        # Perform structural moves with 40% probability
        else if(p>=0.55&&p<0.95) {
            
            B <- prune.and.reattach(B=B)
            
        }
        
        # Perform full relabeling with 5% probability
        else if(p>=0.95) {
            
            # Random relabeling of all clones
            colnames(B) <- c('r',sample(1:(ncol(B)-1)))
            
        }
    }
    # With estimation of error rates within the MCMC
    else {
        # Perform moves on error rates with 10% probability
        if(p<0.10) {
            
            for(err_val in 1:length(alpha)) {
                alpha[[err_val]] <- min(max(rnorm(1,mean=alpha[[err_val]],sd=(alpha[[err_val]]/3)),0.0001),0.9999)
                beta[[err_val]] <- min(max(rnorm(1,mean=beta[[err_val]],sd=(beta[[err_val]]/3)),0.0001),0.9999)
            }
            
        }
        # Perform pairwise relabeling with 50% probability
        else if(p>=0.10&&p<0.60) {
            
            # Relabeling
            B <- relabeling(B=B)
            
        }
        # Perform structural moves with 35% probability
        else if(p>=0.60&&p<0.95) {
            
            B <- prune.and.reattach(B=B)
            
        }
        # Perform full relabeling with 5% probability
        else if(p>=0.95) {
            
            # Random relabeling of all clones
            colnames(B) <- c('r',sample(1:(ncol(B)-1)))
            
        }
    }
    
    return(list(B=B,alpha=alpha,beta=beta))
    
}

# perform node relabeling
relabeling <- function(B) {
    
    # Relabeling
    chosen <- sample(2:ncol(B),2,replace=FALSE)
    tmp <- colnames(B)[chosen[1]]
    colnames(B)[chosen[1]] <- colnames(B)[chosen[2]]
    colnames(B)[chosen[2]] <- tmp
    return(B)
    
}

# perform prune and reattach
prune.and.reattach <- function(B) {
    
    # Change one arch
    is_not_valid <- TRUE
    while(is_not_valid) {
        
        # Select source node
        ch_1 <- sample(x=2:nrow(B),size=1)
        ch_1_gen <- B[ch_1,1:ch_1]
        remaing_node <- as.numeric(which(apply(B[,1:ch_1], c(1), FUN = function(x){!all(x == ch_1_gen)})))
        
        # Chose the target node from the nodes not included in the subtree where ch_1 is the root
        if(length(remaing_node) > 1) {
            
            ch_2 <- sample(x=remaing_node,size=1)
            
        } else if(length(remaing_node)==1) {
            
            ch_2 <- remaing_node
            
        } else {
            # if there aren't any nodes, select a different source
            next
        }

        # A pair of two nodes is valid if the nodes are not already directly connected
        if(!(all(B[ch_1,1:ch_2]==B[ch_2,1:ch_2])&sum(B[ch_1,])==(sum(B[ch_2,])+1))) {
            
            is_not_valid <- FALSE
        }
        
    }
    
    descendent_nodes <- setdiff(1:nrow(B), remaing_node)
    
    # Extract descendent node submatrix
    rem_B <- B[remaing_node,remaing_node,drop=FALSE]
    new_B <- matrix(data = 0L, nrow = nrow(rem_B), ncol = ncol(B))
    colnames(new_B) <- c(colnames(B)[remaing_node], colnames(B)[descendent_nodes])
    new_B[1:length(remaing_node),1:length(remaing_node)] <- rem_B
    gen_ch_2 <- new_B[which(colnames(new_B)==colnames(B)[ch_2]),1:length(remaing_node)]
    desc_B <- cbind(matrix(rep(gen_ch_2,each=length(descendent_nodes)),  nrow = length(descendent_nodes)), 
                    B[descendent_nodes,descendent_nodes,drop=FALSE])
    new_B <- rbind(new_B,desc_B)
    
    return(new_B)
    
}

# Compute attachments matrix C at maximum likelihood given B, D and P
compute_C <- function( B, D, lik_w = rep(1/length(D),length(D)), alpha = 10^-3, beta = 10^-2, different.error.rates = FALSE, marginalize = FALSE ) {
    
    # Initialize return variables
    C_res <- list()
    lik_matrix_res <- list()
    lik_res <- NULL
    joint_lik_res <- 0.0
    
    # Determine indeces to order D such that it matches B
    idx_srt <- as.integer(colnames(B)[2:ncol(B)])
    
    # For each time point
    curr_alpha <- alpha
    curr_beta <- beta
    for(i in 1:length(D)) {
        
        if(different.error.rates) {
            curr_alpha <- alpha[i]
            curr_beta <- beta[i]
        }
        
        # Initialize C for the current time point
        lik_time <- 0
        
        # Go through all the cells
        curr_cells_D <- cbind(rep(1,nrow(D[[i]][,idx_srt,drop=FALSE])),D[[i]][,idx_srt,drop=FALSE])
        
        # Find assignment at maximum log likelihood for current cell
        lik_matrix <- array(0L,c(nrow(D[[i]]),ncol(B)))
        
        for(k in 1:nrow(B)) {
            
            curr_clone_C = matrix(rep(0L,nrow(B)),nrow=1)
            curr_clone_C[1,k] <- 1L
            
            # Save mapping between ordering in B and dataset D
            # Factorization D = C dot B. r_D_tilde represents a combination of mutations and clones
            r_D_tilde <- (curr_clone_C %*% B)*2
            
            sum_cell_clone <- as.matrix(sweep(curr_cells_D,MARGIN=2,r_D_tilde,"+"))
            lik_matrix[,k] <- (curr_beta^Rfast::rowsums(sum_cell_clone==2)) * ((1-curr_beta)^Rfast::rowsums(sum_cell_clone==0)) * ((curr_alpha)^Rfast::rowsums(sum_cell_clone==1)) * ((1-curr_alpha)^Rfast::rowsums(sum_cell_clone==3))
            
        }
        
        C_list_time <- Rfast::rowMaxs(lik_matrix,value=FALSE)
        
        if(marginalize==TRUE) {
            lik_time <- sum(log(Rfast::rowsums(lik_matrix)))
        }
        else {
            lik_time <- sum(log(Rfast::rowMaxs(lik_matrix,value=TRUE)))
        }
        
        storage.mode(C_list_time) <- "integer"

        C_res[[i]] <- (C_list_time-1)
        lik_matrix_res[[i]] <- lik_matrix
        lik_res <- c(lik_res,lik_time)
        joint_lik_res <- joint_lik_res + (lik_time * lik_w[i])
        
    }
    
    return(list(C=C_res,lik=lik_res,joint_lik=joint_lik_res))
    
}

# library(tools)
# library(dplyr)
# library(tidyr)
# library(configr)

ANDx <- function(x, y) {
  z <- x & y
  z[is.na(x) | is.na(y)] <- NA
  return(z)
}

ORx <- function(x, y) {
  z <- x | y
  z[is.na(x) | is.na(y)] <- NA
  return(z)
}

compare_named_lists <- function(l1,l2) {
  if(is.list(l1) && is.list(l2)) { #if they are lists
    if (
      length(unique(names(l1))) == length(l1) &&
      length(unique(names(l2))) == length(l2) && !("" %in% names(l1)) && !("" %in% names(l2))
    ) { #if they are named
      st<-TRUE
      for (i in unique(c(names(l1), names(l2)))) {
        if (i %in% names(l1) && i %in% names(l2)) { #if the same name is present in both
          #log2_print('----')
          #log2_print(l1[[i]])
          #log2_print(l2[[i]])
          if (is.list(l1[[i]]))
            st <- st && compare_named_lists(l1[[i]],l2[[i]])
          else
            st <- st && identical(l1[[i]],l2[[i]])
        }
        else
          return(FALSE)
      }
      return(st)
    }

    return(FALSE)

  }
  else
    return(FALSE)


}

OneInOneOut <- function(InDir, InFilesMask, OutDir, OutExt, Opt) {

	InFiles = list.files(path = InDir, pattern = InFilesMask, no.. = TRUE)
	AllOutFiles <- list.files(path = OutDir, no.. = TRUE)


	InFilesToDo <- c()
	OutFilesToRm <- c()
	OutFilesToRep <- c()
	if ( length(InFiles) >0 ) {
    ActFileContent <- read.config(file = Opt$ActFile)
    StoredFileContent <- read.config(file = Opt$StoredFile)

    OutFiles <- paste0(file_path_sans_ext(InFiles),'.',OutExt)
    log2_print('out')
    log2_print(OutFiles)
    log2_print(file.exists(file.path(OutDir,OutFiles)))
    ResidualOutFiles <- AllOutFiles[!AllOutFiles %in% OutFiles]

    EqFiles <- compare_named_lists(ActFileContent, StoredFileContent)

    if ( !EqFiles ) { # if conf files are non equal
      file.copy(Opt$ActFile, Opt$StoredFile, overwrite = TRUE)
      InFilesToDo <- InFiles
      OutFilesToRm <- ResidualOutFiles
      OutFilesToRep <- OutFiles
    }
		else {
		  #OptStoredFileDateMax = max(file.mtime( c(Opt$StoredFile, Opt$ExtraFiles) ))
			#OptStoredFileDateMin = min(file.mtime( c(Opt$StoredFile, Opt$ExtraFiles) ))
		  if (!is.null(Opt$ExtraFiles))
		    OptExtraDate <- file.mtime(Opt$ExtraFiles)
		  else
		    OptExtraDate <- NULL
			OptStoredFileDateMax = max(c( min(file.mtime( c(Opt$StoredFile, Opt$ActFile))), OptExtraDate ))
			OptStoredFileDateMin = min(c( max(file.mtime( c(Opt$StoredFile, Opt$ActFile))), OptExtraDate ))

			InFilesDate <- file.mtime(file.path(InDir,InFiles))
			RightInFilesTime <- OptStoredFileDateMin > InFilesDate
			OutFilesDate <- file.mtime(file.path(OutDir,OutFiles))
			RightOutFilesTime <- OptStoredFileDateMax < OutFilesDate

			ToDo=!replace_na( ANDx( RightOutFilesTime, RightInFilesTime), FALSE)
			ToRep=!replace_na( ANDx( RightOutFilesTime, RightInFilesTime), TRUE)

			#----------
			log2_print('InFiles -> OptStored -> OutFiles')
			dout <- data.frame(
			  list(
			    "InFiles"=InFiles,
			    BooolIn=RightInFilesTime,
			    InDate=InFilesDate, Rel='<',
			    OptStoredFileDateMin = OptStoredFileDateMin, Rel='<',

			    OptStoredFileDateMax = OptStoredFileDateMax, Rel='<',
			    OutDate = OutFilesDate, Rel='<',
			    BoolOut=RightOutFilesTime,
			    "OutFiles"=OutFiles,

			    "ToDo"=ToDo,
			    "ToRep"=ToRep
			  )
			)
			log2_print(head(dout))
			#----------

			InFilesToDo <- InFiles[ToDo]
			OutFilesToRm <- ResidualOutFiles
			OutFilesToRep <- OutFiles[ToRep]
		}
	}
	log2_print(c('DO',InFilesToDo))
	log2_print(c('REP',OutFilesToRep))
	log2_print(c('RM',OutFilesToRm))

	return(list('InFilesToDo'=InFilesToDo, 'OutFilesToRm'=OutFilesToRm, 'OutFilesToRep'=OutFilesToRep))
}


# c1 <- read.config(file = "conf1")
# c2 <- read.config(file = "conf2")
# Opt=list()
# Opt$StoredFile <- "conf1"
# Opt$ActFile <- "conf2"
# Opt$ExtraFiles=NULL
#
# OneInOneOut( "./in/", InFilesMask = ".*", "./out/", OutExt = "out", Opt = Opt)

ManyInSomeOut <- function(InDir, InFilesMask, OutDir, OutFiles, Opt) {
  InFiles = list.files(path = InDir, pattern = InFilesMask, no.. = TRUE)
  AllOutFiles <- list.files(path = OutDir, no.. = TRUE)

  InFilesToDo <- c()
  OutFilesToRm <- c()
  OutFilesToRep <- c()
  if ( length(InFiles) >0 ) {
    ActFileContent <- read.config(file = Opt$ActFile)
    StoredFileContent <- read.config(file = Opt$StoredFile)

    OutFiles <- file.path(OutDir,OutFiles)

    log2_print('out')
    log2_print(OutFiles)
    log2_print(file.exists(OutFiles))
    ResidualOutFiles <- AllOutFiles[!AllOutFiles %in% OutFiles]

    EqFiles <- compare_named_lists(ActFileContent, StoredFileContent)

    if ( !EqFiles ) { # if conf files are non equal
      file.copy(Opt$ActFile, Opt$StoredFile, overwrite = TRUE)
      InFilesToDo <- InFiles
      OutFilesToRm <- ResidualOutFiles
      OutFilesToRep <- OutFiles
    }
    else {
      if (!is.null(Opt$ExtraFiles))
        OptExtraDate <- file.mtime(Opt$ExtraFiles)
      else
        OptExtraDate <- NULL
      OptStoredFileDateMax = max(c( min(file.mtime( c(Opt$StoredFile, Opt$ActFile))), OptExtraDate ))
      OptStoredFileDateMin = min(c( max(file.mtime( c(Opt$StoredFile, Opt$ActFile))), OptExtraDate ))
      InFilesDate <- file.mtime(file.path(InDir,InFiles))
      RightInFilesTime <- OptStoredFileDateMin > InFilesDate
      OutFilesDate <- file.mtime(OutFiles)
      RightOutFilesTime <- OptStoredFileDateMax < OutFilesDate

      ToDo=any(!replace_na( c(all(RightOutFilesTime), all(RightInFilesTime)), FALSE))
      #ToRep=!replace_na( ANDx( RightOutFilesTime, RightInFilesTime), T)

      #----------
      # log2_print('InFiles -> OptStored -> OutFiles')
      # dout <- data.frame(
      #   list(
      #     "InFiles"=InFiles,
      #     BooolIn=RightInFilesTime,
      #     InDate=InFilesDate, Rel='<',
      #     OptStoredFileDateMin = OptStoredFileDateMin, Rel='<',
      #
      #     OptStoredFileDateMax = OptStoredFileDateMax, Rel='<',
      #     OutDate = OutFilesDate, Rel='<',
      #     BoolOut=RightOutFilesTime,
      #     "OutFiles"=OutFiles,
      #
      #     "ToDo"=ToDo,
      #     "ToRep"=ToRep
      #   )
      # )
      # log2_print(head(dout))
      #----------
      if (ToDo) {
        InFilesToDo <- InFiles
        OutFilesToRm <- ResidualOutFiles
        OutFilesToRep <- OutFiles
      }
    }
  }
  log2_print(c('DO',InFilesToDo))
  log2_print(c('REP',OutFilesToRep))
  log2_print(c('RM',OutFilesToRm))

  return(list('InFilesToDo'=InFilesToDo, 'OutFilesToRm'=OutFilesToRm, 'OutFilesToRep'=OutFilesToRep))
}


SomeInSomeOut <- function(InFiles, OutFiles, Opt) {
  AllOutFiles <- list.files(path = OutDir, no.. = TRUE)

  InFilesToDo <- c()
  OutFilesToRm <- c()
  if ( length(InFiles) >0 ) {
    ActFileContent <- read.config(file = Opt$ActFile)
    StoredFileContent <- read.config(file = Opt$StoredFile)

    ResidualOutFiles <- AllOutFiles[!AllOutFiles %in% OutFiles]

    EqFiles <- compare_named_lists(ActFileContent, StoredFileContent)

    if ( !EqFiles ) { # if conf files are non equal
      file.copy(Opt$ActFile, Opt$StoredFile)
      InFilesToDo <- InFiles
      OutFilesToRm <- c(OutFiles[file.exists(OutFiles)], ResidualOutFiles)
    }
    else {
      #OptStoredFileDateMax = max(file.mtime( c(Opt$StoredFile, Opt$ExtraFiles) ))
      #OptStoredFileDateMin = min(file.mtime( c(Opt$StoredFile, Opt$ExtraFiles) ))
      if (!is.null(Opt$ExtraFiles))
        OptExtraDate <- file.mtime(Opt$ExtraFiles)
      else
        OptExtraDate <- NULL
      OptStoredFileDateMax = max(c( min(file.mtime( c(Opt$StoredFile, Opt$ActFile))), OptExtraDate ))
      OptStoredFileDateMin = min(c( max(file.mtime( c(Opt$StoredFile, Opt$ActFile))), OptExtraDate ))
      InFilesDate <- file.mtime(file.path(InDir,InFiles))
      RightInFilesTime <- OptStoredFileDateMin > InFilesDate
      OutFilesDate <- file.mtime(OutFiles)
      RightOutFilesTime <- OptStoredFileDateMax < OutFilesDate

      ToDo=any(!replace_na( ANDx( RightOutFilesTime, RightInFilesTime), FALSE))
      ToRm=!replace_na( ANDx( RightOutFilesTime, RightInFilesTime), TRUE)

      #----------
      log2_print('InFiles -> OptStored -> OutFiles')
      dout <- data.frame(
        list(
          "InFiles"=InFiles,
          BooolIn=RightInFilesTime,
          InDate=InFilesDate, Rel='<',
          OptStoredFileDateMin = OptStoredFileDateMin, Rel='<',

          OptStoredFileDateMax = OptStoredFileDateMax, Rel='<',
          OutDate = OutFilesDate, Rel='<',
          BoolOut=RightOutFilesTime,
          "OutFiles"=OutFiles,

          "ToDo"=ToDo,
          "ToRm"=ToRm
        )
      )
      log2_print(head(dout))
      #----------

      InFilesToDo <- InFiles
      OutFilesToRm <- c(OutFiles[ToRm], ResidualOutFiles)
    }
  }
  log2_print(paste('DO',InFilesToDo))
  log2_print(paste('remove',OutFilesToRm))
  return(list(InFilesToDo,OutFilesToRm))
}

### depth_computation.R
###

  dp_compute <- function(cellInfo,
                         bam_dir,
                         samtools,
                         data_dir,
                         out_dir,
                         id_column,
                         InFilesToDo = NULL,
                         OutFilesToRm = NULL) {

    ## ################################################################
    ## GET DEPTH FOR EACH SELECTED MUTATION FROM BAMs #####
    ## ################################################################

    dir.create(out_dir, showWarnings = FALSE)
    if (!dir.exists(out_dir))
      showNotification(paste("Filtered data output folder,",
                             out_dir,
                             ", cannot be created"),
                       duration = 10,
                       type = "warning")
    log2_print(data_dir)
    # load data
    snpMut_filt_freq <-
      readRDS(file = paste0(file.path(data_dir, 'snpMut_filt_freq.rds')))

    listMut <-
      strsplit(snpMut_filt_freq$UIDsnp, "_")
    snpMut_filt_freq$Mut_pos <-
      unlist(lapply(listMut, function(m) paste0(m[2], "_", m[3])))

    listMutFormatted <-
      unlist(lapply(unique(strsplit(snpMut_filt_freq$Mut_pos, "_")),
                    function(m) paste0(m[1], ":", m[2], "-", m[2])))

    pattern = ".bam$"
    BAMlist_list <-
      dir(path = bam_dir, pattern = pattern, full.names = TRUE)

    log2_print(InFilesToDo)
    if (!is.null(InFilesToDo)) {
      log2_print('In')
      log2_print(BAMlist_list)
      BAMlist_list <- BAMlist_list[BAMlist_list %in% InFilesToDo]
      log2_print(BAMlist_list)
    }

    if (!is.null(OutFilesToRm)) {
      log2_print('Out')
      ## OutFilesToRm <- file.path(thr_out_dir, OutFilesToRm)
      log2_print(OutFilesToRm)
      ## file.remove(OutFilesToRm)
      log2_print(list.files(path = out_dir))
    }

    if (length(BAMlist_list)==0)
      return()

    BAMlist <- paste(BAMlist_list, collapse = " ")
    BAM = paste0(str_split_fixed(basename(BAMlist_list),
                                 "\\.",
                                 n = 2)[, 1],
                 collapse = "\t")
    header = paste0("CHR\tPOS\t", BAM)

    ## print('BAM')
    ## print(BAM)

    numCores <- detectCores()
    registerDoParallel(numCores)

    res <- foreach(i = 1:length(listMutFormatted), .combine=c) %dopar% {
      ## print(paste0(samtools, " depth -a -Q 20 -r ", listMutFormatted[i], " ", BAMlist))
      system(paste0(samtools,
                    " depth -a -Q 20 -r ",
                    listMutFormatted[i],
                    " ",
                    BAMlist),
             intern = TRUE,
             ignore.stdout = FALSE,
             ignore.stderr = TRUE)
    }
    ## print('res')
    ## print(res)
    ## BAM = gsub(x = paste(unlist(strsplit(x = gsub(x=BAMlist, pattern=bam_dir, ""), split=".bam")), collapse = "\t"), " ","")


    ##return()  #######

    ## ################################################################
    ## MAKE DEPTH MATRIX ####
    ## ################################################################

    depth_tbl <-
      read.table(text = paste0(header,
                               "\n",
                               paste(res, collapse = '\n')),
                 header = TRUE,
                 sep = '\t',
                 stringsAsFactors = FALSE)
    rownames(depth_tbl) <-
      paste(depth_tbl$CHR, depth_tbl$POS, sep = "_")
    depth_tbl <- depth_tbl[, -c(1, 2)]

    snpMut_filt_freq$Mut_pos <-
      paste(snpMut_filt_freq$Chr, snpMut_filt_freq$PosStart, sep = "_")

    mut_tbl <-
      as.data.frame.matrix(table(snpMut_filt_freq$Mut_pos,
                                 snpMut_filt_freq$scID))

    ## cellInfo <- readRDS("data_info.rds")

    ## pattern=".bbq_s4.bam$"

    ## this cycle does not do nothing
    for(i in 1:ncol(depth_tbl)) {
      run <- colnames(depth_tbl)[i]
      ## colnames(depth_tbl)[i] <- as.character(cellInfo$Run[cellInfo$Run == gsub(pattern,"",run)])
      colnames(depth_tbl)[i] <-
        as.character(cellInfo[[id_column]][cellInfo[[id_column]] == run])
    }

    depth_tbl <-
      depth_tbl[order(row.names(depth_tbl)),
                order(colnames(depth_tbl))]
    mut_tbl <-
      mut_tbl[order(row.names(mut_tbl)),
              order(colnames(mut_tbl))]

    ## write final results to text files
    write.table(x = depth_tbl,
                file = file.path(out_dir,
                                 "final_data_depth.txt"),
                sep = '\t',
                row.names = TRUE,
                col.names = TRUE)
    write.table(x = mut_tbl,
                file = file.path(out_dir, "final_data_mut.txt"),
                sep = '\t',
                row.names = TRUE,
                col.names = TRUE)
  }

### end of file -- depth_computation.R

### project_tab.R

### Project reactiveVals ####

inputs[["default_config_dir_std"]] <-
  reactiveVal(create_default_config()) # each time check to create default conf
inputs[["config_dir_std"]] <- reactiveVal(os_conf_subdir)
inputs[["project_folder_std"]] <- reactiveVal()
inputs[["project_folder"]] <- reactiveVal()
inputs[["project_loaded"]] <- reactiveVal(FALSE)


inputs[["pr_folder"]] <- reactiveVal()
inputs[["pr_folder_std"]] <- reactiveVal(.my_actual_wd)
inputs[["pr_name"]] <- reactiveVal()
inputs[["pr_name_std"]] <- reactiveVal()
inputs[["pr_path"]] <- reactiveVal(.my_actual_wd)


inputs[['av_vcf_out_dir']] <- reactiveVal()
inputs[['thr_vcf_in_dir']] <- reactiveVal()
inputs[['thr_out_dir']] <- reactiveVal()
inputs[['dp_out_dir']] <- reactiveVal()
inputs[['va_out_dir']] <- reactiveVal()
inputs[['inf_out_dir']] <- reactiveVal()

### End project reactiveVals ####

### Project functions ####

toggle_inputs <- function(enable_inputs=TRUE) {
  input_list <- reactiveValuesToList(input)
  
  dp_input <- catch_ui_names(input_list, dp_grep_str)
  av_input <- catch_ui_names(input_list, av_grep_str)
  thr_input <- catch_ui_names(input_list, thr_grep_str)
  m_input <- catch_ui_names(input_list, m_grep_str)
  input_list <- c(m_input, av_input, thr_input, dp_input)
  #browser()
  log2_print("enable_inputs")
  log2_print(enable_inputs)
  # Toggle elements
  for(x in names(input_list))
    if(enable_inputs){
      shinyjs::enable(x)
    } else {
      shinyjs::disable(x) 
    }
}

disable_demo_tabs <- function() {
  
  enable_inputs <- FALSE
  
  #browser()
  
  input_list <- reactiveValuesToList(input)
  
  #if(only_buttons){
  #  buttons <- which(sapply(input_list,function(x) {any(grepl('Button',attr(x,"class")))}))
  #  input_list = input_list[buttons]
  #}
  
  dp_input <- catch_ui_names(input_list, dp_grep_str)
  av_input <- catch_ui_names(input_list, av_grep_str)
  thr_input <- catch_ui_names(input_list, thr_grep_str)
  m_input <- catch_ui_names(input_list, m_grep_str)
  
  #thr_grep_str= "^thr_|^`thr_|_thr_"
  #dp_grep_str= "^dp_|^`dp_|_dp_"
  #va_grep_str= "^va_|^`va_|_va_"
  #inf_grep_str= "^inf_|^`inf_|_inf_"
  #m_grep_str= "^m_|^`m_|_m_|sc"
  
  
  
  
  input_list <- c(m_input, av_input, thr_input, dp_input)
  
  #which(sapply(input_list,function(x) {any(grepl('Button',attr(x,"class")))}))
  
  # Toggle elements
  for(x in names(input_list))
    if(enable_inputs){
      shinyjs::enable(x)} else {
        shinyjs::disable(x) }
  
  
}

enable_demo_tabs <- function() {}
  
### End project functions ####

### Project observes ####

observeEvent(input[["pr_name"]], {
  pr_name <- input[["pr_name"]]
  pr_name_ <- inputs[["pr_name"]]
  pr_folder_std_ <- inputs[["pr_folder_std"]]
  pr_name_std_ <- inputs[["pr_name_std"]]
  pr_path_ <- inputs[["pr_path"]]

  pr_name_(pr_name)

  name <- str_replace_all(pr_name,
                          pattern = "[^A-Za-z0-9\\-+_]+",
                          replacement = "_")
  pr_name_std_(name)

  if (str_length(pr_name_std_()) > 0)
    pr_path_(file.path(pr_folder_std_(),
                       pr_name_std_(),
                       ""))
  else
    pr_path_(file.path(pr_folder_std_(), pr_name_std_()))
})


## Original.
## observeEvent(input[["pr_name"]], {
##     inputs[["pr_name"]](input[["pr_name"]])
##
##     name <- str_replace_all(input[["pr_name"]],
##                             pattern = "[^A-Za-z0-9\\-+_]+",
##                             replacement = "_")
##     inputs[["pr_name_std"]](name)
##
##     if (str_length(inputs[["pr_name_std"]]()) > 0)
##         inputs[["pr_path"]](file.path(inputs[["pr_folder_std"]](),
##                                       inputs[["pr_name_std"]](),
##                                       ""))
##         else # MA: Something is fishy here: incorrect indentation...
##             inputs[["pr_path"]](file.path(inputs[["pr_folder_std"]](),
##                                           inputs[["pr_name_std"]]()))
## })

observeEvent(input[["pr_folder"]],{
  inputs[["pr_folder"]](input[["pr_folder"]])
  inputs[["pr_folder_std"]](
    normalizePath(
      parseDirPath(roots = roots_dir,
                   inputs[["pr_folder"]]()),
      mustWork = FALSE
    )
  )

  if (str_length(inputs[["pr_name_std"]]()) > 0)
    inputs[["pr_path"]](file.path(inputs[["pr_folder_std"]](), inputs[["pr_name_std"]](), ""))
  else
    inputs[["pr_path"]](file.path(inputs[["pr_folder_std"]](), inputs[["pr_name_std"]]()))

  updateActionButton(session,
                     'pr_folder',
                     label = normalizePath(set_dir_ui("pr_folder"), mustWork = FALSE))
})

observeEvent(inputs[["pr_path"]](), {
  inputs[["project_loaded"]](FALSE)


  hide(id = "pr_path_div")
  if (!is.null(inputs[["pr_path"]]()))
    if (length(inputs[["pr_path"]]()) > 0) {
      shinyjs::show(id="pr_path_div")

      if (dir.exists(file.path(inputs[["pr_path"]](),os_conf_subdir))) {
        ## if (!load_project(inputs[["pr_path"]](), config_dir="config")) {
        ## warning()
        ## }
        ## else {
        path <- inputs[["pr_path"]]()
        inputs[["pr_folder_std"]](dirname(path))
        inputs[["pr_name"]](basename(path))
        inputs[["pr_name_std"]](basename(path))
        inputs[["project_folder_std"]](path)
        project_folder <- list( "path"="", "root"="")
        project_folder$path[[1]]<-""
        project_folder$path <-
          c(project_folder$path,
            as.list(path_split(path_rel(inputs[["project_folder_std"]](),
                                        start = roots_dir[[".."]]))[[1]]))
        project_folder$root=".."
        inputs[["project_folder"]](project_folder)

        dir <- out_subfolder_compute(inputs[['project_folder']](),'vcf_out')
        inputs[['av_vcf_out_dir']](dir)
        inputs[['thr_vcf_in_dir']](inputs[['av_vcf_out_dir']]())

        dir <- out_subfolder_compute(inputs[['project_folder']](),'filtered_vcf')
        inputs[['thr_out_dir']](dir)

        dir <- out_subfolder_compute(inputs[['project_folder']](),'sam_out')
        inputs[['dp_out_dir']](dir)

        dir  <- out_subfolder_compute(inputs[['project_folder']](), 'filtered_var')
        inputs[['va_out_dir']](dir)

        updateActionButton(session,"pr_folder", label=inputs[["pr_folder_std"]]())
        updateTextInput(session,"pr_name", value=inputs[["pr_name_std"]]())
        updateActionButton(session,"pr_next", label="Load project")
        ## }
      } else {
        #browser()
        updateActionButton(session,"pr_next", label="Create project")
        print("go to proj")
        updateTabsetPanel(session, "main_tabset", selected = "Project")
      }
    }
  #browser()
  if (!is.null(inputs[["demo"]]())) {
    if (inputs[["demo"]]()>=1) {
      inputs[["demo"]](0)

      delay(1000, click("pr_next"))
      #click("pr_next")

      #hide_tab()
      delay(1500, {
        returned_vals <- show_result(rs_())
        rs_(returned_vals$rs)
      })
    }
  }

  #browser()
  if (!is.null(inputs[["reload_project"]]())) {
    #browser()
    if( inputs[["reload_project"]]()>=1) {
      inputs[["reload_project"]](0)

      delay(1000, click("pr_next"))
      #click("pr_next")

      #hide_tab()
      delay(1500, {
        returned_vals <- show_result(rs_())
        rs_(returned_vals$rs)
        #
      })

    }
  }

}, ignoreInit = TRUE)

observeEvent(inputs[["project_loaded"]](), {
  if (!is.null(inputs[["project_loaded"]]()))
    if (inputs[["project_loaded"]]()) {
      shinyjs::hide(id = "pr_next")
      
      log2_print('disable tabs--------------------------------------------------------------------------')
      log2_print(inputs[["demo"]]())
      toggle_inputs(is.null(inputs[["demo"]]()))
      #browser()

      a_sam <- Sys.which("samtools")
      famous_sam <- list()
      famous_sam$path <- c("",path_split(path_rel(dirname(a_sam[[1]]), start = roots_dir[[".."]]))[[1]])
      famous_sam$root <- roots_dir[[".."]]
      if(str_length(a_sam[[1]])>1)
        inputs[["dp_samtools_exec_dir"]](famous_sam)

      a_tool <- Sys.which("annotate_variation.pl")
      famous_tool <- list()
      famous_tool$path <- c("",path_split(path_rel(dirname(a_tool[[1]]), start = roots_dir[[".."]]))[[1]])
      famous_tool$root <- roots_dir[[".."]]
      if(str_length(a_tool[[1]])>1)
        inputs[["av_anovar_exec_dir"]](famous_tool)

      a_neckLACE <- Sys.which("perl")
      if(str_length(a_neckLACE[[1]])==0)
        showNotification("No perl found",
                         duration = 10,
                         type = "error")
    }
  else
  {
    #browser()
    shinyjs::show(id = "pr_next")
    
    if (!is.null(inputs[["demo_remainder"]]())) {
      if (dir.exists(inputs[["demo_remainder"]]())) {
        #dir_delete(inputs[["demo_remainder"]]())
      }
    }
  }
})

observeEvent(input[["pr_next"]], {
  #browser()
  warning_wd <- FALSE
  if(length(inputs[["pr_path"]]()) == 0)
    warning_wd <-TRUE
  else {
    if (dir.exists(inputs[["pr_path"]]())) {
      if (normalizePath(inputs[["pr_path"]]()) == .my_actual_wd) {
        ## pattume non aspetta per la risposta
        ## shinyalert(
        ##  paste("You experiment will be you current working folder,", getwd(), ". Create expriment?"), type = "warning",
        ##  immediate=F)
        ## warning_wd <- input$shinyalert
        log2_print(warning_wd)
        warning_wd <- FALSE
      }
    } else
      warning_wd <- FALSE
  }
  if (!warning_wd) {
    dir.create(inputs[["pr_path"]](), showWarnings = FALSE)
    if (dir.exists(inputs[["pr_path"]]())) {
      inputs[["project_folder_std"]](inputs[["pr_path"]]())
      if (!dir.exists(file.path(inputs[["pr_path"]](),os_conf_subdir))) {
        if (!create_project(inputs[["project_folder_std"]](),
                            inputs[["default_config_dir_std"]]())) {
          showNotification(paste(paste("impossible to create project",
                                       inputs[["pr_path"]]())),
                           duration = 10,
                           type = "error")
        } else {
          inputs[["project_loaded"]](TRUE)
          #showTab(inputId = "main_tabset", target = "SC metadata")
        }
      } else {
        if (!load_project(inputs[["project_folder_std"]]())) {
          showNotification(paste(paste("impossible to load project",
                                       inputs[["pr_path"]]())),
                           duration = 10,
                           type = "error")
          showNotification(paste(paste("check config files")),
                           duration = 10,
                           type = "warning")
        } else {
          inputs[["project_loaded"]](TRUE)
          #showTab(inputId = "main_tabset", target = "SC metadata")
        }
      }


      #browser()
      projs <- update_proj_list(inputs[["pr_path"]]())

      # createmenuitem <- function(x) {
      #   menuSubItem(
      #     text = names(projs)[x],
      #     tabName = str_replace_all(
      #       normalizePath(projs[[x]]),
      #       pattern = .Platform$file.sep,
      #       replacement = ","
      #     )
      #   )
      # }

      submenus <- lapply(seq_along(projs),
                         FUN = function(x) {createmenuitem(x,projs)})

      recent_projects <- menuItem("Recent Projects", submenus, icon = icon("history"))

      output$recent_projects <- renderMenu({recent_projects})


      ## Change to project folder
      setwd(inputs[["project_folder_std"]]())

      project_folder <- list( "path"="", "root"="")
      project_folder$path[[1]]<-""
      project_folder$path <-
        c(project_folder$path,
          as.list(path_split(path_rel(inputs[["project_folder_std"]](),
                                      start = roots_dir[[".."]]))[[1]]))
      project_folder$root=".."
      inputs[["project_folder"]](project_folder)

      dir <- out_subfolder_compute(inputs[['project_folder']](), 'vcf_out')
      inputs[['av_vcf_out_dir']](dir)
      inputs[['thr_vcf_in_dir']](inputs[['av_vcf_out_dir']]())

      dir <- out_subfolder_compute(inputs[['project_folder']](), 'filtered_vcf')
      inputs[['thr_out_dir']](dir)

      dir <- out_subfolder_compute(inputs[['project_folder']](), 'sam_out')
      inputs[['dp_out_dir']](dir)

      dir  <- out_subfolder_compute(inputs[['project_folder']](), 'filtered_var')
      inputs[['va_out_dir']](dir)
    }
  }
  
  #browser()
  if (!is.null(inputs[["demo"]]())) {
    if (inputs[["demo"]]() == -1)
    {
      inputs[["demo"]](NULL)
    }
    else if (inputs[["demo"]]() == 0)
      inputs[["demo"]](-1)
  }
})

### End project observes ####


shinyDirChoose(input,
               "pr_folder",
               roots = roots_dir,
               filetypes = c('', 'txt'),
               defaultRoot = "..")

updateActionButton(session,
                   'pr_folder',
                   label = .my_actual_wd)


### Project outputs ####

output[["pr_path"]] <- renderText(inputs[["pr_path"]]())
output[["pr_title_ui"]] <- renderUI({
  tags$h3(paste("Project:", inputs[["pr_name"]]() ))
})

output[["pr_title_ui_h"]] <- renderUI({
  x <- ""
  if (str_length(inputs[["pr_name"]]()) > 0)
    x <- paste("Project: ", inputs[["pr_name"]]() )
  tags$b(x)
})

output[["pr_path_ui"]] <- renderUI({
  tagList(br(),
          tags$b("Project pathname"),
          verbatimTextOutput("pr_path"),
          br()
  )
})

### End project outputs ####


### end of file -- project_tab.R

### app.R --
###
### Main LACEinterface file.
###
### See file LICENSE in the main folder for licensing and copyright
### information.


### reactlogShow(time = FALSE)

### Libraries needed by the app.

suppressMessages(library(biomaRt, exclude = c("show", "select")))
suppressMessages(library(foreach))
suppressMessages(library(doParallel))
suppressMessages(library(sortable))
suppressMessages(library(shiny,
        exclude = c("runExample",
                    "dataTableOutput",
                    "renderDataTable",
                    "validate",
                    "show")))
suppressMessages(library(shinythemes))
suppressMessages(library(dplyr))
suppressMessages(library(forcats))
suppressMessages(library(tidyr))
suppressMessages(library(readr))
suppressMessages(library(shinyFiles))
suppressMessages(library(shinyjs, exclude = c("show")))
## library(shinyBS)
suppressMessages(library(jsonlite))
suppressMessages(library(configr))
## library(readr)
suppressMessages(library(DT))
suppressMessages(library(stringr))
suppressMessages(library(tools))
suppressMessages(library(fs))
suppressMessages(library(data.table, exclude = c("first", "last", "between")))
suppressMessages(library(htmltools))
suppressMessages(library(shinyBS))
suppressMessages(library(bsplus))
suppressMessages(library(shinydashboard))
suppressMessages(library(shinyvalidate))
suppressMessages(library(logr))
suppressMessages(library(ggplot2))
suppressMessages(library(svglite))
#library(widgetframe)

### Code
### ====

### getWidgetType --

getWidgetType <- function(widgetId) {
  paste0("elem = document.getElementById('", widgetId, "');
            var message;
            if (elem == null) {
              message = 'None'
            } else {
              // RStudio Viewer + IE workaround (dont have .include())
              if (elem.getAttribute('class').indexOf('js-range-slider') > -1) {
                message = 'slider'
              } else if (elem.getAttribute('class').indexOf('shinyFiles') > -1) {
                message = 'shinyFiles'
              } else if (elem.getAttribute('class').indexOf('shinyDirectories') > -1) {
                message = 'shinyDirectories'
              } else if (elem.nodeName == 'SELECT') {
                message = 'select'
              } else if (elem.getAttribute('class').indexOf('shiny-text-output') > -1) {
                message = 'shiny-text-output'
              } else if (elem.getAttribute('class').indexOf('selectized') > -1) {
                message = 'selectized'
              } else {
                message = elem.getAttribute('type');
              }
            }
          Shiny.onInputChange('inputType_", widgetId,"', message)
         "
  )
}


### jsCode --

jsCode <- "shinyBS.addTooltip = function(id, type, opts) {
  var $id = shinyBS.getTooltipTarget(id);
  var dopts = {html: true};
  opts = $.extend(opts, dopts);

  if (type == 'tooltip') {
    $id.tooltip('destroy');
    setTimeout(function() { $id.tooltip(opts); }, 200);
  } else if(type == 'popover') {
    $id.popover('destroy');
    setTimeout(function() { $id.popover(opts); }, 200);
  }
}"

log2_print <- function(x, msg = "") {
  if (str_length(msg)>0)
    log_print(msg, console = FALSE, blank_after = FALSE, hide_notes =TRUE)
  log_print(x, console = FALSE, blank_after = FALSE, hide_notes =FALSE)
}
#log2_print <- function(x, msg = "") {}

### compare_named_lists --
###
### Notes:
### MA: Seems a bit complicated.  What is this supposed to do?

compare_named_lists <- function(l1, l2) {
  if (is.list(l1) && is.list(l2)) { # if they are lists
    if (length(unique(names(l1))) == length(l1) &&
        length(unique(names(l2))) == length(l2) &&
        !("" %in% names(l1)) &&
        !("" %in% names(l2))
    ) { # if they are named
      st <- TRUE
      for (i in unique(c(names(l1), names(l2)))) {

        ## If the same name is present in both..
        ##
        ## Notes:
        ## MA: are we sure there is no 'intersection' operation on
        ## lists in R?

        if (i %in% names(l1) && i %in% names(l2)) {
          if (is.list(l1[[i]]))
            st <- st && compare_named_lists(l1[[i]], l2[[i]])
          else
            st <- st && identical(l1[[i]], l2[[i]])
        }
        else
          return(FALSE)
      }
      return(st)
    }
    return(FALSE)
  }
  else
    return(FALSE)
}


### Help Management
### ===============

source("info_help.R")

### make_shiny_help_popover --

make_shiny_help_popover <- function(for_object,
                                    pop_title,
                                    pop_text,
                                    with_icon = shiny_iconlink(),
                                    server = FALSE) {
  with_icon <- with_icon %>%
    bs_embed_popover(title = pop_title,
                     content = pop_text,
                     placement = "right",
                     trigger = "hover",
                     delay = list("show" = 200, "hide" = 3000),
                     html = TRUE
    )

  for_object <- for_object %>% shinyInput_label_embed(with_icon)
  if (server)
    for_object <-
    tagList(for_object,
            tags$script("$(function () {$('[data-toggle=\"popover\"]').popover()})")
    )
  return(for_object)
}


### make_title_help_popover

make_title_help_popover <- function(for_object,
                                    near_label,
                                    pop_title,
                                    pop_text,
                                    with_icon = shiny_iconlink(),
                                    server = FALSE) {
  with_icon <- with_icon %>%
    bs_embed_popover(title = pop_title,
                     content = pop_text,
                     placement = "right",
                     trigger = "hover",
                     delay= list("show"= 200, "hide"= 3000),
                     html = TRUE
    )
  with_icon <- htmltools::div(class = "pull-right", with_icon)

  tag <- htmltools::div(class = "form-group shiny-input-container",
                        near_label,
                        br(),
                        for_object)

  tag$children[[1]] <-
    tag$children[[1]] %>%
    htmltools::tagAppendChild(with_icon) %>%
    htmltools::tagAppendAttributes(style = "width:100%;")

  if (server)
    tag <-
    tagList(tag,
            tags$script("$(function () {$('[data-toggle=\"popover\"]').popover()})")
    )
  return(tag)
}


### inline --

inline = function (x) {
  tags$div(style="display:inline-block;", x)
}


### ui -- ####

ui <- fluidPage(
  ## singleton(tags$head(tags$script(src = "bs.js"))),

  useShinyjs(),
  withMathJax(),


  ## extendShinyjs(text = jsCode, functions = c("addTooltip")),
  ## extendShinyjs(text = jsCode, functions = c("bsPopover")),

  

  tags$head(tags$style("@import url(https://use.fontawesome.com/releases/v5.7.2/css/all.css);"),
            tags$style(HTML(".bucket-list-container {min-height: 290px; max-height: 300px;}")),
            tags$style(HTML(".rank-list {min-height: 100px; max-height: 240px; overflow-y: scroll;}")),
            #tags$style(HTML(".shinyDirectories, .shinyFiles { width: 100%; direction: rtl; white-space: nowrap; overflow: hidden; text-overflow: ellipsis;}")),
            tags$style(HTML(".shinyDirectories, .shinyFiles { width: 100%; white-space: nowrap; overflow: hidden; text-overflow: ellipsis;}")),
            tags$style(type="text/css", "body {padding-bottom: 70px;}"),
            tags$style(HTML(".content-wrapper, .right-side {background-color: #ffffff;}"))
  ),

  use_bs_popover(),
  
  dashboardPage(
    dashboardHeader(
      title = "LACE 2.0",
      tags$li(class = "dropdown",
              tags$a(uiOutput("pr_title_ui_h")
              )
      )
    ),

    dashboardSidebar(

      sidebarMenu( id = "sidemenu",
        hidden(menuItem("hidden_menu31", tabName = "hidden_menu31")),
        menuItem(
          "Application",
          tabName = "application",
          icon = icon("braille"), #icon("window-maximize"),
          hidden(menuItem("hidden_menu3", tabName = "hidden_menu3")),
          menuSubItem("New Project", tabName = "new_proj", icon = icon("layer-group") ),
          menuSubItem("Quit", tabName = "quit_app", icon = icon("power-off") )#,
          #actionButton("quit_app", "Quit")
        ),
        hidden(menuItem("hidden_menu", tabName = "hidden_menu")),
        menuItem(
          "Projects",
          tabName = "Projects",
          icon = icon("cubes"),
          #expandedName = "ttt",
          menuItemOutput("recent_projects"),
          hidden(menuItem("hidden_menu2", tabName = "hidden_menu2")),
          # menuItem(
          #   "Recent Projects",
          #   tabName = "Recent Projects",
          #   icon = icon("th"),
          #   #expandedName = "projects",
          #   menuSubItem(
          #     "test22",
          #     tabName = ",media,disco,work_folder,Downloads,gitlace,LACEinterface,test2",
          #     #icon = icon("th"),
          #     #inline(actionButton("save_tab", "save")),
          #     #inline(actionButton("load_tab", "load"))
          #   ),
          #   menuSubItem(
          #     "test2",
          #     tabName = ",media,disco,work_folder,Downloads,gitlace,LACEinterface,test2",
          #     #icon = icon("th"),
          #     #inline(actionButton("save_tab", "save")),
          #     #inline(actionButton("load_tab", "load"))
          #   )#,
          #   #hidden(menuItem("hidden_menu4", tabName = "hidden_menu4"))
          # ),
          menuItem(
            "Demo Projects",
            tabName = "Demo Projects",
            icon = icon("database"),
            #expandedName = "demos",
            menuSubItem(
              "Melanoma dataset (Rambow)",
              tabName = "Melanoma dataset",
              #icon = icon("th"),
              #inline(actionButton("save_tab", "save")),
              #inline(actionButton("load_tab", "load"))
            ),
            menuSubItem(
              "Small dataset",
              tabName = "Small dataset",
              #icon = icon("th")#,
              #inline(actionButton("save_tab", "save")),
              #inline(actionButton("load_tab", "load"))
            )#,
            #hidden(menuItem("hidden_menu3", tabName = "hidden_menu3"))
          )#,
          #hidden(menuItem("hidden_menu1", tabName = "hidden_menu1"))
        ),
        menuItem(
          "Current Project",
          tabName = "dashboard",
          icon = icon("cube"),
          hidden(menuItem("hidden_menu2", tabName = "hidden_menu2")),
          #menuSubItem("Close Project", tabName = "close_proj", icon = icon("times-circle") ),
          #actionButton("close_proj", "Close Project"),
          menuItem(
            "Save and Load Current Tab",
            tabName = "dashboard_save",
            icon = icon("th"),
            hidden(menuItem("hidden_menu_SLpre", tabName = "hidden_menu_SLpre")),
            menuSubItem("Save", tabName = "save_tab", icon = icon("download") ),
            menuSubItem("Load", tabName = "load_tab", icon = icon("upload") ),
            menuSubItem("Reset project", tabName = "load_tabs"),
            hidden(menuSubItem("hidden_menu_SL", tabName = "hidden_menu_SL"))
            #inline(actionButton("save_tab", "Save")),
            #inline(actionButton("load_tab", "Load"))
          )
        )#,
        #,
          # menuItem(
          #   "Save and Load All Tabs",
          #   tabName = "dashboard",
          #   icon = icon("th"),
          #   inline(actionButton("save_tab", "save")),
          #   inline(actionButton("load_tab", "load"))
          # ),
          # menuItem(
          #   "Reset Tabs",
          #   tabName = "dashboard",
          #   icon = icon("th"),
          #   inline(actionButton("save_tab", "Clean and restore "))
          # )
        
      ),
      collapsed = FALSE
    ),

    dashboardBody(
      tabItems(
        tabItem(tabName = "hidden_menu"),
        tabItem(tabName = "Small dataset"),
        tabItem(tabName = "Melanoma dataset"),
        tabItem(tabName = "test22"),
        tabItem(tabName = "test2")
      ),
      div(id = "computation_idCol_div",
          h3("Processing ... please wait")
      ),

      fluidRow(
        column(width = 12,
               tabsetPanel(id = "main_tabset",
                           type = "tabs",

                           ## Project tab
                           tabPanel(
                             "Project",
                             br(),
                             uiOutput("pr_title_ui"),
                             br(),
                             helpText(text[["pr_tab_help"]]),
                             br(),br(),


                             shinyDirButton('pr_folder',
                                            'Select folder',
                                            'Please select the project root folder', FALSE) %>%
                               make_title_help_popover (
                                 tags$b("Project root folder"),
                                 "Project folder",
                                 text[["pr_folder"]]
                               ),
                             br(),

                             textInput("pr_name",
                                       "Name",
                                       placeholder = "Your project name") %>%
                               make_shiny_help_popover(
                                 "Project name",
                                 text[["pr_name"]]
                               ),
                             br(),br(),

                             div(id = "pr_path_div",
                                 uiOutput("pr_path_ui")
                             ),

                             actionButton("pr_next", "Create project"),

                             br(),br(),

                           ),


                           ## Metadata tab
                           tabPanel(
                             "SC metadata",
                             br(),
                             tags$h3("Single cell metadata info"),
                             br(),
                             helpText(text[["m_tab_help"]]),
                             br(), br(),
                             shinyFilesButton('sc_metadata_file',
                                              'Select file',
                                              'Please select the experiment metadata file', FALSE) %>%
                               make_title_help_popover (
                                 tags$b("Metadata file"),
                                 "Metadata info file",
                                 text[["sc_metadata_file"]],
                                 server = TRUE
                               ),
                             br(),
                             br(),
                             div(id = "m_idCol_div",
                                 uiOutput("m_idCol")
                             ),

                             tableOutput("m_id_column_tab"),
                             br(),

                             div(id="m_timePointsCol_div",
                                 uiOutput("m_timePointsCol")
                             ),

                             div(id="m_timePoints_div",
                                 uiOutput("m_timePoints")
                             ),
                             br(), br(),

                           ),

                           ## Annotation tab
                           tabPanel(
                             "Annotations",
                             br(),
                             tags$h3("Gene variant annotations"),
                             br(),
                             helpText(text[["av_tab_help"]]),
                             br(), br(),

                             ## runjs(jsCode),
                             shinyDirButton('av_anovar_exec_dir',
                                            'Select folder',
                                            'Please select Annovar executable folder',
                                            FALSE) %>%
                               make_title_help_popover (
                                 tags$b("Annovar executable folder"),
                                 "Annovar",
                                 text[["av_anovar_exec_dir"]]
                               ),
                             br(),
                             br(),

                             shinyDirButton('av_anovar_db_dir',
                                            'Select folder',
                                            'Please select Annovar reference database folder',
                                            FALSE) %>%
                               make_title_help_popover (
                                 tags$b("Annovar reference database"),
                                 "Annovar reference database",
                                 text[["av_anovar_db_dir"]]
                               ),
                             br(),
                             br(),

                             shinyDirButton('av_vcf_in_dir',
                                            'Select folder',
                                            'Please select VCF folder',
                                            FALSE) %>%
                               make_title_help_popover (
                                 tags$b("Variant calling files folder"),
                                 "Variant calling files folder",
                                 text[["av_vcf_in_dir"]]
                               ),
                             br(),
                             br(),

                             ## tags$b("Output folder"),
                             ## br(),
                             ## shinyDirButton('av_out_dir', 'Select folder', 'Please select output folder', FALSE),
                             br(),
                             br(),

                             tags$b("Run"),
                             br(),
                             actionButton("av_exec",
                                          "Annotate variants"),

                             # tags$span(icon("info-circle"),
                             #           id = "icon_av_exec",
                             #           style = "color: blue;"),
                             # bsPopover("icon_av_exec",
                             #           "Load inputs",
                             #           text[["av_exec"]],
                             #           placement = "right",
                             #           trigger = "hover",
                             #           options = list("delay':  {show : 500, hide : 2000}, 'a" = "show")),
                             br(),br(),
                             ## tags$b("Result"),
                             ## br(),
                             ## verbatimTextOutput("av_anovar_exec_dir"),
                             ## verbatimTextOutput("av_anovar_db_dir"),
                             ## verbatimTextOutput("av_vcf_out_dir")
                           ),

                           ## Filters tab
                           tabPanel(
                             "Filters",
                             br(),
                             tags$h3("Quality filters"),
                             br(),
                             helpText(text[["thr_tab_help"]]),

                             br(), br(),
                             numericInput(
                               inputId = "thr_alleles_ratio",
                               label = "Alternate frequency",
                               min = 1,
                               max = 500,
                               step = 1,
                               value = 2
                             ) %>%
                               make_shiny_help_popover(
                                 "Cell alternate allele frequency",
                                 text[["thr_alleles_ratio"]]
                               ),

                             numericInput(
                               inputId = "thr_maf",
                               label = "MAF",
                               min = 0.01,
                               max = 0.5,
                               step = 0.005,
                               value = 0.01
                             ) %>%
                               make_shiny_help_popover(
                                 "Minor alleles frequency",
                                 text[["thr_maf"]]
                               ),

                             numericInput(
                               inputId = "thr_freq",
                               label = "Variant frequency",
                               min = 0.,
                               max = 1,
                               step = 0.01,
                               value = 0.01
                             ) %>%
                               make_shiny_help_popover(
                                 "Sample variant frequency",
                                 text[["thr_freq"]]
                               ),
                             br(),

                             div(id = "thr_bucket_var_list_div",
                                 uiOutput("thr_bucket_var_list")
                             ),
                             br(),br(),br(),br(),


                             tags$b("Run"),
                             br(),
                             actionButton("thr_exec", "Apply filters"),
                             br(),br(),
                             ## tags$b("Result"),
                             ## verbatimTextOutput("thr_filters"),
                             ## verbatimTextOutput('thr_accepted_var'),
                             ## verbatimTextOutput('thr_negleted_var')
                           ),

                           ## Depth tab
                           tabPanel(
                             "SC sampling depths",
                             br(),
                             tags$h3("Single cell sampling depth"),
                             br(),
                             helpText(text[["dp_tab_help"]]),
                             br(), br(),

                             shinyDirButton('dp_samtools_exec_dir',
                                            'Select folder',
                                            'Please select samtools executable folder',
                                            FALSE) %>%
                               make_title_help_popover (
                                 tags$b("Samtools executable folder"),
                                 "Samtools executable folder",
                                 text[["dp_samtools_exec_dir"]]
                               ),
                             br(), br(),

                             shinyDirButton('dp_bam_dir',
                                            'Select folder',
                                            'Please select folder containing the single cell BAM files',
                                            FALSE) %>%
                               make_title_help_popover (
                                 tags$b("BAMs folder"),
                                 "BAMs folder",
                                 text[["dp_bam_dir"]]
                               ),
                             br(), br(),

                             tags$b("Output folder"),
                             br(), br(),

                             tags$b("Run"),
                             br(),
                             actionButton("dp_exec", "Compute depth"),
                             br(), br(),
                             ## tags$b("Result"),
                             ## br(),
                             ## verbatimTextOutput("dp_par"),
                           ),

                           ## Variants tab
                           tabPanel(
                             "Variants",
                             br(),
                             tags$h3("Variant filters"),
                             br(),
                             helpText(text[["va_tab_help"]]),
                             fluidRow(column(4,
                             br(), br(),

                             numericInput(
                               inputId = "va_depth_minimum",
                               label = "Minimum depth",
                               min = 0,
                               max = 500,
                               step = 1,
                               value = 3
                             ) %>%
                               make_shiny_help_popover(
                                 "Minimum depth",
                                 text[["va_depth_minimum"]]
                               ),
                             br(),

                             numericInput(
                               inputId = "va_missing_values_max",
                               label = "Max missing value",
                               min = 0.0,
                               max = 1.0,
                               step = 0.05,
                               value = 0.4
                             ) %>%
                               make_shiny_help_popover(
                                 "Maximumax missing value",
                                 text[["va_missing_values_max"]]
                               ),
                             br(),

                             numericInput(
                               inputId = "va_minumum_median_total", # minimum median depth for total reads
                               label = "Site minumum median depth:",
                               min = 0,
                               max = 500,
                               step = 1,
                               value = 8
                             ) %>%
                               make_shiny_help_popover(
                                 "Minumum median depth per site",
                                 text[["va_minumum_median_total"]]
                               ),
                             br(),

                             numericInput(
                               inputId = "va_minumum_median_mutation", # minimum median depth for reads supporting mutations
                               label = "Mutation minimum median depth",
                               min = 0,
                               max = 500,
                               step = 1,
                               value = 4
                             ) %>%
                               make_shiny_help_popover(
                                 "Minimum median depth per mutation",
                                 text[["va_minumum_median_mutation"]]
                               ),
                             br(),
                             br(),
                             br(),
                             
                             selectizeInput(
                               'va_verified_genes',
                               'Select variant genes',
                               NULL,
                               multiple = TRUE,
                               options = list (
                                 placeholder = 'Choose genes for inferential analysis',
                                 hideSelected = FALSE,
                                 plugins =
                                   list("drag_drop",
                                        "remove_button",
                                        "dropdown_header",
                                        "restore_on_backspace"),
                                 dropdown_header =
                                   list(title = 'a')
                               )
                             ) %>%
                               make_shiny_help_popover(
                                 "Variant gene selection",
                                 text[["va_verified_genes"]]
                               ),
                             ),
                             column(8,
                                    br(),
                                    br(),
                                    #plot
                                    plotOutput('va_filtered_bin_hmap', height = 600)
                             )
                           ),
                           br(),
                           br(),
                           tags$b("Overview of post-filtered mutations"),
                           br(),
                           br(),
                           DTOutput("va_out"),

                           br(), br(),
                           tags$b("Run"),
                           br(),
                           actionButton("va_exec", "Select variants"),
                           br(),
                           br(),
                           ## tags$b("Result"),
                           ## verbatimTextOutput("va_filters")
                             
                             
                           ),


                           ## Inference tab
                           tabPanel(
                             "Inference",
                             br(),
                             tags$h3("Longitudinal tree inference parameters"),
                             br(),
                             helpText(text[["inf_tab_help"]]),

                             br(), br(),
                             numericInput(
                               inputId = "inf_learning_rate",
                               label = "Learning rate",
                               min = 0.01,
                               max = 10,
                               step = 0.1,
                               value = 1
                             ) %>%
                               make_shiny_help_popover(
                                 "Learning rate",
                                 text[["inf_learning_rate"]]
                               ),
                             br(),

                             div(id = "inf_alpha_div",
                                 tagList(
                                   br(),
                                   span() %>%
                                     make_title_help_popover (
                                       tags$b("False positive rates"),
                                       "False positive rate",
                                       text[["inf_alpha"]]
                                     ),
                                   span("Insert a set of false positive rates, one \\(\\alpha\\) for each sampling time point. Fill the last row to add further sets of rates to be used in the inference."),
                                   DTOutput('inf_alpha', width = '100%'),
                                   br()
                                 )
                             ),
                             ## br(),

                             div(id = "inf_beta_div",
                                 tagList(
                                   br(),
                                   span() %>%
                                     make_title_help_popover (
                                       tags$b("False negative rates"),
                                       "False negative rate",
                                       text[["inf_beta"]]
                                     ),
                                   span("Insert a set of false negative rates, one \\(\\beta\\) for each sampling time point. Fill the last row to add further sets of rates to be used in the inference."),
                                   DTOutput('inf_beta', width = '100%'),
                                   br()
                                 )
                             ),
                             br(), br(),

                             numericInput(
                               inputId = "inf_num_iter",
                               label = "Number of iterations",
                               min = 100,
                               max = 4000,
                               step = 100,
                               value = 10000
                             ) %>%
                               make_shiny_help_popover(
                                 "Number of iterations",
                                 text[["inf_num_iter"]]
                               ),
                             br(),

                             numericInput(
                               inputId = "inf_num_rs",
                               label = "Number of restarts",
                               min = 1,
                               max = 100,
                               step = 10,
                               value = 50
                             ) %>%
                               make_shiny_help_popover(
                                 "Number of restarts",
                                 text[["inf_num_rs"]]
                               ),
                             br(),

                             numericInput(
                               inputId = "inf_n_try_bs",
                               label = "Early stopping",
                               min = 100,
                               max = 1000,
                               step = 100,
                               value = 500
                             ) %>%
                               make_shiny_help_popover(
                                 "Early stopping",
                                 text[["inf_n_try_bs"]]
                               ),
                             br(),

                             numericInput(
                               inputId = "inf_num_processes",
                               label = "Number of parallel processes",
                               min = 1,
                               max = 100,
                               step = 5,
                               value = 10
                             ) %>%
                               make_shiny_help_popover(
                                 "Number of parallel processes",
                                 text[["inf_num_processes"]]
                               ),
                             br(),

                             numericInput(
                               inputId = "inf_seed",
                               label = "Seed",
                               min = 0,
                               max = NA,
                               step = 1000,
                               value = 1121
                             ) %>%
                               make_shiny_help_popover(
                                 "Seed",
                                 text[["inf_seed"]]
                               ),
                             br(),


                             checkboxInput("inf_random_tree",
                                           tags$b("Random tree intialization"),
                                           FALSE) %>%
                               make_shiny_help_popover(
                                 "Random tree intialization",
                                 text[["inf_random_tree"]]
                               ),

                             checkboxInput("inf_marginalize",
                                           tags$b("Marginalize"),
                                           FALSE) %>%
                               make_shiny_help_popover(
                                 "Marginalize",
                                 text[["inf_marginalize"]]
                               ),

                             checkboxInput("inf_keep_equivalent",
                                           tags$b("Keep equivalent"),
                                           FALSE) %>%
                               make_shiny_help_popover(
                                 "Keep equivalent",
                                 text[["inf_keep_equivalent"]]
                               ),

                             checkboxInput("inf_check_indistinguishable",
                                           tags$b("Check indistinguishable"),
                                           FALSE) %>%
                               make_shiny_help_popover(
                                 "Check indistinguishable",
                                 text[["inf_check_indistinguishable"]]
                               ),

                             checkboxInput("inf_error_move",
                                           tags$b("Error move"),
                                           FALSE) %>%
                               make_shiny_help_popover(
                                 "Error move",
                                 text[["inf_error_move"]]
                               ),

                             checkboxInput("inf_show",
                                           tags$b("Show results in interface")) %>%
                               make_shiny_help_popover(
                                 "Show results in interface",
                                 text[["inf_show"]]
                               ),

                             br(), br(),br(),
                             tags$b("Run"),
                             br(),
                             actionButton("inf_next", "Run LACE"),
                             br(), br(),
                             tags$b("Result"),
                             verbatimTextOutput("inf_par")
                           ),

                           ## results ####
                           tabPanel(
                             "Longitudinal display",
                             br(),

                             inline(actionLink("res_refresh", "Refresh page")),
                             inline(textOutput("res_lastrefresh")),
                             br(),
                             # #widgetframeOutput('LCT_FP', width = "1000px", height = "800px"),
                             # #LACE:::LACEOutput('LCT_FP'),#, width = 910, height = 800),
                             # {
                             #   #x <-LACE:::LACEOutput('LCT_FP')
                             #   #htmlwidgets::saveWidget(x,"LACE_result.html")
                             #   NULL
                             #   #x
                             # },
                             uiOutput("res"),
                             br(),br(),
                           )
               )
        )
      )
    )
  )
)


### server -- ####

server <- function(input, output, session) {

  setwd(.my_pkg_dir) # per non disperdersi

  inputs <- list()
  inputs[["null_reactiveVal"]] <- reactiveVal("")
  
  observe({
    input$null_reactiveVal > 0
  }, )
  
  types_ <- reactiveVal()

  session$onSessionEnded(function() {
    stopApp()
  })

  ## Source?

  source("check_demo.R", local = TRUE)
  source("load_and_save.R", local = TRUE)
  source("project_mgr.R", local = TRUE)
  source("pipeline_io_ctrl.R")
  source('alpha_beta_table.R', local = TRUE)
  source("annotation.R")
  source("filters_computation.R", local = TRUE)
  source("depth_computation.R", local = TRUE)
  source("step3_explore_lace_input_data.R")
  source("make_lace_final_input.R")



  default_yaml <-
    list(".config_06_inf.yml" = list(),
         ".config_05_va.yml" = list(),
         ".config_04_dp.yml" = list(),
         ".config_03_thr.yml" = list(),
         ".config_02_av.yml" = list(),
         ".config_01_m.yml" = list()
    )

  os_hidden <- "."
  os_conf <- list()
  os_conf[["windows"]] <- "AppData"
  os_conf[["mac"]] <- "Library"
  os_conf[["unix"]] <- ".config"
  os_conf_subdir = "config"
  config_path = os_conf_subdir # to remove


  ## roots_dir=c(root = file.path('..'), "working_dir" = file.path('.'), getVolumes()())
  roots_dir = c("project dir" = file.path("."),
                ".." = file.path(".."),
                "../.." = file.path("..",".."),
                "../../.." = file.path("..","..",".."),
                getVolumes()())

  ## runjs(jsCode)
  ## runjs("console.log(shinyBS.addTooltip)")
  ## js$addTooltip("av_anovar_db_dir", "IL BOTTONE", "")
  ## addPopover(session, "av_anovar_db_dir", "IL BOTTONE2", "e il polpettone", placement = "right")

  observe({
    #browser()
    if (!is.null(input$close)) {
      if (input$close > 0) stopApp()                             # stop shiny
    }
  }, )

  m_loaded_input_ <- reactiveVal()
  av_loaded_input_ <- reactiveVal()
  thr_loaded_input_ <- reactiveVal()
  dp_loaded_input_ <- reactiveVal()
  va_loaded_input_ <- reactiveVal()
  inf_loaded_input_ <- reactiveVal()


  rs_ <- reactiveVal()
  inputs[["res_lastrefresh"]] <- reactiveVal("")
  stime <- reactiveVal(2)
  port <- reactiveVal()
  sopt <- reactiveVal(callr::r_session_options())


  av_grep_str= "^av_|^`av_|_av_"
  thr_grep_str= "^thr_|^`thr_|_thr_"
  dp_grep_str= "^dp_|^`dp_|_dp_"
  va_grep_str= "^va_|^`va_|_va_"
  inf_grep_str= "^inf_|^`inf_|_inf_"
  m_grep_str= "^m_|^`m_|_m_|sc"

  inf_uis = c("inf_num_iter",
              "inf_num_rs",
              "inf_n_try_bs",
              "inf_num_processes",
              "inf_seed",
              "inf_random_tree",
              "inf_marginalize",
              "inf_keep_equivalent",
              "inf_check_indistinguishable",
              "inf_show",
              "inf_learning_rate",
              "inf_error_move")

  va_uis= c('va_depth_minimum',
            'va_missing_values_max',
            'va_minumum_median_total',
            'va_minumum_median_mutation')
  thr_uis = c('thr_alleles_ratio',
              'thr_maf',
              'thr_freq',
              'thr_accepted_var',
              'thr_negleted_var')
  dp_uis= c('dp_samtools_exec_dir', 'dp_bam_dir')
  av_dir_uis = c('av_anovar_exec_dir',
                 'av_anovar_db_dir',
                 'av_vcf_in_dir')




  source("project_tab.R", local = TRUE)
  source("inference_tab.R", local = TRUE)


  setwd(.my_actual_wd)

  ### Dashboard reactiveVals ####
  inputs[["demo"]] <- reactiveVal()
  inputs[["reload_project"]] <- reactiveVal()
  inputs[["demo_remainder"]] <- reactiveVal()
  inputs[["long_job"]] <- reactiveVal(0)



  ### End dashboard reactiveVals ####

  ### Dashboard observers ####

  shinyjs::hide(id="computation_idCol_div")


  observeEvent(inputs[["reload_project"]](), {

    proj_dir <- inputs[["reload_project"]]()

    if (inputs[["reload_project"]]() != 1 && inputs[["reload_project"]]() != 0)
    {
      if (dir.exists(proj_dir)) {
        #browser()
        hide_tab()
        inputs[["reload_project"]](1)
        inputs[["pr_path"]](proj_dir)
      }
      else
        inputs[["reload_project"]](NULL)
    }
    #else
    #  inputs[["reload_project"]](NULL)
  })

  observeEvent( inputs[["demo"]](),{
    if(inputs[["demo"]]() == "Small_dataset" || inputs[["demo"]]() == "Rambow_dataset")
    {
      #browser()
      if(!check_demo()) {
        inputs[["demo"]](NULL)
        toggle_inputs(is.null(inputs[["demo"]]()))
        showNotification("Impossible to download demos.", duration = 10, type = "warning")
      }
      else
      {
        #browser()
        demo_dir<-file.path(.my_pkg_dir,inputs[["demo"]]())
  
        # create tmp folder
        tmp_dir <- tempdir()
        tmp_path <- file.path(tmp_dir,inputs[["demo"]]())
        dir.create(tmp_path, showWarnings = FALSE)
  
        inputs[["demo_remainder"]](tmp_path)
  
        # fill tmp folder
        if (dir.exists(tmp_path) && dir.exists(demo_dir))
        {
          #delay(100, {
            #browser()
            hide_tab()
          #})
  
          dir_copy(demo_dir, tmp_path, overwrite = TRUE)
  
          inputs[["demo"]](1) # temporary project ready
          inputs[["pr_path"]](tmp_path)
  
        }
        else
        {
          #browser()
          inputs[["demo"]](NULL)
          toggle_inputs(is.null(inputs[["demo"]]()))
        }
        # disable tab
        #shinyjs::addClass()
        #shinyjs::disable(selector = '.nav-tabs a[data-value="SC metadata"')
        #shinyjs::disable(selector = '.navbar-nav a[data-value="Annotations"')
        # load output
      }
    }
    #else
    #  inputs[["demo"]](NULL)

  }, ignoreInit = TRUE)


  # observeEvent(input[["sidebarItemExpanded"]], {
  #   print(input[["sidebarItemExpanded"]])
  #
  #   if(input$sidebarItemExpanded == "projects") {
  #
  #   }
  # })


  createmenuitem <- function(x, projs) {
    #browser()
    if (dir.exists(projs[[x]]))
      menuSubItem(
        text = names(projs)[x],
        tabName = str_replace_all(
          normalizePath(projs[[x]]),
          pattern = .Platform$file.sep,
          replacement = ","
        )
      )
    else
      NULL
  }

  observeEvent(input[["sidemenu"]], {
    #browser()
    projs <- update_proj_list(NULL)

    # createmenuitem <- function(x) {
    #   menuSubItem(
    #     text = names(projs)[x],
    #     tabName = str_replace_all(
    #       normalizePath(projs[[x]]),
    #       pattern = .Platform$file.sep,
    #       replacement = ","
    #     )
    #   )
    # }

    submenus <- lapply(seq_along(projs), FUN = function(x) {createmenuitem(x,projs)})
    recent_projects <- menuItem("Recent Projects", submenus,
               icon = icon("history"))

    output$recent_projects <- renderMenu({recent_projects})
    log2_print(recent_projects, msg = "LACEview: recent projects =")

  }, once = TRUE )

  observeEvent( input[["sidemenu"]],{
    log2_print(input[["sidemenu"]], msg = "LACEview:")
    #browser()
    if (input[["sidemenu"]] == "Small dataset") {
      inputs[["demo"]]("Small_dataset")
    }
    if (input[["sidemenu"]] == "Melanoma dataset") {
      inputs[["demo"]]("Rambow_dataset")
    }

    if (input[["sidemenu"]] != "Small dataset" && input[["sidemenu"]] != "Melanoma dataset") {
      proj_dir <- str_replace_all(input[["sidemenu"]], pattern = ",", replacement = .Platform$file.sep)

      inputs[["reload_project"]](proj_dir)
    }
    
    if (input[["sidemenu"]] == "save_tab")
      save_tab()
    if (input[["sidemenu"]] == "load_tab")
      load_tab()
    if (input[["sidemenu"]] == "load_tabs")
      load_tabs()
    
    if (input[["sidemenu"]] == "quit_app")
      stopApp()
    if (input[["sidemenu"]] == "new_proj"){
      #inputs[["pr_path"]]("")
      print("NEW project!!")
      inputs[["pr_folder"]]("")
      inputs[["pr_folder_std"]](.my_actual_wd)
      inputs[["pr_name"]]("")
      inputs[["pr_name_std"]]("")
      
      inputs[["project_folder_std"]](.my_actual_wd)
      project_folder <- list( "path"="", "root"="")
      project_folder$path[[1]]<-""
      project_folder$path <-
        c(project_folder$path,
          as.list(path_split(path_rel(inputs[["project_folder_std"]](),
                                      start = roots_dir[[".."]]))[[1]]))
      project_folder$root=".."
      inputs[["project_folder"]](project_folder)
      
      updateActionButton(session,"pr_folder", label=inputs[["pr_folder_std"]]())
      updateTextInput(session,"pr_name", value=inputs[["pr_name_std"]]())
      #inputs[["pr_path"]](.my_actual_wd) #not necessary
    }
  }, ignoreInit = TRUE)

  # observe({
  #   if (input[["sidemenu"]] == "Small dataset") {
  #     inputs[["demo"]]("Small dataset")
  #   }
  #   if (input[["sidemenu"]] == "Melanoma dataset") {
  #     inputs[["demo"]]("Melanoma dataset")
  #   }
  # })

  save_tab <- function () {

    if(!inputs[["project_loaded"]]()) {
      showNotification("No porject created or loaded.", duration = 10, type = "warning")
      return(NULL)
    }

    config_path <- file.path(inputs[["project_folder_std"]](), os_conf_subdir)

    grep_str <- NULL
    #browser()
    if (input[["main_tabset"]] == "SC metadata") {
      grep_str <- m_grep_str
      config_file <- ".config_01_m.yml"
    } else if (input[["main_tabset"]] == "Annotations") {
      grep_str <- av_grep_str
      config_file <- ".config_02_av.yml"
    } else if (input[["main_tabset"]] == "Filters") {
      grep_str <- thr_grep_str
      config_file <- ".config_03_thr.yml"
    } else if (input[["main_tabset"]] == "SC sampling depths") {
      grep_str <- dp_grep_str
      config_file <- ".config_04_dp.yml"
    } else if (input[["main_tabset"]] == "Variants") {
      grep_str <- va_grep_str
      config_file <- ".config_05_va.yml"
    } else if (input[["main_tabset"]] == "Inference") {
      grep_str <- inf_grep_str
      config_file <- ".config_06_inf.yml"
    }

    if (!is.null(grep_str)) {
      doSave(grep_str, config_path, config_file)

      showNotification(paste(input[["main_tabset"]], "configuration tab saved"),
                       duration = 10)
    }
    updateTabItems(session, "sidemenu", "hidden_menu")
  }




  load_tab <- function() {

    if(!inputs[["project_loaded"]]()) {
      showNotification("No porject created or loaded.", duration = 10, type = "warning")
      return(NULL)
    }

    if(input[["main_tabset"]] == "Project")
      return(NULL)

    config_path <- file.path(inputs[["project_folder_std"]](), os_conf_subdir)

    if (input[["main_tabset"]] == "SC metadata") {
      m_doLoad_a(config_path)
      m_doLoad_b()

    } else if (input[["main_tabset"]] == "Annotations") {
      av_doLoad_a(config_path)
      uis <- av_dir_uis
      for (i in av_dir_uis) {
        av_doLoad_b(i)
      }
    } else if (input[["main_tabset"]] == "Filters") {
      thr_doLoad_a(config_path)
      uis <- thr_uis
      for(i in thr_uis) {
        thr_doLoad_b(i)
      }
    } else if (input[["main_tabset"]] == "SC sampling depths") {
      dp_doLoad_a(config_path)
      uis <- dp_uis
      for(i in dp_uis) {
        dp_doLoad_b(i)
      }

    } else if (input[["main_tabset"]] == "Variants") {
      va_doLoad_a(config_path) #
      uis <- va_uis
      for(i in va_uis) {
        va_doLoad_b(i)

      }
      va_doLoad_c()
    } else if (input[["main_tabset"]] == "Inference") {
      browser()
      inf_doLoad_a(config_path)
      uis <- inf_uis
      for(i in uis) {
        inf_doLoad_b(i)
      }
      inf_doLoad_c()
    }

    showNotification(paste(input[["main_tabset"]], "configuration tab loaded"), duration = 10)
    
    updateTabItems(session, "sidemenu", "hidden_menu")
  }

  load_tabs <- function() {
    
    if(!inputs[["project_loaded"]]()) {
      showNotification("No porject created or loaded.", duration = 10, type = "warning")
      return(NULL)
    }
    
    config_path <- file.path(inputs[["project_folder_std"]](), os_conf_subdir)
    m_doLoad_a(config_path)
    m_doLoad_b()
      
    av_doLoad_a(config_path)
    uis <- av_dir_uis
    for (i in av_dir_uis) {
      av_doLoad_b(i)
    }
    
    thr_doLoad_a(config_path)
    uis <- thr_uis
    for(i in thr_uis) {
      thr_doLoad_b(i)
    }
    
    dp_doLoad_a(config_path)
    uis <- dp_uis
    for(i in dp_uis) {
      dp_doLoad_b(i)
    }
      
    va_doLoad_a(config_path) #
    uis <- va_uis
    for(i in va_uis) {
      va_doLoad_b(i)
    }
    va_doLoad_c()
    
    browser()
    inf_doLoad_a(config_path)
    uis <- inf_uis
    for(i in uis) {
      inf_doLoad_b(i)
    }
    inf_doLoad_c()
    
    showNotification(paste("project", "configuration reset to saved state"), duration = 10)
    
    updateTabItems(session, "sidemenu", "hidden_menu")
  }


  catch_ui_files_priority = 1

  observe({
    save_ui_state(m_grep_str, "config_01_m.yml")
    save_ui_state(av_grep_str, "config_02_av.yml")
    save_ui_state(thr_grep_str, "config_03_thr.yml")
    save_ui_state(dp_grep_str, "config_04_dp.yml")
    save_ui_state(va_grep_str, "config_05_va.yml")
    save_ui_state(inf_grep_str, "config_06_inf.yml")
  },
  priority = catch_ui_files_priority)

  ### End dashboard observers ####

  update_proj_list <- function (project_folder) {
    #browser()
    os_conf_tmp <- file.path(path_home(), os_conf[[.Platform$OS.type]], "LACE")
    if(dir.exists(os_conf_tmp)) {
      if(file.exists(file.path(os_conf_tmp, "projects_list.yml")))
        projects_list <- read.config(file = file.path(os_conf_tmp, "projects_list.yml"))
      else
        projects_list <- list()
    }
    else
      projects_list <- list()

    if(!is.list(projects_list))
      projects_list <- list()

    mask <- (projects_list %in% project_folder)

    if (length(which(mask))>0) {
      projects_list <- projects_list[-which(mask)]
    }

    if (!is.null(project_folder))
      projects_list <- c(list(project_folder), projects_list)

    names(projects_list) <- lapply(
      seq_along(projects_list),
      FUN = function(x) {
        basename(projects_list[[x]])
      })

    write.config(config.dat = projects_list,
                 file.path = file.path(os_conf_tmp, "projects_list.yml"),
                 write.type = "yaml")
    return(projects_list)
  }


  ### Dashboard outputs ####

  ### End dashboard outputs ####

  ### Variational reactiveVals ####


  va_uis <- c('va_depth_minimum',
              'va_missing_values_max',
              'va_minumum_median_total',
              'va_minumum_median_mutation')

  for (va_ui in va_uis)
    inputs[[va_ui]] <- reactiveVal()

  ## inputs[['va_data_in_dir']] <- reactive(inputs[['thr_out_dir']]())
  ## dir  <- out_subfolder_compute(inputs[['project_folder']](), 'filtered_var')
  ## inputs[['va_out_dir']] <- reactive(dir)
  inputs[['va_verified_genes']] <- reactiveVal()
  inputs[['va_list_genes']] <- reactiveVal()



  ## va_data_in_dir_ <- reactive(parseDirPath(roots=roots_dir, inputs[['va_data_in_dir']]()))
  va_depth_minimum_ <-
    reactive(inputs[['va_depth_minimum']]())
  va_missing_values_max_ <-
    reactive(inputs[['va_missing_values_max']]())
  va_minumum_median_total_ <-
    reactive(inputs[['va_minumum_median_total']]())
  va_minumum_median_mutation_ <-
    reactive(inputs[['va_minumum_median_mutation']]())
  va_out_dir_ <-
    reactive(parseDirPath(roots=roots_dir, inputs[['va_out_dir']]()))
  
  va_non_NA_genes_ <- reactiveVal(NULL)
  va_compute_output_ <- reactiveVal(NULL)

  files_ <- reactiveVal(NULL)
  list_genes_ <-
    reactive(inputs[['va_list_genes']]())
  verified_genes_ <-
    reactive(inputs[['va_verified_genes']]())

  va_rvs = reactiveValues(va_buttons = list(), va_observers = list())

  ### End variational reactiveVals ####


  ### Variational functions ####

  va_exec <- function () {
    #browser()

    if (length(va_out_dir_()) == 0)
      return()

    if (!is.integer(inputs[['thr_out_dir']]()) &&
        !is.integer(inputs[['dp_out_dir']]())) {
      if (dir.exists(thr_out_dir_()) &&
          dir.exists(dp_out_dir_())) {
        
        #browser()
        
        valid_genes_names <-
          NA_compute(va_depth_minimum_(),
                   va_missing_values_max_(),
                   thr_out_dir_(),
                   dp_out_dir_(),
                   va_out_dir_(),
                   inputs[['m_time_points']]())
        
        va_non_NA_genes_(valid_genes_names)
        
        #browser()
        
        files <- NA_compute2_load(thr_out_dir_(),
                                  dp_out_dir_(),
                                  va_out_dir_())
        files_(files)
        
      } else
        showNotification(paste("Annotated VCF folder does not exist"),
                         duration = 10,
                         type = "warning")
    } else
      showNotification(paste("Annotated VCF folder is not set"),
                       duration = 10,
                       type = "warning")
  }



  va_exec2 <- function() {
    req(inputs[['thr_out_dir']](), inputs[['dp_out_dir']]())
    if (!is.integer(inputs[['thr_out_dir']]()) &&
        !is.integer(inputs[['dp_out_dir']]()))
      if (dir.exists(thr_out_dir_()) &&
          dir.exists(dp_out_dir_()))

        if (any(!sapply(files_(), is.null))){ # files not found
          
          #browser()
          
          NA_c2 <-
            NA_compute2(va_depth_minimum_(),
                        va_minumum_median_total_(),
                        va_minumum_median_mutation_(),
                        thr_out_dir_(),
                        dp_out_dir_(),
                        va_out_dir_(),
                        inputs[['m_time_points']](),
                        verified_genes_(),
                        va_non_NA_genes_(),
                        files_())
          #browser()
          va_compute_output_(NA_c2$distinct_mutations)
          
          #ggsave(file=file.path(inputs[["project_folder_std"]](),"D.svg"), plot=NA_c2$g, width=10, height=10)
          
          output$va_filtered_bin_hmap <- renderPlot({
            NA_c2$g
          }, height = 600)


        } else {
          if (!is.null(files_()))
            showNotification(paste("Files from previous steps not computed"),
                             duration = 10,
                             type = "warning")
        } else
          showNotification(paste("Sampling depth folder does not exist"),
                           duration = 10,
                           type = "warning")
    else
      showNotification(paste("Sampling depth folder is not set"),
                       duration = 10,
                       type = "warning")
  }

  ### End variational functions ####

  ### Variational observers ####

  observe({
    va_rvs$observers =
      lapply(va_uis,
             function(i) {
               observe({
                 req(inputs[[i]]())
               })
             }
      )
  },
  priority = -1)


  observeEvent(reactiveValuesToList(input), {
    outs <- outputOptions(output)
    lapply(names(outs),
           function(name) {
             outputOptions(output,
                           name,
                           suspendWhenHidden = FALSE)
           })
  },
  once = TRUE,
  priority = -1)


  observe({
    va_observers = lapply(va_uis,
                          function(i) {
                            observeEvent(input[[i]], {
                              ## req(input)
                              inputs[[i]](input[[i]])
                            })

                            output[[i]] <- renderText(inputs[[i]]())
                          })
  },
  priority = -1)



  observeEvent(inputs[['av_anovar_db_dir']](), {
    #browser()
    req(inputs[['av_anovar_db_dir']]())
    ## req(inputs[['thr_out_dir']](), inputs[['dp_out_dir']]())
    ## if (!is.integer(inputs[['thr_out_dir']]()) && !is.integer(inputs[['dp_out_dir']]())) {
    if (!is.integer(inputs[['av_anovar_db_dir']]())) {
      ## if (dir.exists(thr_out_dir_()) && dir.exists(dp_out_dir_())) {
      if (dir.exists(av_anovar_db_dir_())) {
        ## if (file.exists(file.path( thr_out_dir_(), "snpMut_filt_freq.rds"))){
        fa_file <-
          file_path_sans_ext(list.files(path = av_anovar_db_dir_(),
                                        pattern = ".*.fa")[1])
        ref_files <-
          list.files(path = av_anovar_db_dir_(),
                     pattern = ".*.[^f][^a]")
        ref_files2 <-
          file_path_sans_ext(ref_files)
        ref_files2 <-
          ref_files[str_detect(fa_file, ref_files2)]
        ref_files2 <-
          ref_files2[which.min(str_length(ref_files2))]
        ref_files2 <-
          file.path(av_anovar_db_dir_(), ref_files2)
        
        
        if (length(file.exists(ref_files2))>0) {
          if (file.exists(ref_files2)) {
            ref_info <-
              read.table(file = ref_files2,
                         header = FALSE,
                         sep = '\t',
                         stringsAsFactors = FALSE)
            list_gene_symbols <- ref_info[, 13] 
            
            
            ## if (file.exists(file.path( av_anovar_db_dir_(), "snpMut_filt_freq.rds"))){
            ##  print('LOAD FILE2')
            ##  snpMut_filt_freq <- readRDS(file=paste0(file.path( thr_out_dir_(), "snpMut_filt_freq.rds")))
            ##  print(head(snpMut_filt_freq))
          } else {
            showNotification(paste("annovar reference file",
                                   ref_files2,
                                   "not found"),
                             duration = 10,
                             type = "warning")
          }
        }
        
        #not necessary
        if (file.exists(file.path( va_out_dir_(), "snpMut_filt_freq_reduced.rds"))) {
          snpMut_filt_freq <- readRDS(file=paste0(file.path( va_out_dir_(), "snpMut_filt_freq_reduced.rds")))
          inputs[['va_list_genes']](sort(unique(snpMut_filt_freq$Gene)))
        } else
          inputs[['va_list_genes']](list_gene_symbols)
        
        updateSelectizeInput(session,
                             'va_verified_genes',
                             choices = inputs[['va_list_genes']](),
                             selected = inputs[['va_verified_genes']](),
                             server = TRUE)

      }
    }
  })


  observeEvent(input[['va_verified_genes']], {
    #req(input[['va_verified_genes']])
    #browser()
    inputs[['va_verified_genes']](input[['va_verified_genes']])
  }, ignoreNULL = FALSE)

  
  
  
  va_iv <- InputValidator$new()
  va_iv$add_rule("va_depth_minimum", sv_between(0,500))
  va_iv$add_rule("va_missing_values_max", sv_between(0.,1.))
  va_iv$add_rule("va_minumum_median_total", sv_between(0,500))
  va_iv$add_rule("va_minumum_median_mutation", sv_between(0,500))
  va_iv$enable()

  observeEvent(input$va_exec,{
    req(va_iv$is_valid())
    #browser()
    
    va_exec() 
    
    
    if (file.exists(file.path( va_out_dir_(), "snpMut_filt_freq_reduced.rds"))) {
      snpMut_filt_freq <- readRDS(file=paste0(file.path( va_out_dir_(), "snpMut_filt_freq_reduced.rds")))
      inputs[['va_list_genes']](sort(unique(snpMut_filt_freq$Gene)))
    }
    
    updateSelectizeInput(session,
                         'va_verified_genes',
                         choices = inputs[['va_list_genes']](),
                         selected = inputs[['va_verified_genes']](),
                         server = TRUE)
    
    va_exec2()
    #browser()
    #print("a")
  })


  observeEvent({ c(#files_(),
                   va_depth_minimum_(),
                   va_missing_values_max_(),
                   va_minumum_median_total_(),
                   va_minumum_median_mutation_(),
                   verified_genes_()
                   )
  }, 
  { 
    va_exec()
    va_exec2() 
  },
  ignoreNULL = FALSE,
  ignoreInit = TRUE)


  ### End Variational observers ####

  ### Variational outputs ####

  output$va_out <-
    DT::renderDT(va_compute_output_(), server = TRUE)
  #output[['va_verified_genes']] <-
  #  renderText(inputs[['va_verified_genes']]())
  #output[['va_list_genes']] <-
  #  renderPrint(inputs[['va_list_genes']]())
  #output[['va_out_dir']] <-
  #  renderText(parseDirPath(roots = roots_dir,
  #                          inputs[['va_out_dir']]()))

  output[['va_filters']] <-
    renderPrint({
      list(va_data_in_dir = parseDirPath(roots=roots_dir,
                                         inputs[['thr_out_dir']]()),
           depth_minimum = inputs$va_depth_minimum(),
           missing_values_max = inputs$va_missing_values_max(),
           minumum_median_total = inputs$va_minumum_median_total(),
           minumum_median_mutation = inputs$va_minumum_median_mutation(),
           va_out_dir = parseDirPath(roots = roots_dir,
                                     inputs[['va_out_dir']]()),
           va_verified_genes = inputs[['va_verified_genes']]())
    })

  ### End variational outputs ####


  ### Depth reactivaVals ####


  dp_uis <- c('dp_samtools_exec_dir', 'dp_bam_dir')

  dp_rvs = reactiveValues(dp_buttons = list(), dp_observers = list())

  ###
  for (dir_ui in dp_uis) {
    defaultPath <- ""
    # if(dir_ui == "dp_samtools_exec_dir")
    #   defaultPath <- path_rel(dirname(Sys.which("samtools")), start = "..")

    shinyDirChoose(input,
                   dir_ui,
                   roots = roots_dir,
                   filetypes = c('', 'txt'),
                   defaultPath = defaultPath,
                   defaultRoot = "..")
  }

  for (dp_ui in dp_uis)
    inputs[[dp_ui]] <- reactiveVal()

  inputs[['dp_mut_dir']] <-
    reactive(inputs[['thr_out_dir']]())
  ## inputs[['dp_out_dir']] <- reactiveVal()


  dp_samtools_exec_dir_ <-
    reactive(parseDirPath(roots = roots_dir,
                          inputs[['dp_samtools_exec_dir']]()))
  dp_samtools_ <-
    reactive(file.path(dp_samtools_exec_dir_(),
                       'samtools'))
  dp_bam_dir_ <-
    reactive(parseDirPath(roots = roots_dir,
                          inputs[['dp_bam_dir']]()))
  dp_out_dir_ <-
    reactive(parseDirPath(roots = roots_dir,
                          inputs[['dp_out_dir']]()))
  dp_mut_dir_ <-
    reactive(parseDirPath(roots = roots_dir,
                          inputs[['dp_mut_dir']]()))

  ### End depth reactivaVals ####

  ### Depth observers ####

  observe({
    dp_rvs$observers =
      lapply(dp_uis,
             function(i) {
               observe({
                 req(inputs[[i]]())
                 updateActionButton(session, i, label = set_dir_ui(i))
               })
             }
      )
  },
  priority = -1)


  observeEvent(reactiveValuesToList(input), {
    outs <- outputOptions(output)
    lapply(names(outs),
           function(name) {
             outputOptions(output,
                           name,
                           suspendWhenHidden = FALSE)
           })
  },
  once = TRUE,
  priority = -1)

  ## observeEvent(reactiveValuesToList(input),{
  ##   lapply(dp_uis, function(i) {
  ##     click(i)
  ##     hide(paste0(i,'-modal'))
  ##     delay(300,runjs(paste0("$('#",i,"-modal #sF-cancelButton').click()")))
  ##   })
  ## },once = T, priority = -1)


  observe({
    dp_observers = lapply(dp_uis,
                          function(i) {
                            observeEvent(input[[i]], {
                              req(input)
                              inputs[[i]](input[[i]])
                            })

                            output[[i]] <-
                              renderText(parseDirPath(roots = roots_dir,
                                                      inputs[[i]]()))
                          }
    )},
    priority = -1)


  observeEvent(dp_samtools_exec_dir_(), {
    if (!is.integer(input[['dp_samtools_exec_dir']])) {
      if (file.exists(dp_samtools_()))
        showNotification(paste("found samtools"),
                         duration = 10,
                         type = "message")
      else
        showNotification(paste("samtools", 'not present in',
                               dp_samtools_exec_dir_()),
                         duration = 10,
                         type = "warning")
    }
  })

  ### End depth observers ####

  ### Depth functions ####

  dp_exec <- function() {

    if (length(dp_out_dir_()) == 0)
      return()

    Opt = list() # MA:  Perche` maiuscolo il nome della variabile?
    Opt$StoredFile <- file.path(config_path, ".config_04_dp.yml")
    Opt$ActFile <- file.path(config_path, "config_04_dp.yml")
    Opt$ExtraFiles = NULL

    OutFiles <- c("final_data_depth.txt", "final_data_mut.txt")
    chk_files <- ManyInSomeOut(dp_bam_dir_(),
                               InFilesMask = ".*.bam$",
                               OutDir=dp_out_dir_(),
                               OutFiles = OutFiles,
                               Opt = Opt)

    InFilesToDo <- chk_files$InFilesToDo

    ## Use metadata if available
    if (!is.integer(inputs[['sc_metadata_file']]())) {
      if (file.exists(sc_metadata_label_())) {
        if (inputs[['m_id_column']]() %in% colnames(sc_metadata_()))
          InFilesToDo <-
            InFilesToDo[str_split_fixed(InFilesToDo,
                                        "\\.",
                                        n = 2)[, 1] %in%
                          as.character(sc_metadata_()[[inputs[['m_id_column']]()]])]
        else
          showNotification(paste("no valid metadata id column"),
                           duration = 10,
                           type = "warning")
      } else
        showNotification(paste("no valid metadata file"), duration = 10, type = "warning")
    } else
      showNotification(paste("no valid metadata file"), duration = 10, type = "warning")


    InFilesToDo <-
      file.path(dp_bam_dir_(), chk_files$InFilesToDo)
    OutFilesToRm <-
      file.path(dp_out_dir_(), chk_files$OutFilesToRm)

    if (!is.integer(inputs[['dp_samtools_exec_dir']]()) &
        !is.integer(inputs[['dp_bam_dir']]()) &
        !is.integer(inputs[['dp_mut_dir']]())) {
      if (dir.exists(dp_samtools_exec_dir_()) &
          file.exists(dp_samtools_()) &
          dir.exists(dp_bam_dir_()) & dir.exists(dp_mut_dir_())) {
        dp_compute(sc_metadata_(),
                   dp_bam_dir_(),
                   dp_samtools_(),
                   thr_out_dir_(),
                   dp_out_dir_(),
                   inputs[['m_id_column']](),
                   InFilesToDo,
                   OutFilesToRm)
        ## return(read_file(file.path(dp_out_dir_(), 'stdout.log')))
      } else
        showNotification(paste("one of the folders or executable does not exist"),
                         duration = 10,
                         type = "warning")
    } else
      showNotification(paste("one of the folders is not set"),
                       duration = 10,
                       type = "warning")
    return("possible errors")
  }

  ### End depth functions ####

  ### Depth outputs ####

  output$dp_out <- eventReactive(input$dp_exec,{
    dp_exec()
  })



  output[['dp_mut_dir']] <-
    renderText(parseDirPath(roots = roots_dir,
                            inputs[['dp_mut_dir']]()))

  output[['dp_par']] <-
    renderPrint({
      list(samtools =parseDirPath(roots = roots_dir,
                                  inputs$dp_samtools_exec_dir()),
           bam_dir = parseDirPath(roots = roots_dir,
                                  inputs$dp_bam_dir()),
           mut_dir = parseDirPath(roots = roots_dir,
                                  inputs$dp_mut_dir()),
           out_dir = parseDirPath(roots = roots_dir,
                                  inputs$dp_out_dir())
      )
    })

  ### End depth outputs ####




  ### Threshold reactiveVals #### 

  thr_uis = c('thr_alleles_ratio',
              'thr_maf',
              'thr_freq',
              'thr_accepted_var',
              'thr_negleted_var')

  ###
  thr_rvs = reactiveValues(thr_buttons = list(),
                           thr_observers = list())

  for (thr_ui in thr_uis)
    inputs[[thr_ui]] <- reactiveVal()



  thr_alleles_ratio_ <-
    reactive(inputs[['thr_alleles_ratio']]())
  thr_maf_ <-
    reactive(inputs[['thr_maf']]())
  thr_freq_ <-
    reactive(inputs[['thr_freq']]())
  thr_out_dir_ <-
    reactive(parseDirPath(roots = roots_dir,
                          inputs[['thr_out_dir']]()))
  thr_vcf_in_dir_ <-
    reactive(parseDirPath(roots = roots_dir,
                          inputs[['thr_vcf_in_dir']]()))


  exonic_list <- list("frameshift insertion",
                      "frameshift deletion",
                      "frameshift block substitution",
                      "stopgain",
                      "stoploss",
                      "nonframeshift insertion",
                      "nonframeshift deletion",
                      "nonframeshift block substitution",
                      "nonsynonymous SNV",
                      "synonymous SNV",
                      "unknown")

  ### End threshold reactiveVals ####


  ### Threshold functions ####

  
  
  thr_exec <- function() {
    if (length(thr_out_dir_()) == 0)
      return()

    Opt = list()
    Opt$StoredFile <- file.path(config_path, ".config_03_thr.yml")
    Opt$ActFile <- file.path(config_path, "config_03_thr.yml")
    Opt$ExtraFiles = NULL

    ## chk_files <- OneInOneOut( thr_vcf_in_dir_(), InFilesMask = ".*.exonic_variant_function$", thr_out_dir_(), OutExt = "anninput", Opt = Opt)

    OutFiles <- c("cells_aggregate_info.rds",
                  "scMutInfo.rds",
                  "SNPInfo.rds",
                  "snpMut_filt_freq.rds"
    )
    chk_files <- ManyInSomeOut(thr_vcf_in_dir_(),
                               InFilesMask =
                                 ".*.exonic_variant_function$",
                               OutDir=thr_out_dir_(),
                               OutFiles = OutFiles,
                               Opt = Opt)

    InFilesToDo <- chk_files$InFilesToDo

    ## use metadata if available
    if (!is.integer(inputs[['sc_metadata_file']]())) {
      if (file.exists(sc_metadata_label_())){
        if (inputs[['m_id_column']]() %in% colnames(sc_metadata_()))
          InFilesToDo <-
            InFilesToDo[str_split_fixed(InFilesToDo,
                                        "\\.",
                                        n = 2)[, 1] %in%
                          as.character(sc_metadata_()[[inputs[['m_id_column']]()]])]
        else
          showNotification(paste("no valid metadata id column"),
                           duration = 10,
                           type = "warning")
      } else
        showNotification(paste("no valid metadata file"),
                         duration = 10,
                         type = "warning")
    } else
      showNotification(paste("no valid metadata file"),
                       duration = 10,
                       type = "warning")

    InFilesToDo <-
      file.path(thr_vcf_in_dir_(),
                chk_files$InFilesToDo)
    OutFilesToRm <-
      file.path(thr_out_dir_(),
                chk_files$OutFilesToRm)


    if (!is.integer(inputs[['thr_vcf_in_dir']]())) {
      if (dir.exists(thr_vcf_in_dir_())) {


        ## filter1_compute(thr_vcf_in_dir_(), sc_metadata_(), thr_alleles_ratio_(), thr_maf_(), thr_freq_(), thr_out_dir_(), m_id_column_(), m_time_column_())
        filter1_compute(thr_vcf_in_dir_(),
                        sc_metadata_(),
                        thr_alleles_ratio_(),
                        thr_maf_(),
                        thr_freq_(),
                        thr_out_dir_(),
                        inputs[['m_id_column']](),
                        inputs[['m_time_column']](),
                        inputs[['thr_accepted_var']](),
                        InFilesToDo,
                        OutFilesToRm)


        filter2_compute(thr_vcf_in_dir_(),
                        sc_metadata_(),
                        thr_alleles_ratio_(),
                        thr_maf_(),
                        thr_freq_(),
                        thr_out_dir_(),
                        m_id_column_(),
                        m_time_column_(),
                        InFilesToDo,
                        OutFilesToRm)

        snpMut_filt_freq <- filter3_compute(
                        thr_vcf_in_dir_(),
                        sc_metadata_(),
                        thr_alleles_ratio_(),
                        thr_maf_(),
                        thr_freq_(),
                        thr_out_dir_(),
                        m_id_column_(),
                        m_time_column_(),
                        inputs[['m_time_points']](),
                        InFilesToDo,
                        OutFilesToRm)
        
        inputs[['va_list_genes']](sort(unique(snpMut_filt_freq$Gene)))
        
        updateSelectizeInput(session,
                             'va_verified_genes',
                             choices = inputs[['va_list_genes']](),
                             selected = inputs[['va_verified_genes']](),
                             server = TRUE)
        
        return(read_file(file.path(thr_vcf_in_dir_(),
                                   'stdout.log')))
      } else
        showNotification(paste("Filtered VCF folder does not exist"),
                         duration = 10,
                         type = "warning")
    } else
      showNotification(paste("Filtered VCF folder is not set"),
                       duration = 10,
                       type = "warning")
    return("Possible errors") # MA: return or raise an error?
  }

  ### End threshold functions ####

  ### Threshold observers ####
  observe({
    thr_rvs$observers = lapply(thr_uis,
                               function(i) {
                                 observe({
                                   req(inputs[[i]]())
                                   ## updateActionButton(session, i, label = set_dir_ui(i))
                                   updateNumericInput(session, i, value = inputs[[i]]())
                                 })
                               }
    )
  },
  priority = -1)



  observeEvent(reactiveValuesToList(input), {
    outs <- outputOptions(output)
    lapply(names(outs),
           function(name) {
             outputOptions(output,
                           name,
                           suspendWhenHidden = FALSE)
           })
  },
  once = TRUE,
  priority = -1)


  ## # observeEvent(reactiveValuesToList(input),{
  ## #   lapply(thr_uis, function(i) {
  ## #     click(i)
  ## #     hide(paste0(i,'-modal'))
  ## #     delay(300,runjs(paste0("$('#",i,"-modal #sF-cancelButton').click()")))
  ## #   })
  ## # },once = T, priority = -1)

  ###



  ## #dir <- out_subfolder_compute(inputs[['project_folder']](),'filtered_vcf')
  ## #inputs[['thr_out_dir']] <- reactive(dir)



  observe({
    thr_observers = lapply(thr_uis,
                           function(i) {
                             observeEvent(input[[i]], {
                               req(input)
                               inputs[[i]](input[[i]])
                             })

                             output[[i]] <- renderPrint(inputs[[i]]())
                           }
    )
  },
  priority = -1)


  ## Exonic ui
  observeEvent(reactiveValuesToList(input), {
    inputs[['thr_accepted_var']](exonic_list)
    inputs[['thr_negleted_var']](NULL)

  },
  once = TRUE,
  priority = -1)
  
  #observeEvent(input[['thr_accepted_var']],
  #             {
                 #browser()
                 #print(input[['thr_accepted_var']])
                 #print(inputs[['thr_accepted_var']]())
  #             })
  
  #observeEvent(inputs[['thr_accepted_var']](),
  #             {
                 #browser()
                 #print(input[['thr_accepted_var']])
                 #print(inputs[['thr_accepted_var']]())
  #             })

  output$thr_bucket_var_list <-
    renderUI({
      ## showNotification(paste("renderUI[['thr_bucket_var_list']]"), duration = NULL)

      bucket_list_basic <-
        bucket_list(
          header = "Choose the variant functions for inferential analysis",
          add_rank_list(
            text = "Considered exonic variants",
            labels = inputs[['thr_accepted_var']](),
            input_id = "thr_accepted_var",
            options = sortable_options(multiDrag = TRUE, disabled = !is.null(inputs[["demo"]]()))
          ),
          add_rank_list(
            text = "Neglected exonic variants",
            labels = inputs[['thr_negleted_var']](),
            input_id = "thr_negleted_var",
            options = sortable_options(multiDrag = TRUE, disabled = !is.null(inputs[["demo"]]()))
          ),
          group_name = "thr_bucket_var_list",
          orientation = "horizontal", 
          options = sortable_options(disabled = !is.null(inputs[["demo"]]()))
        )

      x <-
        tagList(
          span() %>%
            make_title_help_popover (
              tags$b("Exonic variant function annotations"),
              "List of exonic varian functions",
              text[["thr_bucket_var_list"]],
              server = TRUE
            ),
          bucket_list_basic
        )
      if (loading_()) {
        loaded_(TRUE)
      }
      x
    })

  thr_iv <- InputValidator$new()
  thr_iv$add_rule("thr_alleles_ratio", sv_between(1,55))
  thr_iv$add_rule("thr_maf", sv_between(0.01,0.5))
  thr_iv$add_rule("thr_freq", sv_between(0.,1.))
  thr_iv$enable()
  ### Threshold observers ####

  ### Threshold outputs ####
  output[['thr_vcf_in_dir']] <-
    renderText(parseDirPath(roots = roots_dir,
                            inputs[['thr_vcf_in_dir']]()))
  output[['thr_out_dir']] <-
    renderText(parseDirPath(roots = roots_dir,
                            inputs[['thr_out_dir']]()))
  output[['thr_filters']] <-
    renderPrint({
      list(thr_vcf_in_dir = parseDirPath(roots = roots_dir,
                                         inputs[['thr_vcf_in_dir']]()),
           alleles_ratio = inputs$thr_alleles_ratio(),
           maf = inputs$thr_maf(),
           freq = inputs$thr_freq(),
           thr_out_dir = parseDirPath(roots = roots_dir,
                                      inputs[['thr_out_dir']]()) )
    })

  output$thr_out <- eventReactive(input$thr_exec, {
    req(thr_iv$is_valid())
    thr_exec()
  })

  ### End threshold outputs ####

  ### Annotation reactiveVals ####

  av_dir_uis= c('av_anovar_exec_dir',
                'av_anovar_db_dir',
                'av_vcf_in_dir')

  for (dir_ui in av_dir_uis)
    inputs[[dir_ui]] <- reactiveVal()

  av_rvs = reactiveValues(av_buttons = list(), av_observers = list())

  ## #dir <- out_subfolder_compute(inputs[['project_folder']](),'vcf_out')
  ## #inputs[['av_vcf_out_dir']] <- reactive(dir)

  av_anovar_exec_dir_ <-
    reactive(parseDirPath(roots = roots_dir,
                          inputs[['av_anovar_exec_dir']]()))
  av_anovar_convert_ <-
    reactive(file.path(av_anovar_exec_dir_(),
                       'convert2annovar.pl'))

  av_anovar_db_dir_ <-
    reactive(parseDirPath(roots = roots_dir,
                          inputs[['av_anovar_db_dir']]()))
  av_vcf_in_dir_ <-
    reactive(parseDirPath(roots = roots_dir,
                          inputs[['av_vcf_in_dir']]()))
  av_vcf_out_dir_ <-
    reactive(parseDirPath(roots = roots_dir,
                          inputs[['av_vcf_out_dir']]()))
  av_anovar_annot_ <-
    reactive(file.path(av_anovar_exec_dir_(),
                       'annotate_variation.pl'))

  ### End annotation reactiveVals ####


  ### Annotation functions ####

  set_dir_ui <- function(dir_id,
                         default_dir_label = 'Select folder',
                         input_var = inputs,
                         basedir = "..") { # replace with roots_dir
    this_label <- default_dir_label
    meta <- input_var[[dir_id]]()
    ## if (basedir %in% names(meta))
    if (!is.null(meta))
      if (!is.integer(meta))
        if (dir.exists(parseDirPath(roots = roots_dir, meta)))
          this_label <- parseDirPath(roots = roots_dir, meta)
    return(this_label)
  }


  out_subfolder_compute <- function(out_dir_inp, sub_dir) {
    if ('path' %in% names(out_dir_inp)) {
      tmp <- out_dir_inp[['path']]
      ## tmp <- tmp[-length(tmp)]
      out_dir_inp[['path']] <- c(tmp, sub_dir)
    }
    return (out_dir_inp)
  }

  av_exec <- function() {

    if(length(av_vcf_out_dir_())==0)
      return()

    log2_print('av_exec', msg = "LACEview:")
    log2_print(paste("av_vcf_out_dir_()", av_vcf_out_dir_()), msg = "LACEview:")

    Opt = list()
    Opt$StoredFile <- file.path(config_path,".config_02_av.yml")
    Opt$ActFile <- file.path(config_path,"config_02_av.yml")
    Opt$ExtraFiles = NULL

    chk_files <- OneInOneOut(av_vcf_in_dir_(),
                             InFilesMask = ".*vcf$",
                             av_vcf_out_dir_(),
                             OutExt =
                               "anninput.exonic_variant_function",
                             Opt = Opt)

    InFilesToDo <- chk_files$InFilesToDo

    ## Use metadata if available
    if (!is.integer(inputs[['sc_metadata_file']]())) {
      if (file.exists(sc_metadata_label_())){
        if (inputs[['m_id_column']]() %in% colnames(sc_metadata_()))
          InFilesToDo <-
            InFilesToDo[str_split_fixed(InFilesToDo,
                                        "\\.",
                                        n = 2)[, 1] %in%
                          as.character(sc_metadata_()[[inputs[['m_id_column']]()]])]
        else
          showNotification(paste("no valid metadata id column"),
                           duration = 10,
                           type = "warning")
      }
      else
        showNotification(paste("no valid metadata file"),
                         duration = 10,
                         type = "warning")
    }
    else
      showNotification(paste("no valid metadata file"),
                       duration = 10,
                       type = "warning")


    InFilesToDo <- file.path(av_vcf_in_dir_(),
                             chk_files$InFilesToDo)
    OutFilesToRm <- file.path(av_vcf_out_dir_(),
                              chk_files$OutFilesToRm)

    if (!is.integer(inputs[['av_anovar_exec_dir']]()) &
        !is.integer(inputs[['av_anovar_db_dir']]()) &
        !is.integer(inputs[['av_vcf_in_dir']]())
    ) {

      if (dir.exists(av_anovar_exec_dir_()) &
          file.exists(av_anovar_convert_()) &
          file.exists(av_anovar_annot_()) &
          dir.exists(av_anovar_db_dir_()) &
          dir.exists(av_vcf_in_dir_())
      ) {
        av_compute(av_anovar_convert_(),
                   av_anovar_annot_(),
                   av_anovar_db_dir_(),
                   av_vcf_in_dir_(),
                   av_vcf_out_dir_(),
                   InFilesToDo = InFilesToDo,
                   OutFilesToRm = OutFilesToRm)
        return(read_file(file.path(av_vcf_out_dir_(),
                                   'stdout.log')))
      } else
        showNotification(paste("one of the folders or executable does not exist"),
                         duration = 10,
                         type = "warning")
    } else
      showNotification(paste("one of the folders is not set"),
                       duration = 10,
                       type = "warning")
    return("possible errors") # MA:  CHECK THESE!!!! IS THE CHECK!!!
  }

  ### End annotation functions ####

  ### Annotation observers ####

  for (dir_ui in av_dir_uis) { #av_rvs$buttons???
    defaultPath <- ""
    # if(dir_ui == "av_anovar_exec_dir")
    #   defaultPath <- path_rel(Sys.which("annotate_variation.pl"), start = "..")
    shinyDirChoose(input,
                   dir_ui,
                   roots = roots_dir,
                   defaultPath = defaultPath,
                   filetypes = c('', 'txt'),
                   defaultRoot = "..")
  }

  ## Update buttons
  observe({
    av_rvs$observers = lapply(av_dir_uis,
                              function(i) {
                                observe({
                                  req(inputs[[i]]())
                                  updateActionButton(session,
                                                     i,
                                                     label = set_dir_ui(i))
                                })
                              }
    )
    ## reactiveValuesToList(av_rvs)
  },
  priority = -1)


  ## Make ui updatable
  observeEvent(reactiveValuesToList(input), {
    outs <- outputOptions(output)
    lapply(names(outs),
           function(name) {
             outputOptions(output,
                           name,
                           suspendWhenHidden = FALSE)
           })
  },
  once = TRUE,
  priority = -1)

  ## observeEvent(reactiveValuesToList(input), {
  ##     lapply(av_dir_uis,
  ##            function(i) {
  ##                click(i)
  ##                hide(paste0(i, '-modal'))
  ##                delay(300,
  ##                      runjs(paste0("$('#",
  ##                                   i,
  ##                                   "-modal #sF-cancelButton').click()")))
  ##            })

  ## #delay(1000, runjs("$('.sF-modal modal-dialog modal-lg').keydown(new KeyboardEvent('keydown', {'keyCode': 40}))"))
  ## #delay(1000,"$('#av_anovar_exec_dir-modal').dispatchEvent(new KeyboardEvent('keydown', {'keyCode': 40}))")
  ## #delay(1000,runjs("handleArrowKey('down')"))
  ## },
  ## once = T,
  ## priority = -1)


  observe({
    av_observers = lapply(av_dir_uis,
                          function(i) {
                            observeEvent(input[[i]], {
                              req(input)
                              inputs[[i]](input[[i]])
                            })

                            output[[i]] <-
                              renderText(parseDirPath(roots = roots_dir,
                                                      inputs[[i]]()))
                          }
    )
    ## reactiveValuesToList(av_rvs)
  },
  priority = -1)



  ## observeEvent(inputs[['av_anovar_exec_dir']](),{
  observeEvent(av_anovar_exec_dir_(), {
    if (!is.integer(input[['av_anovar_exec_dir']])) {
      if (file.exists(av_anovar_convert_()))
        showNotification(paste("found convert2annovar.pl"),
                         duration = 10,
                         type = "message")
      else
        showNotification(paste("convert2annovar.pl",
                               'not present in',
                               av_anovar_exec_dir_()),
                         duration = 10,
                         type = "warning")
      if (file.exists(av_anovar_annot_()))
        showNotification(paste("found annotate_variation.pl"),
                         duration = 10,
                         type = "message")
      else
        showNotification(paste("annotate_variation.pl",'not present in',
                               av_anovar_exec_dir_()),
                         duration = 10,
                         type = "warning")
    }
  })

  output$av_out <- eventReactive(input$av_exec, {
    av_exec()
  },
  ignoreInit = TRUE)

  ### End annotation observers ####

  ### Annotation outputs ####

  output[['av_vcf_out_dir']] <-
    renderText(parseDirPath(roots = roots_dir,
                            inputs[['av_vcf_out_dir']]()))
  output$av_anovar_convert <-
    renderText(av_anovar_convert_())
  output$av_anovar_annot <-
    renderText(av_anovar_annot_())

  ### End annotation outputs ####




  ### Metadata reactiveVals ####


  loaded_input_ <- reactiveVal(list())
  loaded_ <- reactiveVal(FALSE)
  m_widgetId_ <- reactiveVal()


  loaded_input <- list()
  initialInputs <- list()
  initialInputs_copy <- list()
  the_files <- list()

  ### End metadata reactiveVals ####

  ### Metadata observers ####

  ## add types to ui #remove
  observeEvent(input, {
    initialInputs <<- reactiveValuesToList(input) # MA: WTF! Ma decidersi sugli assegnamenti????
    for (inp in names(initialInputs)) {
      ## print(inp)
      runjs(getWidgetType(inp))
      ## print(input$inputType)
    }
  },
  once = TRUE,
  priority = 5)


  ## Get types #remove
  observe({
    types <- list()
    for (inp in names(initialInputs))
      types[[inp]]<-input[[paste0('inputType_',inp)]]
    types_(types)
  },
  priority = 4)


  #####



  #####

  ## widgetId output
  observe({
    reactiveValuesToList(input)
    types<-types_()
    for (n in names(initialInputs_copy)) {
      if (n %in% names(types))
        if (types[[n]] == 'shinyFiles')
          if (n %in% names(the_files)) {
            ## initialInputs_copy[[n]] <<- the_files[[n]]
          }
    }
    m_widgetId_(initialInputs_copy)
  },
  priority = catch_ui_files_priority - 1)

  output$m_widgetId <- renderPrint({
    m_widgetId_()
  })


  shinyFileChoose(input,
                  'sc_metadata_file',
                  roots = roots_dir,
                  filetypes = c('rds',''),
                  defaultRoot = "..")


  ## Set_ui_files
  set_file_ui <- function(dir_id,
                          default_dir_label = 'Select file',
                          input_var = inputs,
                          basedir = "..") { #replace with roots_dir
    this_label <- default_dir_label
    meta <- input_var[[dir_id]]()
    ## if (basedir %in% names(meta))
    if (!is.null(meta))
      if (!is.integer(meta))
        if (file.exists(as.character(parseFilePaths(roots = roots_dir,
                                                    meta)[['datapath']])))
          this_label <- as.character(parseFilePaths(roots = roots_dir,
                                                    meta)[['datapath']])
    ## showNotification(this_label, duration = 10)
    return(this_label)
  }

  ## # output$m_sc_metadata_col <- reactive({
  ## #   if (length(sc_metadata_())>0) {
  ## #     sc_metadata_col <- colnames(sc_metadata_())
  ## #   } else {
  ## #     sc_metadata_col <- c("")
  ## #   }
  ## #   sc_metadata_col
  ## # })

  #####
  m_dir_uis = c('sc_metadata_file',
                'm_time_column',
                'm_time_points',
                'm_id_column')
  ## inputs <- list()
  for (dir_ui in m_dir_uis)
    inputs[[dir_ui]] <- reactiveVal()

  loading_ <- reactiveVal(FALSE)
  loaded_ <- reactiveVal(FALSE)


  #####
  sc_metadata_label_ <- reactiveVal("")
  sc_metadata_ <- reactiveVal()
  choices_ <- reactiveVal()

  observeEvent(input[['sc_metadata_file']],{
    if (! loading_()) {
      ## showNotification(paste("input[['sc_metadata_file']]","inputs[['sc_metadata_file']](input[['sc_metadata_file']])"), duration = NULL)

      inputs[['sc_metadata_file']](input[['sc_metadata_file']])

      ## print("set_file_ui('sc_metadata_file')")
      ## print(set_file_ui('sc_metadata_file'))

      sc_metadata_label_(set_file_ui('sc_metadata_file'))

      if (file.exists(sc_metadata_label_()) ) {
        if (file_ext(sc_metadata_label_()) %in% c("rds", "RDS") ) {
          cellInfo <- readRDS( sc_metadata_label_() )
          if (!is.data.frame(cellInfo)){
            showNotification("No metadata file",
                             type = "warning",
                             duration = 10)
            cellInfo <- data.frame()
          }
        } else
          cellInfo <- fread(sc_metadata_label_(), header=TRUE, data.table=FALSE)
      } else {
        showNotification("No metadata file",
                         type = "warning",
                         duration = 10)

        # hideTab(inputId = "main_tabset",
        #         target = "Annotations")
        # hideTab(inputId = "main_tabset",
        #         target = "Filters")
        # hideTab(inputId = "main_tabset",
        #         target = "SC sampling depths")
        # hideTab(inputId = "main_tabset",
        #         target = "Variants")
        # hideTab(inputId = "main_tabset",
        #         target = "Inference")
        # hideTab(inputId = "main_tabset",
        #         target = "Results")

        cellInfo <- data.frame()
      }
      sc_metadata_(cellInfo)

      if ( file.exists(sc_metadata_label_()) )
        choices_(colnames(sc_metadata_()))
      else
        choices_(c(""))

      id_col <- NULL
      time_col <- NULL
      if (length(choices_()) > 2) {
        id_col <- choices_()[1]
        time_col <- choices_()[2]
      }
      inputs[['m_id_column']](id_col)
      inputs[['m_time_column']](time_col)

      time_points <- NULL
      if (!is.null(time_col))
        time_points <- unique(as.character(sc_metadata_()[, time_col]))
      inputs[['m_time_points']](time_points)
    }
  })

  observeEvent(input[['m_id_column']], {
    if (! loading_()) {
      ## showNotification(paste("input[['m_time_column']]","inputs[['m_time_column']](input[['m_time_column']])"), duration = NULL)

      id_col <- input[['m_id_column']]
      inputs[['m_id_column']](id_col)

      if ((inputs[['m_time_column']]() == inputs[['m_id_column']]()) &&
          !("" == inputs[['m_id_column']]()))
        showNotification(paste("Same column for cell Id and time"),
                         type = "warning",
                         duration = 10)
      ## # id_points <- NULL
      ## # if ( id_col %in% colnames(sc_metadata_()) )
      ## #   id_points <- unique(as.character(sc_metadata_()[, id_col]))
      ## # inputs[['m_id_points']](id_points)
    }
  })

  observeEvent(input[['m_time_column']], {
    if (! loading_()) {
      time_col <- input[['m_time_column']]
      inputs[['m_time_column']](time_col)

      time_points <- NULL
      if (time_col %in% colnames(sc_metadata_()) )
        time_points <- unique(as.character(sc_metadata_()[, time_col]))
      inputs[['m_time_points']](time_points)
      if ((inputs[['m_time_column']]() == inputs[['m_id_column']]()) &&
          !(""==inputs[['m_time_column']]()))
        showNotification(paste("Same column for cell Id and time"),
                         type = "warning",
                         duration = 10)
    }
  })


  observeEvent(input[['m_time_points']], {
    if (! loading_()) {
      inputs[['m_time_points']](input[['m_time_points']])
    }
    if (loading_() & loaded_()) {
      loading_(FALSE)
      loaded_(FALSE)
    }
    if (file.exists(sc_metadata_label_()) )
      if (!is.null(input[['m_time_points']]))
        if (length(input[['m_time_points']]) > 1) {
          # showTab(inputId = "main_tabset",
          #         target = "Annotations")
          # showTab(inputId = "main_tabset",
          #         target = "Filters")
          # showTab(inputId = "main_tabset",
          #         target = "SC sampling depths")
          # showTab(inputId = "main_tabset",
          #         target = "Variants")
          # showTab(inputId = "main_tabset",
          #         target = "Inference")
        }
  })

  observeEvent(input[["pr_next"]],{ # The previous does not trigger the event for consecutive load
    if ( file.exists(sc_metadata_label_()) )
      if (!is.null(input[['m_time_points']]))
        if (length(input[['m_time_points']]) > 1) {
          # showTab(inputId = "main_tabset",
          #         target = "Annotations")
          # showTab(inputId = "main_tabset",
          #         target = "Filters")
          # showTab(inputId = "main_tabset",
          #         target = "SC sampling depths")
          # showTab(inputId = "main_tabset",
          #         target = "Variants")
          # showTab(inputId = "main_tabset",
          #         target = "Inference")
        }
    # delay(1000,
    #       updateTabsetPanel(session, "main_tabset", selected = "SC metadata")
    # )

  })


  observeEvent(inputs[['sc_metadata_file']](), {
    ## update the file select button name
    ## showNotification(paste("inputs[['sc_metadata_file']]","updateActionButton"), duration = NULL)
    updateActionButton(session,
                       'sc_metadata_file',
                       label = set_file_ui('sc_metadata_file'))
  })


  ### problema update ma necessario
  observeEvent(inputs[['m_time_column']](), {
    ## update time_col ui
    ## showNotification(paste("inputs[['m_time_column']]","updateSelectInput"), duration = NULL)
    updateSelectInput(session,
                      "m_time_column",
                      ## label = "Time column",
                      choices = choices_(),
                      selected = inputs[['m_time_column']]()
    )
  })

  observeEvent(inputs[['m_id_column']](),{
    ## Update time_col ui
    ## showNotification(paste("inputs[['m_time_column']]","updateSelectInput"), duration = NULL)
    updateSelectInput(session,
                      "m_id_column",
                      ## label = "Cell id column",
                      choices = choices_(),
                      selected = inputs[['m_id_column']]()
    )
  })

  ######

  observe({
    hide(id = "m_idCol_div")
    if (file.exists(sc_metadata_label_()) ) {
      shinyjs::show(id="m_idCol_div")
    }
  })

  observe({
    hide(id = "m_timePointsCol_div")
    if (file.exists(sc_metadata_label_()) ) {
      shinyjs::show(id="m_timePointsCol_div")
    }
  })

  ## Time_column ui
  output$m_idCol <-
    renderUI(selectInput("m_id_column", "Cell id column", c("")) %>%
               make_shiny_help_popover ("Ids column",
                                        text[["m_idCol"]],
                                        server = TRUE
               )
    )

  output$m_timePointsCol <-
    renderUI(selectInput("m_time_column", "Time column", c("")) %>%
               make_shiny_help_popover ("Sampling times column",
                                        text[["m_timePointsCol"]],
                                        server = TRUE
               )
    )


  ## Time_points ui

  output$m_timePoints <-
    renderUI({
      rank_list_basic <-
        rank_list(text = "Drag the time points in chronological order (before the smaller times)",
                  labels = inputs[['m_time_points']](),
                  input_id = "m_time_points", 
                  options = sortable_options(disabled = !is.null(inputs[["demo"]]()))
        )
      x <- rank_list_basic %>%
        make_title_help_popover (tags$b("Sampling points"),
                                 "Sampling time points",
                                 text[["m_timePoints"]],
                                 server = TRUE
        )

      if (loading_()) {
        loaded_(TRUE)
      }
      x
    })


  observe({
    hide(id = "m_timePoints_div")
    if (file.exists(sc_metadata_label_()) )
      shinyjs::show(id = "m_timePoints_div")
  })


  ## output[['sc_metadata_file']] <- renderText(parseFilePaths(roots = roots_dir, inputs[['sc_metadata_file']]()))
  output[['m_id_column_tab']] <-
    renderTable(head(sc_metadata_()[inputs[['m_id_column']]()]),
                bordered = TRUE,
                rownames = TRUE)
  output[['sc_metadata_file']] <-
    renderPrint(inputs[['sc_metadata_file']]())
  output[['m_time_column']] <-
    renderText(inputs[['m_time_column']]())
  output[['m_id_column']] <-
    renderText(inputs[['m_id_column']]())
  output[['m_time_points']] <-
    renderText(inputs[['m_time_points']]())
  ## #output[['m_id_column']] <- renderPrint(inputs[['m_id_column']]())

  ## #output$cellInfo <- renderTable({
  ## #  if ('root'%in%names(input$sc_metadata_file)){
  ## #    inFile <- parseFilePaths(roots = roots_dir[input$sc_metadata_file[['root']]], input$sc_metadata_file)
  ## #    if ( nrow(inFile)) {
  ## #      #cellInfo <- read.csv(as.character(inFile$datapath))
  ## #      cellInfo <- readRDS(as.character(inFile$datapath))
  ## #    }
  ## #  }
  ## #})

  outputOptions(output, "m_timePoints", suspendWhenHidden = FALSE)

  outs <- outputOptions(output)
  lapply(names(outs),
         function(name) {
           outputOptions(output,
                         name,
                         suspendWhenHidden = FALSE)
         })








  #rs_(callr::r_session$new(options = sopt, wait = T))
  #rs_()$supervise(T)

  # rp <- callr::r_bg(
  #   func = long_job,
  #   args = list("port"=port),
  #   supervise = TRUE
  #)

  observeEvent(input[["res_refresh"]], {
    #browser()
    returned_vals <- show_result(rs_())
    rs_(returned_vals$rs)
    inputs[["res_lastrefresh"]](returned_vals$message)
  })



  output[["res_lastrefresh"]] <- renderText(inputs[["res_lastrefresh"]]())


}

shinyApp(ui, server)
## return (shinyApp(ui, server))
## runApp(system.file("shiny", "tooltip_popover_modal", package = "bsplus"))

##}


### end of file -- app.R

# Remove any indistinguishable event from input data
check.indistinguishable <- function( data ) {
    
    # build one data matrix from all time points
    full_data <- NULL
    for(i in 1:length(data)) {
        full_data <- rbind(full_data,data[[i]])
    }
    
    # check for indistinguishable events
    indistinguishable <- as.numeric(which(duplicated(t(full_data))))
    if(length(indistinguishable)>0) {
        # merge names of indistinguishable events
        valid_colnames <- colnames(full_data)[-indistinguishable]
        new_colnames <- valid_colnames
        invalid_colnames <- colnames(full_data)[indistinguishable]
        for(i in invalid_colnames) {
            for(j in valid_colnames) {
                if(all(is.na(full_data[,i])==is.na(full_data[,j]))) { # if NAs in i and j are the same
                    if(all(full_data[,i]==full_data[,j],na.rm=TRUE)) { # if not NAs entries in i and j are the same
                        new_colnames[which(valid_colnames==j)] <- paste0(new_colnames[which(valid_colnames==j)],"|",invalid_colnames[which(invalid_colnames==i)])
                        next
                    }
                }
            }
        }
        # remove indistinguishable events from data
        for(i in 1:length(data)) {
            data[[i]] <- data[[i]][,valid_colnames,drop=FALSE]
            colnames(data[[i]]) <- new_colnames
        }
    }
    
    return(data)
    
}

# Build adjacency matrix from B as returned from the function learn.longitudinal.phylogeny
as.adj.matrix <- function( B ) {
    
    adj_matrix <- array(0L,c((dim(B)[1]-1),(dim(B)[2]-1)))
    rownames(adj_matrix) <- colnames(B)[2:ncol(B)]
    colnames(adj_matrix) <- colnames(B)[2:ncol(B)]
    
    for(rP in 2:(nrow(B)-1)) {
        
        for(rC in ((rP+1):nrow(B))) {
            
            if(all(B[rP,1:rP]==B[rC,1:rP])&&(sum(B[rP,])==(sum(B[rC,])-1))) {
                
                adj_matrix[(rP-1),(rC-1)] <- 1
                
            }
            
        }
        
    }
    
    adj_matrix <- adj_matrix[as.character(sort(as.numeric(rownames(adj_matrix)))),]
    adj_matrix <- adj_matrix[,as.character(sort(as.numeric(colnames(adj_matrix))))]
    
    return(adj_matrix)
    
}

# Build adjacency matrix from B as returned from the function learn.longitudinal.phylogeny (unsorted)
as.adj.matrix.unsorted <- function( B, root = FALSE ) {
    
    if(root) {
        adj_matrix = array(0L,dim(B))
        rownames(adj_matrix) <- colnames(B)[(1:ncol(B))]
        colnames(adj_matrix) <- colnames(B)[(1:ncol(B))]
        for(rP in 1:(nrow(B))-1) {
            for(rC in (rP+1):nrow(B)) {
                if(all(B[rP,1:rP] == B[rC,1:rP]) && sum(B[rP,])==(sum(B[rC,])-1)) {
                    adj_matrix[(rP),(rC)] <- 1
                }   
            }
        }
    }
    else {
        adj_matrix = array(0L,(dim(B) - 1))
        rownames(adj_matrix) <- colnames(B)[(2:ncol(B))]
        colnames(adj_matrix) <- colnames(B)[(2:ncol(B))]
        for(rP in 2:(nrow(B))-1) {
            for(rC in (rP+1):nrow(B)) {
                if(all(B[rP,1:rP] == B[rC,1:rP]) && sum(B[rP,])==(sum(B[rC,])-1)) {
                    adj_matrix[(rP-1),(rC-1)] <- 1
                }   
            }
        }
    }
    
    return(adj_matrix)
    
}

# Build B from an adjacency matrix where we assume clones and mutations to be both ordered
as.B <- function( adj_matrix ) {
    
    n_clones <- dim(adj_matrix)[1]
    B <- diag(n_clones)
    
    for(k in 1:n_clones) {
        
        idx_child <- which(adj_matrix[k,]==1,arr.ind=TRUE)
        
        if(length(idx_child)==1) {
            
            B[idx_child,] <- B[k,] + B[idx_child,]
            
        } 
        else if(length(idx_child)>1) {
            
            B[idx_child,] <- sweep(B[idx_child,],2,B[k,],"+")
            
        }
        
    }
    rownames(B) <- 1:nrow(B)
    colnames(B) <- 1:ncol(B)
    B <- cbind(rep(1,nrow(B)),B)
    B <- rbind(c(1,rep(0,(ncol(B)-1))),B)
    rownames(B)[1] <- 'r'
    colnames(B)[1] <- 'r'
    
    return(B)
    
}

# check that point Q is not on segments A-B
checkQonAB <- function( Ax, Ay, Bx, By, Qx, Qy ) {
    
    if((Qx <= max(Ax,Bx)) && (Qx >= min(Ax, Bx)) && (Qy <= max(Ay, By)) && (Qy >= min(Ay, By))) {
        return(TRUE)
    }
    return(FALSE)
    
}

# check segments orientation
orientation <- function( Ax, Ay, Bx, By, Qx, Qy ) {
    
    res <- ((Qy - Ay) * (Bx - Qx)) - ((Qx - Ax)*(By - Qy))
    if(res > 0) {
        return(1)
    }
    else if(res < 0) {
        return(2)
    }
    else {
        return(0)
    }
    
}

# check if two segments interserct
checkSegmentIntersect <- function( ax1, ay1, ax2, ay2, bx1, by1, bx2, by2 ) {
    
    c1 = orientation(ax1,ay1, bx1,by1, ax2,ay2)
    c2 = orientation(ax1,ay1, bx2,by2, ax2,ay2)
    c3 = orientation(bx1,by1, ax1,ay1, bx2,by2)
    c4 = orientation(bx1,by1, ax2,ay2, bx2,by2)
    
    if((c1 != c2) && (c3 != c4)) {
        return(TRUE)
    }
    if((c1 == 0) && checkQonAB(ax1,ay1, ax2,ay2, bx1,by1)) {
        return(TRUE)
    }
    if((c2 == 0) && checkQonAB(ax1,ay1, ax2,ay2, bx2,by2)) {
        return(TRUE)
    }
    if((c3 == 0) && checkQonAB(bx1,by1, bx2,by2, ax1,ay1)) {
        return(TRUE)
    }
    if((c4 == 0) && checkQonAB(bx1,by1, bx2,by2, ax2,ay2)) {
        return(TRUE)
    }
    
    return(FALSE)
    
}

#' Compute mutation distance among variants from LACE corrected genotype and use it to perform hierarchical clustering.
#' @title compute.mutation.distance
#'
#' @examples
#' data(inference)
#' mutation_distance <- compute.mutation.distance(inference)
#'
#' @param inference Results of the inference by LACE.
#' @return A matrix mutation_distance with the mutation distance among variants computed from LACE corrected genotype and related hierarchical clustering.
#' @export compute.mutation.distance
#' @importFrom stats dist hclust
#'
compute.mutation.distance <- function( inference ) {
    
    mutation_distance <- dist(t(inference$corrected_genotype),method="euclidean")
    mutation_distance <- list(distance_matrix=as.matrix(mutation_distance),hierarchical_clustering=hclust(mutation_distance,method="complete"))
    return(mutation_distance)
    
}

#' Compute error rates for the considered variants comparing observed data to LACE corrected genotype.
#' @title compute.variants.error.rates
#'
#' @examples
#' data(longitudinal_sc_variants)
#' data(inference)
#' variants_error_rates <- compute.variants.error.rates(longitudinal_sc_variants,inference)
#'
#' @param D Mutation data from multiple experiments for a list of driver genes provided as a data matrix per time point. 
#' @param inference Results of the inference by LACE.
#' @return A matrix variants_error_rates with the estimated error rates for the considered variants.
#' @export compute.variants.error.rates
#'
compute.variants.error.rates <- function( D, inference ) {
    
    variants_error_rates <- array(NA,c(ncol(D[[1]]),2))
    rownames(variants_error_rates) <- colnames(D[[1]])
    colnames(variants_error_rates) <- c("Percentage_False_Positives","Percentage_False_Negatives")
    
    corrected_genotype <- inference$corrected_genotype
    observed_genotype <- NULL
    for(i in 1:length(D)) {
        observed_genotype <- rbind(observed_genotype,D[[i]])
    }
    
    for(i in rownames(variants_error_rates)) {
        fn <- length(which(observed_genotype[,i]==0&corrected_genotype[,i]==1))
        fp <- length(which(observed_genotype[,i]==1&corrected_genotype[,i]==0))
        variants_error_rates[i,"Percentage_False_Positives"] <- fp/nrow(observed_genotype)
        variants_error_rates[i,"Percentage_False_Negatives"] <- fn/nrow(observed_genotype)
    }
    
    return(variants_error_rates)
    
}

recursiveDescend <- function( descend ) {
    
    # Check next node,
    next_v <- which(descend$adjM[descend$row_v,] > 0)
    # If not than return
    if(length(next_v)==0) {
        descend$max_deep <- max(descend$l_path)
        return(descend)
    } else if(length(next_v) > 1){
        # brach!
        curr_level <- descend$l_path[descend$ind_lPath]
        for(v in next_v){
            descend$row_v = v
            descend$ind_lPath = descend$ind_lPath + 1
            descend$l_path[descend$ind_lPath] = curr_level + 1
            descend$mut_list <- c(descend$mut_list, as.numeric(v))
            descend <- recursiveDescend(descend = descend)
        }
    } else  {
        descend$l_path[descend$ind_lPath] = descend$l_path[descend$ind_lPath] + 1
        descend$row_v = next_v
        descend$mut_list <- c(descend$mut_list, as.numeric(next_v))
        descend <- recursiveDescend(descend = descend)
    }
    descend$max_deep <- max(descend$l_path)
    return(descend)
    
}

recursiveLongitudinalLayout <- function( idx_Vc, Xc, Yc, cl_df, adjMatrix_base, adjMatrix_overall, mut_TP, labels_show, label_offset) {

    cl_vertex_idx_Vc <- which(cl_df$cl_vertex$names == colnames(adjMatrix_overall)[idx_Vc])
    
    cl_df$cl_vertex$coord.x[cl_vertex_idx_Vc] <- Xc
    cl_df$cl_vertex$coord.y[cl_vertex_idx_Vc] <- Yc
    
    idx_next_v <- which(adjMatrix_overall[idx_Vc,] != 0)
    if(length(idx_next_v) == 0) {
        return(cl_df)
    } else if(length(idx_next_v) > 1){
        
        deep <- c()
        for(v in idx_next_v) {
            descend = list(
                adjM = adjMatrix_overall,
                l_path = c(1),
                ind_lPath = 1,
                mut_list = c(v),
                row_v = v
            )
            deep_tmp <- recursiveDescend(descend)
            dif_clones <- unique(cl_df$cl_vertex$clone[cl_df$cl_vertex$names %in% colnames(adjMatrix_overall)[deep_tmp$mut_list]])
            deep <- c(deep, length(dif_clones))
        }
        ord <- order(-cl_df$cl_vertex$TP[match(names(idx_next_v),cl_df$cl_vertex$names)], deep, decreasing = FALSE)
        
        idx_next_v <- idx_next_v[ord]
        deep <- deep[ord]
        
        offset_X = 0
        for(i in 1:length(idx_next_v)){
            
            Vn_name <- colnames(adjMatrix_overall)[idx_next_v[i]]
            
            if(adjMatrix_overall[idx_Vc,idx_next_v[i]] == 2) {
                X = Xc
                Y = mut_TP[as.character((cl_df$cl_vertex$TP[cl_df$cl_vertex$names == Vn_name]-1))] + 1
                offset_X = offset_X + deep[i] - 1
            } else {

                if(cl_df$cl_vertex$TP[cl_vertex_idx_Vc] < cl_df$cl_vertex$TP[cl_df$cl_vertex$names == Vn_name]) {
                    Y = mut_TP[as.character((cl_df$cl_vertex$TP[cl_df$cl_vertex$names == Vn_name]-1))] + 1
                } else {
                    Y = Yc + 1
                }
                
                X = (Xc + offset_X + 1)
                
                offset_X = offset_X + deep[i]
                
                idx_row_edges = which(cl_df$cl_edges$from == cl_df$cl_vertex$names[cl_vertex_idx_Vc] & cl_df$cl_edges$to == Vn_name)
                mutation_name <- cl_df$cl_vertex$last_mutation[cl_df$cl_vertex$names == Vn_name]
                
                clone_fake_edge <- data.frame(from = cl_df$cl_edges$from[idx_row_edges],
                                              to = paste0('H_',mutation_name),
                                              type = "Parental",
                                              extincion = FALSE,
                                              label = "", 
                                              name = "",
                                              lty = 1)
                
                fake_mutation_edge <- data.frame(from = paste0('H_',mutation_name),
                                                 to = mutation_name,
                                                 type = "Parental",
                                                 extincion = FALSE,
                                                 label = "",
                                                 name = "",
                                                 lty = 1)
                
                mutation_clone_edge <- data.frame(from = mutation_name,
                                                  to = cl_df$cl_edges$to[idx_row_edges],
                                                  type = "Parental",
                                                  extincion = FALSE,
                                                  label = "",
                                                  name = "",
                                                  lty = 1)
                
                fake_vertex <- data.frame(names = paste0('H_',mutation_name), 
                                          branch_level = NA,
                                          branch = NA,
                                          label = "",
                                          last_mutation = "",
                                          TP = NA, 
                                          clone = cl_df$cl_vertex$clone[cl_vertex_idx_Vc],
                                          prevalance = NA,
                                          size = 0,
                                          size2 = 0,
                                          shape = "none",
                                          label.dist = 0,
                                          label.degree = 0,
                                          extincion = 0,
                                          coord.x = X,
                                          coord.y = Yc,
                                          color = NA)
                
                mutation_vertex <- data.frame(names = mutation_name, 
                                              branch_level = NA,
                                              branch = NA,
                                              label = ifelse(test = labels_show %in% c("both", "mutations"), yes = mutation_name, no = ""),
                                              last_mutation = "",
                                              TP = NA, 
                                              clone = cl_df$cl_vertex$clone[cl_vertex_idx_Vc],
                                              prevalance = NA,
                                              size = 6,
                                              size2 = 6,
                                              shape = "square",
                                              label.dist = label_offset,
                                              label.degree = 0,
                                              extincion = 0,
                                              coord.x = X,
                                              coord.y = (Yc + Y)*0.5,
                                              color = cl_df$cl_vertex$color[cl_df$cl_vertex$names == Vn_name])
                
                cl_df$cl_edges <- cl_df$cl_edges[-idx_row_edges,]
                cl_df$cl_edges <- rbind(cl_df$cl_edges, clone_fake_edge, fake_mutation_edge, mutation_clone_edge)
                
                cl_df$cl_vertex <- rbind(cl_df$cl_vertex, fake_vertex, mutation_vertex)
                
            }
            cl_df <- recursiveLongitudinalLayout(idx_next_v[i], X, Y, cl_df, adjMatrix_base, adjMatrix_overall, mut_TP, labels_show, label_offset)
        }
    } else {

        # NO BRANCH
        Vn_name <- colnames(adjMatrix_overall)[idx_next_v]
        
        if(adjMatrix_overall[idx_Vc,idx_next_v] == 2) {
            X = Xc
            Y = mut_TP[as.character((cl_df$cl_vertex$TP[cl_df$cl_vertex$names == Vn_name]-1))] + 1
        } else {
            
            if(cl_df$cl_vertex$TP[cl_vertex_idx_Vc] < cl_df$cl_vertex$TP[cl_df$cl_vertex$names == Vn_name]) {
                Y = mut_TP[as.character((cl_df$cl_vertex$TP[cl_df$cl_vertex$names == Vn_name]-1))] + 1
            } else {
                Y = Yc + 1
            }
            
            X = Xc + 1
            
            idx_row_edges = which(cl_df$cl_edges$from == cl_df$cl_vertex$names[cl_vertex_idx_Vc] & cl_df$cl_edges$to == Vn_name)
            mutation_name <- cl_df$cl_vertex$last_mutation[cl_df$cl_vertex$names == Vn_name]
            clone_fake_edge <- data.frame(from = cl_df$cl_edges$from[idx_row_edges],
                                          to = paste0('H_',mutation_name),
                                          type = "Parental",
                                          extincion = FALSE,
                                          label = "",
                                          name = "",
                                          lty = 1)
            
            fake_mutation_edge <- data.frame(from = paste0('H_',mutation_name),
                                             to = mutation_name,
                                             type = "Parental",
                                             extincion = FALSE,
                                             label = "",
                                             name = "",
                                             lty = 1)
            
            mutation_clone_edge <- data.frame(from = mutation_name,
                                              to = cl_df$cl_edges$to[idx_row_edges],
                                              type = "Parental",
                                              extincion = FALSE,
                                              label = "",
                                              name = "",
                                              lty = 1)
            
            fake_vertex <- data.frame(names = paste0('H_',mutation_name), 
                                      branch_level = NA,
                                      branch = NA,
                                      label = "",
                                      last_mutation = "",
                                      TP = NA,#cl_df$cl_vertex$TP[cl_vertex_idx_Vc],
                                      clone = cl_df$cl_vertex$clone[cl_vertex_idx_Vc],
                                      prevalance = NA,
                                      size = 0,
                                      size2 = 0,
                                      shape = "none",
                                      label.dist = 0,
                                      label.degree = 0,
                                      extincion = 0,
                                      coord.x = X,
                                      coord.y = Yc,
                                      color = NA)
            
            mutation_vertex <- data.frame(names = mutation_name, 
                                          branch_level = NA,
                                          branch = NA,
                                          label = ifelse(test = labels_show %in% c("both", "mutations"), yes = mutation_name, no = ""),
                                          last_mutation = "",
                                          TP = NA,#cl_df$cl_vertex$TP[cl_vertex_idx_Vc],
                                          clone = cl_df$cl_vertex$clone[cl_vertex_idx_Vc],
                                          prevalance = NA,
                                          size = 6,
                                          size2 = 6,
                                          shape = "square",
                                          label.dist = label_offset,
                                          label.degree = 0,
                                          extincion = 0,
                                          coord.x = X,
                                          coord.y = (Yc + Y)*0.5,
                                          color = cl_df$cl_vertex$color[cl_df$cl_vertex$names == Vn_name])
            
            cl_df$cl_edges <- cl_df$cl_edges[-idx_row_edges,]
            cl_df$cl_edges <- rbind(cl_df$cl_edges, clone_fake_edge, fake_mutation_edge, mutation_clone_edge)
            
            cl_df$cl_vertex <- rbind(cl_df$cl_vertex, fake_vertex, mutation_vertex)
            
        }
        cl_df <- recursiveLongitudinalLayout(idx_next_v, X, Y, cl_df, adjMatrix_base, adjMatrix_overall, mut_TP, labels_show, label_offset)
    }
    
    return(cl_df)
    
}

# Draw B
draw.B <- function( B, mut_label = colnames(B)[-1], last_mut_node_label =  TRUE) {
    
    Broot <- data.tree::Node$new('r')
    Broot$mut <- B[1,]
    
    nClone <- nrow(B)
    Clones <- list(Broot)
    
    for(rP in 1:(nrow(B)-1)) {
        
        for(rC in ((rP+1):nrow(B))) {
            
            if(all(Clones[[rP]]$mut[1:rP]==B[rC,1:rP])&&(sum(Clones[[rP]]$mut)==(sum(B[rC,])-1))) {
                if(last_mut_node_label) {
                    mutName <- tail(mut_label[which(B[rC,-1]==1)], 1)
                } else {
                    mutName <- paste(mut_label[which(B[rC,-1]==1)],collapse="")  
                }
                
                Clones[[rC]] <- Clones[[rP]]$AddChild(mutName)
                Clones[[rC]]$mut <- B[rC,]
                
            }
            
        }
        
    }
    
    return(Broot)
    
}

# build B from an adjacency matrix where we assume genotypes and mutations to be both ordered
as.B.trait <- function( adj_matrix, D ) {
    
    # build data structure to save results
    n_clones <- nrow(adj_matrix)
    B <- diag(n_clones)
    
    # build B
    for(k in seq_len(n_clones)) {
        idx_child <- which(adj_matrix[k,]==1,arr.ind=TRUE)
        if(length(idx_child)==1) {
            B[idx_child,] <- B[k,] + B[idx_child,]
        }
        else if(length(idx_child)>1) {
            B[idx_child,] <- sweep(B[idx_child,],2,B[k,],"+")
        }
    }
    rownames(B) <- c("r",seq_len((nrow(B)-1)))
    mycolnames <- "r"
    for(i in 2:nrow(adj_matrix)) {
        mycolnames <- c(mycolnames,as.character(which(rownames(adj_matrix)[i]==colnames(D))))
    }
    colnames(B) <- mycolnames
    
    # return B
    return(B)
    
}

#print to log
#log2_print <- function(x, msg = "") {}
log2_print <- function(x, msg = "") {
  if (str_length(msg)>0)
    log_print(msg, console = FALSE, blank_after = FALSE, hide_notes =TRUE)
  log_print(x, console = FALSE, blank_after = FALSE, hide_notes =FALSE)
}

### load_and_save.R


## Save data

catch_ui_names <- function(inp, grep_str) {
  return(inp[grep(grep_str, names(inp))])
}

doSave <- function(grep_str, config_path, config_file) {
  initialInputs_copy_tmp <- list()
  for (i in names(inputs))
    initialInputs_copy_tmp[[i]] <- inputs[[i]]()

  initialInputs_copy <- catch_ui_names(initialInputs_copy_tmp, grep_str)
  for (n in names(initialInputs_copy) )
    initialInputs_copy <-
    initialInputs_copy[names(initialInputs_copy) %in%
                         paste0(n, "-modal") == FALSE]

  if (dir.exists(config_path))
    write.config(config.dat = initialInputs_copy,
                 file.path = file.path(config_path, config_file),
                 write.type = "yaml")
}



save_ui_state <- function(grep_str, config_file) {
  initialInputs_copy_tmp <- list()
  for (i in names(inputs))
    initialInputs_copy_tmp[[i]] <- inputs[[i]]()
  initialInputs_copy <-  catch_ui_names(initialInputs_copy_tmp, grep_str)
  for ( n in names(initialInputs_copy) )
    initialInputs_copy <-
    initialInputs_copy[names(initialInputs_copy) %in%
                         paste0(n, "-modal") == FALSE]

  if (dir.exists(config_path))
    write.config(config.dat = initialInputs_copy,
                 file.path =
                   file.path(config_path, config_file),
                 write.type = "yaml")
}


## Load data
inf_doLoad_a <- function(config_path) {
  inf_loaded_input_(read.config(file = file.path(config_path,".config_06_inf.yml")))
  loaded_input <- inf_loaded_input_()
  for (id in names(loaded_input)) {
    value <- loaded_input[[id]]
    session$sendInputMessage(id, list(value = value))
  }
}


inf_doLoad_b <- function(i) {
  req(input)
  loaded_input <- inf_loaded_input_()
  if( i %in% names(loaded_input)) {
    if(!is.null(loaded_input[[i]])) {
      #if(!is.integer(loaded_input[[i]])) {
      log2_print(loaded_input[[i]])
      #if (dir.exists(parseDirPath(roots=roots_dir, loaded_input[[i]])))
      inputs[[i]](loaded_input[[i]])
      #}
    }
  }
  else
    inputs[[i]](input[[i]])
}


inf_doLoad_c <- function() {
  req(input)
  loaded_input <- inf_loaded_input_()
  for(i in c("inf_alpha", "inf_beta")) {
    if( i %in% names(loaded_input)) {
      if(!is.null(loaded_input[[i]])) {
        log2_print(loaded_input[[i]])
        inputs[[i]](loaded_input[[i]])
        proxy<-dataTableProxy(i)
        replaceData(proxy, (do.call(cbind, inputs[[i]]())) %>% {as.data.frame((.)[,-1], row.names = (.)[,1])})
      }
    }
    else {
      inputs[[i]](session$userData[["ed_table_react_list"]][[i]]$x$data)
    }
  }
}


va_doLoad_a <- function(config_path) {
  va_loaded_input_(read.config(file = file.path(config_path,".config_05_va.yml")))
  loaded_input <- va_loaded_input_()
  for (id in names(loaded_input)) {
    value <- loaded_input[[id]]
    session$sendInputMessage(id, list(value = value))
  }
}



va_doLoad_b <- function(i) {
  req(input)
  loaded_input <- va_loaded_input_()
  if( i %in% names(loaded_input)) {
    if(!is.null(loaded_input[[i]])) {
      #if(!is.integer(loaded_input[[i]])) {
      log2_print(loaded_input[[i]])
      #if (dir.exists(parseDirPath(roots=roots_dir, loaded_input[[i]])))
      inputs[[i]](loaded_input[[i]])
      #}
    }
  }
  else
    inputs[[i]](input[[i]])
}



va_doLoad_c <- function() {
  req(input)
  loaded_input <- va_loaded_input_()
  if( 'va_verified_genes' %in% names(loaded_input)) {
    if(!is.null(loaded_input[['va_verified_genes']])) {
      #if(!is.integer(loaded_input[[i]])) {
      log2_print(loaded_input[['va_verified_genes']])
      #if (dir.exists(parseDirPath(roots=roots_dir, loaded_input[[i]])))
      inputs[['va_verified_genes']](loaded_input[['va_verified_genes']])
      inputs[['va_list_genes']](loaded_input[['va_list_genes']])

      updateSelectizeInput(session, 'va_verified_genes', 
                           selected = inputs[['va_verified_genes']](), 
                           choices = inputs[['va_list_genes']](), 
                           server = TRUE)
      #}
    }
  }
  else
    inputs[['va_verified_genes']](input[['va_verified_genes']])
}



dp_doLoad_a <- function(config_path) {
  dp_loaded_input_(read.config(file = file.path(config_path,".config_04_dp.yml")))
  loaded_input <- dp_loaded_input_()
  for (id in names(loaded_input)) {
    value <- loaded_input[[id]]
    session$sendInputMessage(id, list(value = value))
  }
}


dp_doLoad_b <- function(i) {
  req(input)
  loaded_input <- dp_loaded_input_()
  if( i %in% names(loaded_input)) {
    if(!is.null(loaded_input[[i]])) {
      if(!is.integer(loaded_input[[i]])) {
        if (dir.exists(parseDirPath(roots=roots_dir, loaded_input[[i]])))
          inputs[[i]](loaded_input[[i]])
      }
    }
  }
  else
    inputs[[i]](input[[i]])
}

thr_doLoad_a <- function(config_path) {
  thr_loaded_input_(read.config(file = file.path(config_path,".config_03_thr.yml")))
  loaded_input <- thr_loaded_input_()
  for (id in names(loaded_input)) {
    value <- loaded_input[[id]]
    session$sendInputMessage(id, list(value = value))
  }
}


thr_doLoad_b <- function(i) {
  req(input)
  loaded_input <- thr_loaded_input_()
  if( i %in% names(loaded_input)) {
    if(!is.null(loaded_input[[i]])) {
      #if(!is.integer(loaded_input[[i]])) {
      #print(loaded_input[[i]])
      #if (dir.exists(parseDirPath(roots=roots_dir, loaded_input[[i]])))
      inputs[[i]](loaded_input[[i]])
      #}
    }
  }
  else
    inputs[[i]](input[[i]])
}


av_doLoad_a <- function(config_path) {
  av_loaded_input_(read.config(file = file.path(config_path,".config_02_av.yml")))
  loaded_input <- av_loaded_input_()
  for (id in names(loaded_input)) {
    value <- loaded_input[[id]]
    session$sendInputMessage(id, list(value = value))
  }
}


av_doLoad_b <- function(i) {
  req(input)
  loaded_input <- av_loaded_input_()
  if( i %in% names(loaded_input)) {
    if(!is.null(loaded_input[[i]])) {
      if(!is.integer(loaded_input[[i]])) { #necessary for directories
        if (dir.exists(parseDirPath(roots=roots_dir, loaded_input[[i]])))
          inputs[[i]](loaded_input[[i]])
        else
          showNotification(paste(parseDirPath(roots=roots_dir, loaded_input[[i]]),'does not exist'), duration = 10, type = "warning")
      }
    }
  }
  else
    inputs[[i]](input[[i]])
}


m_doLoad_a <- function(config_path) {
  m_loaded_input_(read.config(file = file.path(config_path,".config_01_m.yml")))
  loaded_input <- m_loaded_input_()
  #print(loaded_input)
  # for (id in names(loaded_input)) {
  #  #for (id in names(initialInputs)) {
  #   value <- loaded_input[[id]]
  #   session$sendInputMessage(id, list(value = value))
  # }
}



m_doLoad_b <- function() {
  #req(input)
  loading_(TRUE)
  inputs[['sc_metadata_file']](NULL)
  loaded_input <- isolate(m_loaded_input_())
  if (!is.null(loaded_input[['sc_metadata_file']])) {
    if (!is.integer(loaded_input[['sc_metadata_file']])) {
      ## print(loaded_input[['sc_metadata_file']])
      if (file.exists(as.character(parseFilePaths(roots = roots_dir,
                                                  loaded_input[['sc_metadata_file']])[['datapath']])))
        inputs[['sc_metadata_file']](loaded_input[['sc_metadata_file']])
    }
  }
  ## else
  ##  inputs[['sc_metadata_file']](isolate(input[['sc_metadata_file']]))

  ###
  sc_metadata_label_(set_file_ui('sc_metadata_file'))

  if (file.exists(sc_metadata_label_())) {
    if (file_ext(sc_metadata_label_())  %in% c("rds", "RDS") ) {
      cellInfo <- readRDS( sc_metadata_label_() )
      if (!is.data.frame(cellInfo)){
        showNotification("No metadata file", type = "warning", duration = 10)
        cellInfo <- data.frame()
      }
    }
    else
      cellInfo <- fread(sc_metadata_label_(),
                        header = TRUE,
                        data.table = FALSE)
  } else {
    showNotification("No metadata file", type = "warning", duration = 10)
    cellInfo <- data.frame()
  }
  sc_metadata_(cellInfo)

  ## if ( file.exists(sc_metadata_label_()) ) {
  choices_(colnames(sc_metadata_()))
  ## showNotification(paste("input[['sc_metadata_file']]","choises_)"), duration = NULL)
  ## }
  ## else
  ##  choices_(c(""))
  ###


  inputs[['m_id_column']](NULL)
  ## showNotification(paste(inputs[['m_time_column']]()), duration = 10)
  loaded_input <- isolate(m_loaded_input_())
  if (!is.null(loaded_input[['m_id_column']]))
    inputs[['m_id_column']](loaded_input[['m_id_column']])
  ## else
  ##  inputs[['m_time_column']](isolate(input[['m_time_column']]))

  inputs[['m_time_column']](NULL)
  ## showNotification(paste(inputs[['m_time_column']]()), duration = 10)
  loaded_input <- isolate(m_loaded_input_())
  if (!is.null(loaded_input[['m_time_column']]))
    inputs[['m_time_column']](loaded_input[['m_time_column']])
  ## else
  ##  inputs[['m_time_column']](isolate(input[['m_time_column']]))


  inputs[['m_time_points']](NULL)
  ## loaded_input <- isolate(m_loaded_input_())
  if (!is.null(loaded_input[['m_time_points']]))
    inputs[['m_time_points']](loaded_input[['m_time_points']])
  ## else
  ##  inputs[['m_time_points']](isolate(input[['m_time_points']]))
  ## showNotification(paste("loaded", "time points"), duration = NULL)
  ## showNotification(paste(inputs[['m_time_column']]()), duration = NULL)

  loading_(FALSE)
}

## Load project .config.yml configs

inputs[["load_all_configs"]] <- reactiveVal(FALSE)

load_all_configs <- function(project_folder,
                             default_config_dir = NULL,
                             config_dir = os_conf_subdir,
                             default = FALSE,
                             recent = FALSE) {


  inputs[["load_all_configs"]](FALSE)
  browser()
  ## Change to project folder
  setwd(inputs[["project_folder_std"]]())

  config_path <- file.path(inputs[["project_folder_std"]](), os_conf_subdir)

  m_doLoad_a(config_path)
  m_doLoad_b()

  av_doLoad_a(config_path)
  uis <- av_dir_uis
  for (i in av_dir_uis) {
    av_doLoad_b(i)
  }
  thr_doLoad_a(config_path)
  uis <- thr_uis
  for(i in thr_uis) {
    thr_doLoad_b(i)
  }
  dp_doLoad_a(config_path)
  uis <- dp_uis
  for(i in dp_uis) {
    dp_doLoad_b(i)
  }

  va_doLoad_a(config_path) #
  uis <- va_uis
  for(i in va_uis) {
    va_doLoad_b(i)

  }
  va_doLoad_c()

  inf_doLoad_a(config_path)
  uis <- inf_uis
  for(i in uis) {
    inf_doLoad_b(i)
  }
  for(i in c("inf_alpha", "inf_beta")) {
    inf_doLoad_b(i)
  }
  #inf_doLoad_c() # now is called below

  inputs[["load_all_configs"]](TRUE)

  showNotification(paste("config files:", "loaded"),
                   duration = 10)

  return(inputs[["load_all_configs"]]())

}

observeEvent(inputs[["load_all_configs"]](),{
  if(inputs[["load_all_configs"]]()) {
    browser()
    inf_doLoad_c()
    for(i in c("inf_alpha", "inf_beta")) {
      if( i %in% names(inf_loaded_input_())) {
        x <- (do.call(cbind, inputs[[i]]())) %>% {as.data.frame((.)[,-1], row.names = (.)[,1])}
        x <- data.frame(rownames(x), x, check.names=FALSE)
        colnames(x)[1]=""
        session$userData[["ed_table_react_list_obj"]]$obj[[i]]$x$data <- x
        session$userData[["ed_table_react_list"]][[i]]$x$data <- x
        log2_print(session$userData[["ed_table_react_list"]][[i]]$x$data)
      }
    }
  }
})

load_all_configs2 <- function(project_folder,
                              default_config_dir = NULL,
                              config_dir = os_conf_subdir,
                              default = FALSE,
                              recent = FALSE) {

  config_path <- NULL
  if (default)
    config_path <- default_config_dir
  else
    config_path <- file.path(project_folder, config_dir)

  file_default_config_list <- NULL
  if (!recent)
    pattern <- "^\\..*"
  else
    pattern <- ".*"

  if (dir.exists(config_path))
    file_config_list <- list.files(config_path,
                                   pattern = pattern,
                                   all.files = TRUE,
                                   full.names=FALSE,
                                   no..= TRUE )

  file_config_list <- file.path(config_path, file_config_list)


  res <- TRUE
  for (cf in file_config_list) {
    loaded_input <- read.config(file = cf)
    if (basename(cf) == ".config_01_m.yml") {
      ## load_meta(loaded_input)
      m_doLoad_a(config_path)
      m_doLoad_b()
    } else {
      if (is.list(loaded_input))
        for (id in names(loaded_input)) {
          value <- loaded_input[[id]]
          ## session$sendInputMessage(id, list(value = value))

          check <- TRUE
          runjs(getWidgetType(id))

          ## print('----------->>')
          ## print(id)
          ## print((input[[paste0('inputType_',id)]]))

          if (!is.null(input[[paste0('inputType_',id)]])) {
            if (!input[[paste0('inputType_',id)]] %in% c("None")) {
              ## print('-------------')
              ## print(id)
              ## print(input[[paste0('inputType_',id)]])
              if (input[[paste0('inputType_',id)]] %in%
                  c("shinyDirectories", "shinyFiles"))
                check <- dir.exists(parseDirPath(roots=roots_dir, loaded_input[[id]]))
              else if (input[[paste0('inputType_',id)]] %in% c("select")) {
                check <- FALSE
              }
              else if (input[[paste0('inputType_',id)]] %in% c("number")) {
                check <- TRUE

              }
              else if (input[[paste0('inputType_',id)]] %in% c("shiny-text-output")) {
                check <- FALSE

              }
              else if (input[[paste0('inputType_',id)]] %in% c("selectized")) {
                check <- TRUE
              }

              if (id %in% c('thr_accepted_var', 'thr_negleted_var')) {
                check <- TRUE
                value <- loaded_input[[id]]
                inputs[[id]](loaded_input[[id]])
                session$sendInputMessage(id, list(value = value))
              }

              if (id=="va_list_genes") {
                check <- FALSE
                updateSelectizeInput(session,
                                     'va_verified_genes',
                                     choices = inputs[['va_list_genes']](),
                                     server = TRUE)
              }
              if (id=="va_verified_genes") {
                check <- FALSE
                updateSelectizeInput(session, 'va_verified_genes',
                                     selected = inputs[['va_verified_genes']](),
                                     server = TRUE)
              }

              if (check)
                inputs[[id]](loaded_input[[id]])
              else {
                if (input[[paste0('inputType_',id)]] %in%
                    c("shinyDirectories", "shinyFiles"))
                  showNotification(paste(parseDirPath(roots = roots_dir,
                                                      loaded_input[[id]]),
                                         'does not exist'),
                                   duration = 10,
                                   type = "warning")
                else
                  showNotification(paste(id, 'does not exist'),
                                   duration = 10,
                                   type = "warning")
              }
            } else {
              ## print('-------------')
              ## print(id)
              ## print('>>>>>')
            }
          }
        }
      else
        res <- FALSE
    }
  }
  Load_done <- showNotification(paste("config files:", "loaded"),
                                duration = 10)
  return(res)
}


### end of file -- load_and_save.R

#' Plot a longitudinal tree inferred by LACE.
#' @title longitudinal.tree.plot
#'
#' @examples
#' data(inference)
#' clone_labels = c("ARPC2","PRAME","HNRNPC","COL1A2","RPL5","CCT8")
#' longitudinal.tree.plot(inference = inference,
#'                        labels = "clones",
#'                        clone_labels = clone_labels,
#'                        legend_position = "topleft")
#'
#' @param inference Results of the inference by LACE.
#' @param rem_unseen_leafs If TRUE (default) remove all the leafs that have never been observed (prevalence = 0 in each time point)
#' @param show_plot If TRUE (default) output the longitudinal tree to the current graphical device.
#' @param filename Specify the name of the file where to save the longitudinal tree. Dot or graphml formats are supported and are chosen based on the extenction of the filename (.dot or .xml).
#' @param labels_show Specify which type of label should be placed on the tree; options are, 
#'   "mutations": parental edges are labeled with the acquired mutation between the two nodes (genotypes); 
#'   "clones": nodes (genotypes) are labeled with their last acquired mutation; 
#'   "both": either nodes and edges are labeled as specified above; 
#'   "none": no labels will show on the longitudinal tree. 
#' @param clone_labels Character vector that specifies the name of the nodes (genotypes). If it is NULL (default), nodes will be labeled as specified by "label" parameter.
#' @param show_prev If TRUE (default) add to clones label the correspongind prevalance.
#' @param label.cex Specify the size of the labels.
#' @param size Specify size of the nodes. The final area is proportional with the node prevalence.
#' @param size2 Specify the size of the second dimension of the nodes. If NULL (default), it is set equal to "size".
#' @param tk_plot If TRUE, uses tkplot function from igraph library to plot an interactive tree. Default is FALSE.
#' @param tp_lines If TRUE (defaul) the function draws lines between timepoints.
#' @param tp_mark If TRUE (defaul) the function draws different colored area under the nodes in different time points.
#' @param tp_mark_alpha Specify the alpha value of the area drawed when tp_mark = TRUE.
#' @param legend If TRUE (default) a legend will be displayed on the plot.
#' @param legend_position Specify the legend position.
#' @param label_offset Move the mutation labels horizontally (default = 4)
#' @param legend_cex Specify size of the legend text.
#' @return An igraph object g with the longitudinal tree inferred by LACE. 
#' @export longitudinal.tree.plot
#' @rawNamespace import(graphics, except = c("box"))
# #' @import graphics
#' @import grDevices
# '@rawNamespace import(igraph, except = c("crossing","union", "as_data_frame", "groups",  "path"))
# #' @import igraph
#' @import RColorBrewer
#' @import utils
#'
longitudinal.tree.plot <- function( inference, 
                                    rem_unseen_leafs = TRUE,
                                    show_plot = TRUE, 
                                    filename = "lg_output.xml", 
                                    labels_show = "mutations", 
                                    clone_labels = NULL, 
                                    show_prev = TRUE, 
                                    label.cex = 1, 
                                    size = 500, 
                                    size2 = NULL, 
                                    tk_plot = FALSE, 
                                    tp_lines = TRUE, 
                                    tp_mark = TRUE, 
                                    tp_mark_alpha = 0.5, 
                                    legend = TRUE, 
                                    legend_position = "topright", 
                                    label_offset = 4, 
                                    legend_cex = 0.8 ) {
    
    if(is.null(size2)) {
        size2 <- size
    }

    adjMatrix_base <- as.adj.matrix.unsorted(inference$B, root = TRUE)
    
    M_leafs <- which(apply(X = adjMatrix_base, MARGIN = 1, FUN = sum)==0)
    Clone_Mut <- sapply(inference$clones_summary, function(x){tail(x,1)}, USE.NAMES = TRUE)
    C_leafs <- Clone_Mut[match(names(M_leafs), Clone_Mut)]
    
    
    if(rem_unseen_leafs == TRUE) {

        uns_cl_mut <- C_leafs[which(inference$clones_prevalence[match(names(C_leafs),rownames(inference$clones_prevalence)), "Total"] == 0)]

            while(length(uns_cl_mut) > 0) {
                
                # Delete each clone and corresponding mutation from B, clone_prevalence and clone_summary
                del_mut <- as.character(uns_cl_mut)
                del_clone <- names(uns_cl_mut)
                
                inference$B <- inference$B[!(rownames(inference$B) %in% del_clone), !(colnames(inference$B) %in% del_mut)]
                inference$clones_prevalence <- inference$clones_prevalence[!(rownames(inference$clones_prevalence) %in% del_clone),]
                inference$clones_summary <- inference$clones_summary[!(names(inference$clones_summary) %in% del_clone)]
                
                # Repeate
                adjMatrix_base <- as.adj.matrix.unsorted(inference$B, root = TRUE)
                M_leafs <- which(apply(X = adjMatrix_base, MARGIN = 1, FUN = sum)==0)
                Clone_Mut <- sapply(inference$clones_summary, function(x){tail(x,1)}, USE.NAMES = TRUE)
                C_leafs <- Clone_Mut[match(names(M_leafs), Clone_Mut)]
                uns_cl_mut <- C_leafs[which(inference$clones_prevalence[match(names(C_leafs),rownames(inference$clones_prevalence)), "Total"] == 0)]
            }
    }
    
    cl_vertex <- data.frame(prevalance = as.vector(inference$clones_prevalence[,-ncol(inference$clones_prevalence)]))
    cl_vertex$clone <- rep(c(0:(nrow(inference$clones_prevalence)-1)),ncol(inference$clones_prevalence)-1)
    cl_vertex$TP <- rep(1:(ncol(inference$clones_prevalence)-1), each = nrow(inference$clones_prevalence))
    cl_vertex$last_mutation <- rep(c("Root",Clone_Mut), ncol(inference$clones_prevalence)-1)

    if(!is.null(clone_labels)) {
        if(length(clone_labels)!=length(inference$clones_summary)) {
            warning("Label number is different from the number of clones.")
            cl_vertex$label <- ""
        }
        else {
            cl_vertex$label <- rep(c("Root",clone_labels), ncol(inference$clones_prevalence)-1)
        }
    } else if((labels_show == "clones" || labels_show == "both") && labels_show != "none") {
        cl_vertex$label <- rep(c("Root",names(inference$clones_summary)), ncol(inference$clones_prevalence)-1)
    } else {
        cl_vertex$label <- ""
        cl_vertex$label[cl_vertex$last_mutation == "Root"] <- "Root"
    }
    
    cl_vertex$branch <- NA
    cl_vertex$branch_level <- 0
    cl_vertex$names <- paste0("T", cl_vertex$TP, "-", cl_vertex$last_mutation)
    
    # Names must be the first column
    cl_vertex <- cl_vertex[,order(ncol(cl_vertex):1)]
    
    cl_edges <- data.frame(stringsAsFactors = FALSE)
    
    # Processing persistent relations
    for(c in unique(cl_vertex$clone)) {
        t <- 1
        while(t <= max(cl_vertex$TP)) {
            from_cc <- cl_vertex$names[cl_vertex$clone == c & cl_vertex$TP == t]
            if(cl_vertex$prevalance[cl_vertex$names == from_cc] <= 0) {
                t <- t + 1
            }
            else {
                nt <- t + 1
                while(nt <= max(cl_vertex$TP)) {
                    if(max(cl_vertex$prevalance[cl_vertex$clone == c & cl_vertex$TP >= nt]) > 0) {
                        to_cc <- cl_vertex$names[cl_vertex$clone == c & cl_vertex$TP == nt]
                        cl_edges <- rbind(cl_edges, data.frame(from = from_cc, to = to_cc, type = 2L, extincion = FALSE))
                        from_cc <- to_cc
                    }
                    else {
                        to_cc <- cl_vertex$names[cl_vertex$clone == c & cl_vertex$TP == nt]
                        cl_edges <- rbind(cl_edges, data.frame(from = from_cc, to = to_cc, type = 2L, extincion = TRUE))
                        break;
                    }
                    nt <- nt + 1
                }
                t <- nt
            }
        }
    }

    # Processing parental relations
    cl_edges_parental <- data.frame(stringsAsFactors = FALSE)
    
    # Start from one leaf of M_leafs
    for(cs_id in M_leafs) {
        curTP <- max(cl_vertex$TP)
        cp_id <- which(adjMatrix_base[,cs_id]==1)
        while(length(cp_id)>0) {
            
            # Finding first time point in which appear son clone
            TPs <- cl_vertex$TP[cl_vertex$clone == (cs_id-1) & cl_vertex$prevalance > 0 & cl_vertex$TP <= curTP]
            
            if(length(TPs)==0) {
                cs_tp <- curTP
            } else {
                cs_tp <- min(TPs)
            }
            
            # Finding the time point in which parental clone were present most close to the current time point
            TPs <- cl_vertex$TP[cl_vertex$clone == (cp_id-1) & cl_vertex$prevalance > 0 & cl_vertex$TP <= cs_tp]
            
            if(length(TPs) == 0) {
                # If the parental clone has ever prevalence == 0 -> I use the same time point of the son clone
                cp_tp = cs_tp
            } else {
                cp_tp <- max(TPs)
            }
            
            from_cc <- cl_vertex$names[cl_vertex$clone == (cp_id-1) & cl_vertex$TP == cp_tp]
            to_cc <- cl_vertex$names[cl_vertex$clone == (cs_id-1) & cl_vertex$TP == cs_tp]
            mut <- cl_vertex$last_mutation[cl_vertex$clone == (cs_id-1) & cl_vertex$TP == cs_tp]
            cl_edges_parental <- rbind(cl_edges_parental, data.frame(from = from_cc, 
                                                                     to = to_cc, 
                                                                     type = 1L, 
                                                                     extincion = FALSE, 
                                                                     stringsAsFactors = FALSE))
            
            # Now son clone becomes parental clone
            curTP <- cp_tp
            cs_id <- cp_id
            cp_id <- which(adjMatrix_base[,cs_id]==1)
            
        }
    }
    
    # Fix duplicate parental relation (keep earliest) ones
    
    cl_edges_parental$from_cl <- cl_vertex$clone[match(cl_edges_parental$from, cl_vertex$names)]
    cl_edges_parental$to_cl <- cl_vertex$clone[match(cl_edges_parental$to, cl_vertex$names)]
    cl_edges_parental$to_tp <- cl_vertex$TP[match(cl_edges_parental$to, cl_vertex$names)]
    
    cl_edges_parental <- cl_edges_parental[order(cl_edges_parental$to_tp),]
    
    cl_edges_parental <- cl_edges_parental[!duplicated(cl_edges_parental[,c("from_cl", "to_cl")], fromLast = FALSE), c("from","to","type","extincion")]
    
    cl_edges <- rbind(cl_edges, cl_edges_parental)
    
    # Fixing missing persistence relations
    fixing_clones <- data.frame(names = unique(as.character(cl_edges$from)))
    
    fixing_clones$clones <- cl_vertex$clone[match(fixing_clones$names,cl_vertex$names)]
    fixing_clones$TP <- cl_vertex$TP[match(fixing_clones$names,cl_vertex$names)]
    
    fixing_clones <- fixing_clones[duplicated(fixing_clones$clones) | duplicated(fixing_clones$clones, fromLast = TRUE),]
    fixing_clones <- fixing_clones[order(fixing_clones$TP),]
    
    for(fxc in unique(fixing_clones$clones)){
        
        fixing_clones_i <- fixing_clones[fixing_clones$clones == fxc,]
        for(i in 1:(nrow(fixing_clones_i)-1)) {
            
            if(sum(cl_edges$from == fixing_clones_i$names[i] & cl_edges$to == fixing_clones_i$names[i+1] & cl_edges$type ==  2L) == 0) {
                cl_edges <- rbind(cl_edges, data.frame(from = fixing_clones_i$names[i], 
                                                       to = fixing_clones_i$names[i+1],
                                                       type = 2L,
                                                       extincion = FALSE,
                                                       stringsAsFactors = FALSE)
                )
            }
            
        }
        
    }

    # Setting edges labels
    cl_edges$label <- ""
    cl_edges$name <- ""
    
    for(i in 1:nrow(cl_edges)) {
        if(cl_edges$type[i] == 2L) {
            cl_edges$label[i] <- ""
            cl_edges$name[i] <- ""
        }
        else {
            cl_edges$label[i] <- cl_vertex$last_mutation[cl_vertex$names == cl_edges$to[i]]
            cl_edges$name[i] <- cl_edges$label[i]
        }
    }
    
    if(labels_show == "clones" || labels_show == "none") { 
        cl_edges$label <- ""
    }
    
    # Setting the size of the clone circles
    cl_vertex <- cl_vertex[(cl_vertex$names %in% cl_edges$from | cl_vertex$names %in% cl_edges$to),]
    cl_vertex$size <- 2*sqrt(size*cl_vertex$prevalance/pi)
    cl_vertex$size2 <- 2*sqrt(size2*cl_vertex$prevalance/pi)
    cl_vertex$shape <- "circle"
    
    # Setting the size of the extincted clones
    cl_vertex$size[cl_vertex$names %in% as.character(cl_edges$to[cl_edges$extincion])] <- 2*sqrt(size*0.01)
    cl_vertex$size2[cl_vertex$names %in% as.character(cl_edges$to[cl_edges$extincion])] <- 2*sqrt(size2*0.01)
    cl_vertex$shape[cl_vertex$names %in% as.character(cl_edges$to[cl_edges$extincion])] <- "vrectangle"
    
    cl_vertex$label.dist <- -1
    cl_vertex$label.degree <- 0
    
    # Removing labels from persistent clones
    idx_persistent <- sapply(cl_vertex$names, 
                             function(x){
                                 mut_type <- cl_edges$type[cl_edges$to == x]
                                 !(2L %in% mut_type) & length(mut_type) > 0
                             })
    cl_vertex$label[idx_persistent] <- ""
    if(is.null(clone_labels) && labels_show != "clones" && labels_show != "both") {
        cl_vertex$label.dist <- -1
    } else {
        cl_vertex$label.dist[idx_persistent] <- -1 
    }
    
    
    # Add prevalence
    if(show_prev) {
        num_str <- sub("^(-?)0.", "\\1.", sprintf("%.2f", cl_vertex$prevalance))
        cl_vertex$label <- paste0(cl_vertex$label, " (", num_str , ")")
        cl_vertex$label[cl_vertex$prevalance == 0.0] <- ""
        
    }
    
    # Defining extincted clones as 'ext.'
    cl_vertex$extincion <- 0
    cl_vertex$extincion[cl_vertex$names %in% as.character(cl_edges$to[cl_edges$extincion])] <- 1
    
    cl_vertex$label[cl_vertex$extincion == 1] <- ""
    if((labels_show == "clones" || labels_show == "both") && labels_show != "none") {
        cl_vertex$label[cl_vertex$extincion == 1] <- "ext."
        cl_vertex$label.dist[cl_vertex$extincion == 1] <- 1.2  
    }
    
    cl_edges$lty <- ifelse(cl_edges$type == 2L, yes = 2, no = 1)
    
    g <- igraph::graph_from_data_frame(cl_edges, directed=TRUE, vertices=cl_vertex)

    parental_clones <- which(apply(X = adjMatrix_base, MARGIN = 2, function(x) sum(x == 1))==0)
    
    c_br <- 1
    c_level <- 1
    
    while(length(parental_clones)>0) {
        
        c_son <- which(adjMatrix_base[parental_clones[1],]==1)
        
        cl_vertex$branch[cl_vertex$last_mutation == colnames(adjMatrix_base)[parental_clones[1]]] <- c_br
        cl_vertex$branch_level[cl_vertex$last_mutation == colnames(adjMatrix_base)[parental_clones[1]]] <- c_level
        
        if(length(c_son)==0) {
            parental_clones <- parental_clones[-1]
            c_br <- c_br + 1
            c_level <- 1
        }
        else if(length(c_son) > 1) {
            c_br <- c_br + 1
            c_level <- 1
            parental_clones <- c(parental_clones[-1], c_son)
        }
        else {
            c_level <- c_level + 1
            parental_clones <- c(c_son, parental_clones[-1])
        }
        
    }
    
    # Initialize palette
    color_palette <- data.frame(base_color = c("#FF0000","#00FF00","#0000FF","#FF7F00","#8B00FF","#FFFF00","#2E2B5F"), stringsAsFactors = FALSE)
    color_palette$bright <- NA
    color_palette$dark <- NA
    
    for(i in 1:nrow(color_palette)) {
        cp <- colorRampPalette(c("#FFFFFF",color_palette$base_color[i],"#000000"), interpolate = "spline")(100)
        color_palette$bright[i] <- cp[15]
        color_palette$dark[i] <- cp[85]
    }
    
    for(br in unique(cl_vertex$branch[cl_vertex$last_mutation != "Root"])) {
        n_levels <- max(cl_vertex$branch_level[cl_vertex$last_mutation != "Root"])
        
        min_level <- 6 
        
        if(n_levels < min_level) {
            n_levels = min_level
        }
        
        base_colors <- br - 1
        base_colors <- (base_colors - floor(base_colors/nrow(color_palette))*nrow(color_palette)) + 1
        
        cl_vertex$color[cl_vertex$branch == br] <- colorRampPalette(c(color_palette$bright[base_colors],
                                                                      color_palette$dark[base_colors]),
                                                                    interpolate = "linear")(n_levels)[cl_vertex$branch_level[cl_vertex$branch == br]]
        
    }
    
    cl_vertex$color[cl_vertex$names %in% as.character(cl_edges$to[cl_edges$extincion])] <- "#C0C0C0"
    cl_vertex$color[cl_vertex$last_mutation == "Root"] <- "#DDDDDD"
    
    igraph::vertex_attr(graph = g, name = "branch") <- cl_vertex$branch
    igraph::vertex_attr(graph = g, name = "branch_level") <- cl_vertex$branch_level
    igraph::vertex_attr(graph = g, name = "color") <- cl_vertex$color
    
    igraph::delete_edge_attr(graph = g, name = "extincion")

    cl_vertex$coord.x <- NA
    cl_vertex$coord.y <- NA

    g <- igraph::graph_from_data_frame(cl_edges, directed=TRUE, vertices=cl_vertex)
    #adjMatrix_overall <- igraph::get.adjacency(g, sparse = FALSE, attr = "type", names = TRUE)
    adjMatrix_overall <- igraph::as_adjacency_matrix(g, sparse = FALSE, attr = "type", names = TRUE)
    
    # adjMatrix_overall[which(adjMatrix_overall=="")] <- 0
    # adjMatrix_overall[which(adjMatrix_overall=="persistence")] <- 2
    # adjMatrix_overall[which(adjMatrix_overall=="parental")] <- 1
    # storage.mode(adjMatrix_overall) <- "integer"
    
    # Count total level number in each time point
    timepoints <- unique(cl_vertex$TP)
    
    mut_TP <- c()
    for(tp in timepoints) {
        
        col_in <- which(colnames(adjMatrix_base) %in% cl_edges$name[cl_edges$to %in% cl_vertex$names[cl_vertex$TP == tp]])
        
        if(length(col_in) < 2) {
            mut_TP <- c(mut_TP, 0)
        } else {
            
            adjMatrix_base_tp <- adjMatrix_base[col_in, col_in]
            
            idx_anc_c <- which(colSums(adjMatrix_base_tp) == 0)
            
            max_deep_tp = -1
            for(idx in idx_anc_c){
                descend = list(
                    adjM = adjMatrix_base_tp,
                    l_path = 0,
                    ind_lPath = 1,
                    mut_list = NA,
                    row_v = idx
                )
                deep_tmp <- recursiveDescend(descend)$max_deep
                max_deep_tp <- max(max_deep_tp,deep_tmp) 
            }
            mut_TP <- c(mut_TP, max_deep_tp)
        }
    }
    
    
    mut_TP <- cumsum(mut_TP) + 1:length(mut_TP)
    names(mut_TP) <- timepoints
    
    cl_df <- list(cl_vertex = cl_vertex, cl_edges = cl_edges)
    idx_next_v <-  which(colSums(adjMatrix_overall != 0) == 0)
    
    # Root
    Xc = 0
    Yc = 0
    
    # Start building
    cl_df <- recursiveLongitudinalLayout(idx_Vc=idx_next_v, 
                                         Xc=Xc, 
                                         Yc=Yc,
                                         cl_df=cl_df,
                                         adjMatrix_base=adjMatrix_base,
                                         adjMatrix_overall=adjMatrix_overall,
                                         mut_TP=mut_TP,
                                         labels_show=labels_show,
                                         label_offset=label_offset)
    
    cl_vertex <- cl_df$cl_vertex[order(cl_df$cl_vertex$TP, cl_df$cl_vertex$coord.y),]
    cl_edges <- cl_df$cl_edges
    
    
    
    
    
    g_mod <- igraph::graph_from_data_frame(d = cl_edges, directed=TRUE, vertices=cl_vertex)

    time_point_grp <- split(cl_vertex$names[cl_vertex$last_mutation!="Root"], cl_vertex$TP[cl_vertex$last_mutation!="Root"])
    
    ratio = par("din")[1] / par("din")[2]
    
    g_mod$layout <- igraph::norm_coords(as.matrix(cl_vertex[,c("coord.x", "coord.y")]), xmin = -1*ratio, xmax = 1*ratio, ymin = 1, ymax = -1)
    
    if(tk_plot) {
        igraph::tkplot(g_mod,
               rescale = FALSE,
               vertex.label.dist = 0,
               vertex.label.cex = label.cex,
               edge.label.cex = label.cex,
               edge.arrow.size = 0,
        )        
    } else if(show_plot) {
        if(tp_mark) {
            plot(g_mod,
                 mark.groups = time_point_grp,
                 mark.shape = 1,
                 mark.border = 0,
                 mark.col = paste0(RColorBrewer::brewer.pal(9, 'Pastel1'), as.hexmode(round(tp_mark_alpha*255))),
                 rescale = FALSE,
                 vertex.label.cex = label.cex,
                 edge.label.cex = label.cex,
                 edge.arrow.size = 0
            )
        } else {
            plot(g_mod,
                 rescale = FALSE,
                 vertex.label.cex = label.cex,
                 edge.label.cex = label.cex,
                 edge.arrow.size = 0
            )
        } 
        if(legend && tp_mark) {
            legend(x = legend_position, 
                   legend = paste0("Time Point ", names(time_point_grp)), 
                   fill = RColorBrewer::brewer.pal(9, 'Pastel1')[1:length(time_point_grp)], 
                   cex = legend_cex)
        } 
        if(tp_lines) {
            TPs <- sort(unique(cl_vertex$TP))
            
            min_x <- min(g_mod$layout[,1], na.rm = TRUE)
            min_x <- ifelse(test = min_x < 0, yes = min_x*1.3, no = min_x*0.7)
            
            for(i in 1:(length(TPs))) {
                
                if(i < length(TPs)) {
                    range_Y_curr <- range(g_mod$layout[cl_vertex$TP == TPs[i],2], na.rm = TRUE)
                    range_Y_next<- range(g_mod$layout[cl_vertex$TP == TPs[i+1],2], na.rm = TRUE)
                    
                    pos_l <- mean(c(range_Y_curr[1],range_Y_next[2]))
                    
                    abline(h = pos_l)
                }
                pos_text <- mean(range(g_mod$layout[cl_vertex$TP == TPs[i],2], na.rm = TRUE))
                
                text(min_x, pos_text, paste0('TP: ', TPs[i]))
                
            }
            
        }
    }
    
    if(endsWith(x = filename,suffix = ".dot")) {
        igraph::write_graph(graph = g, file = filename, format = "dot")
    } else if(endsWith(x = filename,suffix = ".xml")) {
        igraph::write_graph(graph = g, file = filename, format = "graphml")
    } else if(filename != "" || !is.null(filename) || !is.na(filename)) {
        warning("Filename is not valid. Only dot or graphml format are currently supported.")
    }
    
    return(g)
    
}

### project_mgr.R
###

## Create the user os default config folder
create_default_config_dir_std <- function() {
  if (dir.exists(path_home()))
    if (dir.exists(file.path(path_home(), os_conf[[.Platform$OS.type]]) )){
      if (!dir.exists(file.path(path_home(), os_conf[[.Platform$OS.type]], "LACE")))
        dir.create(file.path(path_home(), os_conf[[.Platform$OS.type]], "LACE") )
      if (dir.exists(file.path(path_home(), os_conf[[.Platform$OS.type]], "LACE") ))
        return(file.path(path_home(), os_conf[[.Platform$OS.type]], "LACE"))
    }
  return(NULL)
  ## logout("impossible to create default configuration folder")
}

## write the .config.yml in the user os default config folder
write_default_yaml <- function(default_config_dir) {
  res <- TRUE
  if (dir.exists(default_config_dir))
    for (y in names(default_yaml)) {
      if (!file.exists(file.path(default_config_dir, y))) {
        write.config(config.dat = default_yaml[[y]],
                     file.path =
                       file.path(default_config_dir, y),
                     write.type = "yaml")
        res <- res & file.exists(file.path(default_config_dir, y))
      }
    }
  else
    res <- FALSE
  return(res)
}

## Call create_default_config_dir_std and write_default_yaml
create_default_config <- function() {
  default_config_dir_std <- create_default_config_dir_std()
  if (!is.null(default_config_dir_std)) {
    if (!write_default_yaml(default_config_dir_std)) {
      ## logout(paste("Not all config files created"))
    }
  } else {
    ## logout(paste("No default configuration folder", default_config_dir_std))
  }
  return(default_config_dir_std)
}


## Copy default yml in the current project config folder
copy_default_configs <- function(project_folder,
                                 default_config_dir,
                                 config_dir = os_conf_subdir,
                                 seletor = NULL) {
  if (dir.exists(default_config_dir)) {
    file_default_config_list <-
      list.files(default_config_dir, all.files = TRUE, full.names=TRUE, no..= TRUE )
    if (dir.exists(project_folder))
      if (dir.exists(file.path(project_folder,config_dir))) {
        res <- file.copy(file_default_config_list,
                         file.path(project_folder,
                                   config_dir),
                         overwrite = FALSE)
        if (all(res))
          return(TRUE)
      }
  }
  return(FALSE)
}

observeEvent(reactiveValuesToList(input), {
  outs <- outputOptions(output)
  lapply(names(outs), function(name) {
    outputOptions(output, name, suspendWhenHidden = FALSE)
  })
},
once = TRUE,
priority = -2)



## Create project folder, copy default configs and load them
create_project <- function(project_folder,
                           default_config_dir,
                           config_dir=os_conf_subdir) {
  if (dir.exists(default_config_dir))
    if (dir.exists(project_folder))
      if (!dir.exists(file.path(project_folder, config_dir))) {
        dir.create(file.path(project_folder, config_dir))
        if (dir.exists(file.path(project_folder, config_dir))) {
          res_tmp <- copy_default_configs(project_folder,
                                      default_config_dir,
                                      config_dir) #overwrite=F
          res <- res_tmp
          res_tmp <- load_all_configs(project_folder,
                                        default_config_dir,
                                        config_dir,
                                        default = FALSE,
                                        recent = FALSE)
          res <- res & res_tmp
          return(res)
        }
      }
  return(FALSE)
}


## Load project configs
load_project <- function(project_folder,
                         config_dir = os_conf_subdir) {
  if (dir.exists(project_folder))
    if (dir.exists(file.path(project_folder, config_dir))) {
      load_all_configs(project_folder, config_dir)
      return(TRUE)
    }
  return(FALSE)
}

## Make .config.yml in user default .config.yml
write_default_configs <-
  function(project_folder,
           default_config_dir,
           config_dir=os_conf_subdir) {
    if (dir.exists(default_config_dir))
      if (dir.exists(project_folder))
        if (dir.exists(file.path(project_folder, config_dir))) {
          file_config_list <-
            list.files(file.path(project_folder,
                                 config_dir),
                       pattern = "^\\..*",
                       full.names = FALSE,
                       no..= TRUE )
          file_config_list <- file.path(project_folder,
                                        file_config_list)
          return(file.copy(file_config_list,
                           default_config_dir))
        }
    return(FALSE)
  }


## Make cofig.yml in project default .config.yml
write_configs <-
  function(project_folder,
           default_config_dir,
           config_dir = os_conf_subdir) {
    if (dir.exists(default_config_dir))
      if (dir.exists(project_folder))
        if (dir.exists(file.path(project_folder, config_dir))) {
          file_config_list <-
            list.files(file.path(project_folder,
                                 config_dir),
                       pattern = "^[^\\.].*",
                       full.names = FALSE,
                       no..= TRUE )
          #file_config_list <- substring(file_config_list,2)
          file_config_list <- paste0(".", file_config_list)
          file_config_list <- file.path(project_folder,file_config_list)
          return(file.copy(file_config_list, default_config_dir))
        }
    return(FALSE)
  } # should exist tab specific write_default_config ?

### end of file -- project_mgr.R

### LACEinterface.R
###
### The LACE GUI to manage longitudinal NGS datasets.
###
### See the file LICENSE in the main folder for licensing
### information.

#' @rawNamespace import(biomaRt, except = c("show","select"))
#'
#' @import foreach
#' @import doParallel
#' @rawNamespace import(curl, except = c("parse_date"))
#' @import sortable
#' @rawNamespace import(shiny, except = c("runExample", "dataTableOutput", "renderDataTable","validate"))
#' @import shinythemes
#' @import dplyr
#' @import forcats
#' @import tidyr
#' @import readr
#' @import shinyFiles
#' @import shinyjs
#' @import jsonlite
#' @import configr
#' @import DT
#' @import stringr
#' @import tools
#' @import fs
#' @rawNamespace import(data.table, except = c("first","last", "between"))
#' @import htmltools
#' @import shinyBS
#' @import bsplus
#' @import shinydashboard
#' @import callr
#' @import shinyvalidate
#' @import logr
#' @import ggplot2
#' @import svglite

## Declare name

utils::globalVariables(".my_actual_wd",
                       package = "LACE",
                       add = FALSE)


#' @md
#' @title LACE Processing and Analysis Interface
#' @description `LACEview` displays a Shiny user interface to
#'     handle the VCF and BAM files processing that is needed to
#'     construct the input for the LACE inference algorithms.
#'     The function generates also the maximum likelihood longitudinal 
#'     clonal tree, and shows the output for further explorations of
#'     the results.
#'
#' @section Installation: 
#' The package is available on GitHub and Bioconductor.
#' LACE 2.0 requires R > 4.1.0 and Bioconductor.
#'
#' To install directly from github run:
#'
#' ```
#' remotes::install_github("https://github.com/BIMIB-DISCo/LACE", 
#'                         dependencies = TRUE)
#' ```
#'
#' @section Dependencies:
#' LACE 2.0 uses *Annovar* and *Samtools suite* as back-ends for variant calling
#' annotation and depth computation, respectively.
#' 
#' *Annovar* is a variant calling software written in *Perl* freely available upon 
#' registration to their website at 
#' [https://annovar.openbioinformatics.org/en/latest/](https://annovar.openbioinformatics.org/en/latest/).
#' 
#' *Perl* can be found and installed at [https://www.perl.org/](https://www.perl.org/). 
#' 
#' *Samtools suite* is a set of tools to handle SAM/BAM/BED file format. It is 
#' freely available at [http://www.htslib.org/](http://www.htslib.org/). To install *Samtools* follow the 
#' instructions in their website. 
#'  

#' 
#' @usage 
#' LACEview()
#' 
#' @return The GUI
#' 
#' @note 
#' The function `LACE` is still available for retrocompatibility.
#' 
#' @import logr
#' 
#' @export
LACEview <- function() {
    appDir <- system.file("shinyapp", package = "LACE")
    if (appDir == "") {
        stop("LACE: Could not find package directory.",
             "Try re-installing `LACE`.",
             call. = FALSE)
    }

    ## Did you remember to set:  options(browser = 'firefox')  ?
    
    .GlobalEnv$.my_actual_wd <- getwd()
    .GlobalEnv$.my_pkg_dir <- appDir
    .GlobalEnv$.my_tmp_file <- file.path(tempdir(), "test.log")
    .GlobalEnv$.my_lf <- logr::log_open(.GlobalEnv$.my_tmp_file)
    
    msg <- ""
    
    print(paste("[info] LACE: active folder =", .GlobalEnv$.my_actual_wd))
    print(paste("[info] LACE: log file =", .GlobalEnv$.my_tmp_file))
    
    answer <- "y"
    
    
    server_pp <- function(input, output) {    
        exec_bool <- reactiveVal(TRUE)
        myOS <- .Platform$OS.type
        
        myAnnovar_script1 <- "convert2annovar.pl"
        myAnnovar_script2 <- "annotate_variation.pl"
        
        if (myOS == "windows") {
            mySamtools <- "samtools.exe"
            myPerl <- "perl.exe"
        }
        else {
            mySamtools <- "samtools"
            myPerl <- "perl"
        }
        
        myAnnovar_script1_path <- Sys.which(myAnnovar_script1)
        myAnnovar_script2_path <- Sys.which(myAnnovar_script2)
        samtools_path <- Sys.which(mySamtools)
        perl_path <- Sys.which(myPerl)
        
        if (str_length(samtools_path) == 0) {
            exec_bool(FALSE)
            msg <- paste(msg, paste("[warning] LACE:", mySamtools, "not found."), sep = "\n")
            msg <- paste(msg, paste("[info] LACE:", mySamtools, "can be set using the UI."), sep = "\n")
        }
        
        if (str_length(perl_path) == 0) {
            exec_bool(FALSE)
            msg <- paste(msg, paste("[warning] LACE:", myPerl, "not found."), sep = "\n")
        }	
        
        if (str_length(myAnnovar_script1_path) == 0) {
            #exec_bool <- FALSE
            msg <- paste(msg, paste("[warning] LACE:", myAnnovar_script1, "not found."), sep = "\n")
            msg <- paste(msg, paste("[info] LACE:", myAnnovar_script1, "folder can be set using the UI."), sep = "\n")
        }
        
        if (str_length(myAnnovar_script2_path) == 0) {
            #exec_bool <- FALSE 
            msg <- paste(msg, paste("[warning] LACE:", myAnnovar_script2, "not found."), sep = "\n")
            msg <- paste(msg, paste("[info] LACE:", myAnnovar_script2, "folder can be set using the UI."), sep = "\n")
        }
        
        if (myOS == "windows") {
            exit_status <- system2("assoc", args = ".pl", 
                                   stdout = TRUE,
                                   stderr = TRUE,
                                   wait = TRUE)
            if (str_starts(exit_status, ".pl")) {
                exec_bool(FALSE)
                msg <- paste(msg, "[warning] LACE:",".pl file extension is not associated to perl.", sep = "\n")
            }
        }
        
        #answer <- "y"
        #answer <- 1
        #answer <- TRUE
        
        #if (!exec_bool) {
        #    print("Required software is not installed or cannot be found.")
        #    answer <- readline(prompt = "Do you wish to continue and start LACE UI? [Y,n]")
        #    #answer <- menu(c("Yes", "No"), title="Do you wish to continue and start LACE UI?")
        #}
        
        
        cat(msg)
        
        ml_txt <- str_replace_all(
          paste(msg, "\n", 
                "Required software is not installed or cannot be found.",
                "Do you wish to continue and start LACE UI?",
                sep = "\n"), 
          pattern = "\n", replacement = "<br>")
        
        ml_txt <- helpText(HTML(ml_txt))
          
        mod_dlg <- modalDialog(
          title = "LACE 2.0",
          ml_txt,
          easyClose = FALSE,
          footer = tagList(
            actionButton("No", "No"),
            actionButton("Yes", "Yes")
          ))
        
        showModal(mod_dlg)
        
        observeEvent(input$No, {
            answer <<- "n"
            stopApp()
        })
        
        observeEvent(input$Yes, {
            answer <<- "y"
            stopApp()
        })
    }        
        
    ui_pp <- basicPage(
        #print("test"),
        #actionButton("show", "Show modal dialog")
    )
    
    
    shiny::runApp(list(ui = ui_pp, server = server_pp))
    
    
    if (!(answer %in% c("y", "Y", "yes", "Yes", "ni", ""))) {
        return(NULL)
    }
    
    
    on.exit({
        setwd(.my_actual_wd)
        rm(list = c(".my_actual_wd"), pos = .GlobalEnv)
        logr::log_close()
        print("LACEview closed")
    })

    shiny::runApp(file.path(appDir, 'app.R'), display.mode = "normal", launch.browser = TRUE)
}


### end of file -- LACEinterface.R

### info_help.R --
###
### Help strings for various bits and pieces of the LACE GUI.
###
### See file LICENSE in the main folder for licensing and copyright
### information.

text = list()


text[["pr_tab_help"]] =
    list("Create a project where the inferential, the longitudinal tree and the
configuration for data filtering can be stored and reloaded. If there
is already a project with the same path, it is automatically
recognized and loaded upon request. Use the sidebar to open recent
projects, or save and reload tab specific configurations.",
      br(),
      br(),
      "In the sidebar you can also choose one of the 'demo projects' to load
some pre-computed data and see the results.")


text[["m_tab_help"]] =
    list("Information regarding the longitudinal single cell experiment. At a
minimum, information about the cells' identifiers and the sample they
belong to is required.  The identifiers are used to identify the file
names to include in the anlysis and the samplings represent the time
points. The cells used in the downstream analysis are the intersection
between the identifiers present in the metadata and the BAM files
stored in the drive.",
         br(),
         br(),
         "You may have to create such a file from scratch. See the (i) button
below.",
         br(),
         br(),
         "If you have loaded one of the 'demo projects', changes on this tab are
not considered.")

text[["av_tab_help"]] =
    list("Annotation of single nucleotide variants for each cell require
Annovar, a database of variant annotations provided by or downloadable
at", tags$a("https://annovar.openbioinformatics.org/en/latest/user-guide/download/"),", and the respective VCF files. If the database of
variant annotations is not suitable for the experiment under analysis,
follow the instructions from Annovar to generate your own
database.",
         br(),
         br(),
         "If you have loaded one of the 'demo projects', changes on this tab are
not considered.")

text[["thr_tab_help"]] =
    list("Single nucleotide variants are not all the same. Some of them are
characterized by low quality score, and mutations regarding exonic
regions may cause specific functional effect on the resulting
proteins. You can use the alternative frequency, minor allele
frequency and the minimum number of cells with the same mutation to
select the SNVs. Choose the functional exonic SNVs to considers e.g.,
if the case, unknown SNVs could be neglected, and reduce the number of
SNPs in the next steps.",
         br(),
         br(),
         "If you have loaded one of the 'demo projects', changes on this tab are
not considered.")

text[["dp_tab_help"]] =
    list("The number of reads per SNV site represents an optimal filter to
retrieve relevant mutations. Depth at specific sites is usually not
provided in standard alignment or variant calling pipelines and are
computationally expensive. Provide the folder with aligned data and,
if not found already, the samtools executable location to compute
depth only on sites passing the filters set in the previous tab.",
         br(),
         br(),
         "If you have loaded one of the 'demo projects', changes on this tab are
not considered.")

text[["va_tab_help"]] =
    list("Relevant variants allows to reproduce more significant longitudinal
clonal tree for the experiment. Apply some filters to retrieve
relevant variants as part of the processing step.",
         br(),
         br(),
         "Press 'Select variant' to do some pre-processing right now, and then,
interactively modify filters and see the results on selected variants
as they are changed. Otherwise, set all the filters and go next to
perform processing and inference all together.")


text[["inf_tab_help"]] =
    list("Set the preferences for the Monte Carlo Markov Chain (MCMC) used in
the inferential step. Add, at least, one false positive 'alpha' and
false negative 'beta' rate for each sampling time point. Multiple sets
of alpha and beta can be added by inserting values on the 'add row'
cells. If multiple rate sets are provided, then the best set of alpha
e beta are evaluated as part of the inference and returned.",
         br(),
         br(),
         "Press 'Run LACE' to save your current configuration, start processing
data and infer the longitudinal clonal tree. If the metadata, and
filters did not change, expensive variant and depth computations are
performed only on cells previously not included.")


text[["pr_name"]] =
    "Give a name to your project\'s experiment. All the results will be stored inside a subfolder with the same name."

text[["pr_folder"]] =
    "Select a folder to store the the project. Path containing project is recognized and saved state is reloaded."

text[["sc_metadata_file"]] =
    "Select file containing at least two columns, one with cell ids (such that id.vcf and id.bam identify the respective files) and the other with the sampling times (\'before drug\', etc.). Accepted file formats are csv and tsv, or rds containg a data.frame object. Headers are required. The list can contains more cells than what are stored in HD."

text[["m_idCol"]] =
    "Select the column containing cell ids. (such that id.vcf and id.bam identify the respective files)"

text[["m_timePointsCol"]] =
    "Select the column containing the sampling times. Sampling time tags can concide with the ids if the case."

text[["m_timePoints"]] =
    "Select and drag one or more sampling points to order them chronologically."

text[["m_doSave"]] =
    "Alpha release: Save state for the current tab."

text[["m_doLoad"]] =
    "Alpha release: Load state for the current tab. Inconguencies are reported (you messed with the config files)."

text[["av_exec"]] =
    "Alpha release: Run current part of the pipeline (in the last tab run the whole pipeline [alpha: did u test it before releasing it?])."

text[["av_anovar_exec_dir"]] =
    'Folder containing the Annovar executables. Annovar is available at this <a href="https://annovar.openbioinformatics.org/en/latest/user-guide/download/">link</a>.'

text[["av_anovar_db_dir"]] =
    "Annovar requires a database with annotations.
     Some are distributed with the software. Refer to the website guide to create your own database of annotated variants."

text[["av_vcf_in_dir"]] =
    "This is the folder where variant calling files are stored.
     File names need to have the format ID.vcf, where ID belongs to the metadata list, otherwise the file is neglected."

text[["thr_maf"]] =
    "Minor allele frequency (MAF) for each referenced SNP included in a default global population, ranging from <1 to <50%.
MAF is a ratio defined as
$MAF={Alleles positive for the variant}/{Total alleles screened}$ . When there are more than two alleles at a variant location, the second most frequent is used to calculate MAF."

text[["thr_alleles_ratio"]] =
    "Alternate allele frequency is the ratio of the obseved alterative allele and total obervations at that locus in a cell."

text[["thr_freq"]] =
    "How frequent is a variant found in a sample. It represents the ratio of cells where the mutation was observed over the total number of cells at a specific sampling time."

text[["thr_bucket_var_list"]] =
    tags$ul(
      tags$li("frameshift insertion: an insertion of one or more nucleotides that cause frameshift changes in protein coding sequence"), 
      tags$li("frameshift deletion: a deletion of one or more nucleotides that cause frameshift changes in protein coding sequence"), 
      tags$li("frameshift block substitution: a block substitution of one or more nucleotides that cause frameshift changes in protein coding sequence"),
      tags$li("stopgain:  a mutation that lead to the immediate creation of stop codon at the variant site"), 
      tags$li("stoploss: a mutation that lead to the immediate elimination of stop codon at the variant site"), 
      tags$li("nonframeshift insertion: an insertion of 3 or multiples of 3 nucleotides that do not cause frameshift changes in protein coding sequence"), 
      tags$li("nonframeshift deletion: a deletion of 3 or mutliples of 3 nucleotides that do not cause frameshift changes in protein coding sequence"),
      tags$li("nonframeshift block substitution: a block substitution of one or more nucleotides that do not cause frameshift changes in protein coding"),
      tags$li("nonsynonymous SNV: a single nucleotide change that cause an amino acid change"), 
      tags$li("synonymous SNV: a single nucleotide change that does not cause an amino acid change"), 
      tags$li("unknown:  unknown function (due to various errors in the gene structure definition in the database file)"))

text[["dp_samtools_exec_dir"]] =
    'Folder containing the Samtools suit executables. Samtools is available at this <a href="https://www.htslib.org/">link</a>.'

text[["dp_bam_dir"]] =
    "This is the folder where Binary Alignment Map files are stored.
File names need to have the format ID.bam, where ID is in the metadata list, otherwise the file is neglected."

text[["va_depth_minimum"]] =
    "Minimum depth to set values to NA."

text[["va_missing_values_max"]] =
    "Maximum number of considered missing data per gene."

text[["va_minumum_median_total"]] =
    "Minimum median depth for total reads."

text[["va_minumum_median_mutation"]] =
    "Minimum median depth for reads supporting mutations."

text[["va_verified_genes"]] =
    "Depending on the experiment, some genes may be more relevant than others.
The list of available genes changes based on the above filters.
Leave empty to consider all the vailable genes."

text[["inf_learning_rate"]] =
    "Parameter to tune the probability of accepting solutions at lower values during mcmc. Value of learning_rate = 1 (default), set a probability proportional to the difference in likelihood; values of learning_rate greater than 1 inclease the chance of accepting solutions at lower likelihood during mcmc while values lower than 1 decrease such probability."

text[["inf_alpha"]] =
    "False positive error rate provided as list of elements; if a vector of alpha (and beta) is provided, the inference is performed for multiple values and the solution at maximum-likelihood is returned."

text[["inf_beta"]] =
    "False negative error rate provided as list of elements; if a vector of beta (and alpha) is provided, the inference is performed for multiple values and the solution at maximum-likelihood is returned."

text[["inf_num_iter"]] =
    "Maximum number of mcmc steps to be performed during the inference."

text[["inf_num_rs"]] =
    "Number of restarts during mcmc inference."

text[["inf_n_try_bs"]] =
    "Number of steps without change in likelihood of best solution after which to stop the mcmc."

text[["inf_num_processes"]] =
    "Number of processes to be used during parallel execution. To execute in single process mode, this parameter needs to be set to either NA or NULL."

text[["inf_seed"]] =
    "Seed for reproducibility."

text[["inf_random_tree"]] =
    "Boolean. Shall I start MCMC search from a random tree? If FALSE (default) and initialization is NULL, search is started from a TRaIT tree (BMC Bioinformatics . 2019 Apr 25;20(1):210. doi: 10.1186/s12859-019-2795-4)."

text[["inf_marginalize"]] =
    "Boolean. If true, the attachment of cells to clones matrix is marginalized when computing likelihood."

text[["inf_keep_equivalent"]] =
    "Boolean. If true, results (the maximum likelihood longitudinal clonal tree, the attachment of cells to clones matrix) at equivalent likelihood with the best returned solution are returned."

text[["inf_check_indistinguishable"]] =
    "Boolean. If true, any indistinguishable event is removed from input data prior inference."

text[["inf_error_move"]] =
    "Boolean. If true, estimation of error rates in the MCMC moves is included."

text[["inf_show"]] =
    "Boolean. Show the interactive interface to explore the output."


### end of file -- info_help.R

# library(shiny)
# library(DT)
# library(stringr)



dt_output = function(title, id) {
  #fluidRow(
  #  column(12, h1(paste0('Table ', sub('.*?([0-9]+)$', '\\1', id), ': ', title)),
  #  hr(), DTOutput(id)
  #))
  DTOutput(id, width = "100%")
}
render_dt = function(data, server = TRUE, editable = 'cell', ...) {
  #renderDT(data, selection = 'none', server = server, editable = editable, ...)
  renderDT(data, server = server)
}



ed_table <- function(id, tb_data, row_basename='row_', session) {

  input <- session$input
  options(htmlwidgets.TOJSON_ARGS = list(na = 'null'))



  tb_data_tmp <- tb_data
  newrow_tmp<-as.list(rep(NA,ncol(tb_data_tmp)))
  names(newrow_tmp) <- colnames(tb_data_tmp)
  tb_data_tmp[nrow(tb_data_tmp)+1,]=newrow_tmp
  r_names <- paste0(row_basename,seq_len(nrow(tb_data_tmp)))
  r_names[length(r_names)] <- "add row"
  rownames(tb_data_tmp) = r_names


  js_key <- c(
    "table.on('key', function(e, datatable, key, cell, originalEvent){",
    "  var targetName = originalEvent.target.localName;",
    "  if(key == 13 && targetName == 'body'){",
    "    $(cell.node()).trigger('dblclick.dt');",
    "  }",
    "});",
    "table.on('keydown', function(e){",
    "  var keys = [9,13,37,38,39,40];",
    "  if(e.target.localName == 'input' && keys.indexOf(e.keyCode) > -1){",
    "     console.log($(e.target).val());",
    "     if((!$(e.target).val() || $(e.target).val().length === 0 )) {$(e.target).val('ee')};",
    "     console.log($(e.target).val());",
    "    $(e.target).trigger('blur');",
    "  }",
    "});",
    "table.on('key-focus', function(e, datatable, cell, originalEvent){",
    "  var targetName = originalEvent.target.localName;",
    "  var type = originalEvent.type;",
    "  if(type == 'keydown' && targetName == 'input'){",
    "    if([9,37,38,39,40].indexOf(originalEvent.keyCode) > -1){",
    "      $(cell.node()).trigger('dblclick.dt');",
    "    }",
    "  }",
    "});"
  )


  isolate({
    #session$userData[['ed_table_react_list']][[id]] <- NULL
    session$userData[['ed_table_react_list']][[id]] <- datatable(
      tb_data_tmp,
      selection = "none",
      editable = list(target = 'cell', disable = list(columns = c(0))),
      callback = JS(js_key),
      extensions = c('Scroller',"KeyTable"),#,"Buttons"),
      options = list(
        keys = TRUE,
        #keys = list(blurable= FALSE, editOnFocus = F),
        select =list(blurable=FALSE),
        #deferRender = F,
        scrollY = 100,
        scroller = TRUE,
        dom = 'Bit' #,
        #buttons = c('copy')
      )
    )
    log2_print('create alfa and beta tables')
    log2_print(session$userData[['ed_table_react_list']][[id]]$x$data)
    #print('step 0')
    #print(ed_table_react_list)
    #print(ed_table_react_list[[id]])
    #print(ed_table_react_list[[id]]$x$data)
    #print('step 0b')


  })


  #print('step 1')
  uis_tmp = list(id)



  ###
  #print ('step a')

  # observeEvent(input[[paste0(id,'_cell_edit')]],{
  #   #browser()
  #   print(id)
  #   ed_table_react_list[[id]]$x$data <<- ed_table_react_list[[id]]$x$data[-1]
  # }, priority = 1, once = TRUE)
  #isolate({
  if (!is.null(session$userData$ed_table_react_list_obj$observers[[paste0(id,'cell_edit')]]))
    session$userData$ed_table_react_list_obj$observers[[paste0(id,'cell_edit')]]$destroy()
  #})

  session$userData$ed_table_react_list_obj$observers[[paste0(id,'cell_edit')]]<-observeEvent(input[[paste0(id,'_cell_edit')]],{

    log2_print('-------------')
    #req(inputs[[paste0(id,'_cell_edit')]])
    # print(input[[paste0(id,'_cell_edit')]])
    # print('aaa')
    #old_e <- inputs[[paste0(id,'_cell_edit_old')]]()

    new_e <- isolate(inputs[[paste0(id,'_cell_edit')]]())
    #log2_print('ccc')
    act_e <- isolate(input[[paste0(id,'_cell_edit')]])
    #log2_print('bbb2')

    # print('old_e')
    # print(old_e)
    # print('act_e')
    #log2_print(act_e)
    # print('new_e')
    #log2_print(new_e)
    if ((new_e$col != act_e$col) | (new_e$row != act_e$row) | (new_e$value != act_e$value)) {
      #print((new_e$col != act_e$col) | (new_e$row != act_e$row) | (new_e$value != act_e$value))
      #print(new_e$col != act_e$col)
      #print(new_e$row != act_e$row)
      #print(new_e$value != act_e$value)
      #print(typeof(new_e$value))
      #print(typeof(act_e$value))
      #log2_print('made equal')
      #log2_print(paste('should pre fire', id))
      #print(act_e)
      #print(new_e)

      #inputs[[paste0(id,'_cell_edit_old')]](act_e)
      inputs[[paste0(id,'_cell_edit')]](act_e)
      #old_e <- inputs[[paste0(id,'_cell_edit_old')]]()


      #print(inputs[[paste0(id,'_cell_edit')]]())

    }
    else {
      #log2_print(paste('should no prefire', id))
    }
  })

  if (!is.null(session$userData$ed_table_react_list_obj$observers[[paste0(id,'cell_edit_s')]]))
    session$userData$ed_table_react_list_obj$observers[[paste0(id,'cell_edit_s')]]$destroy()

  session$userData$ed_table_react_list_obj$observers[[paste0(id,'cell_edit_s')]]<-observeEvent(inputs[[paste0(id,'_cell_edit')]](),{
    #log2_print(paste('pre fire', id))
    log2_print(isolate(inputs[[paste0(id,'_cell_edit')]]()))

    # if(inputs[[paste0(id,'_cell_edit')]]()$row==-1)
    # {
    #   print('return')
    #   return()
    # }
    #
    new_e <-  isolate(inputs[[paste0(id,'_cell_edit')]]())
    act_e <-  isolate(input[[paste0(id,'_cell_edit')]])
    #print(paste('new_e',new_e))
    #print(paste('act_e',act_e))
    #if ((new_e$col == act_e$col) | (new_e$row == act_e$row) | (new_e$value == act_e$value)) {
    #  print('return')
    #  return()
    #}

    #return()
    ev <- isolate(inputs[[paste0(id,'_cell_edit')]]())

    x <- id

    #print(id)
    #print('data')
    #print(session$userData[['ed_table_react_list']][[id]]$x$data)
    mm <- isolate(session$userData[['ed_table_react_list']][[id]]$x$data)
    extra_col <- all(mm[[1]] == rownames(mm))
    if (is.na(extra_col))
      extra_col <- FALSE
    if (extra_col){
      mm <- mm[-1]
      #print('extra_col')
    }
    #print('mm')
    #print(mm)
    #print('step 2')
    #print(mm)
    #browser()
    if(length(as.character(ev$value))>=1)
      if(str_length(as.character(ev$value))>=1) {
        if (input[[paste0(id,'_cell_edit')]]$row == nrow(mm)) {
          #print('step 3a')
          for (j in seq(1, ncol(mm))) {
            if (is.na(mm[nrow(mm),j]))
              mm[nrow(mm),j] <- "0.001"
          }
          newrow<-as.list(rep(NA,ncol(mm)))
          #names(newrow) <- colnames(mm)
          df_newrow <-data.frame(newrow)
          names(df_newrow) <- colnames(mm)

          mm <- rbind(mm, df_newrow)
          rownames(mm)[nrow(mm)] <- "add row"
        }
      }
    #print('step 3')
    r_names <- paste0(row_basename, seq(1:nrow(mm)))
    r_names[length(r_names)] <- "add row"
    rownames(mm) <- r_names

    val_tmp<-isolate(ev)
    if(length(as.character(ev$value))>=1)
      if (str_length(as.character(ev$value))==0)
        val_tmp$value=NA
    #print('step 4')
    if (is.na(mm[val_tmp$row, val_tmp$col]))
      mm[val_tmp$row, val_tmp$col]=val_tmp$value
    mm <- editData(mm, val_tmp, proxy = id, resetPaging = FALSE, rownames=TRUE )

    for (i in seq(1:max(nrow(mm)-1,1))) { # i = val_tmp$row
      canc <-0
      for (j in seq(1, ncol(mm))) {
        if (is.na(mm[i,j])) {
          canc <- canc+1
        }
      }
      #print("2222222222222222222222")
      #print(canc == ncol(mm) && i<nrow(mm))
      #print(nrow(mm))
      if(canc == ncol(mm) && i<nrow(mm)) {
        if (nrow(mm)>2) {
          mm_pre <- NULL
          mm_post <- NULL
          if (i>1)
            mm_pre <- mm[1:i-1,]
          if (i<nrow(mm))
            mm_post <- mm[(i+1):nrow(mm),]
          mm <- rbind(mm_pre, mm_post)
          r_names <- paste0(row_basename, seq(1:nrow(mm)))
          r_names[length(r_names)] <- "add row"
          rownames(mm) <- r_names
          break
        }
        else
        {
          for (j in seq(1, ncol(mm))) {
            if (is.na(mm[nrow(mm),j]))
              mm[1,j] <- "0.001"
          }
        }
      }
    }
    # if(is.na(mm[val_tmp$row, val_tmp$col]))
    #   mm[val_tmp$row, val_tmp$col]=NA


    proxy<-dataTableProxy(id)
    #print('mm')
    #print(mm)

    session$userData[['ed_table_react_list']][[id]]$x$data <- editData(mm, val_tmp, resetPaging = FALSE, rownames=TRUE )
    replaceData(proxy, mm)
    #print('update')
    log2_print(isolate(session$userData[['ed_table_react_list']][[id]]$x$data))
    #print(proxy)
    #session$userData[['ed_table_react_list']][[id]]$x$data
    #print(id)
    #print(ed_table_react_list[[id]]$x$data)
    #print('fine')
  }, ignoreInit = TRUE)
  ###

  return(isolate(session$userData[['ed_table_react_list']][[id]]))

}



# d1 = iris[1:3,1:4]
#
# shinyApp(
#   ui = fluidPage(
#     title = 'Double-click to edit table cells',
#     dt_output('prova', 'mytable'),
#     dt_output('prova', 'mytable2'),
#     DTOutput('inf_alpha')
#   ),
#
#   server = function(input, output, session) {
#
#     empty_table <- data.frame(matrix(ncol = 2, nrow = 0))
#     colnames(empty_table) <- c('T1', 'T2')
#     ed_table_alpha <- ed_table('inf_alpha', empty_table, 'alpha_', session)
#     output$inf_alpha <- render_dt(ed_table_alpha)
#
#     d1_tmp <-d1
#     rownames(d1_tmp) = c('alpha_1','alpha_2','alpha_3')
#     newrow2<-as.list(rep(NA,ncol(d1_tmp)))
#     names(newrow2) <- colnames(d1_tmp)
#     d1_tmp[3,]=newrow2
#
#     #options(DT.options = list(pageLength = 8))
#     ed_table_1 <- ed_table('mytable', d1, 'beta___', session)
#     ed_table_2 <- ed_table('mytable2', d1, 'alpha_', session)
#     output$mytable <- render_dt(ed_table_1)
#     output$mytable2 <- render_dt(ed_table_2)
#     #print(ed_table_1)
#   }
# )

is.zip <- function(filepath){
  result <- tryCatch({
    unzip(filepath, list = TRUE)
    return(TRUE)
  }, error = function(e){
    return(FALSE)
  })
  return(result)
}



download_demo <- function() {
  bStatus <- FALSE
  
  #check if demo already downloaded
  appDir <- system.file("shinyapp", package = "LACE")
  if(stringr::str_length(appDir)>0) {
    if(!dir.exists(file.path(appDir, "Rambow_dataset")) || !dir.exists(file.path(appDir, "Small_dataset"))) {
      bStatus <- FALSE
      
      tmp=tempdir()
      log2_print(tmp, msg = "download_demo: Rambow tmp folder =")
      list.files(tmp)
      
      #download repository
      url <- "https://github.com/gian-asco/LACE-DEMO/archive/refs/tags/1.1.0.zip"
      filen <- file.path(tmp,"LACE_demos.zip")
      tryCatch({
        if (curl::has_internet()) {
          #httr::GET(url, httr::write_disk(filen, overwrite=TRUE), httr::timeout(10))
          file_path <- curl::curl_fetch_disk(url, filen)
          log2_print(file_path, msg = "download_demo: downloading dataset url =")
        }
        #return(TRUE)
      }, error = function(e){
        log2_print("Download error", msg = "download_demo:")
        #return(FALSE)
      })
      
      
      #check file
      if(file.exists(filen)) 
        bStatus <- is.zip(filen)
      else
        bStatus <- FALSE
      
      if(bStatus)
      {
        unzip(filen, exdir=tmp)
        file.copy(from = file.path(tmp, "LACE-DEMO-1.1.0", "Rambow_dataset"), to = appDir, overwrite = TRUE, recursive = TRUE)
        file.copy(from = file.path(tmp, "LACE-DEMO-1.1.0", "Small_dataset"), to = appDir, overwrite = TRUE, recursive = TRUE)
      }
    }
    else
      bStatus <- TRUE
  }
  
  
  return(bStatus)
}

check_demo <- function() {
  bStatus <- FALSE
  appDir <- system.file("shinyapp", package = "LACE")
  if(stringr::str_length(appDir)>0) {
    if(!dir.exists(file.path(appDir, "Rambow_dataset")) || !dir.exists(file.path(appDir, "Small_dataset"))) {
      bStatus <- download_demo()
    }
    else {
      bStatus <- TRUE
      log2_print("demo already downloaded", msg = "check_demo:")
    }
  }
  
  return(bStatus) 
}
  

#if(!check_demo())
#  #print warning and do not proceede
#else
#  #proceed

### inference_tab.R

### Inference reactiveVals ####

inf_uis = c("inf_num_iter",
            "inf_num_rs",
            "inf_n_try_bs",
            "inf_num_processes",
            "inf_seed",
            "inf_random_tree",
            "inf_marginalize",
            "inf_keep_equivalent",
            "inf_check_indistinguishable",
            "inf_show",
            "inf_learning_rate",
            "inf_error_move")


# inf_rvs = reactiveValues(inf_buttons = list(),
#                          inf_observers = list())


for (inf_ui in inf_uis)
  inputs[[inf_ui]] <- reactiveVal()

inputs[['inf_alpha']] <- reactiveVal()
inputs[['inf_beta']] <- reactiveVal()
inputs[['inf_params']] <- reactiveVal()


if (is.null(session$userData[["ed_table_react_list"]])) {
  session$userData[["ed_table_react_list"]] <- reactiveValues()
}

if (is.null(session$userData[["ed_table_react_list_obj"]])) {
  session$userData[["ed_table_react_list_obj"]] <- list()
  session$userData[["ed_table_react_list_obj"]]$obj = reactiveValues()
  session$userData[["ed_table_react_list_obj"]]$observers = list()
}


## Create empty table
if (!is.null(inputs[['m_time_column']])) {
  empty_table <-
    data.frame(matrix(0,
                      ncol = length(inputs[['m_time_column']]()),
                      nrow = 1))
  colnames(empty_table) <-
    c(inputs[['m_time_column']]())
} else {
  empty_table <- data.frame(matrix(0, ncol = 1, nrow = 1))
  colnames(empty_table) <- c('T1')
}

## print(empty_table)
ed_tables <- list()

## Create false positive and false negative UI table set to empty table
ed_tables[['inf_alpha']] <-
  ed_table('inf_alpha', empty_table, 'alpha_', session)

ed_tables[['inf_beta']] <-
  ed_table('inf_beta', empty_table, 'beta_', session)


empty_cell <- data.frame(row=-1, col=-1 , value=0)
inputs[[paste0('inf_alpha','_cell_edit')]] <- reactiveVal(empty_cell)
inputs[[paste0('inf_beta','_cell_edit')]] <- reactiveVal(empty_cell)

### End inference reactiveVals ####

### Inference functions ####

hide_tab <- function() {
  #browser()
  shinyjs::show(id="computation_idCol_div")
  hideTab(inputId = "main_tabset", target = "Project")
  hideTab(inputId = "main_tabset", target = "SC metadata")
  hideTab(inputId = "main_tabset", target = "Annotations")
  hideTab(inputId = "main_tabset", target = "Filters")
  hideTab(inputId = "main_tabset", target = "SC sampling depths")
  hideTab(inputId = "main_tabset", target = "Variants")
  hideTab(inputId = "main_tabset", target = "Inference")
  hideTab(inputId = "main_tabset", target = "Longitudinal display")
}

show_tab <- function(disp = FALSE) {
  #browser()
  shinyjs::hide(id="computation_idCol_div")
  showTab(inputId = "main_tabset", target = "Project")
  showTab(inputId = "main_tabset", target = "SC metadata")
  showTab(inputId = "main_tabset", target = "Annotations")
  showTab(inputId = "main_tabset", target = "Filters")
  showTab(inputId = "main_tabset", target = "SC sampling depths")
  showTab(inputId = "main_tabset", target = "Variants")
  showTab(inputId = "main_tabset", target = "Inference")
  showTab(inputId = "main_tabset", target = "Longitudinal display")
  if (disp)
    #delay(100,
          updateTabsetPanel(session, "main_tabset", selected = "Longitudinal display")
    #)

}


inf_exec <- function() {
  browser()
  if (length(va_out_dir_()) == 0)
    return()
  if (dir.exists(va_out_dir_()))
    if (file.exists(file.path(va_out_dir_(), "D.RData"))) {
      load(file.path(va_out_dir_(),"D.RData"))


      make_numbers <- function(tab) {
        tab_tmp <- as.data.frame(lapply(tab, FUN = as.numeric), check.names = FALSE)
        rownames(tab_tmp) <- rownames(tab)
        return(tab_tmp)
      }

      remove_repeated_rownames <- function(tab) {
        extra_col <- all(tab[[1]] == rownames(tab))
        if (is.na(extra_col))
          extra_col <- FALSE
        if (extra_col) {
          tab <- tab[ , -1, drop = FALSE]
          remove_repeated_rownames(tab)
        }
        return(tab)
      }


      remove_na_row <- function(tab, row_min=2, left_rows=1) {

        row_basename <- rownames(tab)[1]
        if (str_length(row_basename)>0) {
          row_basename <- paste0(str_split_fixed(row_basename, pattern = "_", n = 2)[1], "_")
        }
        else
          row_basename <- "row_"

        keep_row <- c()

        if ( left_rows>0){
          if (left_rows>nrow(tab))
            left_rows <- nrow(tab)-1
          keep_row <- seq((nrow(tab)- left_rows):nrow(tab))
          log2_print("keep_row")
          log2_print(keep_row)
        }
        if ((nrow(tab) - left_rows)>0)
          range <-seq(1:(nrow(tab) - left_rows))
        else
          range <- NULL
        log2_print("range")
        log2_print(range)
        for (i in range) {
          canc <-0
          for (j in seq(1, ncol(tab))) {
            if (is.na(tab[i,j])) {
              canc <- canc+1
            }
            log2_print(canc)
          }
          if(canc < ncol(tab))
            keep_row <- c(keep_row, i)
          else
          {
            for (j in seq(1, ncol(tab))) {
              if (is.na(tab[i,j]))
                tab[i,is.na(tab[i,])] <- 0.0
            }
          }
        }

        log2_print(keep_row)

        tab_tmp <- tab[keep_row, ,drop = FALSE]
        if (nrow(tab_tmp)<=row_min)
          tab_tmp <- tab

        return(tab_tmp)
      }

      put_valid_rates <- function(tab) {

        r_name <- rownames(tab)
        tab_tmp <- as.data.frame(lapply(tab, FUN = abs), check.names = FALSE)
        tab_tmp[tab_tmp>=1.0] <- 0.999
        tab_tmp[is.na(tab_tmp)] <- 0.001
        rownames(tab_tmp) <- r_name

        return(tab_tmp)
      }

      fix_table <- function(tab) {

        tab <- remove_repeated_rownames(tab)
        tab <- make_numbers(tab)
        tab <- remove_na_row(tab, 0, 0)
        tab <- put_valid_rates(tab)

        return (tab)
      }


      #browser()

      alpha <- session$userData[["ed_table_react_list"]][["inf_alpha"]]$x$data
      alpha <- fix_table(alpha)
      r_names <- rownames(alpha)
      alpha <- transpose(as.data.table(alpha))
      colnames(alpha) <- r_names

      beta <- session$userData[["ed_table_react_list"]][["inf_beta"]]$x$data
      beta <- fix_table(beta)
      r_names <- rownames(beta)
      beta <- transpose(as.data.table(beta))
      colnames(beta) <- r_names



      ## Notes:
      ## MA: this is most likely best represented as a R
      ## 'formula'.

      string_tmp <-
        "LACE('D' = D,
                          'learning_rate' = input[['inf_learning_rate']],
                          'num_iter' = input[['inf_num_iter']],
                          'num_rs' = input[['inf_num_rs']],
                          'n_try_bs' = input[['inf_n_try_bs']],
                          'num_processes' = input[['inf_num_processes']],
                          'seed' = input[['inf_seed']],
                          'random_tree' = input[['inf_random_tree']],
                          'marginalize' = input[['inf_marginalize']],
                          'keep_equivalent' = input[['inf_keep_equivalent']],
                          'check_indistinguishable' = input[['inf_check_indistinguishable']],
                          'error_move' = input[['inf_error_move']],
                          'show' = input[['inf_show']],
                          'alpha' = as.list(alpha),
                          'beta' = as.list(beta)
                    )
                    "

      list_par <-
        list('D' = D,
             'learning_rate' = input[['inf_learning_rate']],
             'num_iter' = input[['inf_num_iter']],
             'num_rs' = input[['inf_num_rs']],
             'n_try_bs' = input[['inf_n_try_bs']],
             'num_processes' = input[['inf_num_processes']],
             'seed' = input[['inf_seed']],
             'random_tree' = input[['inf_random_tree']],
             'marginalize' = input[['inf_marginalize']],
             'keep_equivalent' = input[['inf_keep_equivalent']],
             'check_indistinguishable' = input[['inf_check_indistinguishable']],
             'error_move' = input[['inf_error_move']],
             'show' = input[['inf_show']],
             'alpha' = as.list(alpha),
             'beta' = as.list(beta)
        )



      # ##!!! ATTENZIONE QUESTO E' UN TEST FATTO CON 3 CELLULE TUTTE IN TIME_1
      # ## ATTENZIONE LACE1.0 NON ACCETTA che non ci siano cellule ad un tempo
      # ## OVVIO MA ANCHE NO
      # ## QUI SI DUPLICANO I TEMPI VUOTI
      # ## COSI NON DA ERRORE
      # ## METTERE UN CHECK PRIMA DI PROCEDERE
      # D$`before treatment`[1,2] <- 1
      # D$`4d on treatment` <- D$`before treatment` #copia
      # D$`28d on treatment` <- D$`before treatment` #copia
      # D$`57d on treatment` <- D$`before treatment` #copia
      #
      # rownames(D$`4d on treatment`)= c("S1","S2","S3") #nome cellule
      # rownames(D$`28d on treatment`)= c("S4","S5","S6")
      # rownames(D$`57d on treatment`)= c("S9","S8","S7")
      #
      # list_par[["D"]] <- D # must be erased
      #
      # ##!!! fine preparazione ATTENZIONE QUESTO E' UN TEST FATTO CON 3 CELLULE
      # ## CANCELLARE QUANTO SOPRA PRIMA DI MANDARE


      show_tmp <- list_par[["show"]]
      list_par[["show"]] <- FALSE
      inputs[['inf_params']](list_par)

      #browser()
      #inference_res <- LACE::LACE(D, show = F)
      inference_res=do.call('LACE',inputs[['inf_params']]())

      save(inference_res,file=file.path(inputs[["project_folder_std"]](),"inference.RData"))

      returned_vals <- show_result(rs_())
      rs_(returned_vals$rs)
      #inputs[["res_lastrefresh"]](returned_vals$message)
    }
}



show_result <- function (rs, show = TRUE) {


  #browser()


  so <- sopt()
  so$stderr <- "|"
  sopt(so)

  returned_vals <- list()

  if (file.exists(file.path(inputs[["project_folder_std"]](),"inference.RData"))) {
    load(file=file.path(inputs[["project_folder_std"]](),"inference.RData"))

    B <- inference_res$B
    clones_prevalence <- inference_res$clones_prevalence
    C <- inference_res$C
    error_rates <- inference_res$error_rates

    if (!is.null(B) && !is.null(C) && !is.null(clones_prevalence) && !is.null(error_rates))
      if (show) {
        #browser()
        x <- lace_interface(
          B_mat = B,
          clones_prevalence = clones_prevalence,
          C_mat = C,
          error_rates = error_rates
        )
        
        #browser()
        prev_graph <- x$prevalence %>% as.data.frame() %>%  select(-Total) %>% t() %>% as.data.frame()
        prev_graph_w <- prev_graph %>% 
          as_tibble(rownames = "time") %>% 
          pivot_longer(cols = -time, names_to = "clone", values_to = "prevalence") %>% 
          mutate(clone=factor(clone, levels=unique(clone)))
        #prev_graph
        #colours = brewer.pal(n = ncol(prev_graph), name = "Paired")
        colours =  colorRampPalette(brewer.pal(12, "Paired"))(ncol(prev_graph))
        
        pl1 <- ggplot(prev_graph_w, aes(x=time, y=prevalence, color=clone, group=clone, label=round(prevalence, digits=2))) +
          geom_line() + geom_point(size=3) + scale_color_manual(values=colours)
        #+
        #geom_text_repel(box.padding   = 0.7, point.padding = 0.0, segment.color = 'black')
        
        pl2 <- ggplot(prev_graph_w, aes(x=time, y=prevalence, group=clone, color=clone, fill=clone)) + 
          geom_area(alpha=0.8 , size=0.5, colour="black") + 
          geom_point(size=1, colour="black",  position = "stack") +
          scale_fill_manual(values=colours) + 
          theme(
            #panel.background = element_rect(fill = "NA", color = "black"), 
            panel.grid= element_line(colour="grey50", size=0.2)#,
            #panel.ontop = TRUE
            #plot.background = element_rect(fill = "grey50")
          )
        
        
        ggsave(file=file.path(inputs[["project_folder_std"]](),"prevalences.svg"), plot=pl1, width=10, height=10)
        ggsave(file=file.path(inputs[["project_folder_std"]](),"stream.svg"), plot=pl2, width=10, height=10)
        
        
        x <- x[["html"]]
        #browser()



        long_job <- function(port, x) {
          library(shiny)
          library(shinyjs)
          library(htmltools)
          library(htmlwidgets)
          #library(LACE)

          res <- list(
            "ui" = fluidPage(
              useShinyjs(),
              LACE:::LACEOutput("LCT_FP", height = "900px", width = "900px")
            ),
            "server" = function (input, output, session) {
              output$LCT_FP = LACE:::renderLACE(x)
              outputOptions(output,
                            "LCT_FP",
                            suspendWhenHidden = FALSE)
              hide(id = "navbar")
            }
          )

          server_env <- environment(res$server)
          server_env$x <- x
          shiny::runApp(res, display.mode = "normal", launch.browser=FALSE, port = port, quiet = FALSE )
        }

        port(httpuv::randomPort(min = 1024L, max = 49151L, host = "127.0.0.1", n = 20))

        if (class(rs)[1] == "r_session")
        {
          #rs$kill_tree()
          rs$close()
          rs$finalize()
        }

        rs <- callr::r_session$new(options = sopt(), wait = TRUE)
        rs$supervise(TRUE)
        rs$poll_process(4)

        rs$call(long_job, args= list("port" = port(), "x" = x))

        adr <- reactive({
          Sys.sleep(stime())
          log2_print(paste0("http://127.0.0.1:", port()))
          paste0("http://127.0.0.1:", port())
        })


        stime(12)
        log2_print(adr)
        output$res <- renderUI({
          tags$iframe(src=adr(),  height=1000, width=1000, frameborder = "no")
        })

        outputOptions(output,
                      "res",
                      suspendWhenHidden = FALSE)
        #browser()
        stime(15)
        # stime(2) # melius
        # stime(10) # abund
        # stime(10) # quam
        # stime(10)
        # stime(10)

        returned_vals$message <- paste("Last update: ", date())
        returned_vals$rs <- rs

        inputs[["long_job"]](inputs[["long_job"]]()+1)



      }
  } else {
    #returned_vals$message <- paste("Inference results not available. Run LACE first.")
    returned_vals$message <- paste("")

    returned_vals$rs <- rs

    showNotification("Inference results not available. Run LACE first.", duration = 10, type = "warning")
    
    inputs[["long_job"]](inputs[["long_job"]]()+1)

  }

  inputs[["res_lastrefresh"]](returned_vals$message)


  return(returned_vals)

}


observeEvent(inputs[["long_job"]](),{
  #browser()
  show_tab(TRUE)
}, ignoreInit = TRUE)

## inf_go

inf_go <- function() {
  browser()
  # if(inputs[["demo"]]() == "Small_dataset")
  if(!is.null(inputs[["demo"]]()))
  {
    inf_exec()
  } else
  {
    #browser()
    av_exec()
    thr_exec()
    dp_exec()
    va_exec()
    va_exec2()
    inf_exec()
  }

  # shinyjs::hide(id="computation_idCol_div")
  # showTab(inputId = "main_tabset", target = "SC metadata")
  # showTab(inputId = "main_tabset", target = "Annotations")
  # showTab(inputId = "main_tabset", target = "Filters")
  # showTab(inputId = "main_tabset", target = "SC sampling depths")
  # showTab(inputId = "main_tabset", target = "Variants")
  # showTab(inputId = "main_tabset", target = "Inference")
  # browser()
  # delay(2000,
  #       updateTabsetPanel(session, "main_tabset", selected = "Longitudinal display")
  # )

}

### End inference functions ####

### Inference observers ####

observeEvent(reactiveValuesToList(input), {
  outs <- outputOptions(output)
  lapply(names(outs), function(name) {
    outputOptions(output, name, suspendWhenHidden = FALSE)
  })
},
once = TRUE,
priority = -1)


observe({
  inf_observers = lapply(inf_uis,
                         function(i) {
                           observeEvent(input[[i]], {
                             ## req(input)
                             inputs[[i]](input[[i]])
                           })


                           output[[i]] <- renderText(inputs[[i]]())
                         }
  )
},
priority = -1)


observeEvent(session$userData[["ed_table_react_list"]][["inf_alpha"]], {
  ## inputs[["inf_alpha"]](session$userData[['ed_table_react_list']][['inf_alpha']]$x$data)
  ## inputs[["inf_alpha"]](session$userData[['ed_table_react_list_obj']]$obj[["inf_alpha"]]$x$data)
  ## inputs[["inf_alpha"]](do.call(cbind, yaml::yaml.load(yaml::as.yaml(session$userData[['ed_table_react_list_obj']]$obj[["inf_alpha"]]$x$data))) %>% {as.data.frame((.)[,-1], row.names = (.)[,1])})

  #inputs[["inf_alpha"]](session$userData[['ed_table_react_list']][['inf_alpha']]$x$data)
  #data.frame(rownames(session$userData[['ed_table_react_list']][['inf_alpha']]$x$data), session$userData[['ed_table_react_list']][['inf_alpha']]$x$data)
  x <- data.frame(rownames(session$userData[['ed_table_react_list']][['inf_alpha']]$x$data), session$userData[['ed_table_react_list']][['inf_alpha']]$x$data, check.names=FALSE)
  colnames(x)[1]=""
  inputs[["inf_alpha"]](x)
  #inputs[["inf_alpha"]](session$userData[["ed_table_react_list_obj"]]$obj[["inf_alpha"]]$x$data)
})


observeEvent(session$userData[["ed_table_react_list"]][["inf_beta"]], {
  x <- data.frame(rownames(session$userData[['ed_table_react_list']][['inf_beta']]$x$data), session$userData[['ed_table_react_list']][['inf_beta']]$x$data, check.names=FALSE)
  colnames(x)[1]=""
  inputs[["inf_beta"]](x)
})


## Find out interactively the time points and use them as columns
observeEvent(inputs[['m_time_column']](), {
  if (!is.null(inputs[['m_time_points']]) &
      length(inputs[['m_time_points']]()) > 0) {
    empty_table <-
      data.frame(matrix(0.001,
                        ncol = length(inputs[['m_time_points']]()),
                        nrow = 1))
    colnames(empty_table) <- c(inputs[['m_time_points']]())
  } else {
    empty_table <- data.frame(matrix(0.001, ncol = 1, nrow = 1))
    colnames(empty_table) <- c('T1')
  }
  ## ed_tables <- list()
  ## id <- 'inf_alpha'
  for (id in c('inf_alpha', 'inf_beta')) {
    ed_tables[[id]] <-
      ed_table(id,
               empty_table,
               paste0(str_split(id,pattern = '_')[[1]][2],
                      '_'),
               session)
    ## output[[id]] <- render_dt(ed_tables[[id]])
    ## outputOptions(output, id, suspendWhenHidden = FALSE)
  }
})


## Change the order of columns accordingly to time points user
## choice
observeEvent(inputs[['m_time_points']](), {
  req(inputs[['m_time_column']]())
  if(inputs[["load_all_configs"]]()) {
    id <- 'inf_alpha'
    for (id in c('inf_alpha', 'inf_beta')) {
      ## remove extra column
      mm <- isolate(session$userData[["ed_table_react_list"]][[id]]$x$data)
      extra_col <- all(mm[[1]] == rownames(mm))
      if (is.na(extra_col))
        extra_col <- FALSE
      if (extra_col)
        mm <- mm[-1]

      ## Reorder columns
      mm <- mm[,inputs[['m_time_points']](), drop = FALSE]
      session$userData[["ed_table_react_list"]][[id]]$x$data <- mm

      ## Remove extra last row on NAN
      mm <- mm[-nrow(mm),, drop = FALSE]

      ## Create table and render it
      ## output[[id]] <-NULL
      ed_tables[[id]] <-
        ed_table(id,
                 mm,
                 paste0(str_split(id,pattern = '_')[[1]][2], '_'),
                 session)
      log2_print(id)
    }
    id <- 'inf_alpha'
    ## output[['inf_alpha']] <-render_dt(ed_tables[['inf_alpha']])
    session$userData[["ed_table_react_list_obj"]]$obj[[id]] <-
      ed_tables[[id]]
    ## outputOptions(output, id, suspendWhenHidden = FALSE)
    proxy<-dataTableProxy(id)
    replaceData(proxy, mm)

    id <- 'inf_beta'
    ## output[['inf_beta']] <-render_dt(ed_tables[['inf_beta']])
    session$userData[["ed_table_react_list_obj"]]$obj[[id]] <-
      ed_tables[[id]]

    ## outputOptions(output, id, suspendWhenHidden = FALSE)
    proxy<-dataTableProxy(id)
    replaceData(proxy, mm)
    ## refresh()
  }
})

# inputs[["inf_go"]] <- reactiveVal(F)
# observeEvent(inputs[["inf_go"]](), {
#   hideTab(inputId = "main_tabset", target = "SC metadata")
#   shinyjs::hide("SC metadata")
#   shinyjs::hide("Annotations")
#   shinyjs::hide("Filters")
#   shinyjs::hide("SC sampling depths")
#   shinyjs::hide("Variants")
#
#   hideTab(inputId = "main_tabset", target = "Annotations")
#   hideTab(inputId = "main_tabset", target = "Filters")
#   hideTab(inputId = "main_tabset", target = "SC sampling depths")
#   hideTab(inputId = "main_tabset", target = "Variants")
#   hideTab(inputId = "main_tabset", target = "Inference")
#
# }, priority = 5)

observeEvent(input[["inf_next"]], {
  ## inf_exec()
  #browser()

  #print("GGGGGGGGGGGGGGGGGGGGG")
  #inputs[["inf_go"]](!inputs[["inf_go"]]())
  # hideTab(inputId = "main_tabset", target = "Project")
  # hideTab(inputId = "main_tabset", target = "SC metadata")
  # hideTab(inputId = "main_tabset", target = "Annotations")
  # hideTab(inputId = "main_tabset", target = "Filters")
  # hideTab(inputId = "main_tabset", target = "SC sampling depths")
  # hideTab(inputId = "main_tabset", target = "Variants")
  # hideTab(inputId = "main_tabset", target = "Inference")
  # shinyjs::show(id="computation_idCol_div")
  #browser()
  hide_tab()

  delay(2000, inf_go())



},
ignoreInit = TRUE)


observe({
  hide(id = "inf_alpha_div")
  if (file.exists(sc_metadata_label_()) )
    if (!is.null(inputs[['m_time_column']]))
      if (inputs[['m_time_column']]() > 0)
        shinyjs::show(id="inf_alpha_div")
})

observe({
  hide(id = "inf_beta_div")
  if ( file.exists(sc_metadata_label_()) )
    if (!is.null(inputs[['m_time_column']]))
      if (inputs[['m_time_column']]() > 0)
        shinyjs::show(id="inf_beta_div")
})

### End inference observes ####

### Inference outputs ####
output[['inf_alpha']] <-
  DT::renderDT(session$userData[["ed_table_react_list_obj"]]$obj[['inf_alpha']],
               server = TRUE)
outputOptions(output, 'inf_alpha', suspendWhenHidden = FALSE)
output[['inf_beta']] <-
  DT::renderDT(session$userData[["ed_table_react_list_obj"]]$obj[['inf_beta']],
               server = TRUE)
outputOptions(output, 'inf_beta', suspendWhenHidden = FALSE)


output[['inf_par']] <- renderPrint({
  alpha <-
    transpose(as.data.table(session$userData[["ed_table_react_list"]][["inf_alpha"]]$x$data));
  alpha <-
    alpha[, 1:(ncol(alpha) - 1)];
  beta <-
    transpose(as.data.table(session$userData[["ed_table_react_list"]][["inf_beta"]]$x$data));
  beta <-
    beta[,1:(ncol(beta)-1)];

  list("inf_learning_rate" = input[["inf_learning_rate"]],
       "inf_num_iter" = input[["inf_num_iter"]],
       "inf_num_rs" = input[['"inf_num_rs"']],
       "inf_n_try_bs" = input[["inf_n_try_bs"]],
       "inf_num_processes" = input[["inf_num_processes"]],
       "inf_seed" = input[["inf_seed"]],
       "inf_random_tree" = input[["inf_random_tree"]],
       "inf_marginalize" = input[["inf_marginalize"]],
       "inf_keep_equivalent" = input[["inf_keep_equivalent"]],
       "inf_check_indistinguishable" = input[["inf_check_indistinguishable"]],
       "inf_error_move" = input[["inf_error_move"]],
       "inf_show" = input[["inf_show"]],
       "inf_alpha" = as.list(alpha),
       "inf_beta" = as.list(beta) )
})


### End inference outputs ####


### end of file -- inference_tab.R

#' Perform the inference of the maximum likelihood clonal tree from longitudinal data.
#' @title LACE
#'
#' @examples
#' data(longitudinal_sc_variants)
#' inference = LACE(D = longitudinal_sc_variants,
#'                  lik_w = c(0.2308772,0.2554386,0.2701754,0.2435088),
#'                  alpha = list(c(0.10,0.05,0.05,0.05)),
#'                  beta = list(c(0.10,0.05,0.05,0.05)),
#'                  keep_equivalent = TRUE,
#'                  num_rs = 5,
#'                  num_iter = 10,
#'                  n_try_bs = 5,
#'                  num_processes = NA,
#'                  seed = 12345,
#'                  verbose = FALSE,
#'                  show = FALSE)
#'
#' @param D Mutation data from multiple experiments for a list of driver genes. It can be either a list with a data matrix per time point or a SummarizedExperiment object.
#' In this latter, the object must contain two fields: assays and colData. Assays stores one unique data matrix pooling all single cells observed at each time point and colData stores a vector of labels reporting the time point when each single cell was sequenced.
#' Ordering of cells in assays field and colData field must be the same.
#' @param lik_w Weight for each data point. If not provided, weights to correct for sample sizes are used.
#' @param alpha False positive error rate provided as list of elements; if a vector of alpha (and beta) is provided, the inference is performed for multiple values and the solution at
#' maximum-likelihood is returned.
#' @param beta False negative error rate provided as list of elements; if a vector of beta (and alpha) is provided, the inference is performed for multiple values and the solution at
#' maximum-likelihood is returned.
#' @param initialization Binary matrix representing a perfect philogeny clonal tree; clones are rows and mutations are columns.
#' This parameter overrides "random_tree".
#' @param random_tree Boolean. Shall I start MCMC search from a random tree? If FALSE (default) and initialization is NULL, search
#' is started from a TRaIT tree (BMC Bioinformatics . 2019 Apr 25;20(1):210.  doi: 10.1186/s12859-019-2795-4).
#' @param keep_equivalent Boolean. Shall I return results (B and C) at equivalent likelihood with the best returned solution?
#' @param check_indistinguishable Boolean. Shall I remove any indistinguishable event from input data prior inference?
#' @param num_rs Number of restarts during mcmc inference.
#' @param num_iter Maximum number of mcmc steps to be performed during the inference.
#' @param n_try_bs Number of steps without change in likelihood of best solution after which to stop the mcmc.
#' @param learning_rate Parameter to tune the probability of accepting solutions at lower values during mcmc. Value of learning_rate = 1 (default), set a
#' probability proportional to the difference in likelihood; values of learning_rate greater than 1 inclease the chance of accepting solutions at lower likelihood
#' during mcmc while values lower than 1 decrease such probability.
#' @param marginalize Boolean. Shall I marginalize C when computing likelihood?
#' @param error_move Boolean. Shall I include estimation of error rates in the MCMC moves?
#' @param num_processes Number of processes to be used during parallel execution. To execute in single process mode,
#' this parameter needs to be set to either NA or NULL.
#' @param seed Seed for reproducibility.
#' @param verbose Boolean. Shall I print to screen information messages during the execution?
#' @param log_file log file where to print outputs when using parallel. If parallel execution is disabled, this parameter is ignored.
#' @param show Boolean. Show the interactive interface to explore the output.
#'
#' @return A list of 9 elements: B, C, clones_prevalence, relative_likelihoods, joint_likelihood, clones_summary and error_rates. Here, B returns the maximum likelihood longitudinal
#' clonal tree, C the attachment of cells to clones, corrected_genotypes the corrected genotypes and clones_prevalence clones' prevalence; relative_likelihoods and joint_likelihood are respectively
#' the likelihood of the solutions at each individual time points and the joint likelihood; clones_summary provide a summary of association of mutations to clones. In equivalent_solutions,
#' solutions (B and C) with likelihood equivalent to the best solution are returned. Finally error_rates provides the best values of alpha and beta among the considered ones.
#' @export LACE
#' @import parallel
#' @rawNamespace import(SummarizedExperiment, except = c("trim","shift", "show"))
# #' @import SummarizedExperiment
#' @import data.tree
#' @importFrom Rfast rowMaxs
#' @importFrom stats runif rnorm setNames
#'
LACE <- function( D,
                  lik_w = NULL,
                  alpha = NULL,
                  beta = NULL,
                  initialization = NULL,
                  random_tree = FALSE,
                  keep_equivalent = TRUE,
                  check_indistinguishable = TRUE,
                  num_rs = 50,
                  num_iter = 10000,
                  n_try_bs = 500,
                  learning_rate = 1,
                  marginalize = FALSE,
                  error_move = FALSE,
                  num_processes = Inf,
                  seed = NULL,
                  verbose = TRUE,
                  log_file = "",
                  show = TRUE ) {

    # Set the seed
    set.seed(seed)

    # Handle SummarizedExperiment objects
    if(typeof(D)=="S4") {
        curr_data <- t(assays(D)[[1]])
        curr_experiment <- colData(D)[[1]]
        curr_experiment_unique <- unique(curr_experiment)
        curr_D <- list()
        for(i in 1:length(curr_experiment_unique)) {
            curr_D[[as.character(curr_experiment_unique[i])]] <- curr_data[which(curr_experiment==curr_experiment_unique[i]),,drop=FALSE]
        }
        D <- curr_D
    }

    # Set storage mode to integer
    for(i in 1:length(D)) {
        storage.mode(D[[i]]) <- "integer"
    }

    # Remove any indistinguishable event from input data prior inference
    if(check_indistinguishable) {
        D <- check.indistinguishable(D)
    }

    # If initiatilation is NULL and we do not require a random tree, we initializa with TRaIT tree
    if(is.null(initialization) && !random_tree) {

        # set initial tree from where to start MCMC search
        data <- Reduce(rbind,D)
        data[which(is.na(data))] <- 0
        marginal_probs <- matrix(colSums(data,na.rm=TRUE)/nrow(data),ncol=1)
        rownames(marginal_probs) <- colnames(data)
        colnames(marginal_probs) <- "Frequency"
        joint_probs <- array(NA,c(ncol(data),ncol(data)))
        rownames(joint_probs) <- colnames(data)
        colnames(joint_probs) <- colnames(data)
        for (i in seq_len(ncol(data))) {
            for (j in seq_len(ncol(data))) {
                val1 <- data[,i]
                val2 <- data[,j]
                joint_probs[i,j] <- (t(val1)%*%val2)
            }
        }
        joint_probs <- joint_probs/nrow(data)
        adjacency_matrix <- array(0,c(ncol(data),ncol(data)))
        rownames(adjacency_matrix) <- colnames(data)
        colnames(adjacency_matrix) <- colnames(data)
        pmi <- joint_probs
        for(i in seq_len(nrow(pmi))) {
            for(j in seq_len(ncol(pmi))) {
                pmi[i,j] <- log(joint_probs[i,j]/(marginal_probs[i,"Frequency"]*marginal_probs[j,"Frequency"]))
            }
        }
        ordering <- names(sort(marginal_probs[,"Frequency"],decreasing=TRUE))
        adjacency_matrix <- adjacency_matrix[ordering,ordering]
        adjacency_matrix[1,2] = 1
        if(nrow(adjacency_matrix)>2) {
            for(i in 3:nrow(adjacency_matrix)) {
                curr_c <- rownames(adjacency_matrix)[i]
                curr_candidate_p <- rownames(adjacency_matrix)[seq_len((i-1))]
                adjacency_matrix[names(which.max(pmi[curr_candidate_p,curr_c]))[1],curr_c] <- 1
            }
        }
        adjacency_matrix <- rbind(rep(0,nrow(adjacency_matrix)),adjacency_matrix)
        adjacency_matrix <- cbind(rep(0,nrow(adjacency_matrix)),adjacency_matrix)
        adjacency_matrix[1,2] = 1
        initialization <- as.B.trait(adj_matrix=adjacency_matrix,D=D[[1]])

    }

    # Initialize weights to compute weighted joint likelihood
    if(is.null(lik_w)) {
        total_sample_size <- 0
        for(i in 1:length(D)) {
            total_sample_size <- total_sample_size + nrow(D[[i]])
        }
        for(i in 1:length(D)) {
            lik_w <- c(lik_w,(1-(nrow(D[[i]])/total_sample_size)))
        }
        lik_w <- lik_w / sum(lik_w)
    }

    # Initialize error rates alpha and beta
    if(is.null(alpha)) {
        alpha = list(rep(0.01,length(D)))
        beta = list(rep(0.05,length(D)))
    }

    if(verbose) {
        cat(paste0("Starting inference for a total of ",length(alpha)," different values of alpha and beta.","\n"))
    }

    # Setting up parallel execution
    if(num_rs>1) {
        if(is.na(num_processes) || is.null(num_processes) || num_processes == 1) {
            num_processes = 1
        }
        else if(num_processes==Inf) {
            cores <- as.integer((detectCores()-1))
            if(cores < 2) {
                num_processes <- 1
            }
            else {
                num_processes <- min(cores,num_rs)
            }
        }
        else {
            num_processes <- min(num_processes,num_rs)
        }
        if(verbose && num_processes>1) {
            cat("Executing",num_processes,"processes via parallel...","\n")
        }
    }
    else {
        num_processes = 1
    }

    # Now start the inference

    # Sequential computation
    inference <- list()

    for(i in 1:length(alpha)) {

        inference[[i]] <- learn.longitudinal.phylogeny( D = D,
                                                        lik_w = lik_w,
                                                        alpha = alpha[[i]],
                                                        beta = beta[[i]],
                                                        initialization = initialization,
                                                        keep_equivalent = keep_equivalent,
                                                        num_rs = num_rs,
                                                        num_iter = num_iter,
                                                        n_try_bs = n_try_bs,
                                                        learning_rate = learning_rate,
                                                        marginalize = marginalize,
                                                        num_processes = num_processes,
                                                        error_move = error_move,
                                                        seed = round(runif(1)*10000),
                                                        verbose = verbose,
                                                        log_file = log_file)
    }

    # Return the solution at maximum likelihood among the inferrend ones
    lik <- NULL
    for(i in 1:length(inference)) {
        lik <- c(lik,inference[[i]][["joint_lik"]])
    }
    best <- which(lik==max(lik))[1]
    error_rates <- list(alpha=inference[[best]][["alpha"]],beta=inference[[best]][["beta"]])

    # compute corrected genotypes
    inference_B <- inference[[best]][["B"]]
    rownames(inference_B)[1] <- 0
    rownames(inference_B) <- (as.numeric(rownames(inference_B))+1)
    colnames(inference_B)[1] <- 0
    colnames(inference_B) <- (as.numeric(colnames(inference_B))+1)
    inference_attachments <- (unlist(inference[[best]][["C"]])+1)
    inference_attachments_ordered <- inference_attachments
    for(curr_C_index in 1:length(inference_attachments)) {
        inference_attachments_ordered[curr_C_index] <- as.integer(colnames(inference_B)[inference_attachments[curr_C_index]])
    }
    inference_attachments <- inference_attachments_ordered
    idx_B_curr <- sort.int(as.integer(colnames(inference_B)),index.return=TRUE)$ix
    inference_B <- inference_B[idx_B_curr,]
    inference_B <- inference_B[,idx_B_curr]
    inference_C <- array(0L,c(length(inference_attachments),ncol(inference_B)))
    for(curr_C_index in 1:nrow(inference_C)) {
        inference_C[curr_C_index,inference_attachments[curr_C_index]] <- 1L
    }
    corrected_genotypes <- inference_C %*% inference_B
    cells_names <- NULL
    for(cn in 1:length(D)) {
        cells_names <- c(cells_names,rownames(D[[cn]]))
    }
    rownames(corrected_genotypes) <- cells_names
    colnames(corrected_genotypes) <- c("Root",colnames(D[[1]]))
    corrected_genotypes <- corrected_genotypes[,-1,drop=FALSE]

    # Renaming
    B <- inference[[best]][["B"]]
    rownames(B) <- c("Root",paste0("Clone_",1:(nrow(B)-1)))
    colnames(B) <- c("Root",colnames(D[[1]])[as.numeric(colnames(B)[2:ncol(B)])])
    C <- inference[[best]][["C"]]
    names(C) <- paste0("Experiment_",1:length(C))
    for(i in 1:length(C)) {
        tmp <- matrix(C[[i]],ncol=1)
        rownames(tmp) <- rownames(D[[i]])
        colnames(tmp) <- "Clone"
        C[[i]] <- tmp
    }
    relative_likelihoods <- inference[[best]][["lik"]]
    names(relative_likelihoods) <- paste0("Experiment_",1:length(relative_likelihoods))
    joint_likelihood <- inference[[best]][["joint_lik"]]

    # Compute clones' prevalence
    clones_prevalence <- array(NA,c((dim(B)[1]-1),(length(C)+1)))
    rownames(clones_prevalence) <- rownames(B)[2:nrow(B)]
    colnames(clones_prevalence) <- c(paste0("Experiment_",1:length(C)),"Total")
    total_number_cell <- length(unlist(C))
    for(i in 1:dim(clones_prevalence)[1]) {
        clone_number_cell <- 0
        for(j in 1:length(C)) {
            clone_number_cell <- clone_number_cell + length(which(C[[j]]==i))
            clones_prevalence[i,j] <- length(which(C[[j]]==i)) / dim(C[[j]])[1]
        }
        clones_prevalence[i,"Total"] <- clone_number_cell / total_number_cell
    }

    # Include Root in clones_prevalence
    clones_prevalence <- rbind(rep(NA,ncol(clones_prevalence)),clones_prevalence)
    rownames(clones_prevalence)[1] <- "Root"
    clones_prevalence["Root",] <- (1-colSums(clones_prevalence,na.rm=TRUE))

    # Compute clones' summary
    clones_summary <- list()
    Bwor <- B[,-1]
    i = 2
    for(c in rownames(Bwor[-1,])) {
        mut_list <- colnames(Bwor)[Bwor[i,]==1]
        clones_summary[[c]] <- mut_list
        i = i + 1
    }

    # Finally, process equivalent solutions
    if(keep_equivalent == TRUE) {
        equivalent_solutions <- inference[[best]][["equivalent_solutions"]]

        eq_sol_df <- do.call(rbind,  lapply(equivalent_solutions, function(x) {
            adjM <- as.adj.matrix(x$B)
            return(as.vector(adjM))
        }))

        idx_uniq_sol <- which(!duplicated(eq_sol_df))

        equivalent_solutions <- equivalent_solutions[idx_uniq_sol]

        if(length(equivalent_solutions)==1) { # if we have only the best solution (no other equivalent solutions)
            equivalent_solutions <- list()
        } else {

            for(i in 1:length(equivalent_solutions)) {
                rownames(equivalent_solutions[[i]][["B"]]) <- c("Root",paste0("Clone_",rownames(equivalent_solutions[[i]][["B"]])[2:nrow(equivalent_solutions[[i]][["B"]])]))
                colnames(equivalent_solutions[[i]][["B"]]) <- c("Root",colnames(D[[1]])[as.numeric(colnames(equivalent_solutions[[i]][["B"]])[2:ncol(equivalent_solutions[[i]][["B"]])])])
            }
        }

    } else {

        equivalent_solutions <- NULL
    }

    longitudinal_tree <- NULL

    if(show==TRUE) {
        lace_interface(B, clones_prevalence, C, error_rates)
    }
    return(list(B=B,C=C,corrected_genotypes=corrected_genotypes,clones_prevalence=clones_prevalence,relative_likelihoods=relative_likelihoods,joint_likelihood=joint_likelihood,clones_summary=clones_summary,equivalent_solutions=equivalent_solutions,error_rates=error_rates, longitudinal_tree=longitudinal_tree))

}





#' Perform the inference of the maximum likelihood clonal tree from longitudinal data.
#' @title lacedata
#'
#' @examples
#' data(longitudinal_sc_variants)
#' lacedata(D = longitudinal_sc_variants,
#'          lik_w = c(0.2308772,0.2554386,0.2701754,0.2435088),
#'          alpha = list(c(0.10,0.05,0.05,0.05)),
#'          beta = list(c(0.10,0.05,0.05,0.05)),
#'          keep_equivalent = TRUE,
#'          num_rs = 5,
#'          num_iter = 10,
#'          n_try_bs = 5,
#'          num_processes = NA,
#'          seed = 12345,
#'          verbose = FALSE)
#'
#' @param D Mutation data from multiple experiments for a list of driver genes. It can be either a list with a data matrix per time point or a SummarizedExperiment object.
#' In this latter, the object must contain two fields: assays and colData. Assays stores one unique data matrix pooling all single cells observed at each time point and colData stores a vector of labels reporting the time point when each single cell was sequenced.
#' Ordering of cells in assays field and colData field must be the same.
#' @param lik_w Weight for each data point. If not provided, weights to correct for sample sizes are used.
#' @param alpha False positive error rate provided as list of elements; if a vector of alpha (and beta) is provided, the inference is performed for multiple values and the solution at
#' maximum-likelihood is returned.
#' @param beta False negative error rate provided as list of elements; if a vector of beta (and alpha) is provided, the inference is performed for multiple values and the solution at
#' maximum-likelihood is returned.
#' @param initialization Binary matrix representing a perfect philogeny clonal tree; clones are rows and mutations are columns.
#' This parameter overrides "random_tree".
#' @param random_tree Boolean. Shall I start MCMC search from a random tree? If FALSE (default) and initialization is NULL, search
#' is started from a TRaIT tree (BMC Bioinformatics . 2019 Apr 25;20(1):210.  doi: 10.1186/s12859-019-2795-4).
#' @param keep_equivalent Boolean. Shall I return results (B and C) at equivalent likelihood with the best returned solution?
#' @param check_indistinguishable Boolean. Shall I remove any indistinguishable event from input data prior inference?
#' @param num_rs Number of restarts during mcmc inference.
#' @param num_iter Maximum number of mcmc steps to be performed during the inference.
#' @param n_try_bs Number of steps without change in likelihood of best solution after which to stop the mcmc.
#' @param learning_rate Parameter to tune the probability of accepting solutions at lower values during mcmc. Value of learning_rate = 1 (default), set a
#' probability proportional to the difference in likelihood; values of learning_rate greater than 1 inclease the chance of accepting solutions at lower likelihood
#' during mcmc while values lower than 1 decrease such probability.
#' @param marginalize Boolean. Shall I marginalize C when computing likelihood?
#' @param error_move Boolean. Shall I include estimation of error rates in the MCMC moves?
#' @param num_processes Number of processes to be used during parallel execution. To execute in single process mode,
#' this parameter needs to be set to either NA or NULL.
#' @param seed Seed for reproducibility.
#' @param verbose Boolean. Shall I print to screen information messages during the execution?
#' @param log_file log file where to print outputs when using parallel. If parallel execution is disabled, this parameter is ignored.
#'
#' @return shiny interface
#' @export lacedata
#'
lacedata <- function( D,
                      lik_w = NULL,
                      alpha = NULL,
                      beta = NULL,
                      initialization = NULL,
                      random_tree = FALSE,
                      keep_equivalent = TRUE,
                      check_indistinguishable = TRUE,
                      num_rs = 50,
                      num_iter = 10000,
                      n_try_bs = 500,
                      learning_rate = 1,
                      marginalize = FALSE,
                      error_move = FALSE,
                      num_processes = Inf,
                      seed = NULL,
                      verbose = TRUE,
                      log_file = "")  {
  
  show <- TRUE
  
  inp <- list(D=D,
              lik_w = lik_w,
              alpha = alpha,
              beta = beta,
              initialization = initialization,
              random_tree = random_tree,
              keep_equivalent = keep_equivalent,
              check_indistinguishable = check_indistinguishable,
              num_rs = num_rs,
              num_iter = num_iter,
              n_try_bs = n_try_bs,
              learning_rate = learning_rate,
              marginalize = marginalize,
              error_move = error_move,
              num_processes = num_processes,
              seed = seed,
              verbose = verbose,
              log_file = log_file,
              show = show )
  
  inference_res=do.call('LACE',inp)
  
  B <- inference_res$B
  clones_prevalence <- inference_res$clones_prevalence
  C <- inference_res$C
  error_rates <- inference_res$error_rates
  
  if (!is.null(B) && !is.null(C) && !is.null(clones_prevalence) && !is.null(error_rates)) {
    x <- lace_interface(
      B_mat = B,
      clones_prevalence = clones_prevalence,
      C_mat = C,
      error_rates = error_rates
    )
    x <- x[["html"]]
    return(x)
  }
  else
    return(NULL)
}

#' @name longitudinal_sc_variants
#' @title Mutation data from Rambow, Florian, et al. "Toward minimal residual disease-directed therapy in melanoma." Cell 174.4 (2018): 843-855.
#' @description The dataset includes somatic single nucleotide variants at the single cell resolution. SNVs are called from SMARTseq2 fastq obtained from Gene Expression Omnibus database 
#' with the accession number: GSE116237. The dataset includes single cell data from a PDX melanoma model before and on treatment with BRAF and MEK inhibitors. The fastq files are processed 
#' to obtain the mutational profile following GATK best practice (https://gatkforums.broadinstitute.org/gatk/discussion/3891/calling-variants-in-rnaseq) usign the GRCh38 human genome as reference. 
#' Mutation data are stored in an N x M binary matrix with N single cells and M somatic single nucleotide variants. Row names report the ID of the fastq file related to a specific single cell; 
#' columns names report the SNV that are formatted as GeneName_chromosome_position_referenceAllele_alternateAllele. Each matrix entry can be 1 (mutation detected), 0 (mutation absent) or NA 
#' (too low coverage to determine the presence or absence of that mutation). For further details, please refer to the Methods Section and the section 3.1 of supplementary materials of 
#' Ramazzotti, Daniele, et al. "Longitudinal cancer evolution from single cells." bioRxiv (2020). 
#' @docType data
#' @usage data(longitudinal_sc_variants)
#' @format List of mutation data for four time points
#' @source Rambow, Florian, et al. "Toward minimal residual disease-directed therapy in melanoma." Cell 174.4 (2018): 843-855.
#' @return List of mutational data for a total of 475 single cells
NULL

#' @name inference
#' @title Results obtained with the function LACE on the provided input data from Rambow, Florian, et al. "Toward minimal residual disease-directed therapy in melanoma." Cell 174.4 (2018): 843-855.
#' @description Results obtained with the function LACE on the provided input data from Rambow, Florian, et al. "Toward minimal residual disease-directed therapy in melanoma." Cell 174.4 (2018): 843-855.
#' @docType data
#' @usage data(inference)
#' @format Results obtained with the function LACE on the provided input data
#' @return Results obtained with the function LACE on the provided input data
NULL

# #' TimeScape
# #'
# #' \code{timescape} is a tool for visualizing temporal clonal evolution data.
# #'
# #' Interactive components:
# #'   \enumerate{
# #'
# #'     \item Mouseover any clone to view its (i) clone ID and (ii) clonal
# #'     prevalence at each time point.
# #'
# #'     \item Click the view switch button to switch from the traditional
# #'     timescape view to the clonal trajectory view, where each clone
# #'     changes prevalence on its own track.
# #'
# #'     \item Click the download buttons to download a PNG or SVG of the
# #'     view.
# #'
# #'   }
# #'
# #' @param clonal_prev \code{data.frame} Clonal prevalence.
# #'     Required columns are:
# #'     \describe{
# #'
# #'       \item{timepoint:}{\code{character()} time point. Time
# #'          points will be alphanumerically sorted in the view.}
# #'
# #'       \item{clone_id:}{\code{character()} clone id.}
# #'
# #'       \item{clonal_prev:}{\code{numeric()} clonal prevalence.}
# #'
# #'     }
# #' @param tree_edges \code{data.frame} Tree edges of a rooted tree.
# #'     Required columns are:
# #'     \describe{
# #'
# #'       \item{source:}{\code{character()} source node id.}
# #'
# #'       \item{target:}{\code{character()} target node id.}
# #'
# #'     }
# #' @param mutations \code{data.frame} (Optional)  Mutations
# #'     occurring at each clone. Required columns are:
# #'     \describe{
# #'
# #'       \item{chrom:}{\code{character()} chromosome number.}
# #'
# #'       \item{coord:}{\code{numeric()} coordinate of mutation
# #'          on chromosome.}
# #'
# #'       \item{clone_id:}{\code{character()} clone id.}
# #'
# #'       \item{timepoint:}{\code{character()} time point.}
# #'
# #'       \item{VAF:}{\code{numeric()} variant allele frequency
# #'          of the mutation in the corresponding timepoint.}
# #'
# #'     }
# #'     Any additional field will be shown in the mutation table.
# #' @param clone_colours \code{data.frame} Clone ids and their
# #'     corresponding colours. Required columns are:
# #'     \describe{
# #'
# #'       \item{clone_id:}{\code{character()} clone id.}
# #'
# #'       \item{colour:}{\code{character()} the corresponding Hex
# #'          colour for each clone id.}
# #'
# #'     }
# #' @param xaxis_title \code{character()} (Optional) x-axis title.
# #'     Default is "Time Point".
# #' @param yaxis_title \code{character()} (Optional) y-axis title.
# #'     Default is "Clonal Prevalence".
# #' @param phylogeny_title \code{character()} (Optional) Legend
# #'     phylogeny title. Default is "Clonal Phylogeny".
# #' @param alpha \code{numeric()} (Optional) Alpha value for clonal
# #'     sweeps, range [0, 100].
# #' @param genotype_position \code{character()} (Optional) How to
# #'     position the genotypes from ["centre", "stack", "space"].
# #'   \enumerate{
# #'
# #'       \item centre: genotypes are centred with
# #'          respect to their ancestors.
# #'
# #'       \item stack: genotypes are stacked such
# #'          that nogenotype is split at any time point.
# #'
# #'       \item space: genotypes are stacked but
# #'          with a bit of spacing at the bottom.
# #'
# #'     }
# #' @param perturbations \code{data.frame} (Optional) Any
# #'     perturbations that occurred between two time points.
# #'     Required columns are:
# #'     \describe{
# #'
# #'       \item{pert_name:}{\code{character()} the perturbation name.}
# #'
# #'       \item{prev_tp:}{\code{character()} the time point (as labelled
# #'          in clonal prevalence data) BEFORE perturbation.}
# #'
# #'     }
# #' @param sort \code{logical()} (Optional) Whether (TRUE) or not (FALSE)
# #'     to vertically sort the genotypes by their emergence values
# #'     (descending). Default is FALSE. Note that genotype sorting will
# #'     always retain the phylogenetic hierarchy, and this parameter will
# #'     only affect the ordering of siblings.
# #' @param show_warnings \code{logical()} (Optional) Whether or not to show
# #'     any warnings. Default is TRUE.
# #' @param width \code{numeric()} (Optional) Width of the plot. Minimum
# #'     width is 450.
# #' @param height \code{numeric()} (Optional) Height of the plot. Minimum
# #'     height with and without mutations is 500 and 260, respectively.
# #' @export
timescape <- function(clonal_prev,
                      tree_edges,
                      mutations = "NA",
                      clone_colours = "NA",
                      xaxis_title = "Time Point",
                      yaxis_title = "Clonal Prevalence",
                      phylogeny_title = "Clonal Phylogeny",
                      alpha = 50,
                      genotype_position = "stack",
                      perturbations = "NA",
                      sort = FALSE,
                      show_warnings = TRUE,
                      width = 900,
                      height = NULL) {

  # forward options using x
  x = processUserData(clonal_prev,
                      tree_edges,
                      mutations,
                      clone_colours,
                      xaxis_title,
                      yaxis_title,
                      phylogeny_title,
                      alpha,
                      genotype_position,
                      perturbations,
                      sort,
                      show_warnings,
                      width,
                      height)

  return(x)

}

# #' Function to process the user data
# #' @param clonal_prev -- data frame of Clonal prevalence. Note: timepoints will be alphanumerically sorted in the view.
# #'   Format: columns are (1) character() "timepoint" - time point
# #'                       (2) character() "clone_id" - clone id
# #'                       (3) numeric() "clonal_prev" - clonal prevalence.
# #' @param tree_edges -- data frame of Tree edges of a rooted tree.
# #'   Format: columns are (1) character() "source" - source node id
# #'                       (2) character() "target" - target node id.
# #' @param mutations -- data frame (Optional) of Mutations occurring at each clone. Any additional field will be shown in the mutation table.
# #'   Format: columns are (1) character() "chrom" - chromosome number
# #'                       (2) numeric() "coord" - coordinate of mutation on chromosome
# #'                       (3) character() "clone_id" - clone id
# #'                       (4) character() "timepoint" - time point
# #'                       (5) numeric() "VAF" - variant allele frequency of the mutation in the corresponding timepoint.
# #' @param clone_colours -- data frame (Optional) of Clone ids and their corresponding colours
# #'   Format: columns are (1) character() "clone_id" - the clone ids
# #'                       (2) character() "colour" - the corresponding Hex colour for each clone id.
# #' @param xaxis_title -- String (Optional) of x-axis title. Default is "Time Point".
# #' @param yaxis_title -- String (Optional) of y-axis title. Default is "Clonal Prevalence".
# #' @param phylogeny_title -- String (Optional) of Legend phylogeny title. Default is "Clonal Phylogeny".
# #' @param alpha -- Number (Optional) of Alpha value for sweeps, range [0, 100].
# #' @param genotype_position -- String (Optional) of How to position the genotypes from ["centre", "stack", "space"]
# #'   "centre" -- genotypes are centred with respect to their ancestors
# #'   "stack" -- genotypes are stacked such that no genotype is split at any time point
# #'   "space" -- genotypes are stacked but with a bit of spacing at the bottom
# #' @param perturbations -- data frame (Optional) of any perturbations that occurred between two time points.
# #'   Format: columns are (1) character() "pert_name" - the perturbation name
# #'                       (2) character() "prev_tp" - the time point (as labelled in clonal prevalence data)
# #'                                                BEFORE perturbation.
# #' @param sort -- Boolean (Optional) of whether (TRUE) or not (FALSE) to vertically sort the genotypes by their emergence values (descending).
# #'                       Default is FALSE.
# #'                       Note that genotype sorting will always retain the phylogenetic hierarchy, and this parameter will only affect the ordering of siblings.
# #' @param show_warnings -- Boolean (Optional) of  Whether or not to show any warnings. Default is TRUE.
# #' @param width -- Number (Optional) of width of the plot. Minimum width is 450.
# #' @param height -- Number (Optional) of height of the plot. Minimum height with and without mutations is 500 and 260, respectively.
# #' @rdname helpers
# #' @return Returns the ready list of user input data for htmlwidget
processUserData <- function(clonal_prev,
                      tree_edges,
                      mutations,
                      clone_colours,
                      xaxis_title,
                      yaxis_title,
                      phylogeny_title,
                      alpha,
                      genotype_position,
                      perturbations,
                      sort,
                      show_warnings,
                      width,
                      height) {

  # ENSURE MINIMUM DIMENSIONS SATISFIED
  checkMinDims(mutations, height, width)

  # CHECK REQUIRED INPUTS ARE PRESENT
  checkRequiredInputs(clonal_prev, tree_edges)

  # ALPHA VALUE
  checkAlpha(alpha)

  # SORTED GENOTYPES
  if (!is.logical(sort)) {
    stop("Sort parameter must be a boolean.")
  }

  # CLONAL PREVALENCE DATA
  clonal_prev <- checkClonalPrev(clonal_prev)

  # TREE EDGES DATA
  tree_edges <- checkTreeEdges(tree_edges)

  # GENOTYPE POSITIONING
  checkGtypePositioning(genotype_position)

  # CHECK CLONE COLOURS
  checkCloneColours(clone_colours)

  # CHECK PERTURBATIONS
  perturbations <- checkPerts(perturbations)

  # MUTATIONS DATA
  mut_data <- getMutationsData(mutations, tree_edges, clonal_prev, show_warnings)
  mutation_info <- mut_data$mutation_info
  mutation_prevalences <- mut_data$mutation_prevalences
  if (is.data.frame(mutations)) {
    mutations_provided <- TRUE
  }
  else {
    mutations_provided <- FALSE
  }

  # REPLACE SPACES WITH UNDERSCORES
  spaces_replaced <- replaceSpaces(clonal_prev, tree_edges, clone_colours, mutation_info, mutations, mutation_prevalences)
  timepoint_map <- spaces_replaced$timepoint_map
  clone_id_map <- spaces_replaced$clone_id_map
  clonal_prev <- spaces_replaced$clonal_prev
  tree_edges <- spaces_replaced$tree_edges
  mutation_info <- spaces_replaced$mutation_info
  clone_colours <- spaces_replaced$clone_colours
  mutation_prevalences <- spaces_replaced$mutation_prevalences

  # forward options using x
  return(list(
    clonal_prev = jsonlite::toJSON(clonal_prev),
    gtype_tree_edges = jsonlite::toJSON(tree_edges),
    clone_cols = jsonlite::toJSON(clone_colours),
    mutations = jsonlite::toJSON(mutation_info),
    mutation_prevalences = jsonlite::toJSON(mutation_prevalences),
    mutations_provided=mutations_provided, # whether or not mutations are provided
    xaxis_title = as.character(xaxis_title),
    yaxis_title = as.character(yaxis_title),
    phylogeny_title = as.character(phylogeny_title),
    alpha = alpha,
    genotype_position = genotype_position,
    perturbations = jsonlite::toJSON(perturbations),
    sort_gtypes = sort,
    timepoint_map = jsonlite::toJSON(timepoint_map),
    clone_id_map = jsonlite::toJSON(clone_id_map)
  ))
}

# #' Function to check minimum dimensions
# #'
# #' @param mutations -- mutations provided by user
# #' @param height -- height provided by user
# #' @param width -- width provided by user
# #' @rdname helpers
# #' @return None
checkMinDims <- function(mutations, height, width) {

  # set height if not set by user
  if (is.null(height)) {
    if (!is.data.frame(mutations)) { # no mutations
      height = 260
    }
    else { # mutations
      height = 500
    }
  }

  # check height is big enough
  min_width = 450
  if (!is.data.frame(mutations)) { # no mutations
    min_height = 260
  }
  else { # mutations
    min_height = 500
  }

  if (height < min_height) {
    stop("Height must be greater than or equal to ", min_height, "px.")
  }
  if (width < min_width) {
    stop("Width must be greater than or equal to ", min_width, "px.")
  }
}

# #' Function to check required inputs are present
# #'
# #' @param clonal_prev -- clonal_prev provided by user
# #' @param tree_edges -- tree_edges provided by user
# #' @rdname helpers
# #' @return None
checkRequiredInputs <- function(clonal_prev, tree_edges) {

  if (missing(clonal_prev)) {
    stop("Clonal prevalence data frame must be provided.")
  }
  if (missing(tree_edges)) {
    stop("Tree edge data frame must be provided.")
  }
}

# #' check alpha value input is correct
# #'
# #' @param alpha -- alpha provided by user
# #' @rdname helpers
# #' @return None
checkAlpha <- function(alpha) {
  if (!is.numeric(alpha)) {
    stop("Alpha value must be numeric.")
  }

  if (alpha < 0 || alpha > 100) {
    stop("Alpha value must be between 0 and 100.")
  }
}

# #' check clonal_prev parameter data
# #'
# #' @param clonal_prev -- clonal prevalence provided by user
# #' @rdname helpers
# #' @return Clonal prevalence data after checkint it for column names and content types
checkClonalPrev <- function(clonal_prev) {

  # ensure column names are correct
  if (!("timepoint" %in% colnames(clonal_prev)) ||
      !("clone_id" %in% colnames(clonal_prev)) ||
      !("clonal_prev" %in% colnames(clonal_prev))) {
    stop("Clonal prevalence data frame must have the following column names: ",
        "\"timepoint\", \"clone_id\", \"clonal_prev\"")
  }

  # ensure data is of the correct type
  clonal_prev$timepoint <- as.character(clonal_prev$timepoint)
  clonal_prev$clone_id <- as.character(clonal_prev$clone_id)
  clonal_prev$clonal_prev <- as.numeric(as.character(clonal_prev$clonal_prev))

  return(clonal_prev)
}

# #' check tree_edges parameter data
# #'
# #' @param tree_edges -- tree edges provided by user
# #' @rdname helpers
# #' @return Tree edges data after checkint it for column names and content types
checkTreeEdges <- function(tree_edges) {

  # ensure column names are correct
  if (!("source" %in% colnames(tree_edges)) ||
      !("target" %in% colnames(tree_edges))) {
    stop("Tree edges data frame must have the following column names: ",
        "\"source\", \"target\"")
  }

  # ensure data is of the correct type
  tree_edges$source <- as.character(tree_edges$source)
  tree_edges$target <- as.character(tree_edges$target)

  # check for tree rootedness
  sources <- unique(tree_edges$source)
  targets <- unique(tree_edges$target)
  sources_for_iteration <- sources # because we will be changing the sources array over time
  for (i in 1:length(sources_for_iteration)) {
    cur_source <- sources_for_iteration[i]

    # if the source is a target, remove it from the sources list
    if (cur_source %in% targets) {
      sources <- sources[sources != cur_source]
    }
  }

  # if multiple roots are detected, throw error
  if (length(sources) > 1) {
    stop("Multiple roots detected in tree (",paste(sources,collapse=", "),
      ") - tree must have only one root.")
  }

  # if an edge is found whose source and target are equal, throw an error
  if (length(which(as.character(tree_edges$source) == as.character(tree_edges$target))) > 0) {
    stop("One of the tree edges has a source as its own target. Remove this edge.")
  }

  return(tree_edges)
}

# #' check genotype_position parameter
# #'
# #' @param genotype_position -- genotype_position provided by user
# #' @rdname helpers
# #' @return None
checkGtypePositioning <- function(genotype_position) {
  if (!(genotype_position %in% c("stack", "centre", "space"))) {
    stop("Genotype position must be one of c(\"stack\", \"centre\", \"space\")")
  }
}

# #' check clone_colours parameter
# #'
# #' @param clone_colours -- clone_colours provided by user
# #' @rdname helpers
# #' @return None
checkCloneColours <- function(clone_colours) {
  if (is.data.frame(clone_colours)) {

    # ensure column names are correct
    if (!("clone_id" %in% colnames(clone_colours)) ||
        !("colour" %in% colnames(clone_colours))) {
      stop("Node colour data frame must have the following column names: ",
          "\"clone_id\", \"colour\"")
    }
  }
}

# #' check perturbations parameter
# #'
# #' @param perturbations -- perturbations provided by user
# #' @rdname helpers
# #' @return Perturbations after checking them for content types and column names
checkPerts <- function(perturbations) {

  if (is.data.frame(perturbations)) {

    # ensure column names are correct
    if (!("pert_name" %in% colnames(perturbations)) ||
        !("prev_tp" %in% colnames(perturbations))) {
      stop("Perturbations data frame must have the following column names: ",
          "\"pert_name\", \"prev_tp\"")
    }

    # check that columns are of the correct type
    perturbations$pert_name <- as.character(perturbations$pert_name)
    perturbations$prev_tp <- as.character(perturbations$prev_tp)
  }

  return(perturbations)
}

# #' get mutation data
# #'
# #' @param mutations -- mutations data from user
# #' @param tree_edges -- tree edges data from user
# #' @param clonal_prev -- clonal prevalence data from user
# #' @param show_warnings -- show warnings
# #' @rdname helpers
# #' @return List of mutation information and mutation prevalences
getMutationsData <- function(mutations, tree_edges, clonal_prev, show_warnings) {

  if (is.data.frame(mutations)) {

    # ensure column names are correct
    if (!("chrom" %in% colnames(mutations)) ||
        !("coord" %in% colnames(mutations)) ||
        !("clone_id" %in% colnames(mutations)) ||
        !("timepoint" %in% colnames(mutations)) ||
        !("VAF" %in% colnames(mutations))) {
      stop("Mutations data frame must have the following column names: ",
          "\"chrom\", \"coord\", \"clone_id\", \"timepoint\", \"VAF\".")
    }

    # ensure data is of the correct type
    mutations$chrom <- toupper(as.character(mutations$chrom)) # upper case X & Y
    mutations$coord <- as.character(mutations$coord)
    mutations$timepoint <- as.character(mutations$timepoint)
    mutations$clone_id <- as.character(mutations$clone_id)
    mutations$VAF <- as.numeric(as.character(mutations$VAF))

    # check for optional info, and ensure data of correct type
    extra_columns <- colnames(mutations)[which(!(colnames(mutations) %in% c("chrom", "coord", "clone_id", "timepoint", "VAF")))]
    mutations <- data.frame(lapply(mutations, as.character), stringsAsFactors=FALSE)

    # check that all CLONE IDS in the mutations data are present in the tree data
    mutations_clone_ids <- unique(mutations$clone_id)
    tree_edges_clone_ids <- c(unique(tree_edges$source), unique(tree_edges$target))
    clone_ids_missing_from_tree_edges_data <- setdiff(mutations_clone_ids, tree_edges_clone_ids)
    if (length(clone_ids_missing_from_tree_edges_data) > 0) {
      stop("The following clone ID(s) are present in the mutations data but ",
        "are missing from the tree edges data: ",
        paste(clone_ids_missing_from_tree_edges_data, collapse=", "), ".")
    }

    # check that all TIMEPOINTS in the mutations data are present in the clonal prev data
    mutations_tps <- unique(mutations$timepoint)
    clonal_prev_tps <- unique(clonal_prev$timepoint)
    tps_missing_from_clonal_prev_data <- setdiff(mutations_tps, clonal_prev_tps)
    if (length(tps_missing_from_clonal_prev_data) > 0) {
      stop("The following timepoint(s) are present in the mutations data but ",
        "are missing from the clonal prevalence data: ",
        paste(tps_missing_from_clonal_prev_data, collapse=", "), ".")
    }

    # create a location column, combining the chromosome and the coordinate
    mutations$location <- apply(mutations[, c("chrom","coord")], 1 , paste, collapse = ":")

    # coordinate is now a number
    mutations$coord <- as.numeric(as.character(mutations$coord))

    # check X & Y chromosomes are labelled "X" and "Y", not "23", "24"
    num_23 <- mutations[which(mutations$chrom == "23"),]
    if (nrow(num_23) > 0) {
      stop("Chromosome numbered \"23\" was detected in mutations data frame - X and Y chromosomes ",
        "must be labelled \"X\" and \"Y\".")
    }


    # get list of clones in the phylogeny
    clones_in_phylo <- unique(c(tree_edges$source, tree_edges$target))

    # keep only those mutations whose clone ids are present in the phylogeny
    mutations <- mutations[which(mutations$clone_id %in% clones_in_phylo),]

    # MUTATION PREVALENCES DATA

    mutation_prevalences <- mutations

    # keep only those mutations whose clone ids are present in the phylogeny
    mutation_prevalences <- mutation_prevalences[which(mutation_prevalences$clone_id %in% clones_in_phylo),]

    # warn if more than 10,000 rows in data that the visualization may be slow
    if (nrow(mutation_prevalences) > 10000 && show_warnings) {
      log2_print(paste("[WARNING] Number of rows in mutations data exceeds 10,000. ",
        "Resultantly, visualization may be slow. ",
        "It is recommended to filter the data to a smaller set of mutations.", sep=""))
    }

    # compress results
    prevs_split <- split(mutation_prevalences, f = mutation_prevalences$location)

    # reduce the size of the data frame in each list
    prevs_split_small <- lapply(prevs_split, function(prevs) {
      return(prevs[,c("timepoint", "VAF")])
    })


    # MUTATION INFO
    mutation_info <- unique(mutations[,c("chrom","coord","clone_id",extra_columns)])
  }
  else {
    prevs_split_small <- "NA"
    mutation_info <- "NA"
  }

  return(list("mutation_info"=mutation_info, "mutation_prevalences"=prevs_split_small))
}

# #' function to replace spaces with underscores in all data frames & keep maps of original names to space-replaced names
# #' @param clonal_prev -- clonal_prev data from user
# #' @param tree_edges -- tree edges data from user
# #' @param clone_colours -- clone_colours data from user
# #' @param mutation_info -- processed mutation_info
# #' @param mutations -- mutations data from user
# #' @param mutation_prevalences -- mutation_prevalences data from user
# #' @rdname helpers
# #' @return List of data frames with spaces replaced
replaceSpaces <- function(clonal_prev, tree_edges, clone_colours, mutation_info, mutations, mutation_prevalences) {

  # create map of original sample ids to space-replaced sample ids
  timepoint_map <- data.frame(original_timepoint = unique(clonal_prev$timepoint), stringsAsFactors=FALSE)
  timepoint_map$space_replaced_timepoint <- stringr::str_replace_all(timepoint_map$original_timepoint,"\\s+","_")

  # create map of original clone ids to space-replaced clone ids
  clone_id_map <- data.frame(original_clone_id = unique(c(tree_edges$source, tree_edges$target)), stringsAsFactors=FALSE)
  clone_id_map$space_replaced_clone_id <- stringr::str_replace_all(clone_id_map$original_clone_id,"\\s+","_")

  # replace spaces with underscores
  # --> timepoints
  clonal_prev$timepoint <- stringr::str_replace_all(clonal_prev$timepoint,"\\s+","_")
  if (is.data.frame(mutations)) {
    mutation_prevalences <- lapply(mutation_prevalences, function(prevs) {
      prevs$timepoint <- stringr::str_replace_all(prevs$timepoint,"\\s+","_")
      return(prevs)
    })
  }
  # --> clone ids
  clonal_prev$clone_id <- stringr::str_replace_all(clonal_prev$clone_id,"\\s+","_")
  tree_edges$source <- stringr::str_replace_all(tree_edges$source,"\\s+","_")
  tree_edges$target <- stringr::str_replace_all(tree_edges$target,"\\s+","_")
  if (is.data.frame(clone_colours)) {
    clone_colours$clone_id <- stringr::str_replace_all(clone_colours$clone_id,"\\s+","_")
  }
  if (is.data.frame(mutations)) {
    mutation_info$clone_id <- stringr::str_replace_all(mutation_info$clone_id,"\\s+","_")
  }

  return(list("timepoint_map"=timepoint_map,
              "clone_id_map"=clone_id_map,
              "clonal_prev"=clonal_prev,
              "tree_edges"=tree_edges,
              "mutation_info"=mutation_info,
              "clone_colours"=clone_colours,
              "mutation_prevalences"=mutation_prevalences))
}

### filters_computation.R
###


  filter1_compute <- function(vcf_out_dir,
                              cellInfo,
                              thr_alleles_ratio,
                              thr_maf,
                              thr_freq,
                              thr_out_dir,
                              id_column,
                              time_column,
                              variant_functions,
                              InFilesToDo = NULL,
                              OutFilesToRm = NULL) {

    dir.create(thr_out_dir, showWarnings = FALSE)
    if (!dir.exists(thr_out_dir))
      showNotification(paste("Filtered data output folder,",
                             thr_out_dir,
                             ", cannot be created"),
                       duration = 10,
                       type = "warning")

    listSC <- dir(path = vcf_out_dir,
                  pattern = ".exonic_variant_function",
                  ignore.case = TRUE,
                  full.names = TRUE)


    log2_print(InFilesToDo)
    if (!is.null(InFilesToDo)) {
      log2_print('In')
      log2_print(listSC)
      listSC <- listSC[listSC %in% InFilesToDo]
      log2_print(listSC)
    }
    if (!is.null(OutFilesToRm)) {
      log2_print('Out')
      ## OutFilesToRm <- file.path(thr_out_dir, OutFilesToRm)
      log2_print(OutFilesToRm)
      ## file.remove(OutFilesToRm)
      log2_print(list.files(path = thr_out_dir))
    }

    if (length(listSC) == 0)
      return()

    ## Create data structure to save the results
    tbFinal <- setNames(data.frame(matrix(ncol = 12, nrow = 0)),
                        c("scID",
                          "Time",
                          "Gene",
                          "Chr",
                          "PosStart",
                          "PosEnd",
                          "REF",
                          "ALT",
                          "MutType",
                          "rs_ID",
                          "depth",
                          "Allele_Ratio"))
    cn <- colnames(tbFinal)

    ## Consider each cell
    cont <- 0
    for(sc in listSC) {
      ## Get current cell name
      run <- strsplit(x = basename(sc), split = "\\.")[[1]][1]

      ## Read vcf annotated by ANNOVAR for current cell
      tmpTbl <- read.table(file = sc,
                           header = FALSE,
                           sep = "\t",
                           stringsAsFactors = FALSE)
      tmpTbl$scID <- run

      ## Get information about time of experiment
      tmpTime <-
        as.character(cellInfo[[time_column]][cellInfo[[id_column]] == run])
      tmpTbl$Time <- tmpTime

      ## Get genes names
      tmpTbl$V3 <- sapply(as.character(tmpTbl$V3),
                          function(x){strsplit(x, ":")[[1]][1]})

      ## Get the remaining information
      OtherInfo <- tmpTbl[, "V18"]
      tmpTbl <-
        tmpTbl[,
               c("scID",
                 "Time",
                 "V3",
                 "V4",
                 "V5",
                 "V6",
                 "V7",
                 "V8",
                 "V2",
                 "V11")]
      colnames(tmpTbl) <-
        c("sc_ID",
          "Time",
          "Gene",
          "Chr",
          "PStart",
          "PEnd",
          "Ref",
          "Alt",
          "Mut_type",
          "RS_id")

      ## Extract quality and coverage information
      for(m in 1:length(OtherInfo)) {
        info <- unlist(strsplit(OtherInfo[m], ":"))

        if (length(info) < 5) {
          depth <- NA
          RA <- NA
        } else {
          RefAlt <- unlist(strsplit(info[2], ","))
          depth <- sum(as.numeric(RefAlt))
          if (depth == 0) {
            RA <- NA
          } else {
            RA <- as.numeric(RefAlt[2])
          }
        }
        tmpTbl$depth[m] <- depth
        tmpTbl$Allele_Ratio[m] <- RA

      }

      ## Save results for current cell
      colnames(tmpTbl) <- cn
      tbFinal <- rbind(tbFinal,tmpTbl)

      cont = cont + 1
      cat(paste0(cont,
                 " done | ",
                 round(cont / length(listSC),
                       digits = 2),
                 "%\r"))
    }

    tbFinal$UIDsnp <-
      paste0(tbFinal$Gene,
             "_",
             tbFinal$Chr,
             "_",
             tbFinal$PosStart,
             ifelse(tbFinal$rs_ID == '.',
                    "",
                    paste0("_",tbFinal$rs_ID)))
    saveRDS(tbFinal,
            file = paste0(file.path(thr_out_dir,
                                    'cells_aggregate_info.rds')))

    tbFinal <-
      readRDS(file = paste0(file.path(thr_out_dir,
                                      'cells_aggregate_info.rds')))

    ## # remove synonymous mutations and the ones marked as unknown
    ## #modify
    ## #tbFinal <- tbFinal[tbFinal$MutType!="synonymous SNV"& tbFinal$MutType != "unknown", ]
    #browser()
    tbFinal <-
      tbFinal[tbFinal$MutType %in% variant_functions, ]

    u_rsID <-
      as.character(unique(tbFinal$rs_ID[startsWith(as.character(tbFinal$rs_ID),
                                                   prefix="rs")]))

    showNotification(sc, duration = 10, type = "warning")
    ## showNotification(length(u_rsID), duration = 10, type = "warning")

    ensembl_snp = useMart("ENSEMBL_MART_SNP", dataset="hsapiens_snp")
    SNPInfo <- getBM(attributes=c("refsnp_id",
                                  "minor_allele",
                                  "minor_allele_freq",
                                  "polyphen_prediction",
                                  "sift_prediction",
                                  "clinical_significance"),
                     filters = "snp_filter",
                     values = u_rsID,
                     mart = ensembl_snp)

    saveRDS(SNPInfo, file = paste0(file.path(thr_out_dir, 'SNPInfo.rds')))

    scMutInfo <-
      merge.data.frame(tbFinal,
                       SNPInfo,
                       by.x = "rs_ID",
                       by.y = "refsnp_id",
                       all.x = TRUE,
                       all.y = TRUE)
    saveRDS(scMutInfo,
            file = paste0(file.path(thr_out_dir, 'scMutInfo.rds')))
  }


  filter2_compute <- function(vcf_out_dir,
                              cellInfo,
                              thr_alleles_ratio,
                              thr_maf,
                              thr_freq,
                              thr_out_dir,
                              id_column,
                              time_column,
                              InFilesToDo = NULL,
                              OutFilesToRm = NULL) {

    ## ###############################
    ## # START FILTERING STEP ####
    ## ##############################

    listSC <- dir(path = vcf_out_dir,
                  pattern = ".exonic_variant_function",
                  ignore.case = TRUE,
                  full.names = TRUE)

    log2_print(InFilesToDo)
    if (!is.null(InFilesToDo)) {
      log2_print('In')
      log2_print(listSC)
      listSC <- listSC[listSC %in% InFilesToDo]
      log2_print(listSC)
    }
    if (!is.null(OutFilesToRm)) {
      log2_print('Out')
      ## OutFilesToRm <- file.path(thr_out_dir, OutFilesToRm)
      log2_print(OutFilesToRm)
      ## file.remove(OutFilesToRm)
      log2_print(list.files(path = thr_out_dir))
    }

    if (length(listSC)==0)
      return()

    scMutInfo <-
      readRDS(file=paste0(file.path(thr_out_dir, 'scMutInfo.rds')))

    scMutInfo <-
      scMutInfo[((!is.na(scMutInfo$Allele_Ratio)) &
                   (scMutInfo$Allele_Ratio >= thr_alleles_ratio#*scMutInfo$depth
                    )),]

    mutWOrs <- scMutInfo[scMutInfo$rs_ID == ".", ]
    mutWrs <- scMutInfo[scMutInfo$rs_ID != ".", ]

    FilterMutations <- function(sc_r) {

      ## If not a snp, keep it
      if (sc_r["rs_ID"] == ".") {
        return(TRUE)
      }

      ## Filter using MAF
      if (sc_r["minor_allele"] == sc_r["ALT"] &&
          sc_r["minor_allele_freq"] > thr_maf) {
        return(FALSE)
      }
      else if (sc_r["minor_allele"] == sc_r["REF"] &&
               sc_r["minor_allele_freq"] < (1-thr_maf)) {
        return(FALSE)
      }

      return(TRUE)
    }

    #mutWrs_filt <- mutWrs[apply(mutWrs,
    #                            1,
    #                            function(x) FilterMutations(x)), ]
    
    mutWrs_filt <-  mutWrs[
      (!(
        (mutWrs$minor_allele == mutWrs$ALT) & (mutWrs$minor_allele_freq > thr_maf) |
          (mutWrs$minor_allele == mutWrs$REF) & (mutWrs$minor_allele_freq < (1-thr_maf))
      )) & (mutWrs$rs_ID=="."),]
    
    snpMut_filt <- rbind(mutWOrs, mutWrs_filt)
    saveRDS(snpMut_filt,
            file = paste0(file.path(thr_out_dir, 'snpMut_filt.rds')))
  }


  filter3_compute <- function(vcf_out_dir,
                              cellInfo,
                              thr_alleles_ratio,
                              thr_maf,
                              thr_freq,
                              thr_out_dir,
                              id_column,
                              time_column,
                              time_points,
                              InFilesToDo = NULL,
                              OutFilesToRm = NULL) {


    ###########################################################################################################
    ### MAKE CELLS/MUTATIONS MATRIX ####
    ###########################################################################################################

    listSC <- dir(path = vcf_out_dir,
                  pattern = ".exonic_variant_function",
                  ignore.case = TRUE,
                  full.names = TRUE)

    log2_print(InFilesToDo)
    if (!is.null(InFilesToDo)) {
      log2_print('In')
      log2_print(listSC)
      listSC <- listSC[listSC %in% InFilesToDo]
      log2_print(listSC)
    }

    if (!is.null(OutFilesToRm)) {
      log2_print('Out')
      ## OutFilesToRm <- file.path(thr_out_dir, OutFilesToRm)
      log2_print(OutFilesToRm)
      ## file.remove(OutFilesToRm)
      log2_print(list.files(path = thr_out_dir))
    }

    if (length(listSC) == 0)
      return()

    snpMut_filt <-
      readRDS(file = paste0(file.path(thr_out_dir,
                                      'snpMut_filt.rds')))

    ## # consider only cells passing quality checks
    ## #load(file.path('..','valid_clones_mapping.RData'))
    ## #snpMut_filt = snpMut_filt[which(snpMut_filt$scID%in%valid_clones_mapping$Run),]

    Mut_keep = c()
    tblscMut = list()
    for (tn in time_points) {
      tblscMut[[tn]] <-
        table(snpMut_filt$scID[snpMut_filt$Time == tn ],
              snpMut_filt$UIDsnp[snpMut_filt$Time == tn])
      ## Binarization
      tblscMut[[tn]][which(tblscMut[[tn]] > 0)] <- 1

      cs <- colSums(tblscMut[[tn]], na.rm = TRUE) / nrow(tblscMut[[tn]])
      Mut_keep = c(Mut_keep, colnames(tblscMut[[tn]])[cs > thr_freq])
    }

    snpMut_filt_freq <-
      snpMut_filt[snpMut_filt$UIDsnp %in% Mut_keep, ]

    saveRDS(snpMut_filt_freq,
            file = paste0(file.path(thr_out_dir,
                                    'snpMut_filt_freq.rds')))
    
    return(snpMut_filt_freq)
  }

### end of file -- filters_computation.R

data("longitudinal_sc_variants")

context("LACE")

data("inference")
test_that("LACE produces correct output", {
    expect_equal(names(inference),c("B","C","corrected_genotypes","clones_prevalence","relative_likelihoods","joint_likelihood","clones_summary","equivalent_solutions","error_rates"))
})

Lace - Syntax of include statements
===================================

Include statements take a source token to include and an optional list
of definitions which must all be true before the include will take
place.

    include <sourcename> <0-or-more-definitions>

If the include ends with a question mark (`include?`) then should the
sourcename not be available at the time, it will be silently ignored.

Nominally an included ruleset is linearly inserted into the execution stream
there and then.  In practice, while include statements result in compilation of
the rulesets at the same time, the contents of the ruleset will not be run
unless the supplied definitions all pass.  This means that any rules (including
`define`s) in the included ruleset will not be executed unless the definitions
all pass.  As such, use conditional includes carefully.

If an "optional" include source isn't available, it is as though the
definitions did not pass.  If a mandatory include source isn't
available then compilation of the ruleset will fail immediately.

Circular includes are not acceptable.  Not even if there's a set of conditions
which mean that it might never happen.  Any attempt at circular includes will
result in a critical error at compile time because that is when all included
rulesets are loaded in their entirety.

Lace - Helping with Development
===============================

The Lace codebase is divided up in to directories:

    lace/
         lib/
             ... The Lace libraries live here
         test/
             ... All the tests and their data live here
         example/
             ... The example and its data lives here
         extras/
             ... The Lua coverage tool lives here
         doc/
             ... All the documentation lives here

The codebase has a top level `Makefile` which defaults to running the test
suite.  The test suite requires Lua be present and as well as running all the
tests, also runs the Lua coverage tool to produce `luacov.report.out` which
details the coverage of the Lace codebase.

It is a policy that releases must have 100% coverage from the test suite.
Ideally 100% coverage would be attained by the test cases for the given modules
but sometimes cross-module usage is required in order to best provide 100%
coverage.

Any line not covered by the tests will be marked the `***0` in the
`luacov.report.out` file.

If you make substantive non-backward-compatible changes to the API of Lace then
you should increment the ABI number in the main `lib/lace.lua` file.  If you
make bug fixes or backward-compatible improvements then don't worry, the
version number in that file will be incremented during the release process.

If you add more modules to Lace, you should note that you need to update:

1. The `Makefile`'s `MODULES` variable
2. The `test/` directory will need a `test-lace.NEWMODULE.lua` file
3. You will need to alter `lib/lace.lua` to pull it in and update
   the `test/test-lace.lua` test to include a check for the new module
4. You should ensure your new tests in 2 cover the module fully.
5. You should ensure that any changes are shown in the example if possible.
6. You should ensure any changes are reflected in the `docs/` files.

You can check individual test suite coverage by running:

    make MODULES=lace.SOMEMODULE

This will cause the test suite for the named modules *only* to be run, and for
the coverage data for that module to be generated.  It is not policy that a
given module's individual tests MUST cover the module 100%, but if possible it
SHOULD.  It is, as stated before, policy that the full test suite should cover
100% of the modules when run as a whole though.

---
title: "Running LACE"
author: 
  - Daniele Ramazzotti
  - Fabrizio Angaroni
  - Davide Maspero
  - Alex Graudenxi
  - Luca De Sano
  - Gianluca Ascolani
date: "`r format(Sys.time(), '%B %d, %Y')`"
graphics: yes
package: LACE
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{Running LACE}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteDepends{LACE,BiocStyle}
---


## Setup the environement

We now present an example of longitudinal analysis of cancer evolution with *LACE* using single-cell data obtained from Rambow, Florian, et al. "Toward minimal residual disease-directed therapy in melanoma." Cell 174.4 (2018): 843-855. 

As a first step, we load single-cell data for a skin melanoma. The data comprises point mutations for four 
time points: (1) before treatment, (2) 4 days treatment, (3) 28 days treatment and finally (4) 57 days treatment. 

```{r}
library("LACE")
data(longitudinal_sc_variants)
names(longitudinal_sc_variants)
```

We setup the main parameter in oder to perform the inference. First of all, as the four data points may potentially provide sequencing for an unbalanced 
number of cells, we weight each time point proportionally to the sample sizes as follow. We refer to the paper for details. 

```{r}
lik_weights = c(0.2308772,0.2554386,0.2701754,0.2435088)
```

The second main parameter to be defined as input is represented by the false positive and false negative error rates, i.e., alpha and beta. We can specify a 
different rate per time point as a list of rates. When multiple set of rates are provided, *LACE* performs a grid search in order to estimate the best set of error rates. 

```{r}
alpha = list()
alpha[[1]] = c(0.02,0.01,0.01,0.01)
alpha[[2]] = c(0.10,0.05,0.05,0.05)
beta = list()
beta[[1]] = c(0.10,0.05,0.05,0.05)
beta[[2]] = c(0.10,0.05,0.05,0.05)
head(alpha)
head(beta)
```

## Inference

We can now perform the inference as follow. 

```{r}
inference = LACE(D = longitudinal_sc_variants, 
    lik_w = lik_weights, 
    alpha = alpha, 
    beta = beta, 
    keep_equivalent = TRUE, 
    num_rs = 5, 
    num_iter = 10, 
    n_try_bs = 5, 
    num_processes = NA, 
    seed = 12345, 
    verbose = FALSE)
```

We notice that the inference resulting on the command above should be considered only as an example; the parameters num_rs, num_iter and n_try_bs representing the number of 
steps perfomed during the inference are downscaled to reduce execution time. We refer to the Manual for discussion on default values. We provide within the package results 
of inferences performed with correct parameters as RData. 

```{r}
data(inference)
print(names(inference))
```

*LACE* returns a list of nine elements as results. Namely, B and C provide respectively the maximum likelihood longitudinal tree and cells attachments; corrected_genotypes the corrected genotypes, clones_prevalence, the estimated prevalence of any observed clone; relative_likelihoods and joint_likelihood the estimated likelihoods for each time point and the weighted likelihood; clones_summary provide a summary of association of mutations to clones. In equivalent_solutions, solutions (B and C) with likelihood equivalent to the best solution are returned; notice that in the example we disabled this feature by 
setting equivalent_solutions parameter to FALSE. Finally, error rates provide the best error rates (alpha and beta) as estimated by the grid search. 


## Plot

We can plot the inferred model using the function longitudinal.tree.plot. 

```{r fig.width=12, fig.height=8, warning=FALSE, fig.cap="Inferred model"}
clone_labels = c("ARPC2","PRAME","HNRNPC","COL1A2","RPL5","CCT8")
longitudinal.tree = longitudinal.tree.plot(inference = inference, 
                                           labels_show = "clones", 
                                           clone_labels = clone_labels, 
                                           legend_position = "topright")
```

Lace - Lua Access Control Engine
================================

Lace is the core of an access control engine designed to be embedded into other
applications.  It is also designed to be extended by the very applications it
is embedded into.

As such, Lace provides only the core lexing, parsing, error handling and
related functionality of an access control engine, along with some initial
semantics to help the application developer along.

All rules and mechanisms of deciding if access is to be permitted or not are up
to the application author to define.  As such, while this documentation for
Lace will be useful for the application developer; it is recommended that the
applications do not refer their users to the Lace documentation except to
augment that provided in the application specific documentation.

The Lace codebase provides an example of using the library which should be
referred to for getting started with Lace.  However, there is also extensive
documentation on the [[syntax]] of Lace rulesets and also on the
[[compilation]] and [[execution]] phases of access control.

If you wish to assist with Lace development, then see the [[developing]]
document for pointers around the codebase and the test suite.

Lace - Syntax of Allow and Deny statements
==========================================

These access control statements start `allow` or `deny`, followed by a single
token reason to be returned to the caller (thus the reason should be quoted if
it contains spaces), followed by zero or more definition names, all of which
must pass for the statement to execute.

For example:

    allow "Administrators can do anything" is-admin
    deny "Plebs may do nothing" is-pleb
    
    allow ''

As with definitions, includes, etc, if rule names are prefixed with an
exclamation point then their sense is inverted, e.g.

    deny "Only admins may alter hooks" altering-hooks !is-admin

# Copilot Instructions for Lace

## Project Overview

**Lace** is a Rust-based firmware development project focused on building bootloaders and firmware for embedded systems. The project integrates Rust code with U-Boot's C codebase through a library interface called "ulib".

### Key Components

- **lace-boot**: Bootloader implementation supporting multiple platforms (sandbox, EFI, bare metal)
- **lace-util**: Core utility library with hardware identification (CHID), SMBIOS parsing, EDID handling, PE image parsing, and other firmware-related utilities
- **lace-platform**: Platform abstraction layer for different firmware environments
- **lace-util-derive**: Procedural macros for enum utilities
- **tools/**: Helper utilities (pewrap, collect-hwids, fakeedid)
- **ulib/**: U-Boot C codebase (external dependency - DO NOT MODIFY)

### Build System

The project uses a **Cargo workspace** for Rust code and a **custom Makefile** for lace-boot builds that integrate with U-Boot.

**Two build variants exist:**
- **Standard builds**: Link with ulib (U-Boot library) for full firmware functionality
- **Pure builds**: Standalone Rust-only binaries (controlled by `pure` feature flag)

**Supported platforms:**
- `sandbox`: Native Linux binary for testing and development
- `efi`: UEFI applications (efi-x86_app64, efi-x86_app32)
- `metal`: Bare metal ROM images (qemu-x86)

## Code Style and Conventions

### Follow STYLE.md Rigorously

The project has a detailed style guide in `STYLE.md`. Key points:

#### Comments
- Module-level `//!` doc comments explain purpose and design
- Public items require `///` doc comments
- Reference specifications when relevant (e.g., "per RFC 9562", "see SMBIOS spec §7.1")
- Aim for 80 column width
- **Don't comment obvious code** - comment the "why", not the "what"

#### Naming
- Constants: `SCREAMING_SNAKE_CASE`
- Types: `PascalCase`
- Functions/variables: `snake_case`
- Boolean functions: prefix with `is_`, `has_`, `can_`
- Use domain terminology consistently: `chid`, `smbios`, `guid`, `edid`

#### Integer Constants
- Avoid explicit type suffixes unless necessary
- Use underscores for readability: `0x1234_5678`, `1_000_000`
- Use lowercase hex (`0xabc`) for bit patterns, flags, addresses
- Use decimal for counts, sizes, human-meaningful quantities
- Prefer the `bitflags` crate for type-safe flag handling

#### Code Organization
1. Imports at top (formatted by cargo fmt)
2. Constants next
3. Type declarations
4. Functions (public before private)
5. Tests at bottom in `#[cfg(test)] mod test`
6. **Important**: Do not rearrange existing code when making changes

#### Tests
- Naming: `#[test] fn test_<function>_<scenario>()`
- Place unit tests in `#[cfg(test)] mod test` at file bottom
- Use specific assertions with context: `assert_eq!(a, b, "CHID type {} mismatch", i)`
- Test happy path, edge cases, and error conditions

### Licensing

All Lace code (excluding `ulib/`) uses dual licensing:

**Required header for all Rust source files:**
```rust
// SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only
// Copyright (C) 2025, Canonical Ltd.
// Authors: Name <email@canonical.com>
```

**For Python/Shell scripts:**
```python
#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only
# Copyright (C) 2025, Canonical Ltd.
```

**Note**: Files in `ulib/` use GPL-2.0+ and should not be modified as they are part of the U-Boot codebase.

## Building and Testing

### Quick Build Commands

```bash
# Run all pre-commit checks (formatting, linting, compilation, and tests)
pre-commit run --all-files

# Format code only (if not using pre-commit)
./scripts/cargo_ci.py fmt --all

# Run tests only (if not using pre-commit)
./scripts/cargo_ci.py test --workspace --exclude lace-stubble --exclude fakeedid --exclude u-boot-sys

# Only run these if you modified lace-boot:
# Test all board configurations (lace-boot changes only)
./scripts/test-builds.sh

# Build lace-boot for specific board (lace-boot changes only)
make -C lace-boot config all BOARD=sandbox
make -C lace-boot config all BOARD=efi-x86_app64
```

**Note**: Pre-commit hooks automatically run cargo fmt, cargo check, cargo clippy, and cargo test, so you typically only need to run `pre-commit run --all-files` before committing.

### CI Pipeline

GitHub Actions runs four jobs:
1. **check**: `cargo check` for all workspace members
2. **fmt**: Format checking with `rustfmt`
3. **clippy**: Linting with **all warnings as errors** (`-D warnings`)
4. **test**: Unit tests

### The cargo_ci.py Wrapper

The `./scripts/cargo_ci.py` wrapper is used instead of direct `cargo` commands because:
- It splits packages into multiple runs to avoid Cargo feature unification issues
- Use `--workspace` to run on all members, `--exclude` to skip specific crates
- Always exclude `lace-boot` from most workspace commands (it has special build requirements)

## Common Issues and Workarounds

### binman Conflicts

**Issue**: If you have a 'binman' tool from another source (e.g., a git binary manager), it may conflict with U-Boot's binman tool required for qemu-x86 builds.

**Solution**: Remove conflicting binman from PATH or ensure U-Boot's binman (from `ulib/tools/binman`) takes precedence.

### Profile Warnings

You may see warnings like:
```
warning: profiles for the non root package will be ignored, specify profiles at the workspace root
```

This is **expected** and can be ignored. The `panic = "abort"` setting is defined at workspace root in `Cargo.toml`.

### Pure vs Standard Builds

- **Pure builds** are Rust-only and don't link with ulib
- **Standard builds** require ulib and link with U-Boot C code
- qemu-x86 **does not support pure builds** (needs ulib for startup code)
- Use the `pure` feature flag to control which variant is built

### No-std Environment

Most crates use `#![no_std]` as they target firmware environments:
- Use `extern crate alloc;` for heap allocations
- Import from `core::` instead of `std::`
- `lace-util` has an optional `std` feature (default enabled)

## Domain-Specific Knowledge

### Key Concepts

- **CHID (Computer Hardware ID)**: GUIDs derived from system information (SMBIOS, EDID) used for firmware updates (fwupd/LVFS) and device tree selection. Follows Microsoft's CHID specification and RFC 9562.

- **SMBIOS**: System Management BIOS tables containing system information

- **EDID**: Extended Display Identification Data for monitors/panels

- **GUID/UUID**: Globally Unique Identifiers used throughout firmware

- **DTB/FDT**: Device Tree Blob/Flattened Device Tree for hardware description

- **PE Image**: Portable Executable format (used for EFI binaries)

### Critical Files to Understand

- `lace-util/src/chid.rs`: CHID computation algorithm
- `lace-util/src/chid_mapping.rs`: Device tree compatible string mapping
- `lace-util/src/smbios.rs`: SMBIOS table parsing
- `lace-util/src/edid.rs`: EDID parsing
- `lace-boot/build.rs`: Build script handling different targets and ulib linking
- `lace-boot/Makefile`: Complex build system integrating Cargo with U-Boot builds

## Commit Message Format

Follow the conventions in `CONTRIBUTING.md`:

```
component: Short summary in imperative mood

Longer explanation of what and why (not how). Wrap at 72 columns.
Reference issues or specs as needed.
```

### Component Prefixes

- Crate-specific: `lace-util/chid`, `lace-boot`, `lace-platform/efi`
- Repo-wide: `doc`, `ci`, `build`
- Scripts: `scripts` or script name

### Rules

- Capitalize first word after component prefix
- Use present/imperative tense ("Add feature" not "Added feature")
- First line ≤50 chars ideally, 72 max
- No `Signed-off-by` (but you may GPG sign with `git commit -S`)

### Examples

- `lace-util/chid: Add support for EDID panel source`
- `lace-util/smbios: Fix parsing of type 3 tables`
- `ci: Update Rust toolchain to 1.75`

## Development Workflow

### Making Changes

1. **Understand the codebase first** - Read related code and tests
2. **Follow the style guide** - Check `STYLE.md` and existing code patterns
3. **Add/update tests** - Unit tests in `#[cfg(test)] mod test`
4. **Run pre-commit checks before committing**: `pre-commit run --all-files` (runs formatting, linting, compilation checks, and unit tests)
5. **If you modified lace-boot**: Run `./scripts/test-builds.sh` or build specific boards with `make -C lace-boot config all BOARD=<board>`
6. **Commit with proper message format**

**Note**: Pre-commit hooks are configured in `.pre-commit-config.yaml` and run automatically on `git commit` if installed. They include cargo fmt, cargo check, cargo clippy, cargo test, and file quality checks. Always run `pre-commit run --all-files` before using **report_progress** (an internal progress-reporting command used in some automated workflows) to ensure all checks pass; if you do not use that workflow, you can ignore this part of the note.

### Before Submitting PR

- [ ] Code follows STYLE.md
- [ ] All files have proper license headers
- [ ] Tests added for new functionality
- [ ] Pre-commit hooks pass (`pre-commit run --all-files`)
- [ ] Commit messages follow conventions

## Safety and Error Handling

- Document safety invariants for `unsafe` blocks
- Use `Result` and `Option` types appropriately
- Avoid panics in production code paths (remember: `panic = "abort"`)
- For `no_std` code, consider whether panics are acceptable

## Dependencies

- Minimize new dependencies
- Prefer well-maintained crates from the ecosystem
- For `no_std` compatibility, check that dependencies support it
- Use `default-features = false` where appropriate

### Common Dependencies

- `zerocopy`: Zero-copy parsing of binary structures
- `bitflags`: Type-safe bitfield handling
- `fdt`: Flattened Device Tree parsing
- `uefi`, `uefi-raw`: UEFI API bindings
- `clap`: CLI argument parsing (for tools)

## Useful References

- [Project Style Guide](STYLE.md)
- [Contributing Guidelines](CONTRIBUTING.md)
- [Canonical Rust Best Practices](https://canonical.github.io/rust-best-practices/cosmetic-discipline.html)
- Microsoft CHID Specification
- RFC 9562 (UUID/GUID specification)
- UEFI Specification
- SMBIOS Specification

## Tips for Efficient Development

1. **Use `cargo_ci.py` instead of `cargo`** directly for consistency with CI
2. **Exclude lace-boot** from most workspace commands (it has special build requirements)
3. **Read the `lace-boot/README.md`** for detailed build system documentation
4. **Check existing tests** for patterns before writing new ones
5. **Reference specifications** in comments when implementing protocols/standards
6. **Don't rearrange code** unnecessarily - preserve existing organization
7. **Think about `no_std`** - most code runs in firmware environments without a full standard library

Lace - The nitty-gritty of the compilation process
==================================================

When you construct a Lace engine, you give it a compilation callback
set.  That set is used to call you back when Lace encounters something it
needs help compiling.  The structure of it is:

    { _lace = {
       loader = function(compcontext, nametoload) ... end,
       commands = {
          ... = function(compcontext, words...) ... end,
       },
       controltype = {
          ... = function(compcontext, type, words...) ... end,
       }
    } }

Anything outside of the `_lace` entry in the context is considered
fair game structure-wise and can be used by the functions called back
to acquire internal pointers etc.  Note however that the compilation
context will not be passed around during execution so if you need to
remember some of it in order to function properly, then it's up to
you.  Also note that anything not defined above in the `_lace` example
is considered "private" to Lace and should not be touched by non-Lace
code.

In addition, Lace will maintain a `source` entry in the _lace table
with the lexed source which is being compiled and, if we're compiling
an included source, a parent entry with the compilation context of the
parent.  The toplevel field is the compilation context of the top
level compilation.  If parent is nil, then toplevel will equal
compcontext.  Lace also maintains a `linenr` entry with the
currently-being-compiled line number, so that commands and control
types can use that in error reports if necessary.

If `loader` is absent then the `include` statement refuses to include
anything mandatory.  If it is present but returns nil when called then
the `include` statement fails any mandatory includes which do so.

Otherwise the loader is expected to return the 'real' name of the
source and the content of it.  This allows for symbolic lookups.

If Lace encounters a command it does not recognise then it will call
`context._lace.commands[cmdname]` passing in the words representing the line
in question.  It's up to that function to compile the line or else to
return an error.

If Lace encounters a control type during a `define` command which it
does not understand, then it calls the `context._lace.controltype[name]`
function passing in all the remaining arguments.  The control type
function is expected to return a compiled set for the define or else
an error.

To start a Lace engine compiling a ruleset, simply do (pseudocode):

    rules, err = lace.compiler.compile(compcontext, sourcename[, sourcecontent])

If `sourcecontent` is not given, Lace will use the loader in the
`compcontext` to load the source.

If `rules` is `nil`, then `err` is a Lua error.
If `rules` is `false`, then `err` is a formatted error from compilation
Otherwise, `rules` should be a table containing the ruleset.

Internally, once compiled, Lace rulesets are a table of tables.  Each
rule entry has a reference to its source and line number.  It then has
a function pointer for executing this rule, and a set of arguments to
give the rule.  Lace automatically passes the execution context as the
first argument to the rule.  Sub-included rulesets are simply one of
the arguments to the function used to run the rule.

Loader
======

When Lace wishes to load an entry, it calls the `loader` function.  This
is to allow rulesets to be loadable from arbitrary locations such as
files on disk, HTTP URLs, random bits of memory or even out of version
control repositories directly.

The `loader` function is given the compilation context and the name of the
source to load.  Note that while it has the compilation context, the loader
function must be sensitive to the case of the initial load.  Under that
circumstance, the source information in the compilation context will be
unavailable.  The loader function is required to fit the following pseudocode
definition:

    realname, content = loader(compcontext, nametoload)

If `realname` is not a string then content is expected to be an
"internal" error message (see below) which will be augmented with the
calling source position etc and rendered into an error to return to
the caller of `lace.compiler.compile()`.

If `realname` is a string then it is taken to be the real name of the
loaded content (at worst you should return nametoload here) and
`content` is a string representing the contents of that file.

Once it has been loaded by the `loader` function, Lace will compile that
sub-ruleset before continuing with the current ruleset.

Commands
========

When Lace wishes to compile a command for which it has no internal
definition, it will call the command function provided in the
compilation context.  If no such command function is found, it will
produce an error and stop the compilation.

The command functions must fit the following pseudocode definition:

    cmdtab, msg = command_func(compcontext, words...)

If `cmdtab` is not a table, msg should be an "internal" error message
(see below) which will be augmented with the calling source position
etc and rendered into an error to return to the caller of
`lace.compiler.compile()`.

If `cmdtab` is a table, it is taken to be the compiled table
representing the command to run at ruleset execution time.  It should
have the form:

    { fn = exec_function, args = {...} }

Lace will automatically augment that with the source information which
led to the compiled rule for use later.

The `exec_function` is expected to fit the following pseudocode
definition:

    result, msg = exec_function(exec_context, unpack(args))

See [[execution]] for notes on how these `exec_function` functions are meant
to behave.

Control Types
=============

When Lace is compiling a definition rule with a control type it has
not got internally, Lace will call the controltype function associated
with it (or report an error if no such control type is found).

The control type functions must fit the following pseudocode
definition:

    ctrltab, msg = controltype_func(compcontext, type, words...)

If `ctrltab` is not a table, msg should be an "internal" error message
(see below) which will be augmented with the calling source position
etc and rendered into an error to return to the caller of
`lace.compiler.compile()`.

If `ctrltab` is a table, it is taken to be the compiled table
representing the control type to run at ruleset execution time.  It
should have the form:

    { fn = ct_function, args = {...} }

The `ct_function` is expected to fit the following pseudocode
definition:

    result, msg = ct_function(exec_context, unpack(args))

See [[execution]] for notes on how these `ct_function` functions are meant
to behave.

Compiler internal errors
========================

Error messages during compilation are generated by calling:

    return lace.error.error("my message", { x, y })

Where the table is a list of numeric indices of the words which caused the
error.  If words is empty (or nil) then the error is considered to be the
entire line.

Lace will use this information to construct meaningful long error
messages which point at the words in question.  Such as:

    Unknown command name: 'go_fish'
    myruleset :: 6
    go_fish "I have no bananas"
    ^^^^^^^

In the case of control type compilation, the words will automatically be offset
by the appropriate number to account for the define words.  This means you
should always 1-index from your arguments where index 1 is the control type
word index.

The same kind of situation occurs during [[execution]].

Lace - Execution of rulesets
============================

Once compiled, a ruleset is essentially a sequence of functions to call on the
execution context.  The simplest execution context is an empty table.  If Lace
is going to store anything it will use a `_lace` prefix as with [[compilation]]
contexts.  As with compilation, the caller is not permitted to put anything
inside `_lace` nor to rely on its layout.

A few important functions make up the execution engine.  The top level
function is simply:

    result, msg = lace.engine.run(ruleset, exec_context)

This will run `ruleset` with the given execution context and return
a simple result.

If `result` is `nil`, then `msg` is a long-form string error explaining
what went wrong.  It represents a Lua error being caught and as such
you may not want to report it to your users.

If `result` is `false`, then `msg` is a long-form string error
explaining that something returned an error during execution which it
would be reasonable to report to users under most circumstances.

If `result` is `"allow"`, then `msg` is an optional string saying why
the ruleset resulted in an allow.  Ditto for `"deny"`.  Essentially any
string might be a reason.  This is covered below in Commands.

Commands
========

When a command is being run, it is called as:

    result, msg = command_fn(exec_context, unpack(args))

where `args` are the arguments returned during the compilation of the command.

If the function throws an error, that will be caught and processed by
the execution engine.

If `result` is falsehood (`nil`, `false`) then the command is considered to
have failed for some reason and `msg` contains an "internal" error
message to report to the user.  This aborts the execution of the
ruleset.

If `result` is `true`, then the command successfully ran, and execution
continues at the next rule.

If `result` is a string, then the command returned a result.  This
ceases execution of the ruleset and the result and message (which must
be a string explanation) are returned to the caller.  Typically such
results would be "allow" or "deny" although there's nothing forcing
that to be the case.

Control Types
=============

When a control type function is being run, it is called as:

    result, msg = ct_fn(exec_context, unpack(args))

where `args` are the arguments returned when the definition was compiled.

If the function throws an error, it will be caught and processed by
the execution engine.

If `result` is `nil` then msg is an "internal" error, execution will be
stopped and the issue reported to the caller.

If `result` is `false`, the control call failed and returned falsehood.
Anything else and the control call succeeded and returns truth.

Control type functions are called at the point of test, not at the
point of definition.  Control type results are *NOT* cached.  It is up
to the called functions to perform any caching/memoising of results as
needed to ensure suitably performant behaviour.

Helper functions
================

Since sometimes when writing command functions, you need to know if a given
define rule passes, Lace provides a function to do this.  It is bound up in the
behaviour of Lace's internal `define` command and as such, you should treat it
as a black box.

    result, msg = lace.engine.test(exec_context, name)

This, via the magic of the execution context calls through to the
appropriate control type functions, returning their results directly.

This means that it can throw an error in the case of a Lua error,
otherwise it returns the two values as detailed above.

Lace - The syntax of the default statement
==========================================

The `default` statement is unusual in that it has no behaviour at runtime.  At
compile time the `default` statement alters the behaviour of the compiler with
respect to what happens at the end of the ruleset parse.

If, when Lace has finished parsing the ruleset, the last allow or deny was not
unconditional, then the compiler will, in the absence of a `default` statement,
inject a terminal `allow`/`deny` of the opposite sense of the last explicit
operation, unconditionally and with a reason of the empty string.

If a `default` statement was encountered during processing then its chosen
behaviour will be used instead.

The syntax of the `default` statement is:

    default 'allow' <reason>?

or

    default 'deny' <reason>?

If reasons are not provided, the string "Default behaviour" is
substituted.

Once a single `default` statement has been encountered during
compilation; it is an error, and the compiler WILL cease, if it
encounters an additional `default` statement.

Since it's common for rulesets to stem from a single core point, therefore it
is recommended that the application define a `default` policy at the start of
these core statements.

Lace - Syntax of rulesets
=========================

Lace rule files are parsed line-by-line.  There is no provision at
this time for rules to be split across multiple lines.  If you require
such, please [submit a patch][developing].

Lace splits each line into a series of tokens.  Tokens are always
strings and they can be optionally quoted to allow for spaces in them.
Certain tokens then have constraints placed on them so that further
parsing will be unambiguous.

The lexing into tokens is done similarly to how shell does it.  This
means that each whitespace-separated "word" is then subjected to
quoting rules.  Lace does not have many special characters.  The
backslash introduces escaped values in all cases and lace does not
differentiate between single and double quotes.  Backslash escaped
characters sometimes have special meaning if inside quotes.  For
example, the following strings lex in the following ways:

1. Two tokens, one of each word.

          hello world

2. The same as 1.

          hello   world

3. One token with both words separated by one space

          "hello world"

4. Same as 3.

          'hello world'

5. Same as 3 and 4.

          hello\ world

6. One token, consisting of the letters: `uptown`

          up\town

6. One token, consisting of the letters: `up TAB own`

          "up\town"

7. One token, a double-quote character

          \"

8. The same as 7

          '"'

9. The same as 7 and 8.

          "\""


As you can see, the lexing rules are not trivial, but should not come
as a surprise to anyone used to standard command-line lexing
techniques.

Comments in Lace are prefixed by a hash '#', a double-slash '//' or a
double-dash '--'.  The first word of the line must start with one of
these markers (but may contain other text also), for example:

    # This is a comment
    // As is this
    -- And this
    
    #But also this
    //And this
    --And also this

Blank lines are permitted and ignored (except for counting), any
prefixed or postfixed whitespace is deleted before lexing begins.  As
such:

    # This is a comment
      # So is this

This allows for sub-rulesets to be indented as the author pleases,
without altering the meaning of the rules.

The first word of a rule defines its type.  You can think of it as the
command being run.  Lace, by default, provides a small number of rule
types:

1. [Definitions](../syntax-define)
   * This define access control stanzas.  The definitions produced are
     used in further rules to control access.  Lace does not allow any
     name to be reused.
2. [Includes](../syntax-include)
   * Lace can include further rules at any point during a ruleset.  If
     the rules are to be optionally run then Lace cannot perform static
     analysis of the definitions within the ruleset.  Instead it will
     rely on runtime catching of multiple-definitions etc.
3. [Access control statements](../syntax-allow-deny)
   * These are the core functions of Lace.  Namely the allow and deny
     statements.  These use control definitions from earlier in a ruleset
     to determine whether to allow or deny access.  The first allow
     or deny statement which passes will stop execution of the ruleset.
4. [Default statement](../syntax-default)
  * The 'default' statement can only be run once and provides Lace with
    information on what to do in the case of no allow or deny rule passing.

In those files, if you encounter the words WILL, MAY, SHOULD, MUST, or
their negatives, specifically in all-caps, then the meaning of the
words is taken in the spirit of the RFC usage of them.

Lace - Syntax of definition statements
======================================

Definition statements start with one of `def` or `define`.  They may
also start `acl` to look more squiddish.

The rough syntax of a definition statement is:

    define <name> <controltype> <0-or-more-args>

for example:

    define is-admin is-in-group administrators

There are some loose constraints on the name and controltype
arguments.  The name SHOULD NOT contain quote characters and MUST NOT
start with an exclamation point.  The controltype simply SHOULD NOT
contain quotes.

The control types are typically provided by the program which is using
Lace.  However Lace does provide some simple control types which can
be useful and callers are welcome to add them to their engines.

The two simple control types Lace provides are the `allof` and `anyof`
controls.  They, as their arguments, take two or more rule names and match if
all or any (respectively) of their arguments resolve to true.  As with other
parts of Lace, if the rule names are prefixed by exclamation points then their
result is inverted before being tested.

These give you ways to produce common subexpressions of the 'AND' and
'OR' forms to be used in later rules.

# Style Guide

This guide covers aspects not handled by rustfmt. See also
<https://canonical.github.io/rust-best-practices/cosmetic-discipline.html>.

## Comments

### When to comment

- Module-level `//!` doc comments explaining the module's purpose and design
- Public items (`pub fn`, `pub struct`, `pub const`) get `///` doc comments
- Complex algorithms or non-obvious logic get inline `//` comments
- Constants that represent magic values or external specifications
- Safety invariants for `unsafe` blocks

### When not to comment

- Obvious code that reads clearly on its own
- Private helper functions with self-explanatory names and few arguments
- Restating what the code does rather than why

### Comment style

- Aim for 80 columns
- Use complete sentences with proper punctuation for doc comments
- Use sentence fragments for inline comments (no trailing period for
  single-line)
- Reference specifications (e.g., "per RFC 9562", "see SMBIOS spec §7.1")

### Doc comment structure for functions

```rust
/// Verify that data matches an expected CRC32 checksum.
///
/// Longer explanation if needed.
///
/// # Examples
///
/// ```
/// let data = [0x12, 0x34, 0x56, 0x78];
/// assert!(verify_crc32(&data, 0x114));
/// ```
fn verify_crc32(data: &[u8], expected: u32) -> bool {
    ...
}
```

See <https://doc.rust-lang.org/std/fs/fn.hard_link.html> for a fuller example
(see the 'Source' link at the top).

## Tests

### Naming

```rust
#[test]
fn test_<function>_<scenario>() { }
```

### Structure

- Place unit tests in a `#[cfg(test)] mod test` at the bottom of the file
- Place larger / fuzz tests in a separate tests/ directory
- Use real-world test data where possible
- Comment test cases that come from external sources or known-good values

### What to test

- Happy path with representative inputs
- Edge cases (empty input, max values, etc.)
- Error conditions (return `None`, `Err`, etc.)
- Regression tests for fixed bugs

### Assertions

- Use specific assertion macros (`assert_eq!`, `assert_ne!`)
- Include context message: `assert_eq!(a, b, "CHID type {} mismatch", i)`

## Naming Conventions

- Constants: `SCREAMING_SNAKE_CASE`
- Types: `PascalCase`
- Functions/variables: `snake_case`
- Prefix boolean functions with `is_`, `has_`, `can_`
- Use domain terminology consistently (e.g., `chid`, `smbios`, `guid`)

## Integer Constants

### Declaration

Avoid specifying an explicit type for integer constants unless needed:

```rust
const MAGIC: 0x1234_5678;
const REG_OFFSET: u16 = 0x100;
const MAX_ENTRIES: u32 = 256;
```

Use underscores to separate groups of digits for readability:

```rust
const SIZE: 1_000_000;
const FLAGS: 0b1010_0011;
const ADDR: 0xdead_beef_cafe_babe;
```

### Hex vs decimal

- Use lower-case hex (`0xabc`) for bit patterns, flags, addresses, and hardware
  constants
- Use decimal for counts, sizes, and human-meaningful quantities

### Type suffixes

Avoid type suffixes on literals when the type is clear from context:

```rust
// Preferred: type is implied by the constant
const SIZE: 4096;

// Avoid: redundant suffix
const SIZE: u32 = 4096u32;

// OK: suffix to disambiguate in expressions
let x: 1_u64 << 32;

// OK: rely on the inferencer to get the type right
let y = 1 << 32;
```

### Bitmasks and flags

Prefer the `bitflags` crate for type-safe flag handling:

```rust
use bitflags::bitflags;

bitflags! {
    #[derive(Debug, Clone, Copy)]
    pub struct Permissions: u32 {
        const READ = 1 << 0;
        const WRITE = 1 << 1;
        const EXEC = 1 << 2;
    }
}

const DEFAULT_PERMS: Permissions = Permissions::READ.union(Permissions::WRITE);
```

## Code Organization

Remember that these are general guidelines, not fixed requirements. In some
cases constants and types are better with the functions which use them. Use your
discretion. When reviewing code, try to accept what is there unless you have a
good reason for a change.

- Imports go at the top of the file, as formatted by cargo
- Constants come next (assuming they don't need a type declaration)
- Type declarations after that
- Functions after that
- Public items before private items (but don't move functions in existing code
  when making them public / private)
- Tests at the bottom in `mod test`
- Note: Existing code should not be rearranged when making changes

# Contributing to lace

The lace team welcomes contributions via GitHub pull requests. To get your PR
merged, you need to sign [Canonical's contributor license agreement
(CLA)][cla].

Please follow the project [style guide](STYLE.md) for code formatting, comments
and tests.

## Commit Messages

### Format

```
component: Short summary in imperative mood

Longer explanation of what and why (not how). Wrap at 72 columns.
Reference issues or specs as needed.
```

### Component prefix

The component identifies which part of the codebase is affected:

- For crate-specific modules, use `crate/module`: `lace-util/chid`,
  `lace-lib/fdt`
- For top-level crate changes, use the crate name: `lace-util`, `lace-boot`
- For repo-wide files, use a descriptive prefix: `doc`, `ci`, `build`
- For scripts, use `scripts` or the script name: `scripts/vm-run`

Examples:
- `lace-util/chid.rs` → `lace-util/chid: ...`
- `lace-util/src/smbios/mod.rs` → `lace-util/smbios: ...`
- `lace-boot/src/main.rs` → `lace-boot: ...`
- `STYLE.md`, `README.md` → `doc: ...`
- `.github/workflows/` → `ci: ...`

### Rules

- Capitalize the first word after the component prefix
- Use present/imperative tense ("Add feature" not "Added feature")
- First line ≤50 chars ideally, 72 max
- Blank line between summary and body
- Body wrapped at 72 columns
- Do not use `Signed-off-by`
- You may sign your commits (i.e. `git commit -S`) but it is not required

### Examples

- `lace-util/chid: Add support for EDID panel source`
- `lace-util/smbios: Fix parsing of type 3 tables`
- `lace-util/lib: Refactor GUID handling for clarity`

## Development Setup

This repository uses [pre-commit](https://pre-commit.com/) hooks to ensure code
quality. To set up pre-commit hooks:

```bash
pip install pre-commit
pre-commit install
```

The hooks will run automatically on `git commit` and include:
- Code formatting checks (`cargo fmt`)
- Compilation checks (`cargo check`)
- Lint checks (`cargo clippy`)
- Unit tests (`cargo test` on most workspace crates)
- General file checks (trailing whitespace, YAML/TOML syntax, etc.)

**Before committing**, you can run the hooks manually to catch issues early:
```bash
pre-commit run --all-files
```

If pre-commit is installed, it will automatically run on every commit and prevent
commits with failing checks. This ensures all code meets quality standards before
being pushed to the repository.

[cla]: https://ubuntu.com/legal/contributors

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{compute.variants.error.rates}
\alias{compute.variants.error.rates}
\title{compute.variants.error.rates}
\usage{
\method{compute}{variants.error.rates}(D, inference)
}
\arguments{
\item{D}{Mutation data from multiple experiments for a list of driver genes provided as a data matrix per time point.}

\item{inference}{Results of the inference by LACE.}
}
\value{
A matrix variants_error_rates with the estimated error rates for the considered variants.
}
\description{
Compute error rates for the considered variants comparing observed data to LACE corrected genotype.
}
\examples{
data(longitudinal_sc_variants)
data(inference)
variants_error_rates <- compute.variants.error.rates(longitudinal_sc_variants,inference)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frontend.R
\name{lacedata}
\alias{lacedata}
\title{lacedata}
\usage{
lacedata(
  D,
  lik_w = NULL,
  alpha = NULL,
  beta = NULL,
  initialization = NULL,
  random_tree = FALSE,
  keep_equivalent = TRUE,
  check_indistinguishable = TRUE,
  num_rs = 50,
  num_iter = 10000,
  n_try_bs = 500,
  learning_rate = 1,
  marginalize = FALSE,
  error_move = FALSE,
  num_processes = Inf,
  seed = NULL,
  verbose = TRUE,
  log_file = ""
)
}
\arguments{
\item{D}{Mutation data from multiple experiments for a list of driver genes. It can be either a list with a data matrix per time point or a SummarizedExperiment object.
In this latter, the object must contain two fields: assays and colData. Assays stores one unique data matrix pooling all single cells observed at each time point and colData stores a vector of labels reporting the time point when each single cell was sequenced.
Ordering of cells in assays field and colData field must be the same.}

\item{lik_w}{Weight for each data point. If not provided, weights to correct for sample sizes are used.}

\item{alpha}{False positive error rate provided as list of elements; if a vector of alpha (and beta) is provided, the inference is performed for multiple values and the solution at
maximum-likelihood is returned.}

\item{beta}{False negative error rate provided as list of elements; if a vector of beta (and alpha) is provided, the inference is performed for multiple values and the solution at
maximum-likelihood is returned.}

\item{initialization}{Binary matrix representing a perfect philogeny clonal tree; clones are rows and mutations are columns.
This parameter overrides "random_tree".}

\item{random_tree}{Boolean. Shall I start MCMC search from a random tree? If FALSE (default) and initialization is NULL, search
is started from a TRaIT tree (BMC Bioinformatics . 2019 Apr 25;20(1):210.  doi: 10.1186/s12859-019-2795-4).}

\item{keep_equivalent}{Boolean. Shall I return results (B and C) at equivalent likelihood with the best returned solution?}

\item{check_indistinguishable}{Boolean. Shall I remove any indistinguishable event from input data prior inference?}

\item{num_rs}{Number of restarts during mcmc inference.}

\item{num_iter}{Maximum number of mcmc steps to be performed during the inference.}

\item{n_try_bs}{Number of steps without change in likelihood of best solution after which to stop the mcmc.}

\item{learning_rate}{Parameter to tune the probability of accepting solutions at lower values during mcmc. Value of learning_rate = 1 (default), set a
probability proportional to the difference in likelihood; values of learning_rate greater than 1 inclease the chance of accepting solutions at lower likelihood
during mcmc while values lower than 1 decrease such probability.}

\item{marginalize}{Boolean. Shall I marginalize C when computing likelihood?}

\item{error_move}{Boolean. Shall I include estimation of error rates in the MCMC moves?}

\item{num_processes}{Number of processes to be used during parallel execution. To execute in single process mode,
this parameter needs to be set to either NA or NULL.}

\item{seed}{Seed for reproducibility.}

\item{verbose}{Boolean. Shall I print to screen information messages during the execution?}

\item{log_file}{log file where to print outputs when using parallel. If parallel execution is disabled, this parameter is ignored.}
}
\value{
shiny interface
}
\description{
Perform the inference of the maximum likelihood clonal tree from longitudinal data.
}
\examples{
data(longitudinal_sc_variants)
lacedata(D = longitudinal_sc_variants,
         lik_w = c(0.2308772,0.2554386,0.2701754,0.2435088),
         alpha = list(c(0.10,0.05,0.05,0.05)),
         beta = list(c(0.10,0.05,0.05,0.05)),
         keep_equivalent = TRUE,
         num_rs = 5,
         num_iter = 10,
         n_try_bs = 5,
         num_processes = NA,
         seed = 12345,
         verbose = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LACEinterface.R
\name{LACEview}
\alias{LACEview}
\title{LACE Processing and Analysis Interface}
\usage{
LACEview()
}
\value{
The GUI
}
\description{
\code{LACEview} displays a Shiny user interface to
handle the VCF and BAM files processing that is needed to
construct the input for the LACE inference algorithms.
The function generates also the maximum likelihood longitudinal
clonal tree, and shows the output for further explorations of
the results.
}
\note{
The function \code{LACE} is still available for retrocompatibility.
}
\section{Installation}{

The package is available on GitHub and Bioconductor.
LACE 2.0 requires R > 4.1.0 and Bioconductor.

To install directly from github run:

\if{html}{\out{<div class="sourceCode">}}\preformatted{remotes::install_github("https://github.com/BIMIB-DISCo/LACE", 
                        dependencies = TRUE)
}\if{html}{\out{</div>}}
}

\section{Dependencies}{

LACE 2.0 uses \emph{Annovar} and \emph{Samtools suite} as back-ends for variant calling
annotation and depth computation, respectively.

\emph{Annovar} is a variant calling software written in \emph{Perl} freely available upon
registration to their website at
\url{https://annovar.openbioinformatics.org/en/latest/}.

\emph{Perl} can be found and installed at \url{https://www.perl.org/}.

\emph{Samtools suite} is a set of tools to handle SAM/BAM/BED file format. It is
freely available at \url{http://www.htslib.org/}. To install \emph{Samtools} follow the
instructions in their website.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{longitudinal.tree.plot}
\alias{longitudinal.tree.plot}
\title{longitudinal.tree.plot}
\usage{
longitudinal.tree.plot(
  inference,
  rem_unseen_leafs = TRUE,
  show_plot = TRUE,
  filename = "lg_output.xml",
  labels_show = "mutations",
  clone_labels = NULL,
  show_prev = TRUE,
  label.cex = 1,
  size = 500,
  size2 = NULL,
  tk_plot = FALSE,
  tp_lines = TRUE,
  tp_mark = TRUE,
  tp_mark_alpha = 0.5,
  legend = TRUE,
  legend_position = "topright",
  label_offset = 4,
  legend_cex = 0.8
)
}
\arguments{
\item{inference}{Results of the inference by LACE.}

\item{rem_unseen_leafs}{If TRUE (default) remove all the leafs that have never been observed (prevalence = 0 in each time point)}

\item{show_plot}{If TRUE (default) output the longitudinal tree to the current graphical device.}

\item{filename}{Specify the name of the file where to save the longitudinal tree. Dot or graphml formats are supported and are chosen based on the extenction of the filename (.dot or .xml).}

\item{labels_show}{Specify which type of label should be placed on the tree; options are, 
"mutations": parental edges are labeled with the acquired mutation between the two nodes (genotypes); 
"clones": nodes (genotypes) are labeled with their last acquired mutation; 
"both": either nodes and edges are labeled as specified above; 
"none": no labels will show on the longitudinal tree.}

\item{clone_labels}{Character vector that specifies the name of the nodes (genotypes). If it is NULL (default), nodes will be labeled as specified by "label" parameter.}

\item{show_prev}{If TRUE (default) add to clones label the correspongind prevalance.}

\item{label.cex}{Specify the size of the labels.}

\item{size}{Specify size of the nodes. The final area is proportional with the node prevalence.}

\item{size2}{Specify the size of the second dimension of the nodes. If NULL (default), it is set equal to "size".}

\item{tk_plot}{If TRUE, uses tkplot function from igraph library to plot an interactive tree. Default is FALSE.}

\item{tp_lines}{If TRUE (defaul) the function draws lines between timepoints.}

\item{tp_mark}{If TRUE (defaul) the function draws different colored area under the nodes in different time points.}

\item{tp_mark_alpha}{Specify the alpha value of the area drawed when tp_mark = TRUE.}

\item{legend}{If TRUE (default) a legend will be displayed on the plot.}

\item{legend_position}{Specify the legend position.}

\item{label_offset}{Move the mutation labels horizontally (default = 4)}

\item{legend_cex}{Specify size of the legend text.}
}
\value{
An igraph object g with the longitudinal tree inferred by LACE.
}
\description{
Plot a longitudinal tree inferred by LACE.
}
\examples{
data(inference)
clone_labels = c("ARPC2","PRAME","HNRNPC","COL1A2","RPL5","CCT8")
longitudinal.tree.plot(inference = inference,
                       labels = "clones",
                       clone_labels = clone_labels,
                       legend_position = "topleft")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{inference}
\alias{inference}
\title{Results obtained with the function LACE on the provided input data from Rambow, Florian, et al. "Toward minimal residual disease-directed therapy in melanoma." Cell 174.4 (2018): 843-855.}
\format{
Results obtained with the function LACE on the provided input data
}
\usage{
data(inference)
}
\value{
Results obtained with the function LACE on the provided input data
}
\description{
Results obtained with the function LACE on the provided input data from Rambow, Florian, et al. "Toward minimal residual disease-directed therapy in melanoma." Cell 174.4 (2018): 843-855.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{longitudinal_sc_variants}
\alias{longitudinal_sc_variants}
\title{Mutation data from Rambow, Florian, et al. "Toward minimal residual disease-directed therapy in melanoma." Cell 174.4 (2018): 843-855.}
\format{
List of mutation data for four time points
}
\source{
Rambow, Florian, et al. "Toward minimal residual disease-directed therapy in melanoma." Cell 174.4 (2018): 843-855.
}
\usage{
data(longitudinal_sc_variants)
}
\value{
List of mutational data for a total of 475 single cells
}
\description{
The dataset includes somatic single nucleotide variants at the single cell resolution. SNVs are called from SMARTseq2 fastq obtained from Gene Expression Omnibus database 
with the accession number: GSE116237. The dataset includes single cell data from a PDX melanoma model before and on treatment with BRAF and MEK inhibitors. The fastq files are processed 
to obtain the mutational profile following GATK best practice (https://gatkforums.broadinstitute.org/gatk/discussion/3891/calling-variants-in-rnaseq) usign the GRCh38 human genome as reference. 
Mutation data are stored in an N x M binary matrix with N single cells and M somatic single nucleotide variants. Row names report the ID of the fastq file related to a specific single cell; 
columns names report the SNV that are formatted as GeneName_chromosome_position_referenceAllele_alternateAllele. Each matrix entry can be 1 (mutation detected), 0 (mutation absent) or NA 
(too low coverage to determine the presence or absence of that mutation). For further details, please refer to the Methods Section and the section 3.1 of supplementary materials of 
Ramazzotti, Daniele, et al. "Longitudinal cancer evolution from single cells." bioRxiv (2020).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LACEview2.R
\name{lace_interface}
\alias{lace_interface}
\title{LACE Interface}
\usage{
lace_interface(
  B_mat,
  clones_prevalence,
  C_mat,
  error_rates,
  width = NULL,
  height = NULL,
  elementId = NULL,
  info = ""
)
}
\arguments{
\item{B_mat}{(Required). B is the clonal tree matrix where columns are the
clonal mutations and and rows are the clones. The clonal tree matrix should
contain a column and a row named "Root" representing the root of the tree and
the wild type, respectively.
B is a binary matrix where 1 are the mutations associated to the clones.
The wild type column has all ones}

\item{clones_prevalence}{(Required) The clonal prevalence matrix}

\item{C_mat}{(Required) The corrected clonal attachment}

\item{error_rates}{(Required) The false positive alpha and false negative beta
error rates used to infer the clonal tree}

\item{width}{(optional) Size of the window interafce}

\item{height}{(optional) Size of the window interafce}

\item{elementId}{(optional) Element id}

\item{info}{(Optional). HTML formatted text with information regarding
the experiments}
}
\value{
An implementation of the htmlwidgets
}
\description{
This function generates a longitudinal clonal tree and a
graphic interface to explore the data using as input the clonal tree
formatted in the same way as the one produced by LACE during the imputation
steps
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{compute.mutation.distance}
\alias{compute.mutation.distance}
\title{compute.mutation.distance}
\usage{
\method{compute}{mutation.distance}(inference)
}
\arguments{
\item{inference}{Results of the inference by LACE.}
}
\value{
A matrix mutation_distance with the mutation distance among variants computed from LACE corrected genotype and related hierarchical clustering.
}
\description{
Compute mutation distance among variants from LACE corrected genotype and use it to perform hierarchical clustering.
}
\examples{
data(inference)
mutation_distance <- compute.mutation.distance(inference)

}
